Type: Package
Package: connectapi
Title: Utilities for Interacting with the 'Posit Connect' Server API
Version: 0.12.1
Authors@R: c(
    person(given = "Kara",
           family = "Woo",
           role = c("aut", "cre"),
           email = "kara.woo@posit.co"),
    person(given = "Toph",
           family = "Allen",
           role = c("aut"),
           email = "toph@posit.co"),
    person(given = "Neal",
           family = "Richardson",
           role = c("aut")),
    person(given = "Sean",
           family = "Lopp",
           role = c("aut")),
    person(given = "Cole",
           family = "Arendt",
           role = c("aut")),
    person(given = "Posit, PBC",
           role = c("cph", "fnd")))
Description: Provides a helpful 'R6' class and methods for interacting with
    the 'Posit Connect' Server API along with some meaningful utility functions
    for regular tasks. API documentation varies by 'Posit Connect' installation
    and version, but the latest documentation is also hosted publicly at
    <https://docs.posit.co/connect/api/>.
License: MIT + file LICENSE
URL: https://posit-dev.github.io/connectapi/,
        https://github.com/posit-dev/connectapi
BugReports: https://github.com/posit-dev/connectapi/issues
Imports: bit64, fs, glue, httr, mime, jsonlite, lifecycle, magrittr,
        purrr, R6, rlang (>= 0.4.2), tibble, uuid, vctrs (>= 0.3.0),
        base64enc
Suggests: covr, dbplyr, dplyr, ggplot2, gridExtra, httptest, knitr,
        lubridate, progress, rmarkdown, rprojroot, rsconnect, spelling,
        testthat, tidyr, webshot2, withr
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Collate: 'audits.R' 'browse.R' 'connect.R' 'connectapi-package.R'
        'connectapi.R' 'content.R' 'deploy.R' 'get.R' 'git.R'
        'groups.R' 'integrations.R' 'lazy.R' 'page.R' 'parse.R'
        'promote.R' 'ptype.R' 'remote.R' 'runtime-caches.R'
        'schedule.R' 'tags.R' 'utils.R' 'thumbnail.R' 'user.R'
        'utils-pipe.R' 'variant.R'
NeedsCompilation: no
Packaged: 2026-04-29 19:15:15 UTC; kara
Author: Kara Woo [aut, cre],
  Toph Allen [aut],
  Neal Richardson [aut],
  Sean Lopp [aut],
  Cole Arendt [aut],
  Posit, PBC [cph, fnd]
Maintainer: Kara Woo <kara.woo@posit.co>
Repository: CRAN
Date/Publication: 2026-04-29 20:00:03 UTC
Built: R 4.6.0; ; 2026-05-01 21:12:19 UTC; unix

Type: Package
Package: connectapi
Title: Utilities for Interacting with the 'Posit Connect' Server API
Version: 0.12.1
Authors@R: c(
    person(given = "Kara",
           family = "Woo",
           role = c("aut", "cre"),
           email = "kara.woo@posit.co"),
    person(given = "Toph",
           family = "Allen",
           role = c("aut"),
           email = "toph@posit.co"),
    person(given = "Neal",
           family = "Richardson",
           role = c("aut")),
    person(given = "Sean",
           family = "Lopp",
           role = c("aut")),
    person(given = "Cole",
           family = "Arendt",
           role = c("aut")),
    person(given = "Posit, PBC",
           role = c("cph", "fnd")))
Description: Provides a helpful 'R6' class and methods for interacting with
    the 'Posit Connect' Server API along with some meaningful utility functions
    for regular tasks. API documentation varies by 'Posit Connect' installation
    and version, but the latest documentation is also hosted publicly at
    <https://docs.posit.co/connect/api/>.
License: MIT + file LICENSE
URL: https://posit-dev.github.io/connectapi/,
        https://github.com/posit-dev/connectapi
BugReports: https://github.com/posit-dev/connectapi/issues
Imports: bit64, fs, glue, httr, mime, jsonlite, lifecycle, magrittr,
        purrr, R6, rlang (>= 0.4.2), tibble, uuid, vctrs (>= 0.3.0),
        base64enc
Suggests: covr, dbplyr, dplyr, ggplot2, gridExtra, httptest, knitr,
        lubridate, progress, rmarkdown, rprojroot, rsconnect, spelling,
        testthat, tidyr, webshot2, withr
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Collate: 'audits.R' 'browse.R' 'connect.R' 'connectapi-package.R'
        'connectapi.R' 'content.R' 'deploy.R' 'get.R' 'git.R'
        'groups.R' 'integrations.R' 'lazy.R' 'page.R' 'parse.R'
        'promote.R' 'ptype.R' 'remote.R' 'runtime-caches.R'
        'schedule.R' 'tags.R' 'utils.R' 'thumbnail.R' 'user.R'
        'utils-pipe.R' 'variant.R'
NeedsCompilation: no
Packaged: 2026-04-29 19:15:15 UTC; kara
Author: Kara Woo [aut, cre],
  Toph Allen [aut],
  Neal Richardson [aut],
  Sean Lopp [aut],
  Cole Arendt [aut],
  Posit, PBC [cph, fnd]
Maintainer: Kara Woo <kara.woo@posit.co>
Repository: CRAN
Date/Publication: 2026-04-29 20:00:03 UTC
Built: R 4.6.0; ; 2026-04-29 20:51:59 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",connect_tag_tree)
S3method("[",connect_tag_tree)
S3method("[[",connect_tag_tree)
S3method(api_build,op_base_connect)
S3method(api_build,op_head)
S3method(as.data.frame,connect_content_list)
S3method(as.data.frame,connect_integration_list)
S3method(as.data.frame,connect_list_hits)
S3method(as.data.frame,tbl_connect)
S3method(as_tibble,connect_content_list)
S3method(as_tibble,connect_integration_list)
S3method(as_tibble,connect_list_hits)
S3method(connect_vars,op_base)
S3method(connect_vars,op_single)
S3method(connect_vars,tbl_connect)
S3method(dim,tbl_connect)
S3method(dimnames,tbl_connect)
S3method(get_integrations,Connect)
S3method(get_integrations,Content)
S3method(get_integrations,default)
S3method(head,tbl_connect)
S3method(print,connect_integration)
S3method(print,connect_tag_tree)
S3method(print,tbl_connect)
export("%>%")
export(Bundle)
export(Connect)
export(Content)
export(ContentTask)
export(Environment)
export(Task)
export(Vanity)
export(Variant)
export(VariantSchedule)
export(audit_access_open)
export(audit_r_versions)
export(audit_runas)
export(browse_api_docs)
export(browse_connect)
export(browse_dashboard)
export(browse_solo)
export(bundle_dir)
export(bundle_path)
export(bundle_static)
export(connect)
export(content_add_group)
export(content_add_user)
export(content_delete)
export(content_delete_group)
export(content_delete_user)
export(content_item)
export(content_list_by_tag)
export(content_list_guid_has_access)
export(content_list_with_permissions)
export(content_render)
export(content_restart)
export(content_title)
export(content_update)
export(content_update_access_type)
export(content_update_owner)
export(create_integration)
export(create_random_name)
export(create_tag)
export(create_tag_tree)
export(dashboard_url)
export(delete_bundle)
export(delete_integration)
export(delete_runtime_cache)
export(delete_tag)
export(delete_thumbnail)
export(delete_vanity_url)
export(deploy)
export(deploy_current)
export(deploy_repo)
export(deploy_repo_enable)
export(deploy_repo_update)
export(download_bundle)
export(filter_tag_tree_chr)
export(filter_tag_tree_id)
export(get_associations)
export(get_audit_logs)
export(get_aws_content_credentials)
export(get_aws_credentials)
export(get_bundles)
export(get_content)
export(get_content_packages)
export(get_content_permissions)
export(get_content_tags)
export(get_environment)
export(get_group_content)
export(get_group_members)
export(get_group_permission)
export(get_groups)
export(get_integration)
export(get_integrations)
export(get_job_list)
export(get_jobs)
export(get_log)
export(get_my_permission)
export(get_oauth_content_credentials)
export(get_oauth_credentials)
export(get_packages)
export(get_procs)
export(get_runtime_caches)
export(get_runtimes)
export(get_tag_data)
export(get_tags)
export(get_thumbnail)
export(get_timezones)
export(get_usage)
export(get_usage_shiny)
export(get_usage_static)
export(get_user_permission)
export(get_users)
export(get_vanity_url)
export(get_vanity_urls)
export(get_variant)
export(get_variant_default)
export(get_variant_renderings)
export(get_variant_schedule)
export(get_variants)
export(groups_create_remote)
export(has_thumbnail)
export(lock_content)
export(page_cursor)
export(page_offset)
export(poll_task)
export(promote)
export(repo_check_account)
export(repo_check_branches)
export(repo_check_branches_ref)
export(repo_check_manifest_dirs)
export(schedule_describe)
export(search_content)
export(set_content_tag_tree)
export(set_content_tags)
export(set_environment_all)
export(set_environment_new)
export(set_environment_remove)
export(set_integrations)
export(set_run_as)
export(set_schedule)
export(set_schedule_day)
export(set_schedule_dayofmonth)
export(set_schedule_dayofweek)
export(set_schedule_dayweekofmonth)
export(set_schedule_hour)
export(set_schedule_minute)
export(set_schedule_remove)
export(set_schedule_semimonth)
export(set_schedule_week)
export(set_schedule_weekday)
export(set_schedule_year)
export(set_thumbnail)
export(set_vanity_url)
export(swap_vanity_urls)
export(tbl_connect)
export(terminate_jobs)
export(unlock_content)
export(update_integration)
export(user_guid_from_username)
export(users_create_remote)
export(vanity_is_available)
export(variant_render)
export(verify_content_name)
importFrom(lifecycle,deprecate_warn)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(rlang,"%||%")
importFrom(rlang,":=")
importFrom(rlang,arg_match)
importFrom(rlang,is_string)
importFrom(tibble,as_tibble)
importFrom(utils,browseURL)
importFrom(utils,capture.output)
importFrom(utils,compareVersion)
importFrom(utils,head)
importFrom(utils,untar)
importFrom(uuid,UUIDgenerate)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![Lifecycle:
maturing](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#maturing)
[![CRAN
status](https://www.r-pkg.org/badges/version/connectapi)](https://cran.r-project.org/package=connectapi)
[![Codecov test
coverage](https://codecov.io/gh/posit-dev/connectapi/branch/main/graph/badge.svg)](https://app.codecov.io/gh/posit-dev/connectapi?branch=main)
[![R-CMD-check](https://github.com/posit-dev/connectapi/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/posit-dev/connectapi/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

# connectapi <img src='man/figures/logo.svg' align="right" height="139" />

This package provides an R client for the [Posit Connect Server
API](https://docs.posit.co/connect/api/), as well as helpful functions
that utilize the client. It is designed to work with all supported
versions of Connect, though some features may only be available to newer
versions.

## Installation

To install from CRAN:

``` r
install.packages("connectapi")
```

To install the development version:

``` r
remotes::install_github('posit-dev/connectapi')
```

## Client

To create a client:

``` r
library(connectapi)
client <- connect(
  server = "https://connect.example.com",
  api_key = "<SUPER SECRET API KEY>"
)
```

You can also define the following environment variables (in a
`.Renviron` file, for instance):

    CONNECT_SERVER  = https://connect.example.com
    CONNECT_API_KEY = my-secret-api-key

These environment variable values will be used automatically if defined
in your R session.

``` r
library(connectapi)
client <- connect()
```

## Getting Started

Once a client is defined, you can use it to interact with Posit Connect.

### Exploring Data

You can use the `get_` methods to retrieve data from the Posit Connect
server.

``` r
library(connectapi)
client <- connect()

# get data
users <- get_users(client)
groups <- get_groups(client)
usage_shiny <- get_usage_shiny(client)
usage_static <- get_usage_static(client)
some_content <- get_content(client)

# get all content
all_content <- get_content(client, limit = Inf)
```

### Deployment

The `rsconnect` package is usually used for deploying content to
Connect. However, if you want to use programmatic deployment with the
Posit Connect Server API, then these `connectapi` helpers should be
useful!

``` r
library(connectapi)
client <- connect()

# deploying content
# NOTE: a `manifest.json` should already exist from `rsconnect::writeManifest()`

bundle <- bundle_dir("./path/to/directory")
content <- client %>%
  deploy(bundle, name = "my-app-name") %>%
  poll_task()

# set an image for content

content %>%
  set_thumbnail("./my/local/image.png")

content %>%
  set_thumbnail("http://url.example.com/image.png")

# set image and a vanity URL

content %>%
  set_thumbnail("./my/local/image.png") %>%
  set_vanity_url("/my-awesome-app")

# change access_type to "anyone"

content$update(access_type = "all")

# edit another piece of content

client %>%
  content_item("the-content-guid") %>%
  set_vanity_url("/another-awesome-app")

# migrate content to another server

client_prod <- connect(
  server = "prod.example.com",
  api_key = "my-secret-key"
)

prod_bnd <- client %>%
  content_item("the-guid-to-promote") %>%
  download_bundle()

client_prod %>%
  deploy(prod_bnd, title = "Now in Production") %>%
  set_vanity_url("/my-app")

# open a browser to the content item
client_prod %>% browse_dashboard()
client_prod %>% browse_solo()

# open a browser to the docs
client_prod %>% browse_api_docs()
```

## Troubleshooting and FAQ

**Access Denied Errors?**

This is likely due to either (1) `Connect$server` or `Connect$api_key`
being defined improperly or (2) you do not have access to the Posit
Connect cluster to do the operation in question

**Warning about version numbers**

We test the package against a range of versions of Connect, as old as
`2022.10.0`. If your Connect server is older than that, the package may
still work.

We strive to:

- track the latest version of the Posit Connect API
- add new features as they come available and have demand
- maintain backwards compatibility

These priorities are sometimes at odds, and sometimes they create
inconsistencies between versions as a result. To mitigate this, we
recommend:

- Track the version of `connectapi` in use for your applications by
  using `renv`
- Test high value content that uses `connectapi` before updating
  `connectapi` or Posit Connect
- Update Posit Connect to the latest version *first* when an update to
  `connectapi` is needed

**Error - Need to update Posit Connect**

As a helpful clarification for users, we have added error messages to
API requests when the version implemented in the package specifically
introduces a backwards incompatible dependency on older versions of
Posit Connect.

If you get this error message, our recommendation would be:

- Look at [`NEWS.md`](./NEWS.md) to find the moment the change was
  introduced
- Downgrade `connectapi` to the previous version of the package
- (Advanced) Use the “blame” feature on GitHub to track commits and find
  out when the error was introduced

Please feel free to open an Issue if you think there is a bug, or ask a
free-form question on [Posit
Community](https://forum.posit.co/c/posit-professional-hosted/posit-connect/27)

**Other ideas for FAQs or Common Issues?**

Please open an issue or PR! We would love to have your contribution!

## Code of Conduct

Please note that the connectapi project is released with a [Contributor
Code of
Conduct](https://posit-dev.github.io/connectapi/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
#  library(httr)
#  library(connectapi)
#  
#  client <- connect()
#  
#  # notice that TLS verification fails
#  get_users(client)
#  
#  # use a custom Certificate Authority to verify SSL/TLS requests
#  httr::set_config(httr::config(cainfo = "/path/to/my.pem"))
#  
#  # now it should succeed!
#  get_users(client)

## -----------------------------------------------------------------------------
#  # disabling certificate trust (can allow man-in-the-middle attacks, etc.)
#  httr::set_config(httr::config(ssl_verifypeer = 0, ssl_verifyhost = 0))
#  
#  # should work
#  client <- connect()
#  get_users(client)

## -----------------------------------------------------------------------------
#  httr::with_config(
#    httr::config(ssl_verifypeer = 0, ssl_verifyhost = 0),
#    {
#      client <- connect()
#      get_users(client)
#    }
#  )

## -----------------------------------------------------------------------------
#  # for instance, to set custom headers (i.e. to get through a proxy)
#  client$httr_config(httr::add_headers(MY_MAGIC_HEADER="value"))
#  
#  # or to clear sticky cookies if you want to switch nodes in an HA cluster
#  client <- connect()
#  client$server_settings()$hostname
#  client$httr_config(handle = httr::handle(''))
#  
#  # now you have a chance to get a new host
#  client$server_settings()$hostname
#  
#  # use an outbound proxy
#  client$httr_config(httr::use_proxy("http://myproxy.example.com"))

## -----------------------------------------------------------------------------
#  # disables authentication header that is included by default
#  client$using_auth = FALSE
#  
#  # use Kerberos authentication mechanism (requires local credential cache)
#  client$httr_config(httr::authenticate(":", "", type="gssnegotiate"))

#' Get OAuth integrations
#'
#' @description
#' Retrieve OAuth integrations either from the Connect server or associated with a specific content item.
#'
#' If `x` is a `Connect` object, this function lists all OAuth integrations on the server.
#' If `x` is a `Content` object, it returns the integrations associated with that content item.
#'
#' You must have administrator or publisher privileges to use this function.
#'
#' @param x A `Connect` or `Content` R6 object.
#'
#' @return A list of class `connect_integration_list`, where each element is a `connect_integration` object
#'   with the following fields. (Raw API fields are character strings unless noted otherwise):
#'
#'   * `id`: The internal identifier of this OAuth integration.
#'   * `guid`: The GUID of this OAuth integration.
#'   * `created_time`: Timestamp (RFC3339) when the integration was created.
#'   * `updated_time`: Timestamp (RFC3339) when the integration was last updated.
#'   * `name`: A descriptive name.
#'   * `description`: A brief description.
#'   * `template`: The template used to configure the integration.
#'   * `auth_type`: The OAuth flow used.
#'   * `config`: A list with integration-specific config fields.
#'
#' Use [as.data.frame()] or [tibble::as_tibble()] to convert the result to a data frame with parsed types.
#'
#' @seealso
#' [get_integration()], [set_integrations()], [get_associations()]
#'
#' @family oauth integration functions
#'
#' @examples
#' \dontrun{
#' # From a Connect client
#' client <- connect()
#' integrations <- get_integrations(client)
#'
#' # Filter or update specific ones
#' github_integration <- purrr::keep(integrations, \(x) x$template == "github")[[1]]
#'
#' json_payload <- jsonlite::toJSON(list(
#'   description = "Updated Description",
#'   config = list(client_secret = "new-secret")
#' ), auto_unbox = TRUE)
#'
#' client$PATCH(
#'   paste0("v1/oauth/integrations/", github_integration$guid),
#'   body = json_payload
#' )
#'
#' # From a Content item
#' content <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")
#' content_integrations <- get_integrations(content)
#'
#' # Filter content integrations
#' snowflake_integrations <- purrr::keep(content_integrations, ~ .x$template == "snowflake")
#' }
#'
#' @export
get_integrations <- function(x) {
  UseMethod("get_integrations")
}

#' @export
get_integrations.default <- function(x) {
  stop(
    "Cannot get integrations for an object of class '",
    class(x)[1],
    "'. 'x' must be a 'Connect' or 'Content' object."
  )
}

#' @export
get_integrations.Connect <- function(x) {
  error_if_less_than(x$version, "2024.12.0")
  integrations <- x$GET(v1_url("oauth", "integrations"))
  integrations <- purrr::map(integrations, ~ as_integration(.x, client = x))
  class(integrations) <- c("connect_integration_list", class(integrations))
  integrations
}

#' @export
get_integrations.Content <- function(x) {
  error_if_less_than(x$connect$version, "2024.12.0")
  assoc <- x$connect$GET(v1_url(
    "content",
    x$content$guid,
    "oauth",
    "integrations",
    "associations"
  ))
  integrations <- purrr::map(
    assoc,
    ~ get_integration(x$connect, .x$oauth_integration_guid)
  )
  class(integrations) <- c("connect_integration_list", class(integrations))
  integrations
}

#' Convert integrations list to a data frame
#'
#' @description
#' Converts a list returned by [get_integrations()] into a data frame.
#'
#' @param x A `connect_integration_list` object (from [get_integrations()]).
#' @param row.names Passed to [base::as.data.frame()].
#' @param optional Passed to [base::as.data.frame()].
#' @param ... Passed to [base::as.data.frame()].
#'
#' @return A `data.frame` with one row per integration.
#' @export
as.data.frame.connect_integration_list <- function(
  x,
  row.names = NULL, # nolint
  optional = FALSE,
  ...
) {
  integrations_tbl <- as_tibble(x)
  as.data.frame(
    integrations_tbl,
    row.names = row.names,
    optional = optional,
    ...
  )
}

#' Convert integration list to a tibble
#'
#' @description
#' Converts a list returned by [get_integrations()] to a tibble.
#'
#' @param x A `connect_integration_list` object.
#' @param ... Unused.
#'
#' @return A tibble with one row per integration.
#' @export
as_tibble.connect_integration_list <- function(x, ...) {
  parse_connectapi_typed(x, connectapi_ptypes$integrations)
}

# Integration class ----

#' Convert objects to integration class
#'
#' @param x An object to convert to an integration.
#' @param client The Connect client object where the integration comes from.
#'
#' @return An integration object. The object has all the fields from the
#' integrations endpoint (see [get_integrations()]) and a Connect client as a
#' `client` attribute (`attr(x, "client")`)
as_integration <- function(x, client) {
  if (!inherits(x, "list")) {
    stop(
      "Cannot convert object of class '",
      class(x)[1],
      "' to an integration"
    )
  }
  structure(x, class = c("connect_integration", "list"), client = client)
}

#' @export
print.connect_integration <- function(x, ...) {
  cat("Integration:", x$name, "\n")
  cat("GUID:", x$guid, "\n")
  cat("Template:", x$template, "\n")
  invisible(x)
}

#' Get the details of an OAuth integration
#'
#' @description
#' Given the GUID of an OAuth integration available on a Connect server, retrieve
#' its details. You must have administrator or publisher privileges to perform
#' this action.
#'
#' @param client A `Connect` R6 client object.
#' @param guid The GUID of an integration available on the Connect server.
#'
#' @return A `connect_integration` object representing an OAuth integration,
#' which has the following fields:
#'
#'   * `id`: The internal identifier of this OAuth integration.
#'   * `guid`: The GUID of this OAuth integration.
#'   * `created_time`: The timestamp (RFC3339) indicating when this integration
#'     was created.
#'   * `updated_time`: The timestamp (RFC3339) indicating when this integration
#'     was last updated
#'   * `name`: A descriptive name to identify the OAuth integration.
#'   * `description`: A brief text to describe the OAuth integration.
#'   * `template`: The template used to configure this OAuth integration.
#'   * `auth_type`: The authentication type indicates which OAuth flow is used by
#'     this integration.
#'   * `config`: A list with the OAuth integration configuration. Fields
#'     differ between integrations.
#'
#' @seealso [get_integrations()], [get_associations()], [set_integrations()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' x <- get_integration(client, guid)
#' }
#'
#' @family oauth integration functions
#' @export
get_integration <- function(client, guid) {
  validate_R6_class(client, "Connect")
  error_if_less_than(client$version, "2024.12.0")
  as_integration(client$GET(v1_url("oauth", "integrations", guid)), client)
}

# Get and set integrations on content ----

#' Set all OAuth integrations for a content item
#'
#' @description
#' Removes all existing OAuth integrations associated with a content item, and
#' creates associations with the integrations provided. You must have
#' administrator or publisher privileges to perform this action.
#'
#' @param content A `Content` R6 object representing the content item to modify.
#' @param integrations The complete set of integrations to be associated with the
#'   content. May be a single `connect_integration` object, a list of
#'   `connect_integration` objects, or `NULL`. Passing in an empty list or
#'   explicitly passing `NULL` will remove all associated integrations from the
#'   content.
#'
#' @return Invisibly returns `NULL`.
#'
#' @seealso
#' [get_integrations()], [get_integration()], [get_associations()], [content_item()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' content <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")
#'
#' integrations <- get_integrations(client)
#'
#' # Associate a single integration
#' github_integration <- purrr::keep(integrations, \(x) x$template == "github")[[1]]
#' set_integrations(content, github_integration)
#'
#' # Associate multiple integrations at once
#' selected_integrations <- integrations[1:2]
#' set_integrations(content, selected_integrations)
#'
#' # Unset integrations
#' set_integrations(content, NULL)
#' }
#'
#' @family oauth integration functions
#' @family content functions
#' @export
set_integrations <- function(content, integrations) {
  validate_R6_class(content, "Content")
  # Handle a single integration
  if (inherits(integrations, "connect_integration")) {
    integrations <- list(integrations)
  } else if (!is.null(integrations) && !inherits(integrations, "list")) {
    stop(
      "'integrations' must be a 'connect_integration' class object, a list, ",
      "or NULL."
    )
  }
  # Ensure that all the items we've been passed are integrations
  if (!purrr::every(integrations, ~ inherits(.x, "connect_integration"))) {
    stop("All items must be 'connect_integration' objects")
  }

  payload <- purrr::map(integrations, ~ list(oauth_integration_guid = .x$guid))

  content$connect$PUT(
    v1_url(
      "content",
      content$content$guid,
      "oauth",
      "integrations",
      "associations"
    ),
    body = payload
  )
  invisible(NULL)
}

#' Get OAuth associations for a piece of content
#'
#' @description
#' Given a `Content` object, retrieves a list of its
#' OAuth associations. An association contains a content GUID and an association
#' GUID, and indicates that the integration can be used by the content when it
#' runs.
#'
#' @param x A `Content` object
#'
#' @return A list of OAuth integration associations. Each association includes details such as:
#' * `app_guid`: The content item's GUID (deprecated, use `content_guid` instead).
#' * `content_guid`: The content item's GUID.
#' * `oauth_integration_guid`: The GUID of the OAuth integration.
#' * `oauth_integration_name`: The name of the OAuth integration.
#' * `oauth_integration_description`: A description of the OAuth integration.
#' * `oauth_integration_template`: The template used for this OAuth integration.
#' * `oauth_integration_auth_type`: The authentication type (e.g., "Viewer" or "Service Account").
#' * `created_time`: The timestamp when the association was created.
#'
#' @seealso
#' [set_integrations()], [get_integrations()], [get_integration()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Get OAuth associations for an app.
#' my_app <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")
#' my_app_associations <- get_associations(my_app)
#'
#' # Given those associations, retrieve the integrations themselves.
#' my_app_integrations <- purrr::map(
#'   my_app_associations,
#'   ~ get_integration(client, .x$oauth_integration_guid)
#' )
#' }
#'
#' @family oauth integration functions
#' @family content functions
#' @export
get_associations <- function(x) {
  validate_R6_class(x, "Content")
  x$connect$GET(v1_url(
    "content",
    x$content$guid,
    "oauth",
    "integrations",
    "associations"
  ))
}


# Manage integrations ----

#' Create an OAuth integration
#'
#' @description
#' Creates a new OAuth integration on the Posit Connect server. OAuth integrations
#' allow content to access external resources using OAuth credentials.
#'
#' You must have administrator privileges to perform this action.
#'
#' See the Posit Connect documentation on
#' [OAuth integrations](https://docs.posit.co/connect/admin/integrations/oauth-integrations/) for
#' more information.
#'
#' @param client A `Connect` R6 client object.
#' @param name A descriptive name to identify the integration.
#' @param description Optional, default `NULL.` A brief description of the integration.
#' @param template The template to use to configure this integration (e.g.,
#'   "custom", "github", "google", "connect").
#' @param config A list containing the configuration for the integration. The
#'   required fields vary depending on the template selected.
#'
#' @return A `connect_integration` object representing the newly created
#'   integration. See [get_integration()] for details on the returned object.
#'
#' @seealso [get_integrations()], [get_integration()], [update_integration()],
#'   [delete_integration()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Create a GitHub OAuth integration
#' github_integration <- create_integration(
#'   client,
#'   name = "GitHub Integration",
#'   description = "Integration with GitHub for OAuth access",
#'   template = "github",
#'   config = list(
#'     client_id = "your-client-id",
#'     client_secret = "your-client-secret"
#'   )
#' )
#'
#' # Create a custom OAuth integration
#' custom_integration <- create_integration(
#'   client,
#'   name = "Custom API Integration",
#'   description = "Integration with our custom API",
#'   template = "custom",
#'   config = list(
#'     auth_mode = "Confidential",
#'     auth_type = "Viewer",
#'     authorization_uri = "https://api.example.com/oauth/authorize",
#'     client_id = "your-client-id",
#'     client_secret = "your-client-secret",
#'     token_uri = "https://api.example.com/oauth/token"
#'   )
#' )
#' }
#'
#' @family oauth integration functions
#' @export
create_integration <- function(
  client,
  name,
  description = NULL,
  template,
  config
) {
  validate_R6_class(client, "Connect")
  error_if_less_than(client$version, "2024.12.0")
  result <- client$POST(
    v1_url("oauth", "integrations"),
    body = list(
      name = name,
      description = description,
      template = template,
      config = config
    )
  )
  as_integration(result, client)
}

#' Update an OAuth integration
#'
#' @description
#' Updates an existing OAuth integration. All fields except `integration` are optional,
#' and are unchanged if not provided.
#'
#' You must have administrator privileges to perform this action.
#'
#' See the Posit Connect documentation on
#' [OAuth integrations](https://docs.posit.co/connect/admin/integrations/oauth-integrations/) for
#' more information.
#'
#' @param integration A `connect_integration` object (as returned by [get_integrations()],
#'   [get_integration()], or [create_integration()]).
#' @param name A new name for the integration.
#' @param description A new description for the integration.
#' @param template The template to use (generally not changed after creation).
#' @param config A list with updated OAuth integration configuration. If `NULL`
#'   (default), the configuration remains unchanged. You can update individual
#'   configuration fields without affecting others.
#'
#' @return A `connect_integration` object representing the updated OAuth
#'   integration. See [get_integration()] for details on the returned object.
#'
#' @seealso [get_integrations()], [get_integration()], [create_integration()],
#'   [delete_integration()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Get an existing integration
#' integration <- get_integration(client, "your-integration-guid")
#'
#' # Update the integration's name and description
#' updated_integration <- update_integration(
#'   integration,
#'   name = "Updated GitHub Integration",
#'   description = "A more descriptive description."
#' )
#'
#' # Update only the client secret in the configuration
#' updated_integration <- update_integration(
#'   integration,
#'   config = list(
#'     client_secret = "your-new-client-secret"
#'   )
#' )
#' }
#'
#' @family oauth integration functions
#' @export
update_integration <- function(
  integration,
  name = NULL,
  description = NULL,
  template = NULL,
  config = NULL
) {
  if (!inherits(integration, "connect_integration")) {
    stop("'integration' must be a 'connect_integration' object")
  }
  client <- attr(integration, "client")
  validate_R6_class(client, "Connect")
  error_if_less_than(client$version, "2024.12.0")
  result <- client$PATCH(
    v1_url("oauth", "integrations", integration$guid),
    body = list(
      name = name,
      description = description,
      template = template,
      config = config
    )
  )
  as_integration(result, client)
}

#' Delete an OAuth integration
#'
#' @description
#' Deletes an OAuth integration from the Posit Connect server. This permanently
#' removes the integration and any associated content associations.
#'
#' You must have administrator privileges to perform this action.
#'
#' See the Posit Connect documentation on
#' [OAuth integrations](https://docs.posit.co/connect/admin/integrations/oauth-integrations/) for
#' more information.
#'
#' @param integration A `connect_integration` object (as returned by [get_integrations()],
#'   [get_integration()], or [create_integration()]).
#'
#' @return Returns `NULL` invisibly if successful.
#'
#' @seealso [get_integrations()], [get_integration()], [create_integration()],
#'   [update_integration()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Get an integration to delete
#' integration <- get_integration(client, "your-integration-guid")
#'
#' # Delete the integration
#' delete_integration(integration)
#' }
#'
#' @family oauth integration functions
#' @export
delete_integration <- function(integration) {
  if (!inherits(integration, "connect_integration")) {
    stop("'integration' must be a 'connect_integration' object")
  }
  client <- attr(integration, "client")
  validate_R6_class(client, "Connect")
  error_if_less_than(client$version, "2024.12.0")
  client$DELETE(v1_url("oauth", "integrations", integration$guid))
  invisible(NULL)
}

with_mock_dir("2024.12.0", {
  test_that("get_integrations() gets integrations", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    integrations <- get_integrations(client)
    expect_s3_class(integrations, "connect_integration_list")
    expect_equal(integrations[[1]]$name, "GitHub Integration")
    expect_equal(integrations[[2]]$updated_time, "2025-03-25T19:07:01Z")
    expect_equal(integrations[[1]]$config$client_id, "client_id_123")
    expect_s3_class(integrations[[1]], "connect_integration")
  })

  test_that("get_integrations() can be converted to a data frame correctly", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    integrations_df <- get_integrations(client) |>
      as_tibble()
    expect_named(
      integrations_df,
      c(
        "id",
        "guid",
        "created_time",
        "updated_time",
        "name",
        "description",
        "template",
        "auth_type",
        "config"
      )
    )
    expect_equal(
      integrations_df$description,
      c(
        "with refresh support ",
        "The service provides utility to your company"
      )
    )
  })
})

test_that("get_integrations() errs on older Connect versions", {
  client <- MockConnect$new("2024.11.1")
  client$version
  expect_error(
    get_integrations(client),
    "This feature requires Posit Connect version 2024.12.0 but you are using 2024.11.1"
  )
})

test_that("as_integration correctly converts lists to integration objects", {
  client <- MockConnect$new("2024.11.1")
  valid_integration <- list(
    id = "123",
    guid = "abc-123",
    created_time = Sys.time(),
    updated_time = Sys.time(),
    name = "Test Integration",
    description = "A test integration",
    template = "github",
    auth_type = "oauth2",
    config = list(client_id = "client_id")
  )

  result <- as_integration(valid_integration, client)
  expect_s3_class(result, "connect_integration")
  expect_identical(result$guid, valid_integration$guid)
  expect_identical(attr(result, "client"), client)
})

test_that("as_integration.default errors on non-list input", {
  expect_error(as_integration(42), "Cannot convert object of class")
  expect_error(
    as_integration("string"),
    "Cannot convert object of class"
  )
})

test_that("print.integration produces expected output", {
  test_int <- structure(
    list(
      name = "Test Integration",
      guid = "abc-123",
      template = "github"
    ),
    class = c("connect_integration", "list")
  )

  output <- capture.output(result <- print(test_int))

  expect_match(output[1], "Integration: Test Integration")
  expect_match(output[2], "GUID: abc-123")
  expect_match(output[3], "Template: github")
  expect_identical(result, test_int)
})

with_mock_dir("2024.12.0", {
  test_that("get_integration() gets a single integration", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    x <- get_integration(client, "f8688548")
    expect_s3_class(x, "connect_integration")
    expect_equal(x$template, "custom")
    expect_equal(x$guid, "f8688548")
  })
})

test_that("get_integration() errs with old Connect", {
  client <- MockConnect$new("2024.11.1")
  expect_error(
    get_integration(client, "12345678"),
    "This feature requires Posit Connect version 2024.12.0 but you are using 2024.11.1"
  )
})

test_that("set_integrations() sends expected request", {
  with_mock_dir("2024.12.0", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    x <- content_item(client, "12345678")
    y <- get_integration(client, "f8688548")
  })
  without_internet(
    expect_PUT(
      set_integrations(x, y),
      url = "https://connect.example/__api__/v1/content/12345678/oauth/integrations/associations",
      '[{"oauth_integration_guid":"f8688548"}]'
    )
  )
})

test_that("set_integrations() fails when provided the wrong class", {
  with_mock_dir("2024.12.0", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    x <- content_item(client, "12345678")
  })
  expect_error(
    set_integrations(x, "string"),
    "'integrations' must be a 'connect_integration' class object, a list, or NULL"
  )
  expect_error(
    set_integrations(x, list("string")),
    "All items must be 'connect_integration' objects"
  )
})

with_mock_dir("2025.07.0", {
  test_that("get_integrations() works with Content objects", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- content_item(client, "12345678")
    integrations <- get_integrations(content)

    expect_s3_class(integrations, "connect_integration_list")
    expect_equal(length(integrations), 2)
    expect_equal(integrations[[1]]$name, "Integration 1")
    expect_equal(integrations[[2]]$template, "template2")
    expect_s3_class(integrations[[1]], "connect_integration")
  })

  test_that("get_associations() returns association metadata", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- content_item(client, "12345678")
    associations <- get_associations(content)

    expect_type(associations, "list")
    expect_equal(length(associations), 2)
    expect_equal(associations[[1]]$oauth_integration_guid, "0000001")
    expect_equal(associations[[1]]$oauth_integration_name, "Integration 1")
    expect_equal(associations[[1]]$oauth_integration_template, "template1")
    expect_equal(associations[[2]]$oauth_integration_template, "template2")
    expect_true(!is.null(associations[[1]]$created_time))
  })
})


test_that("get_integrations() with Content errs on older Connect versions", {
  client <- MockConnect$new("2024.11.1")
  content <- Content$new(
    connect = client,
    content = list(guid = "12345678")
  )
  expect_error(
    get_integrations(content),
    "This feature requires Posit Connect version 2024.12.0 but you are using 2024.11.1"
  )
})

test_that("get_integrations() fails when provided the wrong class", {
  expect_error(
    get_integrations("string"),
    "Cannot get integrations for an object of class 'character'"
  )
  expect_error(
    get_integrations(list()),
    "Cannot get integrations for an object of class 'list'"
  )
})

with_mock_dir("2025.07.0", {
  client <- Connect$new(server = "https://connect.example", api_key = "fake")

  test_that("create_integration() with bad data returns an error", {
    expect_error(
      create_integration(
        client,
        name = "Connect API Integration",
        description = "Authenticate against the Connect API, but only as a publisher",
        template = "connect",
        config = list(
          max_role = "Not a Role"
        )
      ),
      "The config key max_role must be one of \\(Viewer, Publisher, Admin\\)"
    )
  })

  test_that("create_integration() with good data creates an integration", {
    created <- create_integration(
      client,
      name = "Connect API Integration",
      description = "Authenticate against the Connect API, but only as a publisher",
      template = "connect",
      config = list(
        max_role = "Publisher"
      )
    )
    expect_s3_class(created, "connect_integration")
    expect_equal(created$guid, "60586f1c")
    expect_equal(created$name, "Connect API Integration")
    expect_equal(created$config$max_role, "Publisher")
  })

  test_that("update_integration() with bad data returns an error", {
    created <- create_integration(
      client,
      name = "Connect API Integration",
      description = "Authenticate against the Connect API, but only as a publisher",
      template = "connect",
      config = list(
        max_role = "Publisher"
      )
    )

    expect_error(
      update_integration(
        created,
        config = list(
          max_role = "Not a Role"
        )
      ),
      "The config key max_role must be one of \\(Viewer, Publisher, Admin\\)"
    )
  })

  test_that("update_integration() with good data returns an integration object", {
    created <- create_integration(
      client,
      name = "Connect API Integration",
      description = "Authenticate against the Connect API, but only as a publisher",
      template = "connect",
      config = list(
        max_role = "Publisher"
      )
    )
    updated <- update_integration(
      created,
      description = "Improved description too, and now Viewer role",
      config = list(max_role = "Viewer")
    )
    expect_s3_class(created, "connect_integration")
    expect_equal(updated$guid, "60586f1c")
    expect_equal(
      updated$description,
      "Improved description too, and now Viewer role"
    )
    expect_equal(updated$config$max_role, "Viewer")
  })

  test_that("delete_integration() returns NULL on success", {
    created <- create_integration(
      client,
      name = "Connect API Integration",
      description = "Authenticate against the Connect API, but only as a publisher",
      template = "connect",
      config = list(
        max_role = "Publisher"
      )
    )
    expect_null(delete_integration(created))
  })
})

#' Variant
#'
#' An R6 class that represents a Variant
#'
#' @rdname VariantR6
#' @family R6 classes
#' @export
Variant <- R6::R6Class(
  "Variant",
  inherit = Content,
  public = list(
    #' @field key The variant key.
    key = NULL,
    #' @field variant The variant.
    variant = NULL,
    #' @description Get the underlying variant data.
    #' @description Get and store the (remote) variant data.
    get_variant_remote = function() {
      path <- unversioned_url("variants", self$variant$id)
      self$variant <- self$connect$GET(path)
      self$variant
    },
    #' @description Initialize this variant.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    #' @param key The variant key.
    initialize = function(connect, content, key) {
      super$initialize(connect = connect, content = content)
      # TODO: a better way to GET self
      all_variants <- self$variants()
      if (identical(key, "default")) {
        self$variant <- purrr::keep(all_variants, ~ .x[["is_default"]])[[1]]
      } else {
        self$variant <- purrr::keep(all_variants, ~ .x$key == key)[[1]]
      }
      self$key <- self$variant$key
    },
    #' @description Mail previously rendered content.
    #' @param to Targeting.
    send_mail = function(
      to = c("me", "collaborators", "collaborators_viewers")
    ) {
      warn_experimental("send_mail")
      url <- unversioned_url("variants", self$variant$id, "sender")
      self$connect$POST(
        path = url,
        query = list(
          email = arg_match(to),
          rendering_id = self$variant$rendering_id
        )
      )
    },
    #' @description Get the (remote) schedule data.
    get_schedule = function() {
      self$get_schedule_remote()
    },
    #' @description Get the (remote) schedule data.
    get_schedule_remote = function() {
      warn_experimental("get_schedule_remote")
      url <- unversioned_url("variants", self$variant$id, "schedules")
      res <- self$connect$GET(url)

      if (length(res) == 1) {
        res <- res[[1]]
      }

      if (length(res) > 0) {
        # add the content guid and variant key
        content_guid <- self$content$guid
        variant_key <- self$key
        res <- purrr::list_modify(
          res,
          app_guid = content_guid,
          variant_key = variant_key
        )
      }

      res
    },
    #' @description Get the subscribers.
    get_subscribers = function() {
      warn_experimental("subscribers")
      path <- unversioned_url("variants", self$variant$id, "subscribers")
      self$connect$GET(path)
    },
    #' @description Remove a named subscriber.
    #' @param guid User GUID.
    remove_subscriber = function(guid) {
      warn_experimental("subscribers")
      id <- self$variant$id
      path <- unversioned_url("variants", id, "subscribers", guid)
      self$connect$DELETE(path)
    },
    #' @description Add named subscribers.
    #' @param guids User GUIDs.
    add_subscribers = function(guids) {
      warn_experimental("subscribers")
      path <- unversioned_url("variants", self$variant$id, "subscribers")
      self$connect$POST(path = path, body = guids)
    },
    #' @description Render this variant.
    render = function() {
      warn_experimental("render")
      path <- unversioned_url("variants", self$variant$id, "render")
      res <- self$connect$POST(path)

      # add the content guid and variant key
      content_guid <- self$content$guid
      variant_key <- self$key

      purrr::list_modify(
        res,
        app_guid = content_guid,
        variant_key = variant_key
      )
    },
    #' @description List the renderings of this variant.
    renderings = function() {
      warn_experimental("renderings")
      url <- unversioned_url("variants", self$variant$id, "renderings")
      res <- self$connect$GET(path = url)
      # add the content guid and variant key
      content_guid <- self$content$guid
      variant_key <- self$key

      purrr::map(
        res,
        ~ purrr::list_modify(
          .x,
          app_guid = content_guid,
          variant_key = variant_key
        )
      )
    },
    #' @description Update this variant.
    #' @param ... Target fields and values.
    update_variant = function(...) {
      params <- rlang::list2(...)
      # TODO: allow updating a variant
      url <- unversioned_url("variants", self$variant$id)
      res <- self$connect$POST(url, body = params)
      return(self)
    },
    #' @description Jobs for this variant.
    jobs = function() {
      pre_jobs <- super$jobs()
      purrr::map(
        pre_jobs,
        ~ purrr::list_modify(.x, variant_key = self$key)
      )
    },
    #' @description Return the URL for this variant.
    get_url = function() {
      base_content <- super$get_url()
      glue::glue("{base_content}v{self$key}/")
    },
    #' @description Return the URL associated with one rendering for this variant.
    #' @param rev Rendering identifier.
    get_url_rev = function(rev) {
      base_url <- self$get_url()
      glue::glue("{base_url}_rev{rev}")
    },
    #' @description Return the URL for this variant in the Posit Connect dashboard.
    #' @param pane The pane in the dashboard to link to.
    get_dashboard_url = function(pane = "access") {
      base_content <- super$get_dashboard_url("")
      glue::glue("{base_content}{pane}/{self$variant$id}")
    },
    # nolint start: commented_code_linter
    # TODO: dashboard cannot navigate directly to renderings today
    # get_dashboard_url_rev = function(rev, pane = "") {
    #  base_content <- self$get_dashboard_url("")
    #  glue::glue("{base_content}_rev{rev}")
    # },
    # nolint end
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      super$print(...)
      cat("Variant:\n")
      cat(glue::glue("  get_variant(content, key = '{self$key}' )"), "\n")
      cat("\n")
    }
  )
)

#' VariantTask
#'
#' An R6 class that represents a Variant Task
#'
#' @family R6 classes
VariantTask <- R6::R6Class(
  "VariantTask",
  inherit = Variant,
  public = list(
    #' @field task The task.
    task = NULL,
    #' @field data The variant data.
    data = NULL,
    #' @description Initialize this variant task.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    #' @param key The variant key.
    #' @param task The task data.
    initialize = function(connect, content, key, task) {
      super$initialize(connect = connect, content = content, key = key)
      if (is.null(task$task_id)) {
        task$task_id <- task$id
      }
      self$task <- task
    },
    #' @description Return the underlying task.
    get_task = function() {
      self$task
    },
    #' @description Set the data.
    #' @param data The data.
    add_data = function(data) {
      self$data <- data
      invisible(self)
    },
    #' @description Get the data.
    get_data = function() {
      self$data
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      super$print(...)
      cat("Task: \n")
      cat("  Task ID: ", self$get_task()$task_id, "\n", sep = "")
      cat("\n")
      invisible(self)
    }
  )
)

#' Get Variant
#'
#' `r lifecycle::badge('experimental')` Work with variants
#'
#' - `get_variants()` returns a `tibble` with variant data for a `content_item`
#' - `get_variant_default()` returns the default variant for a `content_item`
#' - `get_variant()` returns a specific variant for a `content_item` (specified by `key`)
#'
#' @param content An R6 Content object. Returned from `content_item()`
#' @param key The Variant key for a specific variant
#'
#' @rdname variant
#'
#' @family variant functions
#' @export
get_variants <- function(content) {
  warn_experimental("get_variants")
  scoped_experimental_silence()
  validate_R6_class(content, "Content")

  variants <- content$variants()

  parse_connectapi_typed(variants, connectapi_ptypes$variant)
}

#' @rdname variant
#' @family variant functions
#' @export
get_variant <- function(content, key) {
  warn_experimental("get_variant")
  scoped_experimental_silence()
  validate_R6_class(content, "Content")
  Variant$new(
    connect = content$connect,
    content = content$content,
    key = key
  )
}

#' @rdname variant
#' @family variant functions
#' @export
get_variant_default <- function(content) {
  get_variant(content, "default")
}

#' Render a Variant
#'
#' `r lifecycle::badge('experimental')` Get details about renderings (i.e. render history)
#' or execute a variant on demand
#'
#' - `get_variant_renderings()` returns all renderings / content for a particular variant. Returns a `tibble`
#' - `variant_render()` executes a variant on demand. Returns a `VariantTask` object
#'
#' @param variant An R6 Variant object. As returned by `get_variant()` or `get_variant_default()`
#'
#' @rdname variant_render
#' @family variant functions
#' @export
get_variant_renderings <- function(variant) {
  warn_experimental("get_variant_renderings")
  scoped_experimental_silence()
  validate_R6_class(variant, "Variant")

  renders <- variant$renderings()
  parse_connectapi_typed(renders, connectapi_ptypes$rendering)
}

#' @rdname variant_render
#' @export
variant_render <- function(variant) {
  warn_experimental("variant_render")
  scoped_experimental_silence()
  validate_R6_class(variant, "Variant")

  rendered <- variant$render()

  VariantTask$new(
    connect = variant$connect,
    content = variant$content,
    key = variant$key,
    task = rendered
  )
}

# TODO
# set_variant_email_viewers <- function() {
#
# }
#
# set_variant_email_collaborators <- function() {
#
# }
#
# set_variant_email_subscribe <- function() {
#
# }
#
# set_variant_email_unsubscribe <- function() {
#
# }

with_mock_api({
  scoped_experimental_silence()

  test_that("Variant requests", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_true(validate_R6_class(content, "Content"))
    variant <- get_variant(content, "12345")
    expect_true(validate_R6_class(variant, "Variant"))

    expect_GET(
      variant$get_variant_remote(),
      "https://connect.example/__api__/variants/12345"
    )

    expect_POST(
      variant$send_mail(),
      "https://connect.example/__api__/variants/12345/sender?email=me&rendering_id=12",
      "{}"
    )

    expect_GET(
      variant$get_schedule(),
      "https://connect.example/__api__/variants/12345/schedules"
    )

    expect_GET(
      variant$get_subscribers(),
      "https://connect.example/__api__/variants/12345/subscribers"
    )
    expect_DELETE(
      variant$remove_subscriber("asdf"),
      "https://connect.example/__api__/variants/12345/subscribers/asdf"
    )

    # TODO: test for add_subscribers, it doesn't look like it could be right
    # (but the API is not documented). Function does not look like it would
    # send valid JSON in the POST body

    expect_POST(
      variant$render(),
      "https://connect.example/__api__/variants/12345/render",
      "{}"
    )

    expect_GET(
      variant$renderings(),
      "https://connect.example/__api__/variants/12345/renderings"
    )

    expect_POST(
      variant$update_variant(key = "value"),
      "https://connect.example/__api__/variants/12345",
      '{"key":"value"}'
    )
  })
})

# Temporarily skip these on newer Connect versions because the manifest for this
# content specifies an old version of R not found in newer images.
skip_if(safe_server_version(client) > "2024.03.0")

# ensure that RSPM is being used so these do not take eternity
rmd_content <- deploy_example(client, "rmarkdown")

# TODO: very hard to test parameterized rmarkdown because creating a
# programmatic variant is not possible

test_that("get_variants works", {
  scoped_experimental_silence()
  vrs <- get_variants(rmd_content)

  expect_equal(nrow(vrs), 1)

  vr <- get_variant(rmd_content, vrs$key)
  expect_true(validate_R6_class(vr, "Variant"))
})

test_that("variant_render works", {
  scoped_experimental_silence()
  vr <- get_variant_default(rmd_content)

  rnd <- variant_render(vr)
  rnd2 <- variant_render(vr)

  expect_true(validate_R6_class(rnd, "VariantTask"))
  # TODO: would be great to be able to "tail the logs", for instance
  # i.e. actually reference the "job" itself...

  # wait for tasks to complete...
  suppressMessages(poll_task(rnd))
  suppressMessages(poll_task(rnd2))
})

test_that("content_render works", {
  rnd <- content_render(rmd_content)

  expect_true(validate_R6_class(rnd, "VariantTask"))

  # wait for tasks to complete...
  suppressMessages(poll_task(rnd))
})

test_that("get_variant_renderings works", {
  scoped_experimental_silence()

  vr <- get_variant_default(rmd_content)

  rnd <- get_variant_renderings(vr)

  expect_gt(nrow(rnd), 1)
})

test_that("get_jobs works", {
  scoped_experimental_silence()
  vr <- get_variant_default(rmd_content)

  all_jobs <- get_jobs(vr)
  expect_gt(nrow(all_jobs), 1)
})

#' Get runtime caches
#'
#' View the runtime caches on a Connect server.
#' Requires Administrator privileges.
#'
#' @param client A `Connect` object.
#'
#' @returns A tibble of runtime caches on the server, showing `language`,
#' `version` and `image_name`. For Connect servers not using off-host execution,
#' `image_name` is `"Local"`.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' get_runtime_caches(client)
#' }
#'
#' @family server management functions
#' @seealso [delete_runtime_cache()]
#' @export
get_runtime_caches <- function(client) {
  validate_R6_class(client, "Connect")
  res <- client$GET(v1_url("system/caches/runtime"))
  purrr::list_rbind(purrr::map(res$caches, ~ tibble::as_tibble(.x)))
}

#' Delete a runtime cache
#'
#' Delete a runtime cache from a Connect server.
#' Requires Administrator privileges.
#'
#' @param client A `Connect` object.
#' @param language The language of the cache, either "R" or "Python".
#' @param version The version of the cache, e.g. "4.3.3".
#' @param image_name Optional. The name of the off-host execution image for
#'   the cache, or "Local" (the default) for native execution caches.
#' @param dry_run Optional, default `FALSE`. If true, perform a dry run of
#'   the deletion.
#'
#' @returns A `Task` object representing the deletion task. If `dry_run` is
#' `TRUE`, returns `NULL` or throws an error if the deletion would fail.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' task <- delete_runtime_cache(client, "R", "4.3.3")
#' poll_task(task)
#' }
#'
#' @family server management functions
#' @seealso [get_runtime_caches()]
#' @export
delete_runtime_cache <- function(
  client,
  language,
  version,
  image_name = "Local",
  dry_run = FALSE
) {
  res <- client$DELETE(
    path = v1_url("system/caches/runtime"),
    body = list(
      language = language,
      version = version,
      image_name = image_name,
      dry_run = dry_run
    ),
    encode = "json",
    parser = "parsed"
  )

  if (dry_run == TRUE) {
    message(glue::glue(
      "Runtime cache deletion dry run did not encounter any errors ",
      '(language = "{language}", version = "{version}", image_name = "{image_name}")'
    ))
    return(invisible(NULL))
  }

  task_data <- client$GET(v1_url("tasks", res$task_id))
  Task$new(client, task_data)
}

with_mock_api({
  client <- connect(server = "https://connect.example", api_key = "fake")

  test_that("get_runtime_caches() gets runtime caches", {
    caches <- get_runtime_caches(client)
    expect_equal(
      get_runtime_caches(client),
      tibble::tibble(
        language = c("R", "Python"),
        version = c("4.3.0", "3.11.3"),
        image_name = c("Local", "Local")
      )
    )
  })

  test_that("delete_runtime_cache() prints message and returns NULL when dry_run == TRUE", {
    expect_message(
      res <- delete_runtime_cache(client, "Python", "3.11.3", dry_run = TRUE),
      paste0(
        "Runtime cache deletion dry run did not encounter any errors ",
        '\\(language = "Python", version = "3.11.3", image_name = "Local"\\)'
      )
    )
    expect_null(res)
  })

  test_that("delete_runtime_cache() returns a task object when dry_run == FALSE", {
    res <- delete_runtime_cache(client, "Python", "3.11.3")
    expect_true(validate_R6_class(res, "Task"))
    expect_equal(res$task$id, "eKm0RFxyzIKR3jnI")
  })

  test_that("delete_runtime_cache() on a nonexistent cache throws a 404 error", {
    expect_error(
      delete_runtime_cache(client, "Python", "3.11.2"),
      "request failed with Client error: \\(404\\) Not Found"
    )
  })
})

structure(
  list(
    url = "__api__/v1/content/8f37d6e0/jobs/waaTO7v75I84S1hQ",
    status_code = 409L,
    headers = structure(
      list(`content-type` = "application/json"),
      class = c("insensitive", "list")
    ),
    content = charToRaw(
      "{\"code\":163,\"error\":\"The specified job cannot be terminated because it is not active\",\"payload\":null}"
    )
  ),
  class = "response"
)

#' Get user information from the Posit Connect server
#'
#' @param src The source object
#' @param page_size the number of records to return per page (max 500)
#' @param prefix Filters users by prefix (username, first name, or last name).
#' The filter is case insensitive.
#' @param limit The max number of records to return
#' @param user_role Optionally filter by user role ("administrator",
#' "publisher", "viewer"). Pass in a vector of multiple roles to match any value
#' (boolean OR). When `NULL` (the default), results are not filtered.
#' @param account_status Optionally filter by account status ("locked",
#' "licensed", "inactive"). Pass a vector of multiple statuses to match any
#' value (boolean OR). When `NULL` (the default), results are not filtered.

#'
#' @return
#' A tibble with the following columns:
#'
#'   * `email`: The user's email
#'   * `username`: The user's username
#'   * `first_name`: The user's first name
#'   * `last_name`: The user's last name
#'   * `user_role`: The user's role. It may have a value of administrator,
#'     publisher or viewer.
#'   * `created_time`: The timestamp (in RFC3339 format) when the user was
#'     created in the Posit Connect server
#'   * `updated_time`: The timestamp (in RFC3339 format) when the user was last
#'     updated in the Posit Connect server
#'   * `active_time`: The timestamp (in RFC3339 format) when the user was last
#'     active on the Posit Connect server
#'   * `confirmed`: When false, the created user must confirm their account
#'     through an email. This feature is unique to password authentication.
#'   * `locked`: Whether or not the user is locked
#'   * `guid`: The user's GUID, or unique identifier, in UUID RFC4122 format
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/users for more information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' # Get all users
#' get_users(client)
#'
#' # Get all licensed users
#' get_users(client, account_status = "licensed")
#'
#' # Get all users who are administrators or publishers
#' get_users(client, user_role = c("administrator", "publisher"))
#' }
#'
#' @export
get_users <- function(
  src,
  page_size = 500,
  prefix = NULL,
  limit = Inf,
  user_role = NULL,
  account_status = NULL
) {
  validate_R6_class(src, "Connect")

  res <- page_offset(
    src,
    src$users(
      page_size = page_size,
      prefix = prefix,
      user_role = user_role,
      account_status = account_status
    ),
    limit = limit
  )

  out <- parse_connectapi_typed(res, connectapi_ptypes$users)

  return(out)
}

#' Get information about content on the Posit Connect server
#'
#' @param src A Connect object
#' @param guid The guid for a particular content item
#' @param owner_guid The unique identifier of the user who owns the content
#' @param name The content name specified when the content was created
#' @param ... Extra arguments. Currently not used.
#' @param .p Optional. A predicate function, passed as-is to `purrr::keep()`
#'   before turning the response into a tibble. Can be useful for performance.
#'
#' @return
#' A tibble with the following columns:
#'   * `guid`: The unique identifier of this content item.
#'   * `name`: A simple, URL-friendly identifier. Allows alpha-numeric
#'     characters, hyphens ("-"), and underscores ("_").
#'   * `title`: The title of this content.
#'   * `description`: A rich description of this content
#'   * `access_type`: Access type describes how this content manages its
#'     viewers. It may have a value of `all`, `logged_in` or `acl`.
#'     The value `all` is the most permissive; any visitor to Posit
#'     Connect will be able to view this content. The value `logged_in`
#'     indicates that all Posit Connect accounts may view the content. The
#'     `acl` value lets specifically enumerated users and groups view the
#'     content. Users configured as collaborators may always view content.
#'   * `connection_timeout`: Maximum number of seconds allowed without data
#'     sent or received across a client connection. A value of 0 means
#'     connections will never time-out (not recommended). When null, the
#'     default `Scheduler.ConnectionTimeout` is used. Applies only to content
#'     types that are executed on demand.
#'   * `read_timeout`: Maximum number of seconds allowed without data received
#'     from a client connection. A value of 0 means a lack of client (browser)
#'     interaction never causes the connection to close. When null, the default
#'     `Scheduler.ReadTimeout` is used. Applies only to content types that are
#'     executed on demand.
#'   * `init_timeout`: The maximum number of seconds allowed for an interactive
#'     application to start. Posit Connect must be able to connect
#'     to a newly launched Shiny application, for example, before this threshold
#'     has elapsed. When null, the default `Scheduler.InitTimeout` is
#'     used. Applies only to content types that are executed on demand.
#'   * `idle_timeout`: The maximum number of seconds a worker process
#'     for an interactive application to remain alive after it goes idle (no
#'     active connections). When null, the default `Scheduler.IdleTimeout`
#'     is used. Applies only to content types that are executed on demand.
#'   * `max_processes`: Specifies the total number of concurrent processes
#'     allowed for a single interactive application. When null, the
#'     default `Scheduler.MaxProcesses` setting is used. Applies only to
#'     content types that are executed on demand.
#'   * `min_processes`: Specifies the minimum number of concurrent
#'     processes allowed for a single interactive application. When null, the
#'     default `Scheduler.MinProcesses` is used. Applies only to content types
#'    that are executed on demand.
#'   * `max_conns_per_process`: Specifies the maximum number of
#'     client connections allowed to an individual process. Incoming connections
#'     which will exceed this limit are routed to a new process or rejected.
#'     When null, the default `Scheduler.MaxConnsPerProcess` is used. Applies
#'     only to content types that are executed on demand.
#'   * `load_factor`: Controls how aggressively new processes are spawned.
#'     When null, the default `Scheduler.LoadFactor` is used. Applies only to
#'     content types that are executed on demand.
#'   * `created_time`: The timestamp (RFC3339) indicating when this
#'     content was created.
#'   * `last_deployed_time`: The timestamp (RFC3339) indicating when
#'     this content last had a successful bundle deployment performed.
#'   * `bundle_id`: The identifier for the active deployment bundle.
#'     Automatically assigned upon the successful deployment of that bundle.
#'   * `app_mode`: The runtime model for this content. Has a value
#'     of `unknown` before data is deployed to this item. Automatically assigned
#'     upon the first successful bundle deployment. Allowed: `api`,
#'     `jupyter-static`, `python-api`, `python-bokeh`, `python-dash`,
#'     `python-streamlit`, `rmd-shiny`, `rmd-static`, `shiny`, `static`,
#'     `tensorflow-saved-model`, `unknown`.
#'   * `content_category`: Describes the specialization of the content
#'     runtime model. Automatically assigned upon the first successful bundle
#'     deployment.
#'   * `parameterized`: True when R Markdown rendered content
#'     allows parameter configuration. Automatically assigned upon the first
#'     successful bundle deployment. Applies only to content with an app_mode
#'     of rmd-static.
#'   * `r_version`: The version of the R interpreter associated
#'     with this content. The value null represents that an R interpreter is
#'     not used by this content or that the R package environment has not been
#'     successfully restored. Automatically assigned upon the successful
#'     deployment of a bundle.
#'   * `py_version`: The version of the Python interpreter
#'     associated with this content. The value null represents that a Python
#'     interpreter is not used by this content or that the Python package
#'     environment has not been successfully restored. Automatically assigned
#'     upon the successful deployment of a bundle.
#'   * `run_as`: The UNIX user that executes this content.
#'     When null, the default Applications.RunAs is used. Applies
#'     only to executable content types - not static.
#'   * `run_as_current_user`: Indicates if this content is allowed
#'     to execute as the logged-in user when using PAM authentication.
#'     Applies only to executable content types - not static.
#'   * `owner_guid`: The unique identifier for the owner
#'   * `content_url`: The URL associated with this content. Computed
#'     from the GUID for this content.
#'   * `dashboard_url`: The URL within the Connect dashboard where
#'     this content can be configured. Computed from the GUID for this content.
#'   * `role`: The relationship of the accessing user to this
#'     content. A value of owner is returned for the content owner. editor
#'     indicates a collaborator. The viewer value is given to users who are
#'     permitted to view the content. A none role is returned for
#'     administrators who cannot view the content but are permitted to view
#'     its configuration. Computed at the time of the request.
#'   * `vanity_url`: The vanity URL associated with this content item.
#'   * `id`: The internal numeric identifier of this content item.
#'   * `tags`: Tags associated with this content item. Each entry is a list
#'     with the following fields:
#'     * `id`: The identifier for the tag.
#'     * `name`: The name of the tag.
#'     * `parent_id`: The identifier for the parent tag. Null if the tag is a
#'       top-level tag.
#'     * `created_time`: The timestamp (RFC3339) indicating when the tag was
#'       created.
#'     * `updated_time`: The timestamp (RFC3339) indicating when the tag was
#'       last updated.
#'   * `owner`: Basic details about the owner of this content item. Each entry
#'     is a list with the following fields:
#'     * `guid`: The user's GUID, or unique identifier, in UUID RFC4122 format.
#'     * `username`: The user's username.
#'     * `first_name`: The user's first name.
#'     * `last_name`: The user's last name.
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/content for more
#' information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' get_content(client)
#' }
#'
#' @export
get_content <- function(
  src,
  guid = NULL,
  owner_guid = NULL,
  name = NULL,
  ...,
  .p = NULL
) {
  validate_R6_class(src, "Connect")

  # The capability to return vanity URLs `vanity_url` was added in Connect
  # v2024.06.0.
  if (compare_connect_version(src$version, "2024.06.0") < 0) {
    include <- "tags,owner"
    content_ptype <- connectapi_ptypes$content[,
      names(connectapi_ptypes$content) != "vanity_url"
    ]
  } else {
    include <- "tags,owner,vanity_url"
    content_ptype <- connectapi_ptypes$content
  }

  res <- src$content(
    guid = guid,
    owner_guid = owner_guid,
    name = name,
    include = include
  )

  if (!is.null(guid)) {
    # convert a single item to a list
    res <- list(res)
  }

  if (!is.null(.p)) {
    res <- res %>% purrr::keep(.p = .p)
  }

  out <- parse_connectapi_typed(res, content_ptype)

  return(out)
}


#' Get Content List with Permissions
#'
#' `r lifecycle::badge('experimental')` These functions are experimental placeholders until the API supports
#' this behavior.
#'
#' `content_list_with_permissions` loops through content and retrieves
#' permissions for each item (with a progress bar). This can take a long time
#' for lots of content! Make sure to use the optional `.p` argument as a predicate
#' function that filters the content list before it is transformed.
#'
#' `content_list_guid_has_access` works with a `content_list_with_permissions`
#' dataset by checking whether a given GUID (either user or group) has access to
#' the content by:
#' - checking if the content has access_type == "all"
#' - checking if the content has access_type == "logged_in"
#' - checking if the provided guid is the content owner
#' - checking if the provided guid is in the list of content permissions (in the "permissions" column)
#'
#' @param src A Connect R6 object
#' @param ... Extra arguments. Currently not used
#' @param .p Optional. A predicate function, passed as-is to `purrr::keep()`. See
#'   `get_content()` for more details. Can greatly help performance by reducing
#'   how many items to get permissions for
#' @param content_list A "content list with permissions" as returned by `content_list_with_permissions()`
#' @param guid A user or group GUID to filter the content list by whether they have access
#'
#' @rdname content_list_with_permissions
#'
#' @export
content_list_with_permissions <- function(src, ..., .p = NULL) {
  warn_experimental("content_list_with_permissions")

  message("Getting content list")
  content_list <- get_content(src, .p = .p)

  message("Getting permission list")
  pb <- optional_progress_bar(
    total = nrow(content_list),
    format = "[:bar] :percent :eta"
  )
  content_list[["permission"]] <- purrr::pmap(
    content_list,
    function(...) {
      pb$tick()
      get_content_permissions(Content$new(connect = src, content = list(...)))
    }
  )

  content_list
}

#' Content List
#'
#' `r lifecycle::badge('experimental')` Get a content list
#'
#' `content_list_by_tag()` retrieves a content list by tag
#'
#' @param src An R6 Connect object
#' @param tag A `connect_tag_tree` object or tag ID
#'
#' @rdname content_list
#' @export
content_list_by_tag <- function(src, tag) {
  validate_R6_class(src, "Connect")
  tag_id <- .get_tag_id(tag)

  res <- src$GET(v1_url("tags", tag_id, "content"))

  out <- parse_connectapi_typed(res, connectapi_ptypes$content)
  return(out)
}

#' @rdname content_list_with_permissions
#' @export
content_list_guid_has_access <- function(content_list, guid) {
  warn_experimental("content_list_filter_by_guid")
  rows_keep <- content_list$access_type %in%
    c("all", "logged_in") |
    content_list$owner_guid == guid |
    purrr::map_lgl(content_list$permission, ~ guid %in% .x$principal_guid)
  content_list[rows_keep, ]
}

#' Get usage information for deployed shiny applications
#'
#' @param src the source object
#' @param content_guid Filter results by content GUID
#' @param min_data_version Filter by data version. Records with a data version
#' lower than the given value will be excluded from the set of results.
#' @param from The timestamp that starts the time window of interest. Any usage
#' information that ends prior to this timestamp will not be returned.
#' Individual records may contain a starting time that is before this if they
#' end after it or have not finished. Must be of class Date or POSIX
#' @param to The timestamp that ends the time window of interest. Any usage
#' information that starts after this timestamp will not be returned.
#' Individual records may contain an ending time that is after this
#' (or no ending time) if they start before it. Must be of class Date or
#' POSIX
#' @param limit The number of records to return.
#' @param previous Retrieve the previous page of Shiny application usage
#' logs relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param nxt Retrieve the next page of Shiny application usage logs
#' relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param asc_order Defaults to TRUE; Determines if the response records
#' should be listed in ascending or descending order within the response.
#' Ordering is by the started timestamp field.
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `content_guid`: The GUID, in RFC4122 format, of the
#'     Shiny application this information pertains to.
#'   * `user_guid`: The GUID, in RFC4122 format, of the user
#'     that visited the application.
#'   * `started`: The timestamp, in RFC3339 format, when the
#'     user opened the application.
#'   * `ended`: The timestamp, in RFC3339 format, when the
#'     user left the application.
#'   * `data_version`: The data version the record was recorded
#'     with. The Shiny Application Events section of the Posit Connect Admin
#'     Guide explains how to interpret data_version values.
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/instrumentation/shiny/usage
#' for more information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' from <- Sys.Date() - lubridate::days(5)
#' get_usage_shiny(client, limit = 20, from = from)
#' }
#'
#' @export
get_usage_shiny <- function(
  src,
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
) {
  validate_R6_class(src, "Connect")

  res <- src$inst_shiny_usage(
    content_guid = content_guid,
    min_data_version = min_data_version,
    from = from,
    to = to,
    limit = limit,
    previous = previous,
    nxt = nxt,
    asc_order = asc_order
  )

  res <- page_cursor(src, res, limit = limit)

  out <- parse_connectapi_typed(res, connectapi_ptypes$usage_shiny)

  return(out)
}

#' Get usage information from deployed static content
#'
#' This function retrieves usage information from static content
#' on the Posit Connect server (e.g. Rmarkdown, Jupyter Notebooks)
#'
#' @param src the source object
#' @param content_guid Filter results by content GUID
#' @param min_data_version Filter by data version. Records with a data version
#' lower than the given value will be excluded from the set of results.
#' @param from The timestamp that starts the time window of interest. Any usage
#' information that ends prior to this timestamp will not be returned.
#' Individual records may contain a starting time that is before this if they
#' end after it or have not finished. Must be of class Date or POSIX
#' @param to The timestamp that ends the time window of interest. Any usage
#' information that starts after this timestamp will not be returned.
#' Individual records may contain an ending time that is after this
#' (or no ending time) if they start before it. Must be of class Date or
#' POSIX
#' @param limit The number of records to return.
#' @param previous Retrieve the previous page of Shiny application usage
#' logs relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param nxt Retrieve the next page of Shiny application usage logs
#' relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param asc_order Defaults to TRUE; Determines if the response records
#' should be listed in ascending or descending order within the response.
#' Ordering is by the started timestamp field.
#'
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `content_guid`: The GUID, in RFC4122 format, of the Shiny
#'     application this information pertains to.
#'   * `user_guid`: The GUID, in RFC4122 format, of the user that
#'     visited the application.
#'   * `variant_key`: The key of the variant the user visited.
#'     This will be null for static content.
#'   * `time`: The timestamp, in RFC3339 format, when the user
#'     visited the content.
#'   * `rendering_id`: The ID of the rendering the user visited.
#'     This will be null for static content.
#'   * `bundle_id`: The ID of the particular bundle used.
#'   * `data_version`: The data version the record was recorded
#'     with. The Rendered and Static Content Visit Events section of the
#'     Posit Connect Admin Guide explains how to interpret data_version
#'     values.
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/instrumentation/content/visits
#' for more information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' from <- Sys.Date() - lubridate::days(5)
#' get_usage_static(client, limit = 20, from = from)
#' }
#'
#' @export
get_usage_static <- function(
  src,
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
) {
  validate_R6_class(src, "Connect")

  res <- src$inst_content_visits(
    content_guid = content_guid,
    min_data_version = min_data_version,
    from = from,
    to = to,
    limit = limit,
    previous = previous,
    nxt = nxt,
    asc_order = asc_order
  )

  res <- page_cursor(src, res, limit = limit)

  out <- parse_connectapi_typed(res, connectapi_ptypes$usage_static)

  return(out)
}

#' Get usage information for deployed content
#'
#' @description
#' Retrieve content hits for all available content on the server. Available
#' content depends on the user whose API key is in use. Administrator accounts
#' will receive data for all content on the server. Publishers will receive data
#' for all content they own or collaborate on.
#'
#' If no date-times are provided, all usage data will be returned.
#'
#' @param client A `Connect` R6 client object.
#' @param content_guid Optional. A single content GUID or a character vector of
#'   GUIDs to filter results. When multiple GUIDs are provided they are
#'   collapsed with `"|"`.
#' @param from Optional date-time (`POSIXct` or `POSIXlt`). Only
#'   records after this time are returned. If not provided, records
#'   are returned back to the first record available.
#' @param to Optional date-time (`POSIXct` or `POSIXlt`). Only records
#'   before this time are returned. If not provided, all records up to
#'   the most recent are returned.
#'
#' @return A list of usage records. Each record is a list with all elements
#'   as character strings unless otherwise specified.
#'
#' * `id`: A string identifier for the hit.
#' * `user_guid`: The user GUID if the visitor is logged-in, `NULL` for
#'   anonymous hits.
#' * `content_guid`: The GUID of the visited content.
#' * `timestamp`: The time of the hit in RFC3339 format.
#' * `data`: A nested list with optional fields:
#'     * `path`: The request path (if recorded).
#'     * `user_agent`: The user agent string (if available).
#'
#' Use [as.data.frame()] or [tibble::as_tibble()] to convert to a flat
#' table with parsed types. In the resulting data frame:
#'
#' * `timestamp` is parsed to `POSIXct`.
#' * `path` and `user_agent` are extracted from the nested `data` field.
#'
#' By default, [as.data.frame()] attempts to extract the nested fields using
#' the \pkg{tidyr} package. If \pkg{tidyr} is not available, or if you want to
#' skip unnesting, call `as.data.frame(x, unnest = FALSE)` to leave `data` as
#' a list-column.
#'
#' @details
#'
#' The data returned by `get_usage()` includes all content types. For Shiny
#' content, the `timestamp` indicates the *start* of the Shiny session.
#' Additional fields for Shiny and non-Shiny are available respectively from
#' [get_usage_shiny()] and [get_usage_static()]. `get_usage_shiny()` includes a
#' field for the session end time; `get_usage_static()` includes variant,
#' rendering, and bundle identifiers for the visited content.
#'
#' When possible, however, we recommend using `get_usage()` over
#' `get_usage_static()` or `get_usage_shiny()`, as it is faster and more efficient.
#'
#' @seealso [as.data.frame.connect_list_hits()], [as_tibble.connect_list_hits()]
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Fetch the last 2 days of hits
#' usage <- get_usage(client, from = Sys.Date() - 2, to = Sys.Date())
#'
#' # Fetch usage after a specified date and convert to a data frame.
#' usage <- get_usage(
#'   client,
#'   from = as.POSIXct("2025-05-02 12:40:00", tz = "UTC")
#' )
#'
#' # Fetch usage for a specific content item
#' usage <- get_usage(client, content_guid = "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")
#'
#' # Fetch all usage
#' usage <- get_usage(client)
#'
#' # Convert to tibble or data frame
#' usage_df <- tibble::as_tibble(usage)
#'
#' # Skip unnesting if tidyr is not installed
#' usage_df <- as.data.frame(usage, unnest = FALSE)
#' }
#'
#' @export
get_usage <- function(client, content_guid = NULL, from = NULL, to = NULL) {
  error_if_less_than(client$version, "2025.04.0")

  if (!is.null(content_guid)) {
    error_if_less_than(client$version, "2026.01.0")
    if (length(content_guid) > 1) {
      content_guid <- paste0(content_guid, collapse = "|")
    }
  }

  usage <- client$GET(
    v1_url("instrumentation", "content", "hits"),
    query = list(
      content_guid = content_guid,
      from = make_timestamp(from),
      to = make_timestamp(to)
    )
  )

  class(usage) <- c("connect_list_hits", class(usage))
  usage
}

#' Convert usage data to a data frame
#'
#' @description
#' Converts an object returned by [get_usage()] into a data frame with parsed
#' column types. By default, extracts `path` and `user_agent` from the `data`
#' field, if available.
#'
#' @param x A `connect_list_hits` object (from [get_usage()]).
#' @param row.names Passed to [base::as.data.frame()].
#' @param optional Passed to [base::as.data.frame()].
#' @param ... Passed to [base::as.data.frame()].
#' @param unnest Logical; if `TRUE` (default), extracts nested fields using
#'   \pkg{tidyr}. Set to `FALSE` to skip unnesting.
#'
#' @return A `data.frame` with one row per usage record.
#' @export
#' @method as.data.frame connect_list_hits
as.data.frame.connect_list_hits <- function(
  x,
  row.names = NULL, # nolint
  optional = FALSE,
  ...,
  unnest = TRUE
) {
  usage_df <- parse_connectapi_typed(x, connectapi_ptypes$usage)
  if (unnest) {
    if (!requireNamespace("tidyr", quietly = TRUE)) {
      stop(
        "`unnest = TRUE` requires tidyr. Install tidyr or set `unnest = FALSE`.",
        call. = FALSE
      )
    }
    usage_df <- tidyr::unnest_wider(
      usage_df,
      "data",
      ptype = list(path = character(0), user_agent = character(0))
    )
  }
  as.data.frame(usage_df, row.names = row.names, optional = optional, ...)
}

#' Convert usage data to a tibble
#'
#' @description
#' Converts an object returned by [get_usage()] to a tibble via
#' [as.data.frame.connect_list_hits()].
#'
#' @param x A `connect_list_hits` object.
#' @param ... Passed to [as.data.frame()].
#'
#' @return A tibble with one row per usage record.
#' @export
#' @importFrom tibble as_tibble
#' @method as_tibble connect_list_hits
as_tibble.connect_list_hits <- function(x, ...) {
  tibble::as_tibble(as.data.frame(x, ...))
}

#' Get Audit Logs from Posit Connect Server
#'
#' @param src The source object
#' @param limit The number of records to return.
#' @param previous Retrieve the previous page of Shiny application usage
#' logs relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param nxt Retrieve the next page of Shiny application usage logs
#' relative to the provided value. This value corresponds to an internal
#' reference within the server and should be sourced from the appropriate
#' attribute within the paging object of a previous response.
#' @param asc_order Defaults to TRUE; Determines if the response records
#' should be listed in ascending or descending order within the response.
#' Ordering is by the started timestamp field.
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `id`: ID of the audit action
#'   * `time`: Timestamp in RFC3339 format when action was taken
#'   * `user_id`: User ID of the actor who made the audit action
#'   * `user_description`: Description of the actor
#'   * `action`: Audit action taken
#'   * `event_description`: Description of action
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/audit_logs for more
#' information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' # get the last 20 audit logs
#' get_audit_logs(client, limit = 20, asc_order = FALSE)
#' }
#'
#' @export
get_audit_logs <- function(
  src,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
) {
  validate_R6_class(src, "Connect")

  res <- src$audit_logs(
    limit = limit,
    previous = previous,
    nxt = nxt,
    asc_order = asc_order
  )

  res <- page_cursor(src, res, limit = limit)

  out <- parse_connectapi_typed(res, connectapi_ptypes$audit_logs)

  return(out)
}

#' Get Real-Time Process Data
#'
#' `r lifecycle::badge('experimental')`
#' This returns real-time process data from the Posit Connect API. It requires
#' administrator privileges to use. NOTE that this only returns data for the
#' server that responds to the request (i.e. in a Highly Available cluster)
#'
#' @param src The source object
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `pid`: The PID of the current process
#'   * `appId`: The application ID
#'   * `appGuid`: The application GUID
#'   * `appName`: The application name
#'   * `appUrl`: The application URL
#'   * `appRunAs`: The application RunAs user
#'   * `type`: The type of process
#'   * `cpuCurrent`: The current CPU usage
#'   * `cpuTotal`: The total CPU usage
#'   * `ram`: The current RAM usage
#'
#' @export
get_procs <- function(src) {
  validate_R6_class(src, "Connect")
  warn_experimental("get_procs")

  scoped_experimental_silence()
  raw_proc_data <- src$procs()

  proc_prep <- purrr::imap(
    raw_proc_data,
    function(x, y) {
      c(list(pid = y), x)
    }
  )
  tbl_data <- parse_connectapi_typed(proc_prep, connectapi_ptypes$procs)

  return(tbl_data)
}

#' Perform an OAuth credential exchange to obtain a visitor's OAuth access token.
#'
#' @param connect A Connect R6 object.
#' @param user_session_token The content visitor's session token. This token
#' can only be obtained when the content is running on a Connect server. The token
#' identifies the user who is viewing the content interactively on the Connect server.
#' Read this value from the HTTP header: `Posit-Connect-User-Session-Token`
#' @param requested_token_type Optional. The requested token type. If unset, will
#' default to `urn:ietf:params:oauth:token-type:access_token`. Otherwise, this can
#' be set to `urn:ietf:params:aws:token-type:credentials` for AWS integrations or
#' `urn:posit:connect:api-key` for Connect API Key integrations.
#' @param audience Optional. The GUID of an OAuth integration associated with
#' this piece of content.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' library(plumber)
#' client <- connect()
#'
#' #* @get /do
#' function(req) {
#'   user_session_token <- req$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
#'   credentials <- get_oauth_credentials(client, user_session_token)
#'
#'   # ... do something with `credentials$access_token` ...
#'
#'   "done"
#' }
#' }
#'
#' @return The OAuth credential exchange response.
#'
#' @details
#' Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-a-viewer-oauth-access-token
#' for more information.
#'
#' @seealso [get_integrations()], [get_oauth_content_credentials()]
#'
#' @export
get_oauth_credentials <- function(
  connect,
  user_session_token,
  requested_token_type = NULL,
  audience = NULL
) {
  validate_R6_class(connect, "Connect")

  if (!is.null(audience)) {
    error_if_less_than(connect$version, "2025.07.0")
  }

  url <- v1_url("oauth", "integrations", "credentials")
  body <- list(
    grant_type = "urn:ietf:params:oauth:grant-type:token-exchange",
    subject_token_type = "urn:posit:connect:user-session-token",
    subject_token = user_session_token,
    requested_token_type = requested_token_type,
    audience = audience
  )
  connect$POST(
    url,
    encode = "form",
    body = body
  )
}

#' Perform an OAuth credential exchange to obtain a content-specific OAuth
#' access token.
#'
#' @param connect A Connect R6 object.
#' @param content_session_token Optional. The content session token. This token
#' can only be obtained when the content is running on a Connect server. The
#' token identifies the service account integration previously configured by
#' the publisher on the Connect server. Defaults to the value found in the file
#' indicated by the `CONNECT_CONTENT_SESSION_TOKEN_FILE` environment variable
#' on Connect >= 2026.02.0, or from the value of the environment variable
#' `CONNECT_CONTENT_SESSION_TOKEN` on earlier versions.
#' @param requested_token_type Optional. The requested token type. If unset,
#' will default to `urn:ietf:params:oauth:token-type:access_token`. Otherwise,
#' this can be set to `urn:ietf:params:aws:token-type:credentials` for AWS
#' integrations or `urn:posit:connect:api-key` for Connect API Key integrations.
#' @param audience Optional. The GUID of an OAuth integration associated with
#' this piece of content.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' library(plumber)
#' client <- connect()
#'
#' #* @get /do
#' function(req) {
#'   credentials <- get_oauth_content_credentials(client)
#'
#'   # ... do something with `credentials$access_token` ...
#'
#'   "done"
#' }
#' }
#'
#' @return The OAuth credential exchange response.
#'
#' @details
#' Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-a-service-account-oauth-access-token
#' for more information.
#'
#' @seealso [get_integrations()], [get_oauth_credentials()]
#'
#' @export
get_oauth_content_credentials <- function(
  connect,
  content_session_token = NULL,
  requested_token_type = NULL,
  audience = NULL
) {
  validate_R6_class(connect, "Connect")
  error_if_less_than(connect$version, "2024.12.0")

  if (!is.null(audience)) {
    error_if_less_than(connect$version, "2025.07.0")
  }

  # First, try CONNECT_CONTENT_SESSION_TOKEN_FILE
  if (is.null(content_session_token)) {
    token_file <- Sys.getenv("CONNECT_CONTENT_SESSION_TOKEN_FILE")
    if (nzchar(token_file)) {
      content_session_token <- readLines(token_file, n = 1, warn = FALSE)
    }
  }
  # If that doesn't exist (it was added in 2026.02.0), look for the token itself
  if (is.null(content_session_token)) {
    content_session_token <- Sys.getenv("CONNECT_CONTENT_SESSION_TOKEN")
    if (nchar(content_session_token) == 0) {
      stop(
        "Could not find the CONNECT_CONTENT_SESSION_TOKEN environment variable."
      )
    }
  }
  url <- v1_url("oauth", "integrations", "credentials")
  body <- list(
    grant_type = "urn:ietf:params:oauth:grant-type:token-exchange",
    subject_token_type = "urn:posit:connect:content-session-token",
    subject_token = content_session_token,
    requested_token_type = requested_token_type,
    audience = audience
  )
  connect$POST(
    url,
    encode = "form",
    body = body
  )
}

#' Obtain a visitor's AWS credentials
#'
#' @param connect A Connect R6 object.
#' @param user_session_token The content visitor's session token. This token
#' can only be obtained when the content is running on a Connect server. The token
#' identifies the user who is viewing the content interactively on the Connect server.
#' Read this value from the HTTP header: `Posit-Connect-User-Session-Token`
#' @param audience Optional. The GUID of an OAuth integration associated with
#' this piece of content.
#'
#' @return The AWS credentials as a list with fields named `access_key_id`,
#' `secret_access_key`, `session_token`, and `expiration`.
#'
#' @details
#' Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-service-account-aws-credentials
#' for more information. See the example below of using this function in a
#' Plumber API with paws to access S3. Any library that allows you to pass
#' AWS credentials will be able to utilize the credentials returned from
#' this function call.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' library(plumber)
#' library(paws)
#' client <- connect()
#'
#' #* @get /do
#' function(req) {
#'   user_session_token <- req$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
#'   aws_credentials <- get_aws_credentials(client, user_session_token)
#'
#'   # Create S3 client with AWS credentials from Connect
#'   svc <- paws::s3(
#'     credentials = list(
#'       creds = list(
#'         access_key_id = aws_credentials$access_key_id,
#'         secret_access_key = aws_credentials$secret_access_key,
#'         session_token = aws_credentials$session_token
#'       )
#'     )
#'   )
#'
#'   # Get object from S3
#'   obj <- svc$get_object(
#'     Bucket = "my-bucket",
#'     Key = "my-data.csv"
#'   )
#'
#'   "done"
#' }
#' }
#'
#' @export
get_aws_credentials <- function(connect, user_session_token, audience = NULL) {
  error_if_less_than(connect$version, "2025.03.0")

  if (!is.null(audience)) {
    error_if_less_than(connect$version, "2025.07.0")
  }

  response <- get_oauth_credentials(
    connect,
    user_session_token,
    requested_token_type = "urn:ietf:params:aws:token-type:credentials",
    audience = audience
  )

  # Extract access token and decode it
  access_token <- rawToChar(base64enc::base64decode(response$access_token))
  credentials <- jsonlite::fromJSON(access_token)

  list(
    access_key_id = credentials$accessKeyId,
    secret_access_key = credentials$secretAccessKey,
    session_token = credentials$sessionToken,
    expiration = credentials$expiration
  )
}

#' Obtain AWS credentials for your content.
#'
#' @param connect A Connect R6 object.
#' @param content_session_token Optional. The content session token. This token
#' can only be obtained when the content is running on a Connect server. The
#' token identifies the service account integration previously configured by
#' the publisher on the Connect server. Defaults to the value from the
#' environment variable: `CONNECT_CONTENT_SESSION_TOKEN`
#' @param audience Optional. The GUID of an OAuth integration associated with
#' this piece of content.
#'
#' @return The AWS credentials as a list with fields named `access_key_id`,
#' `secret_access_key`, `session_token`, and `expiration`.
#'
#' @details
#' Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-service-account-aws-credentials
#' for more information. See the example below of using this function with
#' `paws` to access S3. Any library that allows you to pass AWS credentials
#' will be able to utilize the credentials returned from this function call.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' library(paws)
#'
#' client <- connect()
#' # Pulls the content session token from the environment
#' # when deployed into Connect.
#' aws_credentials <- get_aws_content_credentials(client)
#'
#' # Create S3 client with AWS credentials from Connect
#' svc <- paws::s3(
#'   credentials = list(
#'     creds = list(
#'       access_key_id = aws_credentials$access_key_id,
#'       secret_access_key = aws_credentials$secret_access_key,
#'       session_token = aws_credentials$session_token
#'     )
#'   )
#' )
#'
#' # Get object from S3
#' obj <- svc$get_object(
#'   Bucket = "my-bucket",
#'   Key = "my-data.csv"
#' )
#' }
#'
#' @export
get_aws_content_credentials <- function(
  connect,
  content_session_token = NULL,
  audience = NULL
) {
  error_if_less_than(connect$version, "2025.03.0")

  if (!is.null(audience)) {
    error_if_less_than(connect$version, "2025.07.0")
  }

  response <- get_oauth_content_credentials(
    connect,
    content_session_token,
    requested_token_type = "urn:ietf:params:aws:token-type:credentials",
    audience = audience
  )

  # Extract access token and decode it
  access_token <- rawToChar(base64enc::base64decode(response$access_token))
  credentials <- jsonlite::fromJSON(access_token)

  list(
    access_key_id = credentials$accessKeyId,
    secret_access_key = credentials$secretAccessKey,
    session_token = credentials$sessionToken,
    expiration = credentials$expiration
  )
}

#' Get available runtimes on server
#'
#' Get a table showing available versions of R, Python, Quarto, and Tensorflow
#' on the Connect server.
#'
#' @param client A `Connect` object.
#' @param runtimes Optional. A character vector of runtimes to include. Must be
#' some combination of `"r"`, `"python"`, `"quarto"`, and `"tensorflow"`. Quarto
#' is only supported on Connect >= 2021.08.0, and Tensorflow is only supported
#' on Connect >= 2024.03.0.

#' @return A tibble with columns for `runtime`, `version`, and `cluster_name`
#' and `image_name`. Cluster name and image name are only meaningful on Connect
#' instances running off-host execution.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#' get_runtimes(client, runtimes = c("r", "python", "tensorflow"))
#' }
#'
#' @export
get_runtimes <- function(client, runtimes = NULL) {
  validate_R6_class(client, "Connect")

  # Construct valid runtimes for the Connect version.
  supported <- c(
    "r",
    "python",
    if (compare_connect_version(client$version, "2021.08.0") >= 0) {
      "quarto"
    },
    if (compare_connect_version(client$version, "2024.05.0") >= 0) {
      "tensorflow"
    }
  )
  if (is.null(runtimes)) {
    runtimes <- supported
  } else {
    if (any(!runtimes %in% supported)) {
      stop(glue::glue(
        "`runtimes` must be one of ",
        "{paste(paste0('\"', supported, '\"'), collapse = ', ')}; ",
        "received: {paste(paste0('\"', runtimes, '\"'), collapse = ', ')}."
      ))
    }
  }

  purrr::list_rbind(purrr::map(
    runtimes,
    function(runtime) {
      res <- client$GET(paste0("v1/server_settings/", runtime))
      res_df <- purrr::list_rbind(
        purrr::map(res$installations, ~ tibble::as_tibble(.))
      )
      tibble::add_column(res_df, runtime = runtime, .before = 1)
    }
  ))
}

#' All package dependencies on the server
#'
#' @description Get a data frame of package dependencies used by all content
#' items on the server.
#'
#' @usage get_packages(src, name = NULL, page_size = 100000, limit = Inf)
#'
#' @param src A `Connect` client object.
#' @param name Optional package name to filter by. Python package are normalized
#' during matching; R package names must match exactly.
#' @param page_size Optional. Integer specifying page size for API
#' paging.
#' @param limit Optional. Specify the maximum number of records after which
#' to cease paging.
#'
#' @return A data frame with the following columns:
#'
#' - `language`: Language ecosystem the package belongs to (`r` or `python`)
#' - `language_version`: Version of R or Python used by the content
#' - `name`: Package name
#' - `version`: Package version
#' - `hash`: Package description hash for R packages
#' - `bundle_id`: Identifier for the bundle that depends on this package
#' - `content_id`: Numeric identifier for the content that depends on this
#'   package
#' - `content_guid`: The unique identifier of the content item that depends on
#'   this package
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' packages <- get_packages(client)
#' }
#'
#' @family packages functions
#' @export
get_packages <- function(src, name = NULL, page_size = 100000, limit = Inf) {
  validate_R6_class(src, "Connect")
  error_if_less_than(src$version, "2024.11.0")
  res <- page_offset(
    src,
    src$packages(
      name = name,
      page_size = page_size
    )
  )
  out <- parse_connectapi_typed(res, connectapi_ptypes$packages)

  # Connect is standardizing on using `content_id` and `content_guid`.
  # Handle that name change now in a forward-compatible way.
  if ("app_id" %in% names(out)) {
    out$content_id <- out$app_id
    out$app_id <- NULL
  }
  if ("app_guid" %in% names(out)) {
    out$content_guid <- out$app_guid
    out$app_guid <- NULL
  }

  out
}

#' Get all vanity URLs
#'
#' Get a table of all vanity URLs on the server. Requires administrator
#' privileges.
#'
#' @param client A `Connect` object.
#'
#' @return A tibble with columns for `content_guid`, `path`, and
#' `created_time`.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#' get_vanity_urls(client)
#' }
#'
#' @export
get_vanity_urls <- function(client) {
  res <- client$vanities()
  parse_connectapi_typed(res, connectapi_ptypes$vanities)
}

with_mock_api({
  test_that("we can retrieve the oauth user credentials", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    credentials <- get_oauth_credentials(
      client,
      user_session_token = "user-session-token"
    )
    expect_equal(
      credentials,
      list(
        access_token = "user-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("we can retrieve the oauth content credentials with an explicit token", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_warning(
      credentials <- get_oauth_content_credentials(
        client,
        content_session_token = "content-session-token"
      )
    )
    expect_equal(
      credentials,
      list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("we can retrieve the oauth content credentials with an env var", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      list(CONNECT_CONTENT_SESSION_TOKEN = "content-session-token")
    )

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_warning(
      credentials <- get_oauth_content_credentials(client)
    )
    expect_equal(
      credentials,
      list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("if there is a session token file, we prefer that", {
    token_file <- tempfile()
    cat("content-session-token", file = token_file)
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      list(CONNECT_CONTENT_SESSION_TOKEN_FILE = token_file)
    )

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_warning(
      credentials <- get_oauth_content_credentials(client)
    )
    expect_equal(
      credentials,
      list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("we cannot retrieve the oauth content credentials without a token or env var", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_error(
      expect_warning(get_oauth_content_credentials(client)),
      "Could not find the CONNECT_CONTENT_SESSION_TOKEN environment variable."
    )
  })

  test_that("we can retrieve the AWS viewer credentials", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_warning(
      credentials <- get_aws_credentials(
        client,
        user_session_token = "user-session-token"
      )
    )
    expect_equal(
      credentials,
      list(
        access_key_id = "abc123",
        secret_access_key = "def456",
        session_token = "ghi789",
        expiration = "2025-01-01T00:00:00Z"
      )
    )
  })

  test_that("we can retrieve the AWS content credentials with an explicit token", {
    # get_aws_content_credentials produces multiple warnings about the Posit
    # Connect version; suppress these
    withr::local_options(list(rlib_warning_verbosity = "quiet"))

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))

    credentials <- get_aws_content_credentials(
      client,
      content_session_token = "content-session-token"
    )

    expect_equal(
      credentials,
      list(
        access_key_id = "abc123",
        secret_access_key = "def456",
        session_token = "ghi789",
        expiration = "2025-01-01T00:00:00Z"
      )
    )
  })

  test_that("we can retrieve the AWS content credentials with an env var", {
    withr::local_options(list(rlib_warning_verbosity = "quiet"))
    withr::local_envvar(
      list(CONNECT_CONTENT_SESSION_TOKEN = "content-session-token")
    )

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))

    credentials <- get_aws_content_credentials(client)

    expect_equal(
      credentials,
      list(
        access_key_id = "abc123",
        secret_access_key = "def456",
        session_token = "ghi789",
        expiration = "2025-01-01T00:00:00Z"
      )
    )
  })

  test_that("we cannot retrieve the AWS content credentials without a token or env var", {
    withr::local_options(list(rlib_warning_verbosity = "quiet"))

    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(client, "Connect"))
    expect_error(
      get_aws_content_credentials(client),
      "Could not find the CONNECT_CONTENT_SESSION_TOKEN environment variable."
    )
  })

  test_that("we can retrieve the oauth user credentials with audience", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "user-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )

    credentials <- get_oauth_credentials(
      client,
      user_session_token = "user-session-token",
      audience = "audience"
    )
    expect_equal(
      credentials,
      list(
        access_token = "user-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("we can retrieve the oauth user credentials with audience and req token type", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "user-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )

    credentials <- get_oauth_credentials(
      client,
      user_session_token = "user-session-token",
      requested_token_type = "urn:ietf:params:oauth:token-type:access_token",
      audience = "audience"
    )
    expect_equal(
      credentials,
      list(
        access_token = "user-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("we can retrieve the oauth content credentials with audience", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
    credentials <- get_oauth_content_credentials(
      client,
      content_session_token = "content-session-token",
      audience = "audience",
    )
    expect_equal(
      credentials,
      list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("oauth viewer credentials retrieval with audience on unsupported version fails", {
    client <- MockConnect$new("2025.06.0")

    expect_error(
      get_oauth_credentials(
        client,
        user_session_token = "user-session-token",
        audience = "audience"
      ),
      "ERROR: This feature requires Posit Connect version 2025.07"
    )
  })

  test_that("we can retrieve the oauth content credentials with audience and req token type", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )

    credentials <- get_oauth_content_credentials(
      client,
      content_session_token = "content-session-token",
      requested_token_type = "urn:ietf:params:oauth:token-type:access_token",
      audience = "audience"
    )
    expect_equal(
      credentials,
      list(
        access_token = "content-access-token",
        issued_token_type = "urn:ietf:params:oauth:token-type:access_token",
        token_type = "Bearer"
      )
    )
  })

  test_that("oauth content credentials retrieval with audience on unsupported version fails", {
    client <- MockConnect$new("2025.06.0")

    expect_error(
      get_oauth_content_credentials(
        client,
        content_session_token = "content-session-token",
        audience = "audience"
      ),
      "ERROR: This feature requires Posit Connect version 2025.07"
    )
  })

  test_that("we can retrieve the AWS viewer credentials with audience", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "eyJhY2Nlc3NLZXlJZCI6ICJhYmMxMjMiLCAic2VjcmV0QWNjZXNzS2V5IjogImRlZjQ1NiIsICJzZXNzaW9uVG9rZW4iOiAiZ2hpNzg5IiwgImV4cGlyYXRpb24iOiAiMjAyNS0wMS0wMVQwMDowMDowMFoifQ==", # nolint: line_length_linter
        issued_token_type = "urn:ietf:params:aws:token-type:credentials",
        token_type = "aws_credentials"
      )
    )

    credentials <- get_aws_credentials(
      client,
      user_session_token = "user-session-token",
      audience = "audience"
    )
    expect_equal(
      credentials,
      list(
        access_key_id = "abc123",
        secret_access_key = "def456",
        session_token = "ghi789",
        expiration = "2025-01-01T00:00:00Z"
      )
    )
  })

  test_that("AWS viewer credentials retrieval with audience on unsupported version fails", {
    client <- MockConnect$new("2025.06.0")

    expect_error(
      get_aws_credentials(
        client,
        user_session_token = "user-session-token",
        audience = "audience"
      ),
      "ERROR: This feature requires Posit Connect version 2025.07"
    )
  })

  test_that("we can retrieve the AWS content credentials with audience", {
    client <- MockConnect$new("2025.07.0")
    client$mock_response(
      "POST",
      v1_url("oauth", "integrations", "credentials"),
      content = list(
        access_token = "eyJhY2Nlc3NLZXlJZCI6ICJhYmMxMjMiLCAic2VjcmV0QWNjZXNzS2V5IjogImRlZjQ1NiIsICJzZXNzaW9uVG9rZW4iOiAiZ2hpNzg5IiwgImV4cGlyYXRpb24iOiAiMjAyNS0wMS0wMVQwMDowMDowMFoifQ==", # nolint: line_length_linter
        issued_token_type = "urn:ietf:params:aws:token-type:credentials",
        token_type = "aws_credentials"
      )
    )

    credentials <- get_aws_content_credentials(
      client,
      content_session_token = "content-session-token",
      audience = "audience"
    )

    expect_equal(
      credentials,
      list(
        access_key_id = "abc123",
        secret_access_key = "def456",
        session_token = "ghi789",
        expiration = "2025-01-01T00:00:00Z"
      )
    )
  })

  test_that("AWS content credentials retrieval with audience on unsupported version fails", {
    client <- MockConnect$new("2025.06.0")

    expect_error(
      get_aws_content_credentials(
        client,
        content_session_token = "content-session-token",
        audience = "audience"
      ),
      "ERROR: This feature requires Posit Connect version 2025.07"
    )
  })
})

test_that("get_runtimes() gets all runtimes if none specified", {
  client <- MockConnect$new("2024.10.0")

  client$mock_response(
    "GET",
    "v1/server_settings/r",
    content = list(
      installations = list(
        list(version = "4.3.1", cluster_name = "Local", image_name = "Local"),
        list(version = "4.4.0", cluster_name = "Local", image_name = "Local")
      )
    )
  )
  client$mock_response(
    "GET",
    "v1/server_settings/python",
    content = list(
      installations = list(
        list(version = "3.11.3", cluster_name = "Local", image_name = "Local"),
        list(version = "3.12.4", cluster_name = "Local", image_name = "Local")
      )
    )
  )
  client$mock_response(
    "GET",
    "v1/server_settings/quarto",
    content = list(
      installations = list(
        list(version = "1.4.557", cluster_name = "Local", image_name = "Local"),
        list(version = "1.5.55", cluster_name = "Local", image_name = "Local")
      )
    )
  )
  client$mock_response(
    "GET",
    "v1/server_settings/tensorflow",
    content = list(
      installations = list(
        list(version = "2.16.1", cluster_name = "Local", image_name = "Local")
      )
    )
  )

  expected <- tibble::as_tibble(list(
    runtime = c(
      "r",
      "r",
      "python",
      "python",
      "quarto",
      "quarto",
      "tensorflow"
    ),
    version = c(
      "4.3.1",
      "4.4.0",
      "3.11.3",
      "3.12.4",
      "1.4.557",
      "1.5.55",
      "2.16.1"
    ),
    cluster_name = c(
      "Local",
      "Local",
      "Local",
      "Local",
      "Local",
      "Local",
      "Local"
    ),
    image_name = c(
      "Local",
      "Local",
      "Local",
      "Local",
      "Local",
      "Local",
      "Local"
    )
  ))
  expect_identical(get_runtimes(client), expected)
  expect_identical(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/server_settings/r",
      "GET https://connect.example/__api__/v1/server_settings/python",
      "GET https://connect.example/__api__/v1/server_settings/quarto",
      "GET https://connect.example/__api__/v1/server_settings/tensorflow"
    )
  )
})

test_that("get_runtimes() only specified runtimes", {
  client <- MockConnect$new("2024.10.0")

  client$mock_response(
    "GET",
    "v1/server_settings/python",
    content = list(
      installations = list(
        list(version = "3.11.3", cluster_name = "Local", image_name = "Local"),
        list(version = "3.12.4", cluster_name = "Local", image_name = "Local")
      )
    )
  )
  client$mock_response(
    "GET",
    "v1/server_settings/tensorflow",
    content = list(
      installations = list(
        list(version = "2.16.1", cluster_name = "Local", image_name = "Local")
      )
    )
  )

  expected <- tibble::as_tibble(list(
    runtime = c("python", "python", "tensorflow"),
    version = c("3.11.3", "3.12.4", "2.16.1"),
    cluster_name = c("Local", "Local", "Local"),
    image_name = c("Local", "Local", "Local")
  ))
  expect_identical(get_runtimes(client, c("python", "tensorflow")), expected)
  expect_identical(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/server_settings/python",
      "GET https://connect.example/__api__/v1/server_settings/tensorflow"
    )
  )
})

test_that("get_runtimes() restricts available runtimes based on Connect version.", {
  client <- MockConnect$new("2024.10.0")
  expect_error(
    get_runtimes(client, c("r", "python", "foofy")),
    '`runtimes` must be one of "r", "python", "quarto", "tensorflow"; received: "r", "python", "foofy".'
  )

  client <- MockConnect$new("2024.02.0")
  expect_error(
    get_runtimes(client, "tensorflow"),
    '`runtimes` must be one of "r", "python", "quarto"; received: "tensorflow".'
  )

  client <- MockConnect$new("1.8.3")
  expect_error(
    get_runtimes(client, c("r", "quarto")),
    '`runtimes` must be one of "r", "python"; received: "r", "quarto".'
  )
})

without_internet({
  client <- Connect$new(server = "https://connect.example", api_key = "fake")
  test_that("get_users() works with user_role and account_status", {
    # No filter parameters specified
    expect_GET(
      get_users(client),
      "https://connect.example/__api__/v1/users?page_number=1&page_size=500"
    )

    # Filter just on one parameter
    expect_GET(
      get_users(client, user_role = "administrator"),
      "https://connect.example/__api__/v1/users?page_number=1&page_size=500&user_role=administrator"
    )

    # Filter on two parameters, one requiring concatenation
    expect_GET(
      get_users(
        client,
        user_role = c("administrator", "publisher"),
        account_status = "licensed"
      ),
      paste0(
        "https://connect.example/__api__/v1/users?page_number=1&page_size=500&",
        "user_role=administrator%7Cpublisher&account_status=licensed"
      )
    )
  })
})

test_that("get_vanity_urls() works", {
  with_mock_api({
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    expect_equal(
      get_vanity_urls(client),
      tibble::tibble(
        content_guid = c(
          "39c8d85a-37ae-4b8b-8655-30a06adff2f1",
          "93a3cd6d-5a1b-236c-9808-6045f2a73fb5"
        ),
        path = c(
          "/team-dashboard/",
          "/streamlit/my-app/"
        ),
        created_time = structure(
          c(
            1602623489,
            1677679943
          ),
          tzone = Sys.timezone(),
          class = c("POSIXct", "POSIXt")
        )
      )
    )
  })
})

test_that("get_packages() works as expected with current return value", {
  client <- MockConnect$new("2024.11.0")
  client$mock_response(
    "GET",
    "v1/packages",
    content = list(
      current_page = 1,
      total = 2,
      results = list(
        list(
          language = "python",
          language_version = "3.7.6",
          name = "absl-py",
          version = "0.12.0",
          hash = NULL,
          bundle_id = "9375",
          app_id = "4906",
          app_guid = "9bf33774"
        ),
        list(
          language = "python",
          language_version = "3.7.7",
          name = "absl-py",
          version = "0.8.1",
          hash = NULL,
          bundle_id = "6623",
          app_id = "3652",
          app_guid = "1935b6cb"
        )
      )
    )
  )
  expect_identical(
    get_packages(client),
    tibble::tibble(
      language = c("python", "python"),
      language_version = c("3.7.6", "3.7.7"),
      name = c("absl-py", "absl-py"),
      version = c("0.12.0", "0.8.1"),
      hash = c(NA_character_, NA_character_),
      bundle_id = c("9375", "6623"),
      content_id = c("4906", "3652"),
      content_guid = c("9bf33774", "1935b6cb")
    )
  )
})

test_that("Pagination is wired up correctly for packages method", {
  with_mock_api({
    without_internet({
      client <- Connect$new(
        server = "https://connect.example",
        api_key = "fake"
      )
      expect_GET(
        client$packages(name = "mypkg", page_number = 1),
        "https://connect.example/__api__/v1/packages?name=mypkg&page_number=1&page_size=100000"
      )
    })
  })
})

test_that("get_packages() works as expected with `content_guid` names in API response", {
  # The responses in this test differ from the httptest fixtures used by the
  # prior test by the presence of `content_id` and `content_guid` keys in the
  # return objects.
  client <- MockConnect$new("2024.11.0")

  client$mock_response(
    "GET",
    "v1/packages",
    content = list(
      current_page = 1,
      total = 2,
      results = list(
        list(
          language = "python",
          language_version = "3.7.6",
          name = "absl-py",
          version = "0.12.0",
          hash = NULL,
          bundle_id = "9375",
          app_id = "4906",
          app_guid = "9bf33774",
          content_id = "4906",
          content_guid = "9bf33774"
        ),
        list(
          language = "python",
          language_version = "3.7.7",
          name = "absl-py",
          version = "0.8.1",
          hash = NULL,
          bundle_id = "6623",
          app_id = "3652",
          app_guid = "1935b6cb",
          content_id = "3652",
          content_guid = "1935b6cb"
        )
      )
    )
  )

  expect_identical(
    get_packages(client),
    tibble::tibble(
      language = c("python", "python"),
      language_version = c("3.7.6", "3.7.7"),
      name = c("absl-py", "absl-py"),
      version = c("0.12.0", "0.8.1"),
      hash = c(NA_character_, NA_character_),
      bundle_id = c("9375", "6623"),
      content_id = c("4906", "3652"),
      content_guid = c("9bf33774", "1935b6cb")
    )
  )
})

test_that("get_content only requests vanity URLs for Connect 2024.06.0 and up", {
  with_mock_dir("2024.05.0", {
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    # `$version` is lazy, so we need to call it before `without_internet()`.
    client$version
  })
  without_internet({
    expect_GET(
      get_content(client),
      "http://connect.example/__api__/v1/content?include=tags%2Cowner"
    )
  })

  with_mock_dir("2024.06.0", {
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    # `$version` is lazy, so we need to call it before `without_internet()`.
    client$version
  })
  without_internet({
    expect_GET(
      get_content(client),
      "http://connect.example/__api__/v1/content?include=tags%2Cowner"
    )
  })

  with_mock_dir("2024.07.0", {
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    # `$version` is lazy, so we need to call it before `without_internet()`.
    client$version
  })
  without_internet({
    expect_GET(
      get_content(client),
      "http://connect.example/__api__/v1/content?include=tags%2Cowner%2Cvanity_url"
    )
  })
})

with_mock_dir("2025.04.0", {
  test_that("get_usage() returns usage data in the expected shape", {
    client <- connect(server = "https://connect.example", api_key = "fake")
    usage <- get_usage(
      client,
      from = as.POSIXct("2025-04-01 00:00:01", tz = "UTC")
    )

    expect_s3_class(usage, "connect_list_hits")
    expect_s3_class(usage, "list")

    expect_length(usage, 5)

    # Check first element (raw list, before conversion to data.frame).
    # The id is numeric in the JSON, so it stays numeric in the raw list.
    expect_equal(
      usage[[1]],
      list(
        id = 8966707L,
        user_guid = NULL,
        content_guid = "475618c9",
        timestamp = "2025-04-30T12:49:16.269904Z",
        data = list(
          path = "/hello",
          user_agent = "Datadog/Synthetics"
        )
      )
    )

    # Check conversion to data.frame
    usage_df <- as.data.frame(usage)
    expect_equal(
      usage_df,
      data.frame(
        id = c("8966707", "8966708", "8967206", "8967210", "8966214"),
        user_guid = c(NA, NA, NA, NA, "fecbd383"),
        content_guid = c(
          "475618c9",
          "475618c9",
          "475618c9",
          "475618c9",
          "b0eaf295"
        ),
        timestamp = c(
          parse_connect_rfc3339(c(
            "2025-04-30T12:49:16.269904Z",
            "2025-04-30T12:49:17.002848Z",
            "2025-04-30T13:01:47.40738Z",
            "2025-04-30T13:04:13.176791Z",
            "2025-04-30T12:36:13.818466Z"
          ))
        ),
        path = c("/hello", "/world", "/chinchilla", "/lava-lamp", NA),
        user_agent = c(
          "Datadog/Synthetics",
          NA,
          "Datadog/Synthetics",
          "Datadog/Synthetics",
          NA
        )
      )
    )

    # Check conversion with unnest=FALSE
    usage_df_no_unnest <- as.data.frame(usage, unnest = FALSE)
    expect_equal(
      names(usage_df_no_unnest),
      c("id", "user_guid", "content_guid", "timestamp", "data")
    )
  })

  test_that("Metrics firehose is called with expected parameters", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    # $version is loaded lazily, we need it before calling get_usage()
    client$version

    without_internet({
      expect_GET(
        get_usage(client),
        "https://connect.example/__api__/v1/instrumentation/content/hits"
      )
      expect_GET(
        get_usage(
          client,
          from = as.POSIXct("2025-04-01 00:00:01", tz = "UTC"),
          to = as.POSIXct("2025-04-02 00:00:01", tz = "UTC")
        ),
        paste0(
          "https://connect.example/__api__/v1/instrumentation/content/hits?",
          "from=2025-04-01T00%3A00%3A01Z&to=2025-04-02T00%3A00%3A01Z"
        )
      )
    })
  })
})

with_mock_dir("2026.01.0", {
  test_that("content_guid is passed to the hits endpoint", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    # $version is loaded lazily, we need it before calling get_usage()
    client$version

    without_internet({
      expect_GET(
        get_usage(
          client,
          content_guid = "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"
        ),
        paste0(
          "https://connect.example/__api__/v1/instrumentation/content/hits?",
          "content_guid=aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"
        )
      )
      expect_GET(
        get_usage(
          client,
          content_guid = c(
            "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
            "ffffffff-1111-2222-3333-444444444444"
          )
        ),
        paste0(
          "https://connect.example/__api__/v1/instrumentation/content/hits?",
          "content_guid=aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
          "%7Cffffffff-1111-2222-3333-444444444444"
        )
      )
    })
  })
})

cont1_name <- uuid::UUIDgenerate()
cont1_title <- "Test Content 1"
cont1_guid <- NULL
cont1_bundle <- NULL
cont1_content <- NULL

# get --------------------------------------------

test_that("get_users works", {
  users <- get_users(client)

  expect_s3_class(users, c("tbl_df", "tbl", "data.frame"))
  expect_ptype_equal(users, connectapi_ptypes$users, exact = FALSE)

  # Other tests create users, so specifying the exact number here is conditional
  # on the contents of other tests and the order that tests run in.
  admins <- get_users(client, user_role = "administrator")
  expect_true(nrow(admins) > 0)

  licensed <- get_users(client, account_status = "licensed")
  expect_true(nrow(licensed) > 0)
})

test_that("get_groups works", {
  groups_list <- get_groups(client)
  expect_s3_class(groups_list, c("tbl_df", "tbl", "data.frame"))

  expect_ptype_equal(groups_list, connectapi_ptypes$groups)
})

test_that("get_content works", {
  scoped_experimental_silence()
  content_list <- get_content(client)
  expect_s3_class(content_list, c("tbl_df", "tbl", "data.frame"))

  # various attributes have been added over the years, so exact match
  # doesn't work against all versions of Connect
  expect_ptype_equal(content_list, connectapi_ptypes$content, exact = FALSE)
})

test_that("get_usage_shiny works", {
  shiny_usage <- get_usage_shiny(client)
  expect_s3_class(shiny_usage, c("tbl_df", "tbl", "data.frame"))

  expect_ptype_equal(shiny_usage, connectapi_ptypes$usage_shiny)
})

test_that("get_usage_static works", {
  content_visits <- get_usage_static(client)
  expect_s3_class(content_visits, c("tbl_df", "tbl", "data.frame"))

  # path was added to usage_static in 2024
  expect_ptype_equal(
    content_visits,
    connectapi_ptypes$usage_static,
    exact = FALSE
  )
})

test_that("get_audit_logs works", {
  audit_list <- get_audit_logs(client)
  expect_s3_class(audit_list, c("tbl_df", "tbl", "data.frame"))

  # This is different on older versions, not sure it's worth worrying about how
  skip_if_connect_older_than(client, "2022.09.0")
  expect_ptype_equal(audit_list, connectapi_ptypes$audit_logs)
})

test_that("get_procs works", {
  scoped_experimental_silence()
  proc_data <- get_procs(client)

  # TODO: This is not a great test, since no processes are running
  # we could always start a content restoration...
  expect_s3_class(proc_data, "tbl_df")
  expect_ptype_equal(proc_data, connectapi_ptypes$procs)
})

# experimental --------------------------------------------

test_that("content_list_with_permissions works", {
  scoped_experimental_silence()

  rlang::with_options(
    progress_enabled = FALSE,
    cl <- content_list_with_permissions(client)
  )

  expect_true("permission" %in% names(cl))
  expect_s3_class(cl, "tbl_df")
})

test_that("content_list_with_permissions predicate works", {
  scoped_experimental_silence()

  # deploy a static app so we know it is not empty
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  uniq_id <- uuid::UUIDgenerate()
  deployed <- deploy(client, bnd, uniq_id)

  rlang::with_options(
    progress_enabled = FALSE,
    cl <- content_list_with_permissions(
      client,
      .p = ~ .x$guid == deployed$content$guid
    )
  )

  expect_true("permission" %in% names(cl))
  expect_s3_class(cl, "tbl_df")
  expect_equal(nrow(cl), 1)
})

test_that("content_list_guid_has_access works", {
  scoped_experimental_silence()

  # deploy a static app so we know it is not empty
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  uniq_id <- uuid::UUIDgenerate()
  deployed <- deploy(client, bnd, uniq_id)

  rlang::with_options(
    progress_enabled = FALSE,
    cl <- content_list_with_permissions(client)
  )

  my_guid <- client$me()$guid
  filt <- content_list_guid_has_access(cl, my_guid)
  expect_true("permission" %in% names(filt))
  expect_true(nrow(filt) <= nrow(cl))

  expect_true(nrow(filt) > 0)
  expect_true(deployed$content$guid %in% filt$guid)
})

test_that("content_list_by_tag works", {
  skip("not yet tested")
})

library(testthat)
library(connectapi)

if (nzchar(Sys.getenv("CONNECTAPI_INTEGRATED"))) {
  # Make sure one server is up
  tryCatch(
    httr::content(httr::GET(paste0(Sys.getenv("CONNECT_SERVER"), "/__ping__"))),
    error = function(e) {
      stop("Server 1 is not healthy")
    }
  )

  test_dir(
    rprojroot::find_package_root_file("tests/integrated"),
    reporter = "summary",
    package = "connectapi",
    load_package = "installed"
  )
} else {
  message(
    "Not running integration tests. Set environment variable CONNECTAPI_INTEGRATED=true to run integration tests"
  )
}

structure(
  list(
    url = "__api__/applications/23456789/image",
    status_code = 404L,
    content = list(
      code = 4,
      error = "The requested object does not exist.",
      payload = NULL
    )
  ),
  class = "response"
)

#' Check to see if a vanity URL is currently in use
#'
#' `r lifecycle::badge('experimental')`
#'
#' @param connect A Connect R6 object
#' @param vanity string of the vanity URL to check
#'
#' @return logical indicating if the vanity URL is available.
#'
#' @family audit functions
#' @export
vanity_is_available <- function(connect, vanity) {
  current_vanities <- connect$vanities()
  current_vanity_paths <- purrr::map_chr(current_vanities, "path")

  # In case a full URL has been given, prune it down to just the path
  vanity <- sub(paste0("^", connect$server), "", vanity)
  # and make sure it has a leading and trailing slash
  vanity <- sub("^/?(.*[^/])/?$", "/\\1/", vanity)

  !(vanity %in% current_vanity_paths)
}

#' Audit R Versions
#'
#' `r lifecycle::badge('experimental')`
#'
#' @param content `data.frame` of content information, as from [get_content()]
#'
#' @return A plot that shows the R version used by content over time and in
#'   aggregate.
#' @family audit functions
#' @export
audit_r_versions <- function(content) {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("ggplot2 is required for this function")
  }
  if (!requireNamespace("gridExtra", quietly = TRUE)) {
    stop("gridExtra is required for this function")
  }

  timeline <- content[
    !is.na(content$r_version),
    c("r_version", "last_deployed_time")
  ]

  # histogram
  p1 <- ggplot2::ggplot(timeline) +
    ggplot2::geom_bar(ggplot2::aes(r_version)) +
    ggplot2::theme_minimal() +
    ggplot2::labs(
      title = "Distribution of Content by R Version",
      x = NULL,
      y = NULL
    )

  # timeline
  p2 <- ggplot2::ggplot(timeline) +
    ggplot2::geom_point(
      pch = 4,
      ggplot2::aes(x = last_deployed_time, color = r_version, y = r_version)
    ) +
    ggplot2::theme_minimal() +
    ggplot2::labs(
      title = "Content by Time",
      x = "Last Updated",
      y = NULL,
      color = ""
    )

  gridExtra::grid.arrange(p1, p2, ncol = 2)
}

#' Audit Run As Settings
#'
#' `r lifecycle::badge('experimental')`
#'
#' @param content `data.frame` of content information, as from [get_content()]
#'
#' @return A data frame with the app name and the Run As user if the Run As user
#'   is not the default
#' @family audit functions
#' @export
audit_runas <- function(content) {
  content$run_as <- ifelse(
    content$run_as_current_user,
    "current user",
    content$run_as
  )
  content <- content[!is.na(content$run_as), c("name", "run_as")]
  names(content) <- c("app_name", "run_as_user")
  content
}

# type can be all, logged_in, acl
#' Audit Access Controls
#'
#' `r lifecycle::badge('experimental')`
#'
#' @param content `data.frame` of content information, as from [get_content()]
#' @param type One of "all" or "logged_in". If "all", return a list of apps
#'   whose access control is set to "Everyone". If "logged_in", return a list of
#'   apps whose access control is set to "All logged in users"
#'
#' @family audit functions
#' @export
audit_access_open <- function(content, type = "all") {
  content$name[content$access_type == type]
}

without_internet({
  test_that("audit_logs query params", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$audit_logs(),
      "https://connect.example/__api__/v1/audit_logs?limit=500&ascOrder=true"
    )

    expect_GET(
      con$audit_logs(limit = 1000, previous = "asdf", nxt = "qwer"),
      "https://connect.example/__api__/v1/audit_logs?limit=500&previous=asdf&ascOrder=true&next=qwer"
    )
  })
})

with_mock_api({
  test_that("audit_r_versions does not error", {
    skip_if_not_installed("ggplot2")
    skip_if_not_installed("gridExtra")
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- get_content(con)

    withr::with_pdf(nullfile(), {
      expect_s3_class(audit_r_versions(content), "gtable")
    })
  })

  test_that("audit_access_open", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- get_content(con)

    expect_identical(audit_access_open(content, "all"), "My-Streamlit-app")
    expect_identical(
      audit_access_open(content, "logged_in"),
      c("team-admin-dashboard", "Performance-Data-1671216053560")
    )
  })

  test_that("audit_runas", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    content <- get_content(con)
    expect_equal(
      audit_runas(content),
      tibble::tibble(
        app_name = "My-Streamlit-app",
        run_as_user = "current user"
      )
    )
  })

  test_that("vanity_is_available", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(vanity_is_available(con, "not-currently/used"))
    expect_false(vanity_is_available(con, "/streamlit/my-app/"))
    expect_false(vanity_is_available(con, "streamlit/my-app"))
    expect_false(vanity_is_available(
      con,
      "https://connect.example/streamlit/my-app"
    ))
  })
})

#
# This is a Shiny web application. You can run the application by clicking
# the 'Run App' button above.
#
# Find out more about building applications with Shiny here:
#
#    http://shiny.rstudio.com/
#

library(shiny)

# Define UI for application that draws a histogram
ui <- fluidPage(
  # Application title
  titlePanel("Old Faithful Geyser Data"),

  # Sidebar with a slider input for number of bins
  sidebarLayout(
    sidebarPanel(
      sliderInput("bins", "Number of bins:", min = 1, max = 50, value = 30)
    ),

    # Show a plot of the generated distribution
    mainPanel(
      plotOutput("distPlot")
    )
  )
)

# Define server logic required to draw a histogram
server <- function(input, output) {
  output$distPlot <- renderPlot({
    # generate bins based on input$bins from ui.R
    x <- faithful[, 2]
    bins <- seq(min(x), max(x), length.out = input$bins + 1)

    # draw the histogram with the specified number of bins
    hist(x, breaks = bins, col = "darkgray", border = "white")
  })
}

# Run the application
shinyApp(ui = ui, server = server)

#' Git Repository Helpers
#'
#' `r lifecycle::badge('experimental')` These functions help use Posit Connect's
#' configured authorization to query available branches and subdirectories for
#' deployment using `deploy_repo()`
#'
#' - `repo_check_account()` messages whether an account is in use, and then returns that account
#' - `repo_check_branches()` retrieves which branches are available, returning in a named list
#' - `repo_check_manifest_dirs()` retrieves which directories contain a
#'   `manifest.json`, returning in a named list
#'
#' @param client A Connect R6 object
#' @param host The git repository host (with schema). For example, "https://github.com"
#' @param repository The git repository to explore or consider deploying
#' @param branch The git branch to explore for subdirectories
#'
#' @rdname git
#' @name git
#' @seealso connectapi::deploy_repo
#' @family content functions
#' @export
repo_check_account <- function(client, host) {
  validate_R6_class(client, "Connect")
  warn_experimental("git")
  account <- client$repo_account(host)
  # TODO: add messaging about which account is being used...
  if (nchar(account$username) > 0) {
    message(glue::glue("Accessing '{host}' with account: {account$username}"))
  } else {
    message(glue::glue("Accessing '{host}' anonymously."))
  }
  return(account)
}

#' @rdname git
#' @export
repo_check_branches <- function(client, repository) {
  validate_R6_class(client, "Connect")
  warn_experimental("git")
  branches_raw <- client$repo_branches(repository)
  branches_task <- Task$new(connect = client, task = branches_raw)

  task_res <- poll_task(branches_task, callback = NULL)
  branch_entries <- parse_repo_branches(task_res$get_data())
  branches <- purrr::map(branch_entries, ~ .x$branch)
  purrr::set_names(branches, branches)
}

#' @rdname git
#' @export
repo_check_branches_ref <- function(client, repository) {
  validate_R6_class(client, "Connect")
  warn_experimental("git")
  branches_raw <- client$repo_branches(repository)
  branches_task <- Task$new(connect = client, task = branches_raw)

  task_res <- poll_task(branches_task, callback = NULL)
  branch_entries <- parse_repo_branches(task_res$get_data())
  branches <- purrr::map(branch_entries, ~ .x$branch)
  refs <- purrr::map_chr(branch_entries, ~ .x$ref)
  purrr::set_names(refs, branches)
}

# Gracefully handle different formats of the /repo/branches task result
parse_repo_branches <- function(task_data) {
  if (identical(task_data$type, "git-repo-ref-branch-array")) {
    task_data$data
  } else if (identical(task_data$type, "git-repo-branches-result")) {
    task_data$data$branches
  } else {
    stop(
      glue::glue(
        "Unexpected task result type '{task_data$type}' returned from the ",
        "/repo/branches endpoint."
      )
    )
  }
}

#' @rdname git
#' @export
repo_check_manifest_dirs <- function(client, repository, branch) {
  validate_R6_class(client, "Connect")
  warn_experimental("git")
  manifest_dirs_raw <- client$repo_manifest_dirs(
    repo = repository,
    branch = branch
  )
  manifest_dirs_task <- Task$new(connect = client, task = manifest_dirs_raw)

  task_res <- poll_task(manifest_dirs_task, callback = NULL)
  task_data <- task_res$get_data()
  stopifnot(identical(task_data$type, "git-repo-branch-manifest-dirs-array"))
  manifest_dirs <- purrr::map(task_data$data, ~.x)
  purrr::set_names(manifest_dirs, manifest_dirs)
}

#' Deploy a Git Repository
#'
#' `r lifecycle::badge('experimental')` Deploy a git repository directly to Posit Connect,
#' using Posit Connect's "pull-based" "git-polling" method of deployment.
#'
#' - `deploy_repo_enable()` enables (or disables) Posit Connect's git polling for a piece of content
#' - `deploy_repo_update()` triggers an update of the content from its git
#'   repository, if any are present
#'
#' @param client A Connect R6 object
#' @param repository The git repository to deploy
#' @param branch The git branch to deploy
#' @param subdirectory The subdirectory to deploy (must contain a `manifest.json`)
#' @param name The "name" / unique identifier for the content. Defaults to a random character string
#' @param title The "title" of the content
#' @param content An R6 Content object (i.e. the result of `content_item()`)
#' @param enabled Whether Connect will enable automatic polling for repository updates
#' @param ... Additional options for defining / specifying content attributes
#'
#' @return A ContentTask object, for use with `poll_task()` (if you want to follow the logs)
#'
#' @seealso connectapi::poll_task, connectapi::repo_check_branches, connectapi::repo_check_manifest_dirs
#'
#' @family content functions
#' @rdname deploy_repo
#' @export
deploy_repo <- function(
  client,
  repository,
  branch,
  subdirectory,
  name = create_random_name(),
  title = name,
  ...
) {
  validate_R6_class(client, "Connect")

  content_metadata <- content_ensure(
    connect = client,
    name = name,
    title = title,
    ...,
    .permitted = c("new")
  )

  deployed_content <- content_item(client, content_metadata$guid)
  deployed_content$repo_set(
    repository = repository,
    branch = branch,
    directory = subdirectory
  )

  task <- deployed_content$deploy()

  ContentTask$new(connect = client, content = content_metadata, task = task)
}

#' @rdname deploy_repo
#' @export
deploy_repo_enable <- function(content, enabled = TRUE) {
  validate_R6_class(content, "Content")

  content$repo_enable(enabled)
  content$get_content_remote()
  content
}

#' @rdname deploy_repo
#' @export
deploy_repo_update <- function(content) {
  validate_R6_class(content, "Content")

  con <- content$connect
  repo_data <- content$repository()
  if (is.null(repo_data)) {
    stop(glue::glue(
      "Content item '{content$content$guid}' is not git-backed content"
    ))
  }
  branch_status <- repo_check_branches_ref(con, repo_data$repository)

  if (!repo_data$branch %in% names(branch_status)) {
    stop(glue::glue(
      "Branch '{repo_data$branch}' was no longer found on repository '{repo_data$repository}'"
    ))
  }
  if (
    identical(repo_data$last_known_commit, branch_status[[repo_data$branch]])
  ) {
    message(glue::glue(
      "No changes were found in the Git repository: {repo_data$repository}@{repo_data$branch}"
    ))
    return(content)
  }
  task <- content$deploy()

  ContentTask$new(connect = con, content = content$content, task = task)
}

skip_if_connect_older_than(client, "2022.12.0")

cont1_name <- uuid::UUIDgenerate()
cont1_title <- "Test Content 1"
cont1_guid <- NULL
cont1_bundle <- NULL

test_that("git deployment works", {
  scoped_experimental_silence()
  cont0 <- deploy_repo(
    client,
    "https://github.com/posit-dev/connectapi",
    "main",
    "tests/testthat/examples/static"
  )
  expect_true(validate_R6_class(cont0, "Content"))
  expect_true(validate_R6_class(cont0, "ContentTask"))

  new_name <- uuid::UUIDgenerate()
  cont1 <- deploy_repo(
    client,
    "https://github.com/posit-dev/connectapi",
    "main",
    "tests/testthat/examples/static",
    new_name,
    cont1_title
  )
  expect_true(validate_R6_class(cont1, "Content"))
  expect_true(validate_R6_class(cont1, "ContentTask"))

  deploy_again <- deploy_current(cont1)
  expect_true(validate_R6_class(deploy_again, "Content"))
  expect_true(validate_R6_class(deploy_again, "ContentTask"))
})

test_that("repo_check_account works", {
  scoped_experimental_silence()
  expect_message(
    acc <- repo_check_account(client, "https://github.com"),
    "anonymous"
  )

  expect_true(nchar(acc$username) == 0)
})

test_that("repo_check_branches works", {
  scoped_experimental_silence()
  expect_message(
    expect_error(repo_check_branches(client, "https://github.com")),
    "not found"
  )

  br <- repo_check_branches(
    client,
    "https://github.com/posit-dev/connectapi"
  )
  expect_true("main" %in% br)
})

test_that("repo_check_branches_ref works", {
  scoped_experimental_silence()
  expect_message(
    expect_error(repo_check_branches_ref(client, "https://github.com")),
    "not found"
  )

  br <- repo_check_branches_ref(
    client,
    "https://github.com/posit-dev/connectapi"
  )
  expect_type(br, "character")
  expect_true("main" %in% names(br))
  expect_false("main" %in% br)
  expect_true(all(nchar(br) == 40))
})

test_that("repo_check_manifest_dirs works", {
  scoped_experimental_silence()
  expect_message(
    expect_error(repo_check_manifest_dirs(
      client,
      "https://github.com",
      "main"
    )),
    "not found"
  )

  drs <- repo_check_manifest_dirs(
    client,
    "https://github.com/posit-dev/connectapi",
    "main"
  )
  expect_true("tests/testthat/examples/static" %in% drs)
})

test_that("deploy_repo_enable works", {
  scoped_experimental_silence()

  new_name <- uuid::UUIDgenerate()
  cont1 <- deploy_repo(
    client,
    "https://github.com/posit-dev/connectapi",
    "main",
    "tests/testthat/examples/static",
    new_name,
    "deploy_repo_enable test"
  )
  expect_true(validate_R6_class(cont1, "Content"))
  expect_true(validate_R6_class(cont1, "ContentTask"))

  polling_enabled <- function(content) {
    git <- content$repository()
    isTRUE(git$polling)
  }
  # TODO: flaky... how to be safer?
  Sys.sleep(5) # sleep for deployment...?
  expect_true(polling_enabled(cont1))
  res <- deploy_repo_enable(cont1, FALSE)
  expect_false(polling_enabled(cont1))
  res <- deploy_repo_enable(cont1, TRUE)
  expect_true(polling_enabled(cont1))
})

test_that("deploy_repo_update works", {
  scoped_experimental_silence()

  # this is really hard to test... because we need a git repo that changes

  new_name <- uuid::UUIDgenerate()
  cont1 <- deploy_repo(
    client,
    "https://github.com/posit-dev/connectapi",
    "main",
    "tests/testthat/examples/static",
    new_name,
    "deploy_repo_update test - good"
  )
  expect_true(validate_R6_class(cont1, "Content"))
  expect_true(validate_R6_class(cont1, "ContentTask"))

  res <- deploy_repo_update(cont1)
  expect_true(validate_R6_class(res, "Content"))

  wrong_branch <- uuid::UUIDgenerate()
  cont2 <- deploy_repo(
    client,
    "https://github.com/posit-dev/connectapi",
    "master-not-a-real-branch",
    "tests/testthat/examples/static",
    wrong_branch,
    "deploy_repo_update test - wrong branch"
  )
  expect_true(validate_R6_class(cont2, "Content"))
  expect_true(validate_R6_class(cont2, "ContentTask"))

  res <- expect_error(deploy_repo_update(cont2), "master-not-a-real-branch")

  not_git_name <- uuid::UUIDgenerate()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  cont3 <- deploy(client, bnd, not_git_name)

  expect_error(deploy_repo_update(cont3), "not git-backed content")
})

structure(
  list(
    url = "__api__/v1/oauth/integrations",
    status_code = 400L,
    headers = structure(
      list(`content-type` = "application/json"),
      class = c("insensitive", "list")
    ),
    content = charToRaw(
      "{\"code\":228,\"error\":\"The config key max_role must be one of (Viewer, Publisher, Admin)\",\"payload\":null}"
    )
  ),
  class = "response"
)

structure(
  list(
    url = "__api__/v1/applications/34567890/thumbnail",
    status_code = 404L,
    content = list(
      code = 17,
      error = "The requested item could not be removed because it does not exist.",
      payload = NULL
    )
  ),
  class = "response"
)

library(testthat)
library(connectapi)

multi_reporter <- MultiReporter$new(
  reporters = list(
    ProgressReporter$new(max_failures = 1000),
    CheckReporter$new(file = fs::path("check-results-check.txt"))
  )
)

test_check("connectapi", reporter = multi_reporter)

#' VariantSchedule
#'
#' An R6 class that represents a Schedule
#' @family R6 classes
#' @export
VariantSchedule <- R6::R6Class(
  "VariantSchedule",
  # TODO: would be cool to have multiple inheritance...
  inherit = Variant,
  public = list(
    #' @field schedule_data The schedule data.
    schedule_data = NULL,
    #' @description Initialize this schedule.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    #' @param key The variant key.
    #' @param schedule The schedule data.
    initialize = function(connect, content, key, schedule) {
      super$initialize(connect = connect, content = content, key = key)
      # TODO: need to validate schedule (needs an ID)
      self$schedule_data <- schedule
    },
    #' @description Perform an HTTP GET request of the named API path. Returns an object parsed from the HTTP response.
    #' @param path API path.
    GET = function(path) {
      self$connect$GET(path)
    },
    #' @description Perform an HTTP POST request of the named API path. Returns an object parsed from the HTTP response.
    #' @param path API path.
    #' @param body The HTTP payload.
    POST = function(path, body) {
      self$connect$POST(path = path, body = body)
    },
    #' @description Perform an HTTP DELETE request of the named API path. Returns the HTTP response object.
    #' @param path API path.
    DELETE = function(path) {
      self$connect$DELETE(path = path)
    },
    #' @description Set the schedule for this variant
    #' @param ... Schedule fields.
    set_schedule = function(...) {
      warn_experimental("set_schedule")
      params <- rlang::list2(...)
      if ("start_time" %in% names(params)) {
        params$start_time <- make_timestamp(params$start_time)
      }
      if ("next_run" %in% names(params)) {
        params$next_run <- make_timestamp(params$next_run)
      }
      if (self$is_empty()) {
        params <- purrr::list_modify(
          params,
          app_id = self$variant$app_id,
          variant_id = self$variant$id
        )
        path <- unversioned_url("schedules")
      } else {
        path <- unversioned_url("schedules", self$get_schedule()$id)
      }
      cli <- self$connect
      res <- cli$POST(path = path, body = params)

      self$schedule_data <- res
      return(self)
    },
    #' @description Return if this variant has a schedule.
    is_empty = function() {
      if (length(self$schedule_data) == 0) {
        TRUE
      } else {
        FALSE
      }
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      super$print(...)
      cat("Schedule:\n")
      cat("  get_variant_schedule(variant)\n\n")
      if (self$is_empty()) {
        cat("  WARNING: No schedule defined\n")
      } else {
        cat(c("", paste0(" ", self$describe_schedule(), "\n")))
      }
    },
    #' @description Get the schedule data.
    get_schedule = function() {
      return(self$schedule_data)
    },
    #' @description Get and store the (remote) schedule data.
    get_schedule_remote = function() {
      sch <- super$get_schedule_remote()
      self$schedule_data <- sch
      return(self$schedule_data)
    },
    #' @description Description of the associated schedule.
    describe_schedule = function() {
      # TODO: create a human readable description of schedule
      if (!self$is_empty()) {
        rawdata <- self$get_schedule()
        schdata <- jsonlite::fromJSON(rawdata$schedule)
        # TODO: translate dayofweek "Days" to something more usable
        plural <- ifelse(
          ifelse(is.null(schdata$N), FALSE, schdata$N > 1),
          "s",
          ""
        )
        desc <- switch(
          rawdata$type,
          "minute" = glue::glue("Every {schdata$N} minute{plural}"),
          "hour" = glue::glue("Every {schdata$N} hour{plural}"),
          "day" = glue::glue("Every {schdata$N} day{plural}"),
          "weekday" = glue::glue("Every weekday"),
          "week" = glue::glue("Every {schdata$N} week{plural}"),
          "dayofweek" = glue::glue(
            "On week days {glue::glue_collapse(schdata$Days, ', ')}"
          ),
          "semimonth" = ifelse(
            schdata$First == "TRUE",
            "On the 1st and 15th of each month",
            "On the 14th and Last day of each month"
          ),
          "dayofmonth" = glue::glue(
            "Every {schdata$N} month{plural} on day {schdata$Day}"
          ),
          "dayweekofmonth" = glue::glue(
            "Every {schdata$N} month{plural} on week {schdata$Week}, day {schdata$Day}"
          ),
          "year" = glue::glue("Every {schdata$N} year{plural}"),
          "Unknown schedule"
        )
        tz_offset <- .get_offset(self$connect, rawdata$timezone)
        c(
          desc,
          # TODO: a nice way to print out relative times...?
          glue::glue("Starting {rawdata$start_time} ({tz_offset})"),
          glue::glue("Next Run {rawdata$next_run} ({tz_offset})")
        )
      }
    }
  )
)

.get_offset <- function(connect, timezone) {
  res <- purrr::keep(connect$timezones, ~ .x[["timezone"]] == timezone)
  if (length(res) != 1) {
    stop(glue::glue("ERROR: timezone '{timezone}' not found"))
  }
  return(res[[1]][["offset"]])
}

#' Get a Variant Schedule
#'
#' `r lifecycle::badge('experimental')` Gets the schedule associated with a Variant.
#'
#' @param variant A Variant object, as returned by `get_variant()` or `get_variant_default()`
#'
#' @return A VariantSchedule object
#'
#' @rdname get_variant_schedule
#' @family schedule functions
#' @export
get_variant_schedule <- function(variant) {
  warn_experimental("get_schedule")
  scoped_experimental_silence()
  validate_R6_class(variant, "Variant")

  content_details <- variant$content
  connect_client <- variant$connect

  variant_key <- variant$key
  variant_schedule <- variant$get_schedule_remote()

  VariantSchedule$new(
    connect = connect_client,
    content = content_details,
    key = variant_key,
    schedule = variant_schedule
  )
}

#' Set a Schedule
#'
#' `r lifecycle::badge('experimental')` Sets the schedule for a given Variant. Requires a
#' `Schedule` object (as returned by `get_variant_schedule()`)
#'
#' - `set_schedule()` is a raw interface to Posit Connect's `schedule` API
#' - `set_schedule_*()` functions provide handy wrappers around `set_schedule()`
#' - `set_schedule_remove()` removes a schedule / un-schedules a variant
#'
#' Beware, using `set_schedule()` currently uses the Posit Connect `schedule` API
#' directly, and so can be a little clunky. Using the `set_schedule_*()` is generally
#' recommended.
#'
#' @param .schedule A schedule object. As returned by `get_variant_schedule()`
#' @param email Whether to send emails on this schedule
#' @param activate `r lifecycle::badge('deprecated')` This parameter no longer
#'   has any effect due to changes in the Connect API and will be removed in a
#'   future release.
#' @param schedule A JSON blob (as a string) describing the schedule. See "More Details"
#' @param start_time The start time of the schedule
#' @param n The "number of" iterations
#' @param day The day of the week (0-6) or day of the month (0-31)
#' @param days The days of the week (0-6)
#' @param week The week of the month (0-5)
#' @param first [logical] Whether to execute on the 1st and 15th (TRUE) or 14th and last (FALSE)
#' @param timezone The timezone to use for setting the schedule. Defaults to `Sys.timezone()`
#' @param ... Scheduling parameters
#'
#' @return An updated Schedule object
#'
#' @rdname set_schedule
#' @family schedule functions
#' @export
set_schedule <- function(
  .schedule,
  ...
) {
  warn_experimental("set_schedule")
  scoped_experimental_silence()
  validate_R6_class(.schedule, "VariantSchedule")
  params <- rlang::list2(...)

  if ("activate" %in% names(params)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule(activate)")
    params$activate <- NULL
  }

  # TODO: check whether this schedule actually exists...
  # TODO: fix capitalization if "day" or "days" or "first" or "n" is provided
  # TODO: if type = "weekday", make sure "schedule" is turned into a {} JSON blob properly

  # because "schedule" has to be a JSON blob, which is confusing
  if ("schedule" %in% names(params)) {
    orig_schedule <- params$schedule
    if (is.list(params$schedule)) {
      params$schedule <- jsonlite::toJSON(params$schedule, auto_unbox = TRUE)
    }
    if (!(is_string(params$schedule) && jsonlite::validate(params$schedule))) {
      stop(glue::glue(
        "The schedule you provided is invalid: {capture.output(str(orig_schedule))}"
      ))
    }
  }

  if ("type" %in% names(params) && !"schedule" %in% names(params)) {
    warning(
      "Specifying 'type' without 'schedule' can cause unexpected results. ",
      "Different schedule 'type's have different 'schedule' requirements"
    )
  }

  if ("type" %in% names(params) && !params$type %in% schedule_types) {
    stop(glue::glue(
      "Invalid `type` provided. Should be one of `schedule_types`: {params$type}"
    ))
  }

  # update the existing schedule rather than (likely) erroring
  # this could create some weird edge cases... (see warnings / errors above)
  final_params <- purrr::list_modify(.schedule$get_schedule_remote(), !!!params)

  .schedule$set_schedule(!!!final_params)
}

schedule_types <- c(
  "minute",
  "hour",
  "day",
  "weekday",
  "week",
  "dayofweek",
  "semimonth",
  "dayofmonth",
  "dayweekofmonth",
  "year"
)

#' @rdname set_schedule
#' @export
set_schedule_minute <- function(
  .schedule,
  n = 30,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_minute(activate)")
  }
  set_schedule(
    .schedule,
    type = "minute",
    schedule = list(N = n),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_hour <- function(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_hour(activate)")
  }
  set_schedule(
    .schedule,
    type = "hour",
    schedule = list(N = n),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_day <- function(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_day(activate)")
  }
  set_schedule(
    .schedule,
    type = "day",
    schedule = list(N = n),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_weekday <- function(
  .schedule,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_weekday(activate)")
  }
  set_schedule(
    .schedule,
    type = "weekday",
    schedule = "{}",
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_week <- function(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_week(activate)")
  }
  set_schedule(
    .schedule,
    type = "week",
    schedule = list(N = n),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_dayofweek <- function(
  .schedule,
  days,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_dayofweek(activate)")
  }
  set_schedule(
    .schedule,
    type = "dayofweek",
    schedule = list(Days = days),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_semimonth <- function(
  .schedule,
  first = TRUE,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_semimonth(activate)")
  }
  set_schedule(
    .schedule,
    type = "semimonth",
    schedule = list(First = first),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_dayofmonth <- function(
  .schedule,
  n = 1,
  day = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_dayofmonth(activate)")
  }
  set_schedule(
    .schedule,
    type = "dayofmonth",
    schedule = list(N = n, Day = day),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_dayweekofmonth <- function(
  .schedule,
  n = 1,
  day = 1,
  week = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_dayweekofmonth(activate)")
  }
  set_schedule(
    .schedule,
    type = "dayweekofmonth",
    schedule = list(N = n, Day = day, Week = week),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

#' @rdname set_schedule
#' @export
set_schedule_year <- function(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
) {
  if (!missing(activate)) {
    lifecycle::deprecate_warn("0.11.0", "set_schedule_year(activate)")
  }
  set_schedule(
    .schedule,
    type = "year",
    schedule = list(N = n),
    start_time = start_time,
    email = email,
    timezone = timezone
  )
}

example_schedules <- list(
  list(type = "minute", schedule = list(N = 15)),
  list(type = "hour", schedule = list(N = 2)),
  list(type = "day", schedule = list(N = 3)),
  list(type = "weekday", schedule = "{}"),
  list(type = "week", schedule = list(N = 2)),
  list(type = "dayofweek", schedule = list(Days = list(1))),
  list(type = "dayofweek", schedule = list(Days = list(0, 1, 2, 3, 4, 5, 6))),
  list(type = "semimonth", schedule = list(First = TRUE)),
  list(type = "semimonth", schedule = list(First = FALSE)),
  list(type = "dayofmonth", schedule = list(N = 3, Day = 4)),
  list(type = "dayweekofmonth", schedule = list(N = 3, Day = 1, Week = 4)),
  list(type = "year", schedule = list(N = 2))
)

#' @rdname set_schedule
#' @export
set_schedule_remove <- function(.schedule) {
  validate_R6_class(.schedule, "VariantSchedule")
  cli <- .schedule$connect
  path <- unversioned_url("schedules", .schedule$get_schedule()$id)
  cli$DELETE(path = path)
  get_variant(.schedule, .schedule$key)
}

#' @rdname set_schedule
#' @export
schedule_describe <- function(.schedule) {
  cat(.schedule$describe_schedule(), sep = "\n")
  invisible(.schedule)
}


#' Get TimeZones
#'
#' Get the available timezones from the server.
#'
#' @param connect An R6 Connect object
#'
#' @return A TimeZone vector to be used for setting time zones
#'
#' @family schedule functions
#' @export
get_timezones <- function(connect) {
  raw_tz <- connect$timezones

  tz_values <- purrr::map_chr(raw_tz, ~ .x[["timezone"]])
  tz_display <- purrr::map_chr(
    raw_tz,
    ~ glue::glue("{.x[['timezone']]} ({.x[['offset']]})")
  )

  return(as.list(rlang::set_names(tz_values, tz_display)))
}

test_that("get_timezones() gets timeszones from v1 url when available", {
  client <- MockConnect$new()

  client$mock_response(
    method = "GET",
    path = "v1/timezones",
    content = list(
      list(timezone = "Africa/Abidjan", offset = "+00:00"),
      list(timezone = "Africa/Accra", offset = "+00:00")
    )
  )

  expect_equal(
    get_timezones(client),
    list(
      `Africa/Abidjan (+00:00)` = "Africa/Abidjan",
      `Africa/Accra (+00:00)` = "Africa/Accra"
    )
  )
  expect_equal(
    client$call_log,
    "GET https://connect.example/__api__/v1/timezones"
  )
})

test_that("get_timezones() gets timeszones from unversioned url when v1 returns 404", {
  client <- MockConnect$new()

  client$mock_response(
    method = "GET",
    path = "v1/timezones",
    content = "404 page not found",
    status_code = 404L,
    headers = c("Content-Type" = "text/plain; charset=utf-8")
  )
  client$mock_response(
    method = "GET",
    path = "timezones",
    content = list(
      list(timezone = "Africa/Abidjan", offset = "+00:00"),
      list(timezone = "Africa/Accra", offset = "+00:00")
    )
  )

  expect_equal(
    get_timezones(client),
    list(
      `Africa/Abidjan (+00:00)` = "Africa/Abidjan",
      `Africa/Accra (+00:00)` = "Africa/Accra"
    )
  )
  expect_equal(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/timezones",
      "GET https://connect.example/__api__/timezones"
    )
  )
})

test_that("get_timezones() raises 404 error when v1 and unversioned return 404", {
  client <- MockConnect$new()

  client$mock_response(
    method = "GET",
    path = "v1/timezones",
    content = "404 page not found",
    status_code = 404L,
    headers = c("Content-Type" = "text/plain; charset=utf-8")
  )
  client$mock_response(
    method = "GET",
    path = "timezones",
    content = "404 page not found",
    status_code = 404L,
    headers = c("Content-Type" = "text/plain; charset=utf-8")
  )

  expect_error(
    get_timezones(client),
    regexp = "timezones request failed with Client error"
  )
  expect_equal(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/timezones",
      "GET https://connect.example/__api__/timezones"
    )
  )
})

# Temporarily skip these on newer Connect versions because the manifest for this
# content specifies an old version of R not found in newer images.
skip_if(safe_server_version(client) > "2024.03.0")

rmd_content <- deploy_example(client, "rmarkdown")

## Test Schedule -------------------------------------------------------

d_var <- (function(rmd_content) {
  scoped_experimental_silence()
  get_variant_default(rmd_content)
})(rmd_content)
d_var_sch <- (function(d_var) {
  scoped_experimental_silence()
  get_variant_schedule(d_var)
})(d_var)

invisible(purrr::map(
  example_schedules,
  function(.x, sch) {
    test_that(
      glue::glue("example schedule {.x$type} works"),
      {
        scoped_experimental_silence()
        sch <- get_variant_schedule(sch)
        sch <- set_schedule(sch, !!!.x, start_time = Sys.time())
        expect_true(validate_R6_class(sch, "VariantSchedule"))
        set_schedule_remove(sch)
      }
    )
  },
  sch = d_var_sch
))

test_that("set_schedule_remove is idempotent", {
  scoped_experimental_silence()
  expect_true(validate_R6_class(set_schedule_remove(d_var_sch), "Variant"))
  expect_true(validate_R6_class(set_schedule_remove(d_var_sch), "Variant"))
  expect_true(validate_R6_class(set_schedule_remove(d_var_sch), "Variant"))
})

test_that("schedule helpers work", {
  scoped_experimental_silence()
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_minute(get_variant_schedule(d_var_sch))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_hour(get_variant_schedule(d_var_sch))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_day(get_variant_schedule(d_var_sch))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_weekday(get_variant_schedule(
        d_var_sch
      ))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_week(get_variant_schedule(d_var_sch))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_dayofweek(
        get_variant_schedule(d_var_sch),
        c(0, 1, 6)
      )),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_semimonth(get_variant_schedule(
        d_var_sch
      ))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_semimonth(
        get_variant_schedule(d_var_sch),
        FALSE
      )),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_dayofmonth(get_variant_schedule(
        d_var_sch
      ))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_dayweekofmonth(get_variant_schedule(
        d_var_sch
      ))),
      "Variant"
    )
  )
  expect_true(
    validate_R6_class(
      set_schedule_remove(set_schedule_year(get_variant_schedule(d_var_sch))),
      "Variant"
    )
  )
})

test_that("schedule display works", {
  skip("need a way to make this less time sensitive (with next run)")
  scoped_experimental_silence()

  tzs <- get_timezones(client)

  set_schedule_remove(d_var_sch)
  tmp <- set_schedule_day(
    get_variant_schedule(d_var_sch),
    start_time = as.POSIXct("2022-01-01 00:00:00"),
    n = 2,
    timezone = tzs$`Universal (+00:00)`
  )
  expect_snapshot_output(schedule_describe(tmp))
})

test_that("get_schedules works", {
  scoped_experimental_silence()
  # TODO: add a helper that makes this prettier

  tmp <- set_schedule_day(d_var_sch, n = 5)
  expect_true(validate_R6_class(tmp, "Variant"))

  res <- client$schedules()
  expect_gte(length(res), 1)

  set_schedule_remove(d_var_sch)
})

# Today set to 100MB. Should see if we can get this from Connect
max_bundle_size <- "100M"

#' Bundle
#'
#' An R6 class that represents a bundle
#'
#' @family R6 classes
#' @export
Bundle <- R6::R6Class(
  "Bundle",
  public = list(
    #' @field path The bundle path on disk.
    path = NULL,
    #' @field size The size of the bundle.
    size = NULL,

    #' @description Initialize this content bundle.
    #' @param path The bundle path on disk.
    initialize = function(path) {
      self$path <- path
      self$size <- fs::file_size(path = path)
      if (
        fs::file_exists(path) && self$size > fs::as_fs_bytes(max_bundle_size)
      ) {
        warning(glue::glue(
          "Bundle size is greater than {max_bundle_size}. ",
          "Please ensure your bundle is not including too much."
        ))
      }
    },

    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      cat("Posit Connect Bundle: \n")
      cat("  Path: ", self$path, "\n", sep = "")
      cat("  Size: ", capture.output(self$size), "\n", sep = "")
      cat("\n")
      cat('bundle_path("', self$path, '")', "\n", sep = "")
      cat("\n")
      invisible(self)
    }
  )
)

#' Task
#'
#' An R6 class that represents a Task
#'
#' @family R6 classes
#' @export
Task <- R6::R6Class(
  "Task",
  public = list(
    #' @field connect The Connect instance.
    connect = NULL,
    #' @field task The task.
    task = NULL,
    #' @field data The task data.
    data = NULL,
    #' @description Initialize this task.
    #' @param connect The `Connect` instance.
    #' @param task The task data.
    initialize = function(connect, task) {
      validate_R6_class(connect, "Connect")
      self$connect <- connect
      # TODO: need to validate task (needs task_id)
      if ("id" %in% names(task) && !"task_id" %in% names(task)) {
        # deal with different task interfaces on Connect
        task$task_id <- task$id
      }
      self$task <- task
    },
    #' @description Return the underlying task.
    get_task = function() {
      self$task
    },
    #' @description Set the data.
    #' @param data The data.
    add_data = function(data) {
      self$data <- data
      invisible(self)
    },
    #' @description Get the data.
    get_data = function() {
      self$data
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      cat("Posit Connect Task: \n")
      cat("  Task ID: ", self$get_task()$task_id, "\n", sep = "")
      cat("\n")
      invisible(self)
    }
  )
)

#' ContentTask
#'
#' An R6 class that represents a Task for a piece of Content
#'
#' @family R6 classes
#' @export
ContentTask <- R6::R6Class(
  "ContentTask",
  inherit = Content,
  # implements the "Task" interface too
  public = list(
    #' @field task The task.
    task = NULL,
    #' @field data The task data.
    data = NULL,
    #' @description Initialize this task.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    #' @param task The task data.
    initialize = function(connect, content, task) {
      validate_R6_class(connect, "Connect")
      self$connect <- connect
      # TODO: need to validate content
      self$content <- content
      # TODO: need to validate task (needs task_id)
      self$task <- task
    },
    #' @description Return the underlying task.
    get_task = function() {
      self$task
    },
    #' @description Set the data.
    #' @param data The data.
    add_data = function(data) {
      self$data <- data
      invisible(self)
    },
    #' @description Get the data.
    get_data = function() {
      self$data
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      cat("Posit Connect Content Task: \n")
      cat("  Content GUID: ", self$content$guid, "\n", sep = "")
      cat(
        "  URL: ",
        self$content$dashboard_url,
        "\n",
        sep = ""
      )
      cat("  Task ID: ", self$get_task()$task_id, "\n", sep = "")
      cat("\n")
      invisible(self)
    }
  )
)

#' Vanity
#'
#' An R6 class that represents a Vanity URL
#'
#' @family R6 classes
#' @export
Vanity <- R6::R6Class(
  "Vanity",
  inherit = Content,
  public = list(
    #' @field vanity The vanity.
    vanity = NULL,
    #' @description Initialize this vanity.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    #' @param vanity The vanity data.
    initialize = function(connect, content, vanity) {
      validate_R6_class(connect, "Connect")
      self$connect <- connect
      # TODO: validate content
      self$content <- content
      # TODO: validate vanity (needs path_prefix)
      self$vanity <- vanity
    },
    #' @description Return the underlying vanity.
    get_vanity = function() {
      self$vanity
    },

    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      cat("Posit Connect Content Vanity URL: \n")
      cat("  Content GUID: ", self$content$guid, "\n", sep = "")
      cat("  Vanity URL: ", self$get_vanity()$path, "\n", sep = "")
      cat("\n")
      invisible(self)
    }
  )
)


#' Define a bundle from a Directory
#'
#' Creates a bundle from a target directory.
#'
#' @param path The path to the directory to be bundled
#' @param filename The output bundle path
#'
#' @return Bundle A bundle object
#'
#' @family deployment functions
#' @export
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' bundle_dir(system.file("tests/testthat/examples/shiny/", package = "connectapi"))
bundle_dir <- function(
  path = ".",
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz")
) {
  # TODO: check for manifest.json
  stopifnot(fs::dir_exists(path))
  message(glue::glue("Bundling directory ({path})"))

  before_wd <- getwd()
  setwd(path)
  on.exit(expr = setwd(before_wd), add = TRUE)

  check_bundle_contents(".")
  utils::tar(
    tarfile = filename,
    files = ".",
    compression = "gzip",
    tar = "internal"
  )

  tar_path <- fs::path_abs(filename)

  Bundle$new(path = tar_path)
}

check_bundle_contents <- function(dir) {
  all_contents <- fs::path_file(fs::dir_ls(dir))
  if (!"manifest.json" %in% all_contents) {
    stop(glue::glue(
      "ERROR: no `manifest.json` file found in {dir}. Please generate with `rsconnect::writeManifest()`"
    ))
  }
  if ("packrat.lock" %in% all_contents) {
    warning(glue::glue(
      "WARNING: `packrat.lock` file found in {dir}. This can have unexpected consequences."
    ))
  }
  if ("packrat" %in% all_contents) {
    warning(glue::glue(
      "WARNING: `packrat` directory found in {dir}. This can have unexpected consequences"
    ))
  }
}

#' Define a bundle from a static file (or files)
#'
#' Defines a bundle from static files. It copies all files to a temporary
#' directory, generates a basic manifest file (using the first file as the
#' "primary"), and bundles the directory.
#'
#' NOTE: the `rsconnect` package is required for this function to work properly.
#'
#' @param path The path to a file (or files) that will be used for the static bundle
#' @param filename The output bundle path
#'
#' @return Bundle A bundle object
#'
#' @family deployment functions
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' bundle_static(system.file("logo.png", package = "connectapi"))
#'
#' @export
bundle_static <- function(
  path,
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz")
) {
  tmpdir <- fs::file_temp("bundledir")
  dir.create(tmpdir, recursive = TRUE)
  all_files <- fs::file_copy(path = path, new_path = paste0(tmpdir, "/"))
  rlang::check_installed(
    "rsconnect",
    "the `rsconnect` package needs to be installed to use `bundle_static()`"
  )
  # TODO: error if these files are not static?
  # TODO: a better way to get the primary document besides `all_files[[1]]`?
  rsconnect::writeManifest(
    appDir = tmpdir,
    appPrimaryDoc = fs::path_file(all_files[[1]])
  )
  bundle_dir(tmpdir, filename = filename)
}

#' Define a bundle from a path (a path directly to a tar.gz file)
#'
#' @param path The path to a .tar.gz file
#'
#' @return Bundle A bundle object
#'
#' @family deployment functions
#' @export
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' bundle_path(system.file("tests/testthat/examples/static.tar.gz", package = "connectapi"))
bundle_path <- function(path) {
  # TODO: need a check on filetype
  # TODO: a way to check that the .tar.gz has a manifest.json?
  # TODO: err if the file path does not exist
  tar_path <- fs::path_abs(path)
  message(glue::glue("Bundling path {path}"))

  Bundle$new(path = tar_path)
}

#' Download a Bundle from Deployed Connect Content
#'
#' Downloads a Content item's active bundle, or (optionally) one of its other bundles.
#'
#' @param content A Content object
#' @param filename Optional. The output bundle path
#' @param bundle_id Optional. A string representing the bundle_id to download.
#'   If NULL, will use the currently active bundle.
#' @param overwrite Optional. Default FALSE. Whether to overwrite the target location if it already exists
#'
#' @return Bundle A bundle object
#'
#' @family deployment functions
#' @export
download_bundle <- function(
  content,
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz"),
  bundle_id = NULL,
  overwrite = FALSE
) {
  validate_R6_class(content, "Content")

  from_content <- content$get_content_remote()
  if (is.null(bundle_id)) {
    if (is.null(from_content$bundle_id)) {
      stop(
        glue::glue(
          "This content has no bundle_id.",
          "It has never been successfully deployed.",
          "See {content$get_dashboard_url()} for more information.",
          .sep = " "
        )
      )
    }
    bundle_id <- from_content$bundle_id
  }

  message("Downloading bundle")
  content$bundle_download(
    bundle_id = bundle_id,
    filename = filename,
    overwrite = overwrite
  )

  Bundle$new(path = filename)
}

#' Deploy a bundle
#'
#' Deploys a bundle (tarball) to an Posit Connect server. If not provided,
#' `name` (a unique identifier) will be an auto-generated alphabetic string. If
#' deploying to an existing endpoint, you can set `name` or `guid` to the
#' desired content.
#'
#' This function accepts the same arguments as `connectapi::content_update()`.
#'
#' `deploy_current()` is a helper to easily redeploy the currently active bundle
#' for an existing content item.
#'
#' @param connect A Connect object
#' @param bundle A Bundle object
#' @param name The unique name for the content on the server
#' @param title optional The title to be used for the content on the server
#' @param guid optional The GUID if the content already exists on the server
#' @param ... Additional arguments passed along to the content creation
#' @param .pre_deploy An expression to execute before deploying the new bundle.
#'   The variables `content` and `bundle_id` are supplied
#' @param content A Content object
#'
#' @return Task A task object
#'
#' @seealso connectapi::content_update
#' @family deployment functions
#' @export
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # beware bundling big directories, like `renv/`, `data/`, etc.
#' bnd <- bundle_dir(".")
#'
#' deploy(client, bnd)
#' }
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' client <- connect()
#' bnd <- bundle_path(system.file("tests/testthat/examples/static.tar.gz", package = "connectapi"))
#' deploy(client, bnd)
#'
deploy <- function(
  connect,
  bundle,
  name = create_random_name(),
  title = name,
  guid = NULL,
  ...,
  .pre_deploy = {}
) {
  validate_R6_class(bundle, "Bundle")
  validate_R6_class(connect, "Connect")

  con <- connect

  message("Getting content endpoint")
  content <- content_ensure(
    connect = con,
    name = name,
    title = title,
    guid = guid,
    ...
  )

  message("Uploading bundle")
  # upload
  new_bundle_id <- con$content_upload(
    bundle_path = bundle$path,
    guid = content$guid
  )[["id"]]

  pre_deploy_expr <- rlang::enexpr(.pre_deploy)
  rlang::eval_bare(
    pre_deploy_expr,
    env = rlang::env(
      content = content_item(con, content$guid),
      bundle_id = new_bundle_id
    )
  )

  message("Deploying bundle")
  # deploy
  task <- con$content_deploy(guid = content$guid, bundle_id = new_bundle_id)

  ContentTask$new(connect = con, content = content, task = task)
}

#' @rdname deploy
#' @export
deploy_current <- function(content) {
  validate_R6_class(content, "Content")
  res <- content$deploy()
  return(ContentTask$new(
    connect = content$connect,
    content = content,
    task = res
  ))
}

#' Set the Vanity URL
#'
#' Set the vanity URL for a piece of content.
#'
#' @param content A Content object
#' @param url The path component of the URL
#' @param force optional. Default FALSE. Whether to force-reassign a vanity URL that already exists
#'
#' @return An updated Content object
#'
#' @examples
#' \dontrun{
#' bnd <- bundle_dir("~/my/directory")
#' connect() %>%
#'   deploy(bnd) %>%
#'   set_vanity_url("a/vanity/url")
#' }
#'
#' @family content functions
#' @export
set_vanity_url <- function(content, url, force = FALSE) {
  validate_R6_class(content, "Content")
  con <- content$connect
  error_if_less_than(con$version, "1.8.6")
  guid <- content$content$guid

  scoped_experimental_silence()
  # TODO: Check that the URL provided is appropriate

  res <- con$PUT(
    path = v1_url("content", guid, "vanity"),
    body = list(
      path = url,
      force = force
    )
  )

  Vanity$new(
    connect = con,
    content = content$get_content_remote(),
    vanity = res
  )
}

#' Delete the Vanity URL
#'
#' Delete the vanity URL for a piece of content.
#'
#' @param content A Content object
#'
#' @family content functions
#' @export
delete_vanity_url <- function(content) {
  con <- content$connect
  error_if_less_than(con$version, "1.8.6")
  guid <- content$content$guid

  con$DELETE(v1_url("content", guid, "vanity"), parser = "parsed")

  content
}

#' Get the Vanity URL
#'
#' Get the vanity URL for a piece of content.
#'
#' @param content A Content object
#'
#' @return A character string (or NULL if not defined)
#'
#' @family content functions
#' @export
get_vanity_url <- function(content) {
  validate_R6_class(content, "Content")
  con <- content$connect
  error_if_less_than(con$version, "1.8.6")
  guid <- content$content$guid

  van <- tryCatch(
    {
      con$GET(v1_url("content", guid, "vanity"))
    },
    error = function(e) {
      # TODO: check to ensure that this error was expected
      return(NULL)
    }
  )
  if (is.null(van)) {
    return(NULL)
  }

  return(van$path)
}

#' Swap Vanity URLs
#'
#' Swap the vanity URLs of two pieces of content.
#'
#' @param content_a A Content object
#' @param content_b A Content object
#'
#' @returns A list of the new vanity URLs for `content_a` and `content_b`
#'
#' @family content functions
#' @export
swap_vanity_urls <- function(content_a, content_b) {
  # TODO: Add prompt if in an interactive session
  # TODO: Add pretty print output of what is happening

  validate_R6_class(content_a, "Content")
  validate_R6_class(content_b, "Content")

  vanity_a <- get_vanity_url(content_a)
  vanity_b <- get_vanity_url(content_b)

  if (is.null(vanity_a) && is.null(vanity_b)) {
    warning("Neither content has a vanity URL")
  } else {
    tryCatch(
      delete_vanity_url(content_a),
      error = function(e) {
        stop(
          "Unable to modify the vanity URL for content_a: ",
          e$message,
          call. = FALSE
        )
      }
    )
    tryCatch(
      delete_vanity_url(content_b),
      error = function(e) {
        set_vanity_url(content_a, vanity_a)
        stop(
          "Unable to modify the vanity URL for content_b: ",
          e$message,
          call. = FALSE
        )
      }
    )
    if (!is.null(vanity_a)) {
      set_vanity_url(content_b, vanity_a)
    }
    if (!is.null(vanity_b)) {
      set_vanity_url(content_a, vanity_b)
    }
    vanity_a <- get_vanity_url(content_a)
    vanity_b <- get_vanity_url(content_b)
  }

  return(
    list(
      content_a = vanity_a,
      content_b = vanity_b
    )
  )
}

#' Poll Task
#'
#' Polls a task, waiting for information about a deployment. If the task has
#' results, the output will be a modified "Task" object with `task$get_data()`
#' available to retrieve the results.
#'
#' For a simple way to silence messages, set `callback = NULL`
#'
#' @param task A Task object
#' @param wait The interval to wait between polling
#' @param callback A function to be called for each message received. Set to NULL for no callback
#'
#' @return Task The Task object that was input
#'
#' @family deployment functions
#' @export
poll_task <- function(task, wait = 1, callback = message) {
  validate_R6_class(task, c("Task", "ContentTask", "VariantTask"))
  con <- task$connect

  all_task_data <- list()

  finished <- FALSE
  code <- -1
  first <- 0
  while (!finished) {
    task_data <- con$task(task$get_task()$task_id, wait = wait, first = first)
    finished <- task_data[["finished"]]
    code <- task_data[["code"]]
    first <- task_data[["last"]]

    if (!is.null(callback)) {
      lapply(task_data[["output"]], callback)
    }
    all_task_data <- c(all_task_data, task_data[["output"]])
  }

  if (code != 0) {
    msg <- task_data[["error"]]
    # print the logs if there is no callback
    if (is.null(callback)) {
      lapply(all_task_data, message)
    }
    stop(msg)
  }
  if (!is.null(task_data[["result"]])) {
    task$add_data(task_data[["result"]])
  }
  task
}

#' Build a Dashboard URL from a Content Item
#'
#' Returns the URL for the content dashboard (opened to the selected pane).
#'
#' @param content [Content] A Content object
#' @param pane character The pane in the dashboard to link to
#'
#' @return character The dashboard URL for the content provided
#'
#' @family content functions
#' @export
dashboard_url <- function(content, pane = "") {
  content$get_dashboard_url(pane = pane)
}

testthat::skip_if_not_installed("dbplyr")

cont1_name <- uuid::UUIDgenerate()
cont1_title <- "Test Content 1"
cont1_guid <- NULL
cont1_bundle <- NULL
cont1_content <- NULL

bnd_static <- bundle_dir(rprojroot::find_package_root_file(
  "tests/testthat/examples/static"
))
tmp_content <- deploy(client, bnd_static)

test_that("error on bad 'src' object", {
  expect_error(
    tbl_connect("bad_src", "users"),
    "src.*Connect"
  )
})

test_that("error on bad 'from' value", {
  expect_error(
    tbl_connect(client, "bad_from")
  )
})

test_that("users works", {
  users <- tbl_connect(client, "users")
  expect_s3_class(users, c("tbl_connect", "tbl_lazy", "tbl"))

  users_local <- users %>% dplyr::collect()
  expect_s3_class(users_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(users)))
  expect_type(colnames(users), "character")
  expect_gt(length(colnames(users)), 1)

  expect_ptype_equal(users_local, connectapi_ptypes$users, exact = FALSE)
})

test_that("usage_static works", {
  content_visits <- tbl_connect(client, "usage_static")
  expect_s3_class(content_visits, c("tbl_connect", "tbl_lazy", "tbl"))

  content_visits_local <- content_visits %>% dplyr::collect()
  expect_s3_class(content_visits_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(content_visits)))
  expect_type(colnames(content_visits), "character")
  expect_gt(length(colnames(content_visits)), 1)

  # path was added in 2024
  expect_ptype_equal(
    content_visits_local,
    connectapi_ptypes$usage_static,
    exact = FALSE
  )
})

test_that("usage_shiny works", {
  shiny_usage <- tbl_connect(client, "usage_shiny")
  expect_s3_class(shiny_usage, c("tbl_connect", "tbl_lazy", "tbl"))

  shiny_usage_local <- shiny_usage %>% dplyr::collect()
  expect_s3_class(shiny_usage_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(shiny_usage)))
  expect_type(colnames(shiny_usage), "character")
  expect_gt(length(colnames(shiny_usage)), 1)

  expect_ptype_equal(shiny_usage_local, connectapi_ptypes$usage_shiny)
})

test_that("content works", {
  scoped_experimental_silence()
  content_list <- tbl_connect(client, "content")
  expect_s3_class(content_list, c("tbl_connect", "tbl_lazy", "tbl"))

  content_list_local <- content_list %>% dplyr::collect()
  expect_s3_class(content_list_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(content_list)))
  expect_type(colnames(content_list), "character")
  expect_gt(length(colnames(content_list)), 1)

  # various attributes have been added over the years, so exact match
  # doesn't work against all versions of Connect
  expect_ptype_equal(
    content_list_local,
    connectapi_ptypes$content,
    exact = FALSE
  )
})

test_that("groups works", {
  scoped_experimental_silence()
  groups_list <- tbl_connect(client, "groups")
  expect_s3_class(groups_list, c("tbl_connect", "tbl_lazy", "tbl"))

  groups_list_local <- groups_list %>% dplyr::collect()
  expect_s3_class(groups_list_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(groups_list)))
  expect_type(colnames(groups_list), "character")
  expect_gt(length(colnames(groups_list)), 1)

  expect_ptype_equal(groups_list_local, connectapi_ptypes$groups, exact = FALSE)
})

test_that("audit_logs works", {
  scoped_experimental_silence()
  audit_list <- tbl_connect(client, "audit_logs")
  expect_s3_class(audit_list, c("tbl_connect", "tbl_lazy", "tbl"))

  audit_list_local <- audit_list %>% dplyr::collect()
  expect_s3_class(audit_list_local, c("tbl_df", "tbl", "data.frame"))

  expect_true(is.na(nrow(audit_list)))
  expect_type(colnames(audit_list), "character")
  expect_gt(length(colnames(audit_list)), 1)

  # This is different on older versions, not sure it's worth worrying about how
  skip_if_connect_older_than(client, "2022.09.0")
  expect_ptype_equal(audit_list_local, connectapi_ptypes$audit_logs)
})

# setup ---------------------------------------------------

cont1_name <- uuid::UUIDgenerate()
cont1_title <- "Test Content 1"
cont1_guid <- NULL
cont1_bundle <- NULL
cont1_content <- NULL

# bundle ---------------------------------------------------

test_that("bundle_static deploys", {
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  uniq_id <- uuid::UUIDgenerate()
  deployed <- deploy(client, bnd, uniq_id)

  expect_true(validate_R6_class(bnd, "Bundle"))
  expect_true(validate_R6_class(deployed, "Content"))

  deployed2 <- deploy(client, bnd, uniq_id)
  expect_true(validate_R6_class(deployed2, "Content"))
})

test_that("bundle_dir deploys", {
  dir_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static"
  )
  tmp_file <- fs::file_temp(pattern = "bundle", ext = ".tar.gz")
  bund <- bundle_dir(path = dir_path, filename = tmp_file)

  expect_equal(tmp_file, bund$path)

  # with a name / title
  tsk <- deploy(
    connect = client,
    bundle = bund,
    name = cont1_name,
    title = cont1_title
  )

  cont1_guid <<- tsk$content$guid
  cont1_content <<- tsk

  # how should we test that deployment happened?
  expect_true(validate_R6_class(tsk, "Content"))
  expect_equal(tsk$content$name, cont1_name)
  expect_equal(tsk$content$title, cont1_title)

  expect_true(validate_R6_class(tsk, "ContentTask"))
  expect_gt(nchar(tsk$get_task()$task_id), 0)

  # with a guid
  tsk2 <- deploy(connect = client, bundle = bund, guid = cont1_guid)
  expect_true(validate_R6_class(tsk2, "Content"))
  expect_equal(tsk2$content$name, cont1_name)
  expect_equal(tsk2$content$title, cont1_title)
  expect_equal(tsk2$content$guid, cont1_guid)
})

test_that("bundle_path deploys", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  expect_equal(tar_path, as.character(bund$path))

  # deploy to a new endpoint
  tsk <- deploy(connect = client, bundle = bund)

  # TODO: how should we test that deployment happened?
  expect_true(validate_R6_class(tsk, "Content"))
})

test_that("download_bundle works", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)
  poll_task(tsk)
  downloaded <- download_bundle(tsk)

  # TODO: is shasum always available...? A way to do without the shell?
  expect_equal(
    system(
      glue::glue("shasum {downloaded$path} | cut -d ' ' -f 1"),
      intern = TRUE
    ),
    system(glue::glue("shasum {bund$path} | cut -d ' ' -f 1"), intern = TRUE)
  )
})

test_that("delete_bundle() and get_bundles() work", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)
  poll_task(tsk)
  first_bnd <- tsk$get_content_remote()$bundle_id
  my_guid <- tsk$content$guid

  tsk <- deploy(connect = client, bundle = bund, guid = my_guid)
  poll_task(tsk)
  second_bnd <- tsk$get_content_remote()$bundle_id

  all_bnd <- get_bundles(tsk)

  # check get_bundles() output
  expect_true(first_bnd %in% all_bnd[["id"]])
  expect_true(second_bnd %in% all_bnd[["id"]])
  expect_length(all_bnd[["id"]], 2)
  expect_s3_class(all_bnd, "tbl_df")

  # check delete_bundle() functions
  res <- delete_bundle(tsk, first_bnd)
  expect_true(validate_R6_class(res, "Content"))
  expect_length(get_bundles(res)[["id"]], 1)

  expect_error(
    expect_message(
      delete_bundle(tsk, second_bnd),
      "active bundle"
    ),
    "Bad Request"
  )
  expect_length(get_bundles(res)[["id"]], 1)
})

# deploy ---------------------------------------------------

test_that("strange name re-casing does not break things", {
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  testname <- "test_Test_45"
  deploy1 <- deploy(client, bnd, testname)
  deploy2 <- deploy(client, bnd, testname)

  testname2 <- "test_Test"
  deploy(client, bnd, testname2)
  deploy(client, bnd, testname2)
})

test_that(".pre_deploy hook works", {
  scoped_experimental_silence()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  deployed <- deploy(client, bnd, uuid::UUIDgenerate(), .pre_deploy = {
    content %>% set_vanity_url(glue::glue("pre_deploy_{bundle_id}"))
  })

  active_bundle <- deployed$get_content_remote()$bundle_id
  expect_equal(
    get_vanity_url(deployed),
    as.character(glue::glue("/pre_deploy_{active_bundle}/"))
  )
})

test_that("deploy_current works", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)
  poll_task(tsk)

  created <- tsk$get_content_remote()$created_time
  first_deploy <- tsk$get_content_remote()$last_deployed_time

  # unrelated changes do not modify
  tsk$update(title = "test deploy_current")
  expect_equal(tsk$get_content_remote()$last_deployed_time, first_deploy)

  Sys.sleep(1)

  # a redeploy does
  res <- deploy_current(tsk)
  expect_true(validate_R6_class(res, "ContentTask"))

  expect_true(tsk$get_content_remote()$last_deployed_time > first_deploy)
})

# vanity_url ---------------------------------------------------

test_that("set_vanity_url works", {
  new_name <- uuid::UUIDgenerate()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  cont1 <- deploy(client, bnd, name = new_name)
  res <- set_vanity_url(cont1, new_name)

  expect_true(validate_R6_class(res, "Vanity"))
  expect_equal(res$get_vanity()$path, paste0("/", new_name, "/"))

  res2 <- set_vanity_url(cont1, paste0(new_name, "update"))
  expect_true(validate_R6_class(res2, "Vanity"))
  expect_equal(res2$get_vanity()$path, paste0("/", new_name, "update/"))
})

test_that("set_vanity_url force works", {
  new_name <- uuid::UUIDgenerate()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  cont <- deploy(client, bnd, name = new_name)
  res <- set_vanity_url(cont, new_name)

  another_name <- uuid::UUIDgenerate()
  cont2 <- deploy(client, bnd, name = another_name)

  expect_error(suppressMessages(
    set_vanity_url(cont2, new_name, force = FALSE),
    "409"
  ))

  res2 <- set_vanity_url(cont2, new_name, force = TRUE)
  expect_identical(
    get_vanity_url(cont2),
    paste0("/", new_name, "/")
  )

  expect_null(suppressMessages(get_vanity_url(cont)))
})

test_that("get_vanity_url works", {
  tmp_content_name <- uuid::UUIDgenerate()
  tmp_content_prep <- content_ensure(client, name = tmp_content_name)
  tmp_content <- Content$new(connect = client, content = tmp_content_prep)

  # without a vanity
  curr_vanity <- suppressMessages(get_vanity_url(tmp_content))
  expect_null(curr_vanity)

  # with a vanity
  res <- set_vanity_url(tmp_content, tmp_content_name)
  existing_vanity <- get_vanity_url(tmp_content)
  expect_type(existing_vanity, "character")
  expect_equal(existing_vanity, paste0("/", tmp_content_name, "/"))
})

test_that("delete_vanity_url works", {
  tmp_content_name <- uuid::UUIDgenerate()
  tmp_content_prep <- content_ensure(client, name = tmp_content_name)
  tmp_content <- Content$new(connect = client, content = tmp_content_prep)

  # create a vanity
  res <- set_vanity_url(tmp_content, tmp_content_name)
  expect_true(validate_R6_class(res, "Vanity"))
  expect_equal(res$get_vanity()$path, paste0("/", tmp_content_name, "/"))

  # delete the vanity
  res <- delete_vanity_url(tmp_content)
  expect_true(validate_R6_class(res, "Content"))
  expect_error(validate_R6_class(res, "Vanity"), "R6 Vanity")

  # get the vanity
  res <- get_vanity_url(tmp_content)
  expect_null(res)
})

test_that("swap_vanity_urls works", {
  tmp_content_a_name <- uuid::UUIDgenerate()
  tmp_content_a_prep <- content_ensure(client, name = tmp_content_a_name)
  tmp_content_a <- Content$new(
    connect = client,
    content = tmp_content_a_prep
  )

  tmp_content_b_name <- uuid::UUIDgenerate()
  tmp_content_b_prep <- content_ensure(client, name = tmp_content_b_name)
  tmp_content_b <- Content$new(
    connect = client,
    content = tmp_content_b_prep
  )

  # warns with no vanity urls
  res <- suppressMessages(expect_warning(swap_vanity_urls(
    tmp_content_a,
    tmp_content_b
  )))
  expect_null(res[["content_a"]])
  expect_null(res[["content_b"]])

  # works with just one vanity url (content_a)
  set_content_a <- set_vanity_url(tmp_content_a, tmp_content_a_name)
  swap_res <- suppressMessages(swap_vanity_urls(tmp_content_a, tmp_content_b))

  expect_null(swap_res$content_a)
  expect_identical(swap_res$content_b, paste0("/", tmp_content_a_name, "/"))

  # works with both vanity urls
  swap_res2 <- suppressMessages(swap_vanity_urls(tmp_content_a, tmp_content_b))

  expect_identical(swap_res2$content_a, paste0("/", tmp_content_a_name, "/"))
  expect_identical(swap_res2$content_b, swap_res$content_a)

  # works with just one vanity url (content_b)
  delete_vanity_url(tmp_content_a)
  expect_null(get_vanity_url(tmp_content_a))
  set_vanity_url(tmp_content_b, tmp_content_b_name)

  swap_res3 <- swap_vanity_urls(tmp_content_a, tmp_content_b)

  expect_identical(swap_res3$content_a, paste0("/", tmp_content_b_name, "/"))
  expect_null(swap_res3$content_b)
})

# misc functions ---------------------------------------------------

test_that("poll_task works and returns its input", {
  expect_message(
    res <- poll_task(cont1_content)
  )
  expect_equal(res, cont1_content)
})

test_that("download_bundle works", {
  bnd <- download_bundle(content_item(client, cont1_guid))

  expect_true(validate_R6_class(bnd, "Bundle"))
})

test_that("download_bundle throws an error for undeployed content", {
  cont_prep <- content_ensure(client)
  cont <- content_item(client, cont_prep$guid)

  expect_error(
    download_bundle(cont),
    "This content has no bundle_id"
  )
})

test_that("dashboard_url resolves properly", {
  cont <- content_item(client, cont1_guid)

  dash_url <- dashboard_url(cont)

  skip("not yet tested")
})

test_that("deployment timestamps respect timezone", {
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  myc <- deploy(client, bnd)
  myc_guid <- myc$content$guid

  # will fail without the png package
  invisible(tryCatch(client$GET(url = myc$get_url()), error = function(e) {}))

  all_usage <- get_usage_static(client, content_guid = myc_guid)
  for (i in 1:5) {
    if (nrow(all_usage) == 0) {
      # We may need to wait a beat for the metrics to show up.
      # Retry a few times just in case.
      # This did not show up testing against Connect versions <= 2022.09.0,
      # but 2023.03.0 and newer seemed to hit this
      Sys.sleep(1)
      all_usage <- get_usage_static(client, content_guid = myc_guid)
    }
  }
  expect_equal(nrow(all_usage), 1)

  # we just did this, so it should be less than 1 minute ago...
  # (really protecting against being off by hours b/c of timezone differences)
  expect_lt(
    Sys.time() - all_usage$time,
    lubridate::make_difftime(60, "seconds")
  )
})

# thumbnail ---------------------------------------------------

test_that("set_thumbnail works with local images", {
  scoped_experimental_silence()
  img_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/logo.png"
  )

  res <- set_thumbnail(cont1_content, img_path)

  expect_true(validate_R6_class(res, "Content"))
})

test_that("get_thumbnail works", {
  scoped_experimental_silence()
  img_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/logo.png"
  )

  tmp_img <- fs::file_temp(pattern = "img", ext = ".png")
  get_thumbnail(cont1_content, tmp_img)

  expect_identical(
    readBin(img_path, "raw"),
    readBin(tmp_img, "raw")
  )

  # works again (i.e. does not append data)
  get_thumbnail(cont1_content, tmp_img)
  expect_identical(
    readBin(img_path, "raw"),
    readBin(tmp_img, "raw")
  )

  # works with no path
  auto_path <- get_thumbnail(cont1_content)
  expect_identical(
    readBin(img_path, "raw"),
    readBin(auto_path, "raw")
  )
  expect_identical(fs::path_ext(auto_path), "png")
})

test_that("has_thumbnail works with an image", {
  scoped_experimental_silence()

  expect_true(has_thumbnail(cont1_content))
})

test_that("delete_thumbnail works", {
  scoped_experimental_silence()

  expect_true(validate_R6_class(delete_thumbnail(cont1_content), "Content"))
  expect_false(has_thumbnail(cont1_content))

  # works again - i.e. if no image available
  expect_true(validate_R6_class(delete_thumbnail(cont1_content), "Content"))
})

test_that("has_thumbnail works with no image", {
  scoped_experimental_silence()

  expect_false(has_thumbnail(cont1_content))
})

test_that("get_thumbnail returns NA if no image", {
  scoped_experimental_silence()

  tmp_img <- fs::file_temp(pattern = "img", ext = ".png")
  response <- get_thumbnail(cont1_content, tmp_img)

  expect_false(identical(tmp_img, response))
  expect_true(is.na(response))
})

test_that("set_thumbnail works with remote paths", {
  scoped_experimental_silence()

  # This test fails with the 1.8.8.2 image. The failure is to do with
  # downloading the favicon, used in the test as the remote image.
  skip_if_connect_older_than(cont1_content$connect, "2021.09.0")

  res <- set_thumbnail(
    cont1_content,
    glue::glue("{cont1_content$connect$server}/connect/__favicon__")
  )

  expect_true(validate_R6_class(res, "Content"))
})

test_that("bundle_dir errors if no manifest.json", {
  expect_error(
    suppressMessages(bundle_dir(rprojroot::find_testthat_root_file(
      "bundle_dir_tests",
      "no_manifest"
    ))),
    "no `manifest.json`"
  )
})

test_that("bundle_dir warns if packrat.lock", {
  expect_warning(
    suppressMessages(bundle_dir(rprojroot::find_testthat_root_file(
      "bundle_dir_tests",
      "include_packrat_lock"
    ))),
    "`packrat.lock` file found"
  )
})

test_that("bundle_dir warns if packrat folder", {
  expect_warning(
    suppressMessages(bundle_dir(rprojroot::find_testthat_root_file(
      "bundle_dir_tests",
      "include_packrat"
    ))),
    "`packrat` directory found"
  )
})

test_that("bundle_dir not fooled by subfolders", {
  expect_error(
    suppressMessages(bundle_dir(rprojroot::find_testthat_root_file(
      "bundle_dir_tests",
      "subfolders"
    ))),
    "no `manifest.json`"
  )
})

test_that("bundle_dir works with relative paths", {
  bnd <- suppressMessages(bundle_dir(fs::path_rel(rprojroot::find_testthat_root_file(
    "bundle_dir_tests",
    "doc"
  ))))
  expect_true(fs::file_exists(bnd$path))
})

test_that("bundle_dir errors for nonexistent paths", {
  expect_error(
    suppressMessages(
      bundle_dir(rprojroot::find_testthat_root_file(
        "bundle_dir_tests",
        "does_not_exist_not_real"
      ))
    ),
    "not TRUE"
  )
})

test_that("swap_vanity_urls() fails gracefully", {
  with_mock_api({
    client <- connect(server = "https://connect.example", api_key = "fake")

    works <- content_item(client, "951bf3ad")
    fails <- content_item(client, "c3426b0b")

    ## swap_vanity_urls emits multiple warnings about Posit Connect version;
    ## suppress these
    withr::local_options(list(rlib_warning_verbosity = "quiet"))
    expect_error(
      swap_vanity_urls(fails, works),
      paste(
        "Unable to modify the vanity URL for content_a:",
        "https://connect\\.example/__api__/v1/content/c3426b0b/vanity",
        "request failed with Client error: \\(403\\) Forbidden"
      )
    )
    expect_error(
      swap_vanity_urls(works, fails),
      paste(
        "Unable to modify the vanity URL for content_b:",
        "https://connect\\.example/__api__/v1/content/c3426b0b/vanity",
        "request failed with Client error: \\(403\\) Forbidden"
      )
    )
  })
})

expect_ptype_equal <- function(actual, expected, exact = TRUE) {
  if (!exact) {
    # Keep only the columns from each that are in the other
    shared_names <- intersect(names(actual), names(expected))
    actual <- actual[, shared_names]
    expected <- expected[, shared_names]
  }
  expect_equal(vctrs::vec_ptype(actual), vctrs::vec_ptype(expected))
}

skip_if_connect_older_than <- function(client, version) {
  current <- numeric_version(simplify_version(safe_server_version(client)))
  if (current < numeric_version(version)) {
    skip(paste("Requires Connect >=", version))
  }
}

deploy_example <- function(connect, name, ...) {
  example_dir <- rprojroot::find_package_root_file(
    "tests",
    "testthat",
    "examples",
    name
  )
  bundle_file <- fs::file_temp(pattern = name, ext = ".tar.gz")
  bund <- bundle_dir(path = example_dir, filename = bundle_file)

  tsk <- deploy(
    connect = connect,
    bundle = bund,
    name = name,
    ...
  )

  guid <- tsk$content$guid
  content <- content_item(tsk$connect, guid)

  # TODO: a smarter, noninteractive wait...
  suppressMessages(poll_task(tsk))
  content
}

#' User
#'
#' Get user details
#'
#' `user_guid_from_username()` is a helper to retrieve a user GUID, given the
#' user's username. It is useful in Shiny applications for using `session$user`
#'
#' @param client A Connect R6 object
#' @param username The user's username
#'
#' @rdname user
#'
#' @export
user_guid_from_username <- function(client, username) {
  validate_R6_class(client, "Connect")

  user <- client$users(prefix = username)
  res <- user$results
  if (length(res) == 0) {
    stop("ERROR: user not found")
  } else if (length(res) > 1) {
    filt <- purrr::keep(res, ~ .x$username == username)
    if (length(filt) == 1) {
      return(filt[[1]]$guid)
    } else {
      warning(
        "WARNING: multiple users found, but a unique exact match could not be found. Returning the first"
      )
      return(res[[1]]$guid)
    }
  } else {
    return(res[[1]]$guid)
  }
}

#' Get group information from the Posit Connect server
#'
#' @param src The source object.
#' @param page_size The number of records to return per page (max 500).
#' @param prefix Filters groups by prefix (group name).
#' The filter is case insensitive.
#' @param limit The number of groups to retrieve before paging stops. Default
#' is to return all results; however, for Connect server versions older than
#' 2025.04.0, `limit` is capped at 500 when `prefix` is provided.
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `guid`: The unique identifier of the group
#'   * `name`: The group name
#'   * `owner_guid`: The group owner's unique identifier. When using LDAP or
#'     Proxied authentication with group provisioning enabled this property
#'     will always be null.
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/groups for more information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' # get all groups
#' get_groups(client, limit = Inf)
#' }
#'
#' @family groups functions
#' @export
get_groups <- function(src, page_size = 500, prefix = NULL, limit = Inf) {
  validate_R6_class(src, "Connect")

  # Prior to 2025.04.0, the `v1/groups` endpoint ignored page_number when
  # `prefix` was provided, always returning the first page.
  # So for those versions, we should request the max page size needed to
  # satisfy `limit`, and make them equal, so that page_offset() only fetches
  # one page.
  # Proper pagination added in https://github.com/posit-dev/connect/pull/31086
  if (
    !is.null(prefix) && compare_connect_version(src$version, "2025.04.0") < 0
  ) {
    # The max page size allowed by the server is 500, so in this version of
    # Connect, that's the maximum number of groups we can get.
    page_size <- limit <- min(500, limit)
  }
  res <- page_offset(
    src,
    src$groups(page_size = page_size, prefix = prefix),
    limit = limit
  )

  parse_connectapi_typed(res, connectapi_ptypes$groups)
}

#' Get users within a specific group
#'
#' @param src A Connect client object
#' @param guid A group GUID identifier
#'
#' @return
#' A tibble with the following columns:
#'
#'   * `email`: The user's email
#'   * `username`: The user's username
#'   * `first_name`: The user's first name
#'   * `last_name`: The user's last name
#'   * `user_role`: The user's role. It may have a value of administrator,
#'     publisher or viewer.
#'   * `created_time`: The timestamp (in RFC3339 format) when the user
#'     was created in the Posit Connect server
#'   * `updated_time`: The timestamp (in RFC3339 format) when the user
#'     was last updated in the Posit Connect server
#'   * `active_time`: The timestamp (in RFC3339 format) when the user
#'     was last active on the Posit Connect server
#'   * `confirmed`: When false, the created user must confirm their
#'     account through an email. This feature is unique to password
#'     authentication.
#'   * `locked`: Whether or not the user is locked
#'   * `guid`: The user's GUID, or unique identifier, in UUID RFC4122 format
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/groups/-group_guid-/members
#' for more information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' # get the first 20 groups
#' groups <- get_groups(client)
#'
#' group_guid <- groups$guid[1]
#'
#' get_group_members(client, guid = group_guid)
#' }
#'
#' @family groups functions
#' @export
get_group_members <- function(src, guid) {
  validate_R6_class(src, "Connect")

  res <- src$group_members(guid)

  parse_connectapi(res$results)
}

#' Get content access permissions for a group or groups
#'
#' @param src A Connect client object
#' @param guids A character vector of group guids
#'
#' @return
#' A tibble with the following columns:

#'   * `group_guid`: The group's GUID
#'   * `group_name`: The group's name
#'   * `content_guid`: The content item's GUID
#'   * `content_name`: The content item's name
#'   * `content_title`: The content item's title
#'   * `access_type`: The access type of the content item ("all", "logged_in", or "acl")
#'   * `role`: The access type that members of the group have to the
#'     content item, "publisher" or "viewer".
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' # Get a data frame of groups
#' groups <- get_groups(client)
#'
#' # Get permissions for a single group by passing in the corresponding row.
#' get_group_content(client, groups[1, "guid"])
#' dplyr::filter(groups, name = "research_scientists") %>%
#'   dplyr::pull(guid) %>%
#'   get_group_content(client, .)
#'
#' # Get permissions for all groups by passing in all group guids.
#' get_group_content(client, groups$guid)
#' }
#'
#' @family groups functions
#' @export
get_group_content <- function(src, guids) {
  validate_R6_class(src, "Connect")

  purrr::list_rbind(
    purrr::map(
      guids,
      ~ get_one_groups_content(src = src, guid = .x)
    )
  )
}

get_one_groups_content <- function(src, guid) {
  validate_R6_class(src, "Connect")

  res <- src$group_content(guid)
  if (length(res) == 0) {
    return(tibble::tibble(
      group_guid = NA_character_,
      group_name = NA_character_,
      content_guid = NA_character_,
      content_name = NA_character_,
      content_title = NA_character_,
      access_type = NA_character_,
      role = NA_character_
    ))
  }
  parsed <- parse_connectapi_typed(res, connectapi_ptypes$group_content)

  permissions_df <- purrr::list_rbind(
    purrr::map(
      parsed$permissions,
      ~ tibble::as_tibble(unlist(
        purrr::keep(
          .x,
          ~ .x[["principal_guid"]] == guid
        ),
        recursive = FALSE
      ))
    )
  )

  tibble::tibble(
    group_guid = guid,
    group_name = permissions_df$principal_name,
    content_guid = parsed$content_guid,
    content_name = parsed$content_name,
    content_title = parsed$content_title,
    access_type = parsed$access_type,
    role = permissions_df$principal_role
  )
}

# Helpers to make it easier to identify where we're calling public APIs and not
v1_url <- function(...) {
  paste("v1", ..., sep = "/")
}

unversioned_url <- function(...) {
  paste(..., sep = "/")
}

# Make an alias of unversioned_url to help us be able to search for the places
# where we call the unversioned API, but not be misled by places we've left in
# for backwards compatibility.
unversioned_fallback_url <- unversioned_url

valid_page_size <- function(x, min = 1, max = 500) {
  # This could be changed to error if x is outside the range
  min(max(min, x), max)
}

generate_R6_print_output <- # nolint: object_name_linter
  function() {
    con <- Connect$new(server = "http://test_host", api_key = "test_key")
    bnd <- Bundle$new(path = "/test/path")

    ex_content <- list(
      guid = "content-guid",
      title = "content-title",
      dashboard_url = "http://test_host/connect/#/apps/content-guid/",
      url = "http://content-url"
    )
    cnt1 <- Content$new(connect = con, ex_content)

    ex_task <- list(task_id = "task-id")
    tsk1 <- ContentTask$new(connect = con, content = ex_content, task = ex_task)

    ex_vanity <- list(path_prefix = "vanity-prefix")
    van1 <- Vanity$new(connect = con, content = ex_content, vanity = ex_vanity)

    obj_list <- list(con, bnd, cnt1, tsk1, van1)

    unlist(mapply(
      function(.x, .y) {
        c(
          "----------------------------",
          .y,
          "----------------------------",
          capture.output(print(.x))
        )
      },
      .x = obj_list,
      .y = lapply(obj_list, function(x) {
        class(x)[[1]]
      }),
      SIMPLIFY = FALSE
    ))
  }

is_R6_class <- # nolint: object_name_linter
  function(instance, class) {
    return(R6::is.R6(instance) && inherits(instance, class))
  }

validate_R6_class <- # nolint: object_name_linter
  function(instance, class) {
    obj <- rlang::enquo(instance)
    if (!R6::is.R6(instance) || !inherits(instance, class)) {
      stop(paste(
        rlang::quo_text(obj),
        "must be an R6",
        glue::glue_collapse(class, sep = " or "),
        "object"
      ))
    }
    invisible(TRUE)
  }

# super useful examples
# https://github.com/tidyverse/tibble/blob/master/R/compat-lifecycle.R
warn_experimental <- function(name) {
  if (
    rlang::is_true(rlang::peek_option(
      "connectapi_disable_experimental_warnings"
    ))
  ) {
    return(invisible(NULL))
  }
  warn_once(
    msg = glue::glue(
      "The `{name}` function is experimental and subject to change without warning in a future release"
    ),
    id = paste0(name, "-experimental")
  )
}

scoped_experimental_silence <- function(frame = rlang::caller_env()) {
  rlang::local_options(
    .frame = frame,
    connectapi_disable_experimental_warnings = TRUE
  )
}

warn_dire <- function(name) {
  if (rlang::is_true(rlang::peek_option("connectapi_disable_dire_warnings"))) {
    return(invisible(NULL))
  }
  warn_once(
    msg = glue::glue(
      "DO NOT USE IN PRODUCTION - The {name} function is for internal testing purposes only"
    ),
    id = paste0(name, "-dire")
  )
}

scoped_dire_silence <- function(frame = rlang::caller_env()) {
  rlang::local_options(
    .frame = frame,
    connectapi_disable_dire_warnings = TRUE
  )
}

warn_once <- function(msg, id = msg, ...) {
  if (rlang::is_true(rlang::peek_option("connectapi_disable_warnings"))) {
    return(invisible(NULL))
  }

  rlang::warn(msg, .frequency = "once", .frequency_id = id, ...)
}

check_connect_license <- function(url) {
  if (is_R6_class(url, "Connect")) {
    url <- url$server
  }
  res <- httr::GET(glue::glue("{url}/__ping__"))
  if (res$status_code == 402) {
    stop(glue::glue("ERROR: The Connect server's license is expired ({url})"))
  }
  invisible(res)
}

simplify_version <- function(version) {
  sub("([0-9]+\\.[0-9]+\\.[0-9]+).*", "\\1", version)
}

safe_server_settings <- function(client) {
  srv <- tryCatch(
    {
      client$server_settings()
    },
    error = function(e) {
      message(
        glue::glue(
          "Problem talking to Posit Connect at {client$server}/__api__/server_settings"
        )
      )
      stop(e)
    }
  )
  return(srv)
}

safe_server_version <- function(client) {
  version <- safe_server_settings(client)$version
  if (is.null(version) || nchar(version) == 0) {
    version <- NA
  }
  version
}

error_if_less_than <- function(using_version, tested_version) {
  if (is.na(using_version)) {
    msg <- glue::glue(
      "WARNING: This feature requires Posit Connect version {tested_version} ",
      "but the server version is not exposed by this Posit Connect instance. ",
      "You may experience errors when using this functionality."
    )
    warn_once(msg)
    return(invisible())
  }
  comp <- compare_connect_version(using_version, tested_version)
  if (comp < 0) {
    msg <- glue::glue(
      "ERROR: This feature requires Posit Connect version {tested_version} ",
      "but you are using {using_version}. Please upgrade Posit Connect."
    )
    stop(msg)
  }
  invisible()
}

compare_connect_version <- function(using_version, tested_version) {
  as.integer(
    compareVersion(
      simplify_version(using_version),
      simplify_version(tested_version)
    )
  )
}

warn_untested_connect <- function(
  using_version,
  minimum_tested_version = "1.8.8.2"
) {
  comp <- compare_connect_version(using_version, minimum_tested_version)
  if (!is.na(using_version) && (comp < 0)) {
    warn_once(
      glue::glue(
        "You are using an older version of Posit Connect ",
        "({using_version}) than is tested ({minimum_tested_version}). ",
        "Some APIs may not function as expected."
      ),
      id = "old-connect"
    )
  }
  invisible()
}

token_hex <- function(n) {
  raw <- as.raw(sample(0:255, n, replace = TRUE))
  paste(as.character(raw), collapse = "")
}

# Checks to see if an http status contains a 404 error, and that that 404
# response does not contain an error code indicating that an endpoint was called
# but encountered a 404 for some other reason.
endpoint_does_not_exist <- function(res) {
  httr::status_code(res) == "404" &&
    !("code" %in% names(httr::content(res, as = "parsed")))
}

# Extract the Connect error code from a response
error_code <- function(res) {
  cont <- httr::content(res)
  if ("code" %in% names(cont)) {
    cont[["code"]]
  } else {
    # Return -1 if there is no error, just so we aren't comparing NA or NULL
    -1
  }
}

# Returns `TRUE` if we're running on Connect as determined by the
# `RSTUDIO_PRODUCT` or `POSIT_PRODUCT` env var, else `FALSE`.
on_connect <- function() {
  Sys.getenv("RSTUDIO_PRODUCT") == "CONNECT" ||
    Sys.getenv("POSIT_PRODUCT") == "CONNECT"
}

# Returns `TRUE` if we're running on Connect Cloud as determined by the
# `POSIT_PRODUCT` env var, else `FALSE`.
on_connect_cloud <- function() {
  Sys.getenv("POSIT_PRODUCT") == "CONNECT_CLOUD"
}

# Returns `TRUE` if running via testthat
is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

# Calls `message()` if not running in testthat. Useful for messages that would
# otherwise appear very frequently in test output.
message_if_not_testing <- function(...) {
  if (!is_testing()) {
    message(...)
  }
}

test_that("simplify_version works", {
  expect_identical(simplify_version("1.8.2-4"), "1.8.2")
  expect_identical(simplify_version("1.8.2.1-4"), "1.8.2")
  expect_identical(simplify_version("10.70.204.1-4"), "10.70.204")
  expect_identical(simplify_version("10.0.0.0-4"), "10.0.0")
  expect_identical(simplify_version("2026.01.0-dev+15-z596093a50w"), "2026.01.0")
})

test_that("compare_connect_version works", {
  expect_identical(
    compare_connect_version("2024.09.0", "1.8.6"),
    1L
  )
  expect_identical(
    compare_connect_version("2025.01.1", "2025.01.0"),
    1L
  )
  expect_identical(
    compare_connect_version("2025.01.0", "2025.01.0"),
    0L
  )
  expect_identical(
    compare_connect_version("2024.08.2", "2024.09"),
    -1L
  )
  expect_identical(
    compare_connect_version(NA, "2024.09"),
    -1L
  )
})

test_that("error_if_less_than errors works as expected", {
  withr::local_options(list(rlib_warning_verbosity = "verbose"))
  expect_silent(error_if_less_than("2024.09.0", "1.8.6"))
  expect_error(
    error_if_less_than("2024.08.2", "2024.09"),
    "ERROR: This feature requires Posit Connect version 2024.09"
  )
  expect_warning(
    error_if_less_than(NA, "2024.09"),
    "WARNING: This feature requires Posit Connect version 2024.09"
  )
})

test_that("warn_untested_connect works", {
  withr::local_options(list(rlib_warning_verbosity = "verbose"))
  # silent for patch version changes
  expect_silent(warn_untested_connect("1.8.2.1-10", "1.8.2-4"))

  # silent if newer
  expect_silent(warn_untested_connect("1.8.2-4", "1.8.0.5-1"))

  # warnings for minor version changes
  expect_warning(warn_untested_connect("1.8.2-4", "2.8.0.5-1"), "older")
  rlang::reset_warning_verbosity("old-connect")
})

test_that("warn_untested_connect warning snapshot", {
  # warning messages seem to cause issues in different environments based on color codes
  skip_on_cran()
  withr::local_options(list(rlib_warning_verbosity = "verbose"))
  # No warning
  expect_snapshot(capture_warning(warn_untested_connect("2022.02", "2022.01")))
  expect_snapshot(capture_warning(warn_untested_connect("2022.01", "2022.02")))
})

test_that("message_if_not_testing messages appropriately", {
  expect_no_message(message_if_not_testing("a message"))

  withr::with_envvar(list(TESTTHAT = ""), {
    expect_message(message_if_not_testing("a message"), "a message")
  })
})

# Temporarily skip these on newer Connect versions because the manifest for this
# content specifies an old version of R not found in newer images.
skip_if(safe_server_version(client) > "2024.03.0")

# ensure that RSPM is being used so these do not take eternity
shiny_content <- deploy_example(client, "shiny")

test_that("set_run_as works with a good linux user", {
  scoped_experimental_silence()
  res <- set_run_as(shiny_content, "rstudio-connect")
  expect_equal(
    res$content$run_as,
    "rstudio-connect"
  )

  skip("TODO: failing because of a bug in Connect")
  res2 <- set_run_as(shiny_content, NULL)
  expect_null(res2$content$run_as)
})

test_that("set_run_as fails with a bad linux user", {
  scoped_experimental_silence()
  expect_error(
    suppressMessages(
      set_run_as(shiny_content, "fake-user")
    ),
    "400"
  )
})

test_that("set_run_as works for run_as_current_user", {
  scoped_experimental_silence()
  res <- set_run_as(
    shiny_content,
    run_as = NULL,
    run_as_current_user = TRUE
  )

  expect_true(
    shiny_content$content$run_as_current_user
  )

  res2 <- set_run_as(
    shiny_content,
    run_as = NULL,
    run_as_current_user = FALSE
  )

  expect_false(
    shiny_content$content$run_as_current_user
  )
})

structure(
  list(
    url = "content/01234567/__thumbnail__",
    status_code = 200L,
    headers = structure(list(
      `content-type` = "image/jpeg"
    )),
    content = as.raw(c(
      0x4e,
      0x41
    ))
  ),
  class = "response"
)

structure(
  list(
    request <- list(url = "__api__/v1/system/caches/runtime"),
    url = "__api__/v1/system/caches/runtime",
    status_code = 404L,
    content = charToRaw(
      "{\"code\":4,\"error\":\"Cache does not exist\",\"payload\":null}"
    )
  ),
  class = "response"
)

structure(
  list(
    url = "__api__/v1/content/c3426b0b/vanity",
    status_code = 403L,
    content = charToRaw(
      "{\"code\":22,\"error\":\"You don't have permission to perform this operation.\",\"payload\":null}"
    )
  ),
  class = "response"
)

NA_datetime_ <- # nolint: object_name_linter
  vctrs::new_datetime(NA_real_, tzone = Sys.timezone())
NA_list_ <- # nolint: object_name_linter
  list(list())

connectapi_ptypes <- list(
  users = tibble::tibble(
    "email" = NA_character_,
    "username" = NA_character_,
    "first_name" = NA_character_,
    "last_name" = NA_character_,
    "user_role" = NA_character_,
    "created_time" = NA_datetime_,
    "updated_time" = NA_datetime_,
    "active_time" = NA_datetime_,
    "confirmed" = FALSE,
    "locked" = FALSE,
    "external_id" = NA_character_,
    "guid" = NA_character_
  ),
  groups = tibble::tibble(
    "guid" = NA_character_,
    "name" = NA_character_,
    "owner_guid" = NA_character_,
    "gid" = NA_character_
  ),
  usage_shiny = tibble::tibble(
    "content_guid" = NA_character_,
    "user_guid" = NA_character_,
    "started" = NA_datetime_,
    "ended" = NA_datetime_,
    "data_version" = NA_integer_
  ),
  usage_static = tibble::tibble(
    "content_guid" = NA_character_,
    "user_guid" = NA_character_,
    "variant_key" = NA_character_,
    "time" = NA_datetime_,
    "rendering_id" = NA_character_,
    "bundle_id" = NA_character_,
    "data_version" = NA_integer_
  ),
  usage = tibble::tibble(
    "id" = NA_character_,
    "user_guid" = NA_character_,
    "content_guid" = NA_character_,
    "timestamp" = NA_datetime_,
    "data" = NA_list_
  ),
  content = tibble::tibble(
    "guid" = NA_character_,
    "name" = NA_character_,
    "title" = NA_character_,
    "description" = NA_character_,
    "access_type" = NA_character_,
    "connection_timeout" = NA_integer_,
    "read_timeout" = NA_integer_,
    "init_timeout" = NA_integer_,
    "idle_timeout" = NA_integer_,
    "max_processes" = NA_integer_,
    "min_processes" = NA_integer_,
    "max_conns_per_process" = NA_integer_,
    "load_factor" = NA_real_,
    "created_time" = NA_datetime_,
    "last_deployed_time" = NA_datetime_,
    "bundle_id" = NA_character_,
    "app_mode" = NA_character_,
    "content_category" = NA_character_,
    "parameterized" = FALSE,
    "cluster_name" = NA_character_,
    "image_name" = NA_character_,
    "r_version" = NA_character_,
    "py_version" = NA_character_,
    "quarto_version" = NA_character_,
    "run_as" = NA_character_,
    "run_as_current_user" = FALSE,
    "owner_guid" = NA_character_,
    "content_url" = NA_character_,
    "dashboard_url" = NA_character_,
    "app_role" = NA_character_,
    "vanity_url" = NA_character_,
    "id" = NA_character_,
    "owner" = NA_list_,
    "tags" = NA_list_,
  ),
  content_old = tibble::tibble(
    "id" = NA_integer_,
    "guid" = NA_character_,
    "access_type" = NA_character_,
    "connection_timeout" = NA_real_,
    "read_timeout" = NA_real_,
    "init_timeout" = NA_real_,
    "idle_timeout" = NA_real_,
    "max_processes" = NA_integer_,
    "min_processes" = NA_integer_,
    "max_conns_per_process" = NA_integer_,
    "load_factor" = NA_real_,
    "url" = NA_character_,
    "vanity_url" = NA,
    "name" = NA_character_,
    "title" = NA_character_,
    "bundle_id" = NA_integer_,
    # (1=shiny, 2=shiny Rmd, 3=source Rmd, 4=static, 5=api, 6=tensorflow, 7=python, 8=flask, 9=dash, 10=streamlit)
    "app_mode" = NA_integer_,
    "content_category" = NA_character_,
    "has_parameters" = NA,
    "created_time" = NA_datetime_,
    "last_deployed_time" = NA_datetime_,
    "r_version" = NA_character_,
    "py_version" = NA_character_,
    "build_status" = NA_integer_,
    "run_as" = NA_character_,
    "run_as_current_user" = NA,
    "description" = NA_character_,
    "app_role" = NA_character_,
    "owner_first_name" = NA_character_,
    "owner_last_name" = NA_character_,
    "owner_username" = NA_character_,
    "owner_guid" = NA_character_,
    "owner_email" = NA_character_,
    "owner_locked" = NA,
    "is_scheduled" = NA,
    "git" = NA_list_
  ),
  audit_logs = tibble::tibble(
    "id" = NA_character_,
    "time" = NA_datetime_,
    "user_id" = NA_character_,
    "user_guid" = NA_character_,
    "user_description" = NA_character_,
    "action" = NA_character_,
    "event_description" = NA_character_
  ),
  procs = tibble::tibble(
    pid = NA_character_,
    appId = NA_integer_,
    appGuid = NA_character_,
    appName = NA_character_,
    appUrl = NA_character_,
    appRunAs = NA_character_,
    type = NA_character_,
    cpuCurrent = NA_real_,
    cpuTotal = NA_integer_,
    ram = fs::as_fs_bytes(NA_integer_)
  ),
  variant = tibble::tibble(
    id = NA_integer_,
    app_id = NA_integer_,
    key = NA_character_,
    bundle_id = NA_integer_,
    is_default = NA,
    name = NA_character_,
    email_collaborators = NA,
    email_viewers = NA,
    created_time = NA_datetime_,
    rendering_id = NA_integer_,
    render_time = NA_datetime_,
    render_duration = bit64::NA_integer64_,
    visibility = NA_character_,
    owner_id = NA_integer_
  ),
  rendering = tibble::tibble(
    id = NA_integer_,
    app_id = NA_integer_,
    variant_id = NA_integer_,
    bundle_id = NA_integer_,
    job_key = NA_character_,
    render_time = NA_datetime_,
    render_duration = bit64::as.integer64(NA_integer_),
    active = NA,
    app_guid = NA_character_,
    variant_key = NA_character_,
  ),
  jobs = tibble::tibble(
    id = NA_character_,
    ppid = NA_character_,
    pid = NA_character_,
    key = NA_character_,
    remote_id = NA_character_,
    content_id = NA_character_,
    content_guid = NA_character_,
    app_id = NA_character_,
    app_guid = NA_character_,
    variant_id = NA_character_,
    bundle_id = NA_character_,
    start_time = NA_datetime_,
    end_time = NA_datetime_,
    last_heartbeat_time = NA_datetime_,
    queued_time = NA_datetime_,
    queue_name = NA_character_,
    tag = NA_character_,
    exit_code = NA_integer_,
    status = NA_integer_,
    hostname = NA_character_,
    cluster = NA_character_,
    image = NA_character_,
    run_as = NA_character_,
  ),
  bundles = tibble::tibble(
    id = NA_character_,
    content_guid = NA_character_,
    created_time = NA_datetime_,
    r_version = NA_character_,
    py_version = NA_character_,
    active = NA,
    size = fs::as_fs_bytes(NA_integer_),
    metadata = NA_list_,
  ),
  permissions = tibble::tibble(
    id = NA_character_,
    content_guid = NA_character_,
    principal_guid = NA_character_,
    principal_type = NA_character_,
    role = NA_character_
  ),
  group_content = tibble::tibble(
    content_guid = NA_character_,
    content_name = NA_character_,
    content_title = NA_character_,
    access_type = NA_character_,
    permissions = NA_list_
  ),
  job_termination = tibble::tibble(
    app_id = NA_integer_,
    app_guid = NA_character_,
    job_key = NA_character_,
    job_id = NA_character_,
    result = NA_character_,
    code = NA_integer_,
    error = NA_character_
  ),
  vanities = tibble::tibble(
    content_guid = NA_character_,
    path = NA_character_,
    created_time = NA_datetime_
  ),
  job_log = tibble::tibble(
    source = NA_character_,
    timestamp = NA_datetime_,
    data = NA_character_
  ),
  packages = tibble::tibble(
    language = NA_character_,
    language_version = NA_character_,
    name = NA_character_,
    version = NA_character_,
    hash = NA_character_,
    bundle_id = NA_character_,
    app_id = NA_character_,
    app_guid = NA_character_,
  ),
  content_packages = tibble::tibble(
    language = NA_character_,
    name = NA_character_,
    version = NA_character_,
    hash = NA_character_
  ),
  integrations = tibble::tibble(
    id = NA_character_,
    guid = NA_character_,
    created_time = NA_datetime_,
    updated_time = NA_datetime_,
    name = NA_character_,
    description = NA_character_,
    template = NA_character_,
    auth_type = NA_character_,
    config = NA_list_
  )
)

#' Create a Remote User
#'
#' The remote user creation workflow involves authentication providers like LDAP
#' that involve a queryable identity store. This helper wraps the API calls
#' necessary to retrieve information about and then create such a user. It
#' functions with a "fuzzy match" `prefix` by default, but if you want to
#' instantiate users directly, you should set `exact = TRUE`.
#'
#' NOTE: there can be problems with usernames that are not unique. Please open
#' an issue if you run into any problems.
#'
#' @param connect An R6 Connect object.
#' @param prefix character. The prefix of the user name to search for.
#' @param expect number. Optional. The number of responses to expect for this search.
#' @param check boolean. Optional. Whether to check for local existence first.
#' @param exact boolean. Optional. Whether to only create users whose username
#'   exactly matches the provided `prefix`.
#'
#' @return The results of creating the users.
#'
#' @export
users_create_remote <- function(
  connect,
  prefix,
  expect = 1,
  check = TRUE,
  exact = FALSE
) {
  expect <- as.integer(expect)
  if (check) {
    local_users <- get_users(connect, prefix = prefix)
    if (exact) {
      local_users <- local_users[local_users["username"] == prefix, ]
    }
    if (nrow(local_users) > 0) {
      if (!exact) {
        message(glue::glue(
          "At least one user with username prefix '{prefix}' already exists"
        ))
      } else {
        message(glue::glue("A user with username '{prefix}' already exists"))
      }
      return(local_users)
    }
  }

  remote_users <- connect$users_remote(prefix = prefix)
  remote_users_res <- remote_users[["results"]]
  if (exact) {
    remote_users_res <- purrr::keep(
      remote_users_res,
      ~ .x[["username"]] == prefix
    )
  }
  if (length(remote_users_res) != expect) {
    message(glue::glue(
      "Found {length(remote_users_res)} remote users. Expected {expect}"
    ))
    if (length(remote_users_res) > 0) {
      user_str <- toString(purrr::map_chr(remote_users_res, ~ .x[["username"]]))
      message(glue::glue("Users found: {user_str}"))
    }
    stop(
      "The expected user(s) were not found. Please specify a more accurate 'prefix'"
    )
  }
  purrr::map(
    remote_users_res,
    function(.x, src) {
      message(glue::glue("Creating remote user: {.x[['username']]}"))
      src$users_create_remote(temp_ticket = .x[["temp_ticket"]])
    },
    src = connect
  )
  message("Done creating remote users")
  return(get_users(connect, prefix = prefix))
}

#' Create a Remote Group
#'
#' @param connect An R6 Connect object.
#' @param prefix character. The prefix of the user name to search for.
#' @param expect number. Optional. The number of responses to expect for this search.
#' @param check boolean. Optional. Whether to check for local existence first.
#' @param exact boolean. Optional. Whether to only create groups whose name
#'   exactly matches the provided `prefix`.
#'
#' @return The results of creating the groups.
#'
#' @export
groups_create_remote <- function(
  connect,
  prefix,
  expect = 1,
  check = TRUE,
  exact = FALSE
) {
  expect <- as.integer(expect)
  if (check) {
    local_groups <- get_groups(connect, prefix = prefix)
    if (exact) {
      local_groups <- local_groups[local_groups["name"] == prefix, ]
    }
    if (nrow(local_groups) > 0) {
      if (!exact) {
        message(glue::glue(
          "At least one group with name prefix '{prefix}' already exists"
        ))
      } else {
        message(glue::glue("A group with the name '{prefix}' already exists"))
      }
      return(local_groups)
    }
  }

  remote_groups <- connect$groups_remote(prefix = prefix)
  remote_groups_res <- remote_groups[["results"]]
  if (exact) {
    remote_groups_res <- purrr::keep(
      remote_groups_res,
      ~ .x[["name"]] == prefix
    )
  }
  if (length(remote_groups_res) != expect) {
    message(glue::glue(
      "Found {length(remote_groups_res)} remote groups. Expected {expect}"
    ))
    if (length(remote_groups_res) > 0) {
      groups_found <- glue::glue_collapse(
        purrr::map_chr(remote_groups_res, ~ .x[["name"]]),
        sep = ", "
      )
      message(glue::glue("Groups found: {groups_found}"))
    }
    stop(
      "The expected group(s) were not found. Please specify a more accurate 'prefix'"
    )
  }
  purrr::map(
    remote_groups_res,
    function(.x, src) {
      message(glue::glue("Creating remote group: {.x[['name']]}"))
      src$groups_create_remote(temp_ticket = .x[["temp_ticket"]])
    },
    src = connect
  )
  message("Done creating remote groups")
  get_groups(connect, prefix = prefix)
}

test_that("groups_create_remote: succeed when no local group exists", {
  client <- MockConnect$new()
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups/remote",
    content = list(
      results = list(list(
        name = "Everyone",
        guid = NULL,
        temp_ticket = "fake"
      )),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "PUT",
    path = "v1/groups",
    content = list(
      guid = "1c1ab604-4a6a-4d07-9477-a88ac08386cd",
      name = "Everyone",
      owner_guid = NULL
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(
        list(
          guid = "1c1ab604-4a6a-4d07-9477-a88ac08386cd",
          name = "Everyone",
          owner_guid = NULL
        )
      ),
      current_page = 1L,
      total = 1L
    )
  )

  suppressMessages(
    expect_message(
      res <- groups_create_remote(client, "Everyone"),
      "Creating remote group"
    )
  )
  expect_equal(res$name, "Everyone")
  expect_equal(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups/remote",
      "PUT https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups"
    )
  )
})

test_that("groups_create_remote: succeed without checking local groups if check is FALSE", {
  client <- MockConnect$new()
  client$mock_response(
    method = "GET",
    path = "v1/groups/remote",
    content = list(
      results = list(list(
        name = "Everyone",
        guid = NULL,
        temp_ticket = "fake"
      )),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "PUT",
    path = "v1/groups",
    content = list(
      guid = "1c1ab604-4a6a-4d07-9477-a88ac08386cd",
      name = "Everyone",
      owner_guid = NULL
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(
        list(
          guid = "1c1ab604-4a6a-4d07-9477-a88ac08386cd",
          name = "Everyone",
          owner_guid = NULL
        )
      ),
      current_page = 1L,
      total = 1L
    )
  )

  suppressMessages(
    expect_message(
      res <- groups_create_remote(client, "Everyone", check = FALSE),
      "Creating remote group"
    )
  )
  expect_equal(res$name, "Everyone")
  expect_equal(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/groups/remote",
      "PUT https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups"
    )
  )
})

test_that("groups_create_remote: err if number of remote groups != `expect`", {
  client <- MockConnect$new()
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups/remote",
    content = list(
      results = list(list(
        name = "Everyone",
        guid = NULL,
        temp_ticket = "fake"
      )),
      current_page = 1L,
      total = 1L
    )
  )

  expect_error(
    res <- suppressMessages(groups_create_remote(client, "Everyone", expect = 2)),
    "The expected group\\(s\\) were not found. Please specify a more accurate 'prefix'"
  )
  expect_equal(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups/remote"
    )
  )
})

test_that("groups_create_remote: create groups if multiple found and n == `expect`", {
  client <- MockConnect$new()
  client$mock_response(
    method = "GET",
    path = "v1/groups/remote",
    content = list(
      results = list(
        list(
          name = "Everyone",
          guid = NULL,
          temp_ticket = "fake"
        ),
        list(
          name = "Everyone two",
          guid = NULL,
          temp_ticket = "fake"
        )
      ),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(),
      current_page = 1L,
      total = 1L
    )
  )
  client$mock_response(
    method = "PUT",
    path = "v1/groups",
    content = list(
      guid = "fake-guid-1",
      name = "Everyone",
      owner_guid = NULL
    )
  )
  client$mock_response(
    method = "PUT",
    path = "v1/groups",
    content = list(
      guid = "fake-guid-2",
      name = "Everyone two",
      owner_guid = NULL
    )
  )
  client$mock_response(
    method = "GET",
    path = "v1/groups",
    content = list(
      results = list(
        list(
          guid = "fake-guid-1",
          name = "Everyone",
          owner_guid = NULL
        ),
        list(
          guid = "fake-guid-2",
          name = "Everyone two",
          owner_guid = NULL
        )
      ),
      current_page = 1L,
      total = 1L
    )
  )

  suppressMessages(
    expect_message(
      res <- groups_create_remote(client, "Everyone", expect = 2),
      "Creating remote group"
    )
  )
  expect_identical(
    res$name,
    c("Everyone", "Everyone two")
  )
  expect_identical(
    client$call_log,
    c(
      "GET https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups/remote",
      "PUT https://connect.example/__api__/v1/groups",
      "PUT https://connect.example/__api__/v1/groups",
      "GET https://connect.example/__api__/v1/groups"
    )
  )
})

with_mock_api({
  mock_dir_client <- Connect$new(
    server = "https://connect.example",
    api_key = "fake"
  )

  test_that("groups_create_remote: err when local and remote groups return no matches", {
    expect_error(
      res <- suppressMessages(groups_create_remote(mock_dir_client, "Nothing")),
      "The expected group\\(s\\) were not found. Please specify a more accurate 'prefix'"
    )
  })

  test_that("groups_create_remote: message when local group exists and check is TRUE", {
    expect_message(
      res <- groups_create_remote(mock_dir_client, "Everyone"),
      "At least one group with name prefix 'Everyone' already exists"
    )
    expect_identical(res$name, "Everyone Else")
  })

  test_that("groups_create_remote: message when local group exists and check is TRUE", {
    expect_message(
      res <- groups_create_remote(mock_dir_client, "Everyone"),
      "At least one group with name prefix 'Everyone' already exists"
    )
    expect_identical(res$name, "Everyone Else")
  })

  test_that("groups_create_remote: only consider exact matches when exact is TRUE", {
    suppressMessages(
      expect_message(
        groups_create_remote(mock_dir_client, "Art", exact = TRUE),
        "Creating remote group"
      )
    )
  })

  test_that("groups_create_remote: error if expected group not found", {
    expect_error(
      suppressMessages(groups_create_remote(mock_dir_client, "Art")),
      "The expected group\\(s\\) were not found"
    )
  })
})

# Manage Tags ------------------------------------------

# TODO: this definitely needs some cleanup... there are several disparate data structures:
# 1 - the raw "table" of JSON data returned from the public API
# 2 - the "tag-tree" of JSON data returned from the "old" / private API
# 3 - the output of `tag_tree()` that is optimized for usage by R
# 4 - the output of `connect_tag_tree()`, which has the same data structure, but has classes associated

# TODO: we should see if we can cut out 2 and 3 as intermediate states that exist...
# - see if (2) can be skipped now by refactoring...
# - see if (3) can be lumped into (4)?

#' Get all Tags on the server
#'
#' Tag manipulation and assignment functions
#'
#' @param src The source object
#' @param name The name of the tag to create
#' @param parent optional. A `connect_tag_tree` object (as returned by `get_tags()`) pointed at the parent tag
#' @param content An R6 Content object, as returned by `content_item()`
#' @param tag A `connect_tag_tree` object (as returned by `get_tags()`)
#' @param tags A `connect_tag_tree` object (as returned by `get_tags()`)
#' @param ids A list of `id`s to filter the tag tree by
#' @param pattern A regex to filter the tag tree by (it is passed to `grepl`)
#' @param ... Additional arguments
#'
#' Manage tags (requires Administrator role):
#' - `get_tags()` - returns a "tag tree" object that can be traversed with `tag_tree$tag1$childtag`
#' - `get_tag_data()` - returns a tibble of tag data
#' - `create_tag()` - create a tag by specifying the Parent directly
#' - `create_tag_tree()` - create tag(s) by specifying the "desired" tag tree
#' hierarchy
#' - `delete_tag()` - delete a tag (and its children). WARNING: will
#' disassociate any content automatically
#'
#' Manage content tags:
#' - `get_content_tags()` - return a `connect_tag_tree` object corresponding to
#' the tags for a piece of content.
#' - `set_content_tag_tree()` - attach a tag to content by specifying the
#' desired tag tree
#' - `set_content_tags()` - Set multiple tags at once by providing
#' `connect_tag_tree` objects
#'
#' Search a tag tree:
#' - `filter_tag_tree_chr()` - filters a tag tree based on a regex
#' - `filter_tag_tree_id()` - filters a tag tree based on an id
#'
#' @export
#' @rdname tags
get_tags <- function(src) {
  validate_R6_class(src, "Connect")

  connect_tag_tree(tag_tree(src$get_tag_tree()), NULL)
}

#' @export
#' @rdname tags
get_tag_data <- function(src) {
  validate_R6_class(src, "Connect")

  res <- src$get_tag_tree()

  tag_tbl <- parse_tags_tbl(res)

  return(tag_tbl)
}

# TODO: this is hard to "use" directly because what it returns is not a tag... maybe create a Tag R6 class?
#' @export
#' @rdname tags
create_tag <- function(src, name, parent = NULL) {
  validate_R6_class(src, "Connect")
  if (is.null(parent) || is.numeric(parent)) {
    parent_id <- parent
  } else if (inherits(parent, "connect_tag_tree")) {
    if (is.null(parent[["id"]])) {
      stop("must specify a `parent` tag, and not the entire tag tree")
    }
    parent_id <- parent[["id"]]
  } else {
    stop("`parent` must be an ID or a connect_tag_tree object")
  }
  res <- src$tag_create(name = name, parent_id = parent_id)
  print(filter_tag_tree_id(get_tags(src), res))
  cat("\n")
  return(src)
}

# TODO: try without quotes...
#' @export
#' @rdname tags
create_tag_tree <- function(src, ...) {
  validate_R6_class(src, "Connect")

  params <- rlang::list2(...)

  results <- purrr::reduce(
    params,
    function(.parent, .x, con) {
      res <- con$tag_create_safe(.x, .parent)
      return(res[["id"]])
    },
    con = src,
    .init = NULL
  )
  print(filter_tag_tree_id(get_tags(src), results))
  cat("\n")
  return(src)
}

#' @export
#' @rdname tags
delete_tag <- function(src, tag) {
  if (is.numeric(tag)) {
    tag_id <- tag
  } else if (inherits(tag, "connect_tag_tree")) {
    if (is.null(tag[["id"]])) {
      stop(
        "`tag` must reference some tag specifically, and not the entire tag tree"
      )
    }
    tag_id <- tag[["id"]]
  } else {
    stop("`tag` must be an ID or a connect_tag_tree object")
  }
  src$tag_delete(id = tag_id)
  return(src)
}

# connect_tag_tree S3 object --------------------------------------------

# TODO: Need to find a way to denote categories...?
# error  : chr "Cannot assign a category to an app"
# TODO: Need to protect against a bad data structure...
# TODO: possible that you could decouple this from a connect server and get strange results
#       (i.e. build a tag tree from server A, use it to "set_content_tags" for server B - ids would not match)
connect_tag_tree <- function(tag_data, filter = "filtered") {
  structure(tag_data, class = c("connect_tag_tree", "list"), filter = filter)
}

#' @export
print.connect_tag_tree <- function(x, ...) {
  if (!is.null(attr(x, "filter"))) {
    cat(glue::glue("Posit Connect Tag Tree ({attr(x, 'filter')})"))
    cat("\n")
  } else {
    cat("Posit Connect Tag Tree\n")
  }
  if (length(x) > 0) {
    recursive_tag_print(x, "")
  } else {
    cat("  < No tags >\n")
  }
}

#' @export
`$.connect_tag_tree` <- function(x, y) {
  res <- NextMethod("$")
  if (is.list(res)) {
    connect_tag_tree(res)
  } else {
    res
  }
}

#' @export
`[[.connect_tag_tree` <- function(x, ...) {
  res <- NextMethod("[[")
  if (is.list(res)) {
    connect_tag_tree(res)
  } else {
    res
  }
}

#' @export
`[.connect_tag_tree` <- function(x, i, j) {
  res <- NextMethod("[")
  warn_once(
    "`[` drops the `connect_tag_tree` class. Use `$` or `[[` instead",
    id = "[.connect_tag_tree"
  )
  res
}

# Content Tags ----------------------------------------------------

#' @export
#' @rdname tags
get_content_tags <- function(content) {
  validate_R6_class(content, "Content")
  ctags <- content$tags()
  # TODO: find a way to build a tag tree from a list of tags

  tagtree <- get_tags(content$connect)
  res <- filter_tag_tree_id(
    tagtree,
    purrr::map_chr(ctags, ~ as.character(.x$id))
  )
  attr(res, "filter") <- "content"
  res
}

#' @export
#' @rdname tags
set_content_tag_tree <- function(content, ...) {
  validate_R6_class(content, "Content")

  params <- rlang::list2(...)
  if (length(params) == 1) {
    stop(
      "cannot assign a category to an app. Please specify an additional tag level"
    )
  }

  tags <- get_tags(content$connect)

  # check that tags exist
  tmp <- purrr::pluck(tags, !!!params)
  if (!is.null(tmp[["id"]])) {
    # only use the "bottom most" ID to tag
    content$tag_set(tmp[["id"]])
  } else {
    stop("the tag tree specified was not found")
  }
  print(get_content_tags(content))
  cat("\n")
  return(content)
}

#' @export
#' @rdname tags
set_content_tags <- function(content, ...) {
  validate_R6_class(content, "Content")
  new_tags <- rlang::list2(...)
  purrr::map(
    new_tags,
    function(.x) {
      content$tag_set(.get_tag_id(.x))
    }
  )
  print(get_content_tags(content))
  cat("\n")
  content
}

.get_tag_id <- function(.x) {
  if (inherits(.x, "connect_tag_tree") && !"id" %in% names(.x)) {
    print(.x)
    stop("this tag does not have an 'id'. Is it a tag list?")
  }
  if (inherits(.x, "connect_tag_tree")) {
    return(.x[["id"]])
  } else {
    return(.x)
  }
}

set_content_tags_remove <- function() {
  # TODO
}

set_content_tag_tree_remove <- function() {
  # TODO
}

# Filter ----------------------------------------------------

#' @export
#' @rdname tags
filter_tag_tree_id <- function(tags, ids) {
  stopifnot(inherits(tags, "connect_tag_tree"))
  flt <- recursive_filter_id(tags = tags, ids = ids)
  if (!is.null(flt)) {
    flt
  } else {
    connect_tag_tree(list())
  }
}

#' @export
#' @rdname tags
filter_tag_tree_chr <- function(tags, pattern) {
  stopifnot(inherits(tags, "connect_tag_tree"))

  flt <- recursive_filter_chr(tags = tags, pattern = pattern)
  if (!is.null(flt)) {
    flt
  } else {
    connect_tag_tree(list())
  }
}

# Tree Structure Creation ----------------------------------

# input is the output of `get_tag_data()` or `con$tag()`
tag_tree_from_data <- function(tag_data) {
  parsed <- tag_tree_parse_data(tag_data)

  connect_tag_tree(tag_tree(parsed))
}

tag_tree_parse_data <- function(tag_data) {
  if (tibble::is_tibble(tag_data)) {
    tag_data <- purrr::transpose(tag_data)
  }
  base_categories <- purrr::keep(
    tag_data,
    ~ is.null(.x[["parent_id"]]) || is.na(.x[["parent_id"]])
  )

  output <- purrr::map(
    base_categories,
    tag_tree_parse_data_impl,
    tag_data = tag_data
  )

  return(output)
}

tag_tree_parse_data_impl <- function(target, tag_data) {
  filtered_data <- purrr::keep(
    tag_data,
    ~ !is.null(.x[["parent_id"]]) &&
      !is.na(.x[["parent_id"]]) &&
      .x[["parent_id"]] == target[["id"]]
  )

  # recurse through the tree
  output <- purrr::map(
    filtered_data,
    tag_tree_parse_data_impl,
    tag_data = tag_data
  )

  # what we get back needs to become "children"
  target[["children"]] <- output

  return(target)
}

# Utils ----------------------------------------------------

recursive_filter_id <- function(tags, ids) {
  tags_noname <- tags
  tags_noname$name <- NULL
  tags_noname$id <- NULL
  recurse_res <- purrr::map(tags_noname, ~ recursive_filter_id(.x, ids))
  rr_nonull <- purrr::keep(recurse_res, ~ !is.null(.x))
  if ((!is.null(tags$id) && tags$id %in% ids) || length(rr_nonull) > 0) {
    if (!is.null(tags[["name"]])) {
      name_add <- list(name = tags$name, id = tags$id)
    } else {
      name_add <- list()
    }
    connect_tag_tree(c(name_add, rr_nonull))
  } else {
    NULL
  }
}

recursive_filter_chr <- function(tags, pattern) {
  tags_noname <- tags
  tags_noname$name <- NULL
  tags_noname$id <- NULL
  recurse_res <- purrr::map(tags_noname, ~ recursive_filter_chr(.x, pattern))
  rr_nonull <- purrr::keep(recurse_res, ~ !is.null(.x))
  if (!is.null(tags[["name"]])) {
    name_match <- any(purrr::map_lgl(pattern, ~ grepl(.x, tags$name)))
  } else {
    name_match <- FALSE
  }
  if (name_match || length(rr_nonull) > 0) {
    if (!is.null(tags[["name"]])) {
      name_add <- list(name = tags$name, id = tags$id)
    } else {
      name_add <- list()
    }
    connect_tag_tree(c(name_add, rr_nonull))
  } else {
    NULL
  }
}

recursive_find_tag <- function(tags, tag, parent_id = NULL) {
  tags_noname <- tags
  tags_noname$name <- NULL
  tags_noname$id <- NULL
  recurse_res <- purrr::map_chr(
    tags_noname,
    ~ as.character(recursive_find_tag(.x, tag, parent_id))
  )
  recurse_res_any <- recurse_res[!is.na(recurse_res)]
  if (length(recurse_res_any) == 0) {
    recurse_res_any <- NA_real_
  }
  names(recurse_res_any) <- NULL

  if (!is.na(recurse_res_any)) {
    return(recurse_res_any)
  } else if (is.null(parent_id) && !is.null(tags$name) && tags$name == tag) {
    res <- tags$id
    names(res) <- NULL
    return(res)
  } else if (
    !is.null(parent_id) && tags$id == parent_id && tag %in% names(tags_noname)
  ) {
    res <- tags[[tag]]$id
    names(res) <- NULL
    return(res)
  } else {
    return(NA_real_)
  }
}


recursive_tag_print <- function(x, indent) {
  x_noname <- x
  x_noname$name <- NULL
  x_noname$id <- NULL
  ch <- box_chars()
  # print a "single level tag"
  if (length(x_noname) == 0 && nchar(indent) == 0) {
    if (!is.null(x$name)) {
      cat(indent, pc(ch$l, ch$h, ch$h, " "), x$name, "\n", sep = "")
    }
  } else if (nchar(indent) == 0) {
    if (!is.null(x$name)) {
      cat(x$name, "\n", sep = "")
    }
  }
  purrr::map2(
    x_noname,
    seq_along(x_noname),
    function(.y, .i, list_length) {
      if (.i == list_length) {
        cat(indent, pc(ch$l, ch$h, ch$h, " "), .y$name, "\n", sep = "")
        recursive_tag_print(.y, paste0(indent, "   "))
      } else {
        cat(indent, pc(ch$j, ch$h, ch$h, " "), .y$name, "\n", sep = "")
        recursive_tag_print(.y, paste0(indent, pc(ch$v, "   ")))
      }
    },
    list_length = length(x_noname)
  )
  invisible(x)
}

recursive_tag_restructure <- function(.x) {
  if (length(.x$children) > 0) {
    rlang::set_names(
      list(c(
        purrr::flatten(purrr::map(
          .x$children,
          recursive_tag_restructure
        )),
        id = as.character(.x$id),
        name = .x$name
      )),
      .x$name
    )
  } else {
    rlang::set_names(
      list(list(id = as.character(.x$id), name = .x$name)),
      .x$name
    )
  }
}

tag_tree <- function(.x) {
  purrr::flatten(purrr::map(.x, recursive_tag_restructure))
}

parse_tags_tbl <- function(x) {
  parsed_tags <- purrr::list_rbind(purrr::map(
    x,
    ~ {
      out <- tibble::tibble(
        id = as.character(.x$id),
        name = .x$name,
        created_time = .x$created_time,
        updated_time = .x$updated_time,
        parent_id = ifelse(is.null(.x$parent_id), NA_character_, .x$parent_id)
      )

      if (length(.x$children) > 0) {
        child <- parse_tags_tbl(.x$children)
        out <- rbind(out, child)
      }

      return(out)
    }
  ))

  return(parsed_tags)
}


# HELPER FUNCTIONS FOR tag_tree FROM fs
pc <- function(...) {
  paste0(..., collapse = "")
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }
  get("is_latex_output", asNamespace("knitr"))()
}

is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (!is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

# These are derived from https://github.com/r-lib/cli/blob/e9acc82b0d20fa5c64dd529400b622c0338374ed/R/tree.R#L111
box_chars <- function() {
  if (is_utf8_output()) {
    list(
      "h" = "\u2500", # horizontal
      "v" = "\u2502", # vertical
      "l" = "\u2514",
      "j" = "\u251C"
    )
  } else {
    list(
      "h" = "-", # horizontal
      "v" = "|", # vertical
      "l" = "\\",
      "j" = "+"
    )
  }
}

parent_tag_name <- uuid::UUIDgenerate(use.time = TRUE)
child_tag_name <- uuid::UUIDgenerate(use.time = TRUE)

parent_tag <- NULL
child_tag <- NULL

content_name <- uuid::UUIDgenerate(use.time = TRUE)

tag_content <- NULL

check_tag_exists <- function(con, id) {
  res <- tryCatch(suppressMessages(con$tag(id)), error = function(e) {
    return(e)
  })
  if (is.numeric(res[["id"]])) {
    TRUE
  } else if (regexpr("simpleError", res) && regexpr("(404) Not Found", res)) {
    FALSE
  } else {
    stop("error retrieving tag")
  }
}

# Tests ---------------------------------------

test_that("create tags works", {
  parent_tag <<- client$tag_create(parent_tag_name)
  child_tag <<- client$tag_create(child_tag_name, parent_tag$id)

  expect_identical(parent_tag$name, parent_tag_name)
  expect_identical(child_tag$name, child_tag_name)
  expect_identical(child_tag$parent_id, parent_tag$id)

  # tag_id lookup works
  client$get_tags(TRUE)
  expect_equal(child_tag$id, client$get_tag_id(child_tag_name))
})

## Test high level functions --------------------------------------------------

test_that("get_tags works", {
  atags <- get_tags(client)
  expect_s3_class(atags, "connect_tag_tree")
})

test_that("create_tag and delete_tag works", {
  ptag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_2 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_3 <- uuid::UUIDgenerate(use.time = TRUE)

  capture.output(res <- create_tag(client, ptag_1))
  expect_true(validate_R6_class(res, "Connect"))

  tags <- get_tags(client)
  capture.output(create_tag(client, ctag_1, tags[[ptag_1]]))

  tags <- get_tags(client)
  capture.output(create_tag(client, ctag_2, tags[[ptag_1]][[ctag_1]]))

  tags <- get_tags(client)
  capture.output(create_tag(
    client,
    ctag_3,
    tags[[ptag_1]][[ctag_1]][[ctag_2]]
  ))

  tags <- get_tags(client)

  delete_tag(client, tags[[ptag_1]][[ctag_1]][[ctag_2]][[ctag_3]])
  expect_false(check_tag_exists(
    client,
    tags[[ptag_1]][[ctag_1]][[ctag_2]][[ctag_3]][["id"]]
  ))

  delete_tag(client, tags[[ptag_1]][[ctag_1]][[ctag_2]])
  expect_false(check_tag_exists(
    client,
    tags[[ptag_1]][[ctag_1]][[ctag_2]][["id"]]
  ))

  delete_tag(client, tags[[ptag_1]][[ctag_1]])
  expect_false(check_tag_exists(client, tags[[ptag_1]][[ctag_1]][["id"]]))

  res <- delete_tag(client, tags[[ptag_1]])
  expect_false(check_tag_exists(client, tags[[ptag_1]][["id"]]))

  expect_true(validate_R6_class(res, "Connect"))
})

test_that("delete_tag errs for whole tree", {
  alltags <- get_tags(client)

  expect_error(
    delete_tag(client, alltags),
    "entire tag tree"
  )
})

test_that("con$tag with id returns just one record", {
  ptag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  capture.output(a1 <- create_tag_tree(client, ptag_1, ctag_1))

  tags1 <- get_tags(client)
  res <- client$tag(tags1[[ptag_1]][[ctag_1]][["id"]])
  expect_equal(
    names(res),
    c("id", "name", "parent_id", "created_time", "updated_time")
  )
})

test_that("create_tag_tree works", {
  ptag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_2 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_3 <- uuid::UUIDgenerate(use.time = TRUE)

  capture.output(a1 <- create_tag_tree(client, ptag_1, ctag_1))
  expect_true(validate_R6_class(a1, "Connect"))

  tags1 <- get_tags(client)

  capture.output(
    a2 <- create_tag_tree(client, ptag_1, ctag_1, ctag_2, ctag_3)
  )
  expect_true(validate_R6_class(a2, "Connect"))

  tags2 <- get_tags(client)

  expect_identical(tags1[[ptag_1]][["id"]], tags2[[ptag_1]][["id"]])

  delete_tag(client, tags2[[ptag_1]])
  expect_error(
    suppressMessages(client$tag(tags2[[ptag_1]][["id"]])),
    "Not Found"
  )
})


test_that("get_content_tags and set_content_tags works", {
  ptag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1_2 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_2_1 <- uuid::UUIDgenerate(use.time = TRUE)

  app1 <- deploy(
    client,
    bundle_dir(rprojroot::find_package_root_file(
      "tests",
      "testthat",
      "examples",
      "static"
    ))
  )

  capture.output(
    tmp1 <- create_tag_tree(client, ptag_1, ctag_1_1, ctag_1_2)
  )
  capture.output(tmp2 <- create_tag_tree(client, ptag_1, ctag_2_1))

  expect_true(validate_R6_class(tmp1, "Connect"))
  expect_true(validate_R6_class(tmp2, "Connect"))

  ct <- get_content_tags(app1)
  expect_length(ct, 0)

  all_tags <- get_tags(client)

  c1o <- capture.output(
    c1 <- set_content_tags(
      app1,
      all_tags[[ptag_1]][[ctag_1_1]][[ctag_1_2]]
    )
  )
  expect_identical(c1, app1)
  expect_length(get_content_tags(app1), 1)

  c2o <- capture.output(
    c2 <- set_content_tags(
      app1,
      all_tags[[ptag_1]][[ctag_1_1]][[ctag_1_2]],
      all_tags[[ptag_1]][[ctag_2_1]]
    )
  )
  expect_identical(c2, app1)
  expect_length(get_content_tags(app1)[[ptag_1]], 4) # 2 tags, id, name

  # TODO: use newer way to delete tags
  app1$tag_delete(get_content_tags(app1)[[ptag_1]][["id"]])
  expect_length(get_content_tags(app1), 0)
})

test_that("set_content_tag_tree works", {
  ptag_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1_1 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_1_2 <- uuid::UUIDgenerate(use.time = TRUE)
  ctag_2_1 <- uuid::UUIDgenerate(use.time = TRUE)

  app1 <- deploy(
    client,
    bundle_dir(rprojroot::find_package_root_file(
      "tests",
      "testthat",
      "examples",
      "static"
    ))
  )

  capture.output(
    tmp1 <- create_tag_tree(client, ptag_1, ctag_1_1, ctag_1_2)
  )
  capture.output(tmp2 <- create_tag_tree(client, ptag_1, ctag_2_1))

  expect_true(validate_R6_class(tmp1, "Connect"))
  expect_true(validate_R6_class(tmp2, "Connect"))

  ct <- get_content_tags(app1)
  expect_length(ct, 0)

  c1o <- capture.output(
    c1 <- set_content_tag_tree(
      app1,
      ptag_1,
      ctag_1_1,
      ctag_1_2
    )
  )
  expect_identical(c1, app1)
  expect_length(get_content_tags(app1), 1)

  c2o <- capture.output(
    c2 <- set_content_tag_tree(
      app1,
      ptag_1,
      ctag_2_1
    )
  )
  expect_identical(c2, app1)
  expect_length(get_content_tags(app1)[[ptag_1]], 4) # 2 tags, id, name

  # TODO: use newer way to delete tags
  app1$tag_delete(get_content_tags(app1)[[ptag_1]][["id"]])
  expect_length(get_content_tags(app1), 0)
})

simple_tag_tree <- connect_tag_tree(
  list(
    hi = list(
      name = "hi",
      id = 1,
      ho = list(name = "ho", id = 2),
      silver = list(name = "silver", id = 3),
      away = list(name = "away", id = 4)
    )
  )
)

test_that("works with no input", {
  expect_output(
    print(connect_tag_tree(list(), NULL)),
    "No tags"
  )
  expect_output(
    print(connect_tag_tree(list())),
    "No tags"
  )
})

test_that("print method ends in a newline", {
  c0 <- capture.output({
    print(connect_tag_tree(list()))
    cat("hi")
  })
  expect_length(c0, 3)
  expect_equal(c0[3], "hi")

  c1 <- capture.output({
    print(simple_tag_tree)
    cat("max")
  })
  expect_identical(c1[length(c1)], "max")
})

test_that("print methods work as expected", {
  skip("not tested yet")
})

test_that("$ works as expected", {
  expect_s3_class(simple_tag_tree$hi, "connect_tag_tree")
  expect_s3_class(simple_tag_tree$hi$ho, "connect_tag_tree")
  expect_named(simple_tag_tree$hi$ho, c("name", "id"))
  expect_output(print(simple_tag_tree$hi), "filtered")
  expect_type(simple_tag_tree$hi$ho$name, "character")
  expect_type(simple_tag_tree$hi$ho$id, "double")
})

test_that("[ works as expected", {
  # drops the connect_tag_tree class
  # because maintaining the structure becomes hard...
  withr::local_options(list(rlib_warning_verbosity = "verbose"))
  expect_warning(
    res <- simple_tag_tree["hi"],
    "drops the `connect_tag_tree` class"
  )
  expect_false(inherits(res, "connect_tag_tree"))
  expect_type(res, "list")
})

test_that("[[ works as expected", {
  expect_s3_class(simple_tag_tree[["hi"]], "connect_tag_tree")
  expect_s3_class(simple_tag_tree[["hi"]][["ho"]], "connect_tag_tree")
  expect_named(simple_tag_tree[["hi"]][["ho"]], c("name", "id"))
  expect_output(print(simple_tag_tree[["hi"]]), "filtered")
  expect_type(simple_tag_tree[["hi"]][["ho"]][["name"]], "character")
  expect_type(simple_tag_tree[["hi"]][["ho"]][["id"]], "double")
})

test_that("filter_tag_tree_chr works as expected", {
  tt <- simple_tag_tree

  expect_length(filter_tag_tree_chr(tt, "hi"), 1)
  expect_length(filter_tag_tree_chr(tt, "ho"), 1)
  expect_length(filter_tag_tree_chr(tt, "ho")[["hi"]], 3) # name, id, ho

  expect_length(filter_tag_tree_chr(tt, "(ho)|(away)")[["hi"]], 4) # name, id, ho, away
})

test_that("filter_tag_tree_id works as expected", {
  tt <- simple_tag_tree

  expect_length(filter_tag_tree_id(tt, 1), 1)
  expect_length(filter_tag_tree_id(tt, 2), 1)
  expect_length(filter_tag_tree_id(tt, 2)[["hi"]], 3) # name, id, ho

  expect_length(filter_tag_tree_id(tt, c(2, 4))[["hi"]], 4) # name, id, ho, away
})

test_that("filter handles no responses", {
  tt <- simple_tag_tree

  expect_length(filter_tag_tree_chr(tt, "something"), 0)
  expect_length(filter_tag_tree_id(tt, 45), 0)
})

test_that("filter handles no input", {
  tt <- simple_tag_tree

  expect_length(filter_tag_tree_chr(tt, character()), 0)
})

test_that("filter handles no input", {
  tt <- simple_tag_tree

  expect_length(filter_tag_tree_id(tt, integer()), 0)
})

test_that("restructure from tag_data works", {
  ref_time <- Sys.time()
  ref_time_str <- format(ref_time, "%Y-%m-%dT%H:%M:%SZ")
  tag_data <- list(
    list(
      id = "1",
      name = "First Tag",
      parent_id = NA,
      created_time = ref_time_str,
      updated_time = ref_time_str
    ),
    list(
      id = "2",
      name = "Second Tag",
      parent_id = NA,
      created_time = ref_time_str,
      updated_time = ref_time_str
    ),
    list(
      id = "3",
      name = "Third Tag",
      parent_id = "1",
      created_time = ref_time_str,
      updated_time = ref_time_str
    ),
    list(
      id = "4",
      name = "Fourth Tag",
      parent_id = "3",
      created_time = ref_time_str,
      updated_time = ref_time_str
    ),
    list(
      id = "5",
      name = "Fifth Tag",
      parent_id = "2",
      created_time = ref_time_str,
      updated_time = ref_time_str
    )
  )
  t1 <- tag_tree_from_data(tag_data)

  tbl <- purrr::map_df(tag_data, identity)

  t2 <- tag_tree_from_data(tbl)

  expect_identical(t1, t2)
})

# because format(NULL, "%Y-%m") == "NULL"
safe_format <- function(expr, ...) {
  if (is.null(expr)) {
    return(NULL)
  } else {
    return(format(expr, ...))
  }
}

datetime_to_rfc3339 <- function(input) {
  tmp <- format(input, format = "%Y-%m-%dT%H:%M:%OS5%z")
  ln <- nchar(tmp)
  paste0(substr(tmp, 0, ln - 2), ":", substr(tmp, ln - 1, ln))
}

make_timestamp <- function(input) {
  if (is.character(input)) {
    # TODO: make sure this is the right timestamp format
    return(input)
  }

  # In the call to `safe_format`:
  # - The format specifier adds a literal "Z" to the end of the timestamp, which
  #   tells Connect "This is UTC".
  # - The `tz` argument tells R to produce times in the UTC time zone.
  # - The `usetz` argument says "Don't concatenate ' UTC' to the end of the string".
  safe_format(input, "%Y-%m-%dT%H:%M:%SZ", tz = "UTC", usetz = FALSE)
}

ensure_columns <- function(.data, ptype, strict = FALSE) {
  # Given a prototype, ensure that all columns are present and cast to the correct type.
  # If a column is missing in .data, it will be created with all missing values of the correct type.
  # If a column is present in both, it will be cast to the correct type.
  # If a column is present in .data but not in ptype, it will be left as is.
  # If `strict == TRUE`, include only columns present in the ptype, in the order they occur.
  for (i in names(ptype)) {
    .data <- ensure_column(.data, ptype[[i]], i)
  }

  if (strict) {
    .data <- .data[, names(ptype), drop = FALSE]
  }

  .data
}

ensure_column <- function(data, default, name) {
  stopifnot(length(default) == 1)
  col <- data[[name]]
  scoped_experimental_silence()
  if (rlang::is_null(col)) {
    col <- vctrs::vec_rep(default, nrow(data))
    col <- vctrs::vec_cast(col, default)
  } else {
    if (
      vctrs::vec_is(default, NA_datetime_) && !vctrs::vec_is(col, NA_datetime_)
    ) {
      # manual fix because vctrs::vec_cast cannot cast double -> datetime or char -> datetime
      col <- coerce_datetime(col, default, name = name)
    }

    if (inherits(default, "fs_bytes") && !inherits(col, "fs_bytes")) {
      col <- coerce_fsbytes(col, default)
    }

    if (inherits(default, "integer64") && !inherits(col, "integer64")) {
      col <- bit64::as.integer64(col)
    }

    if (is.character(default) && (is.integer(col) || is.double(col))) {
      if (is.double(col)) {
        col <- format(col, scientific = FALSE, trim = TRUE)
      } else {
        col <- as.character(col)
      }
    }

    if (inherits(default, "list") && !inherits(col, "list")) {
      col <- list(col)
    }

    col <- vctrs::vec_cast(col, default, x_arg = name)
  }
  data[[name]] <- col
  data
}

parse_connectapi_typed <- function(data, ptype, strict = FALSE) {
  ensure_columns(parse_connectapi(data), ptype, strict)
}

# Build a tibble column-by-column instead of row-by-row (via list_rbind).
# This avoids type conflicts when the same field is NULL in some rows and
# non-NULL in others: NULL -> NA, and unlist() coerces that NA to match the
# type of the non-null values in the same column. ensure_columns() handles
# any further type coercion (e.g. character -> POSIXct) after this step.
parse_connectapi <- function(data) {
  if (length(data) == 0) return(tibble::tibble())

  all_names <- unique(unlist(lapply(data, names)))
  cols <- stats::setNames(lapply(all_names, function(nm) {
    # NULL / missing fields become NA; unlist() will coerce to the right type
    values <- lapply(data, function(row) row[[nm]] %||% NA)
    if (any(vapply(values, function(v) is.list(v) || length(v) > 1, logical(1)))) {
      # List column: wrap scalars so every element is a list
      lapply(values, function(v) if (is.list(v)) v else list(v))
    } else {
      # Scalar column: simplify to a vector
      unlist(values)
    }
  }), all_names)
  tibble::as_tibble(cols)
}

coerce_fsbytes <- function(x, to, ...) {
  if (is.numeric(x)) {
    fs::as_fs_bytes(x)
  } else {
    vctrs::stop_incompatible_cast(x = x, to = to, x_arg = "x", to_arg = "to")
  }
}

# name - optional. Must be named, the name of the variable / column being converted
coerce_datetime <- function(x, to, ...) {
  tmp_name <- rlang::dots_list(...)[["name"]]
  if (is.null(tmp_name) || is.na(tmp_name) || !is.character(tmp_name)) {
    tmp_name <- "x"
  }

  if (is.null(x)) {
    as.POSIXct(character(), tz = tzone(to))
  } else if (is.numeric(x)) {
    vctrs::new_datetime(as.double(x), tzone = tzone(to))
  } else if (is.character(x)) {
    parse_connect_rfc3339(x)
  } else if (inherits(x, "POSIXct")) {
    x
  } else if (
    all(is.logical(x) & is.na(x)) && length(is.logical(x) & is.na(x)) > 0
  ) {
    NA_datetime_
  } else {
    vctrs::stop_incompatible_cast(
      x = x,
      to = to,
      x_arg = tmp_name,
      to_arg = "to"
    )
  }
}

# nolint start: commented_code_linter
# Parses a character vector of dates received from Connect, using use RFC 3339,
# returning a vector of POSIXct datetimes.
#
# R parses character timestamps as ISO 8601. When specifying %z, it expects time
# zones to be specified as `-1400` to `+1400`.
#
# Connect's API sends times in a specific RFC 3339 format: indicating time zone
# offsets with `-14:00` to `+14:00`, and zero offset with `Z`.
# https://github.com/golang/go/blob/54fe0fd43fcf8609666c16ae6d15ed92873b1564/src/time/format.go#L86
# For example:
# - "2023-08-22T14:13:14Z"
# - "2023-08-22T15:13:14+01:00"
# - "2020-01-01T00:02:03-01:00"
# nolint end
parse_connect_rfc3339 <- function(x) {
  # Convert timestamps with offsets to a format recognized by `strptime`.
  x <- gsub("([+-]\\d\\d):(\\d\\d)$", "\\1\\2", x)
  x <- gsub("Z$", "+0000", x)

  # Parse with an inner call to `strptime()`, which returns a POSIXlt object,
  # and convert that to `POSIXct`.
  #
  # We must specify `tz` in the inner call to correctly compute date math.
  # Specifying `tz` when in the outer call just changes the time zone without
  # doing any date math!
  #
  # > xlt [1] "2024-08-29 16:36:33 EDT" tzone(xlt) [1] "America/New_York"
  # as.POSIXct(xlt, tz = "UTC") [1] "2024-08-29 16:36:33 UTC"
  format_string <- "%Y-%m-%dT%H:%M:%OS%z"
  as.POSIXct(x, format = format_string, tz = Sys.timezone())
}

vec_cast.POSIXct.double <- # nolint: object_name_linter
  function(x, to, ...) {
    warn_experimental("vec_cast.POSIXct.double")
    vctrs::new_datetime(x, tzone = tzone(to))
  }

vec_cast.POSIXct.character <- # nolint: object_name_linter
  function(x, to, ...) {
    as.POSIXct(x, tz = tzone(to))
  }

tzone <- function(x) {
  attr(x, "tzone")[[1]] %||% ""
}

vec_cast.character.integer <- # nolint: object_name_linter
  function(x, to, ...) {
    as.character(x)
  }

new_datetime <- function(x = double(), tzone = "") {
  tzone <- tzone %||% ""
  if (is.integer(x)) {
    x <- as.double(x)
  }
  stopifnot(is.double(x))
  stopifnot(is.character(tzone))
  structure(x, tzone = tzone, class = c("POSIXct", "POSIXt"))
}

test_that("coerce_fsbytes fills the void", {
  expect_s3_class(coerce_fsbytes(1L, fs::as_fs_bytes(NA_integer_)), "fs_bytes")
  expect_s3_class(coerce_fsbytes(1, fs::as_fs_bytes(NA_integer_)), "fs_bytes")
  expect_error(
    coerce_fsbytes(data.frame(), fs::as_fs_bytes(NA_integer_)),
    class = "vctrs_error_incompatible_type"
  )
})

test_that("coerce_datetime fills the void", {
  chardate <- "2023-10-25T17:04:08Z"
  numdate <- as.double(Sys.time())
  expect_s3_class(coerce_datetime(chardate, NA_datetime_), "POSIXct")
  expect_s3_class(
    coerce_datetime(c(chardate, chardate), NA_datetime_),
    "POSIXct"
  )
  expect_s3_class(coerce_datetime(numdate, NA_datetime_), "POSIXct")
  expect_s3_class(coerce_datetime(c(numdate, numdate), NA_datetime_), "POSIXct")
  expect_s3_class(coerce_datetime(NA_datetime_, NA_datetime_), "POSIXct")
  expect_s3_class(
    coerce_datetime(c(NA_datetime_, NA_datetime_), NA_datetime_),
    "POSIXct"
  )
  expect_s3_class(coerce_datetime(NA_integer_, NA_datetime_), "POSIXct")
  expect_s3_class(
    coerce_datetime(c(NA_integer_, NA_integer_), NA_datetime_),
    "POSIXct"
  )
  expect_s3_class(coerce_datetime(NA, NA_datetime_), "POSIXct")
  expect_s3_class(coerce_datetime(c(NA, NA), NA), "POSIXct")
  expect_s3_class(coerce_datetime(NULL, NA), "POSIXct")

  expect_error(
    coerce_datetime(data.frame(), NA_datetime_),
    class = "vctrs_error_incompatible_type"
  )
  expect_error(
    coerce_datetime(list(), NA_datetime_, name = "list"),
    class = "vctrs_error_incompatible_type"
  )

  expect_error(
    coerce_datetime(NA_complex_, NA_datetime_, name = "complexity"),
    class = "vctrs_error_incompatible_type"
  )
})

test_that("parse_connect_rfc3339() parses timestamps with offsets as expected", {
  x_mixed <- c(
    "2023-08-22T14:13:14Z",
    "2020-01-01T01:02:03Z",
    "2023-08-22T15:13:14+01:00",
    "2020-01-01T00:02:03-01:00"
  )

  x_zero_offset <- c(
    "2023-08-22T14:13:14Z",
    "2020-01-01T01:02:03Z"
  )

  x_plus_one <- c(
    "2023-08-22T15:13:14+01:00",
    "2020-01-01T02:02:03+01:00"
  )

  x_minus_one <- c(
    "2023-08-22T13:13:14-01:00",
    "2020-01-01T00:02:03-01:00"
  )

  single_zero_offset <- "2023-08-22T14:13:14Z"

  single_offset <- "2023-08-22T15:13:14+01:00"

  withr::local_envvar(TZ = "America/New_York")
  expected <- as.POSIXct(strptime(
    c(
      "2023-08-22T14:13:14+0000",
      "2020-01-01T01:02:03+0000"
    ),
    format = "%Y-%m-%dT%H:%M:%S%z",
    tz = Sys.timezone()
  ))
  expect_identical(parse_connect_rfc3339(x_mixed), rep(expected, 2))
  expect_identical(parse_connect_rfc3339(x_zero_offset), expected)
  expect_identical(parse_connect_rfc3339(x_plus_one), expected)
  expect_identical(parse_connect_rfc3339(x_minus_one), expected)
  expect_identical(parse_connect_rfc3339(single_zero_offset), expected[1])
  expect_identical(parse_connect_rfc3339(single_offset), expected[1])

  withr::local_envvar(TZ = "UTC")
  expected <- as.POSIXct(strptime(
    c(
      "2023-08-22T14:13:14+0000",
      "2020-01-01T01:02:03+0000"
    ),
    format = "%Y-%m-%dT%H:%M:%S%z",
    tz = Sys.timezone()
  ))
  expect_identical(parse_connect_rfc3339(x_mixed), rep(expected, 2))
  expect_identical(parse_connect_rfc3339(x_zero_offset), expected)
  expect_identical(parse_connect_rfc3339(x_plus_one), expected)
  expect_identical(parse_connect_rfc3339(x_minus_one), expected)
  expect_identical(parse_connect_rfc3339(single_zero_offset), expected[1])
  expect_identical(parse_connect_rfc3339(single_offset), expected[1])

  withr::local_envvar(TZ = "Asia/Tokyo")
  expected <- as.POSIXct(strptime(
    c(
      "2023-08-22T14:13:14+0000",
      "2020-01-01T01:02:03+0000"
    ),
    format = "%Y-%m-%dT%H:%M:%S%z",
    tz = Sys.timezone()
  ))
  expect_identical(parse_connect_rfc3339(x_mixed), rep(expected, 2))
  expect_identical(parse_connect_rfc3339(x_zero_offset), expected)
  expect_identical(parse_connect_rfc3339(x_plus_one), expected)
  expect_identical(parse_connect_rfc3339(x_minus_one), expected)
  expect_identical(parse_connect_rfc3339(single_zero_offset), expected[1])
  expect_identical(parse_connect_rfc3339(single_offset), expected[1])
})


test_that("parse_connect_rfc3339() handles fractional seconds", {
  withr::local_envvar(TZ = "UTC")
  expected <- as.POSIXct(strptime(
    c(
      "2024-12-06T19:09:29.948016766+0000",
      "2024-12-06T19:09:29.948070345+0000"
    ),
    format = "%Y-%m-%dT%H:%M:%OS%z",
    tz = Sys.timezone()
  ))

  x <- c("2024-12-06T19:09:29.948016766Z", "2024-12-06T19:09:29.948070345Z")

  expect_identical(parse_connect_rfc3339(x), expected)
})

test_that("make_timestamp produces expected output", {
  x_mixed <- c(
    "2023-08-22T14:13:14Z",
    "2020-01-01T01:02:03Z",
    "2023-08-22T15:13:14+01:00",
    "2020-01-01T00:02:03-01:00"
  )

  x_zero_offset <- c(
    "2023-08-22T14:13:14Z",
    "2020-01-01T01:02:03Z"
  )

  x_plus_one <- c(
    "2023-08-22T15:13:14+01:00",
    "2020-01-01T02:02:03+01:00"
  )

  x_minus_one <- c(
    "2023-08-22T13:13:14-01:00",
    "2020-01-01T00:02:03-01:00"
  )

  single_zero_offset <- "2023-08-22T14:13:14Z"

  single_offset <- "2023-08-22T15:13:14+01:00"

  outcome <- c(
    "2023-08-22T14:13:14Z",
    "2020-01-01T01:02:03Z"
  )

  withr::local_envvar(TZ = "America/New_York")
  expect_equal(
    make_timestamp(coerce_datetime(x_mixed, NA_datetime_)),
    rep(outcome, 2)
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_zero_offset, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_plus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_minus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_zero_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(make_timestamp(outcome), outcome)

  withr::local_envvar(TZ = "UTC")
  expect_equal(
    make_timestamp(coerce_datetime(x_mixed, NA_datetime_)),
    rep(outcome, 2)
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_zero_offset, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_plus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_minus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_zero_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(make_timestamp(outcome), outcome)

  withr::local_envvar(TZ = "Asia/Tokyo")
  expect_equal(
    make_timestamp(coerce_datetime(x_mixed, NA_datetime_)),
    rep(outcome, 2)
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_zero_offset, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_plus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(x_minus_one, NA_datetime_)),
    outcome
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_zero_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(
    make_timestamp(coerce_datetime(single_offset, NA_datetime_)),
    outcome[1]
  )
  expect_equal(make_timestamp(outcome), outcome)
})

test_that("make_timestamp is safe for strings", {
  expect_equal(make_timestamp("hello"), "hello")
  expect_equal(make_timestamp(rep("hello", 5)), rep("hello", 5))

  expect_equal(make_timestamp(NA_character_), NA_character_)
})

test_that("make_timestamp converts to character", {
  expect_type(make_timestamp(NA_datetime_), "character")
})

test_that("ensure_column works with lists", {
  list_chk_null <- ensure_column(tibble::tibble(), NA_list_, "hello")
  expect_s3_class(list_chk_null, "tbl_df")
  expect_type(list_chk_null$hello, "list")

  list_chk_same <- ensure_column(
    tibble::tibble(hello = list(list(1, 2, 3), list(1, 2, 3, 4))),
    NA_list_,
    "hello"
  )
  expect_s3_class(list_chk_same, "tbl_df")
  expect_type(list_chk_same$hello, "list")
})

test_that("ensure_column works with POSIXct", {
  time_chk_null <- ensure_column(tibble::tibble(), NA_datetime_, "hello")
  expect_s3_class(time_chk_null, "tbl_df")
  expect_s3_class(time_chk_null$hello, "POSIXct")

  time_chk_some <- ensure_column(
    tibble::tibble(one = c(1, 2, 3)),
    NA_datetime_,
    "hello"
  )
  expect_s3_class(time_chk_some, "tbl_df")
  expect_s3_class(time_chk_some$hello, "POSIXct")

  skip("Ahh! this fails presently. Are double -> POSIXct conversions allowed?")
  time_chk_convert <- ensure_column(
    tibble::tibble(hello = c(1, 2, 3)),
    NA_datetime_,
    "hello"
  )
  expect_s3_class(time_chk_convert, "tbl_df")
  expect_s3_class(time_chk_convert$hello, "POSIXct")
})

test_that("converts length one list", {
  hm <- ensure_column(tibble::tibble(one = "hi"), NA_list_, "one")
  expect_type(hm$one, "list")
})

test_that("parse_connectapi handles mixed null/non-null character values", {
  data <- list(
    list(guid = "aaa", bundle_id = NULL, name = "first"),
    list(guid = "bbb", bundle_id = "123", name = "second")
  )

  result <- parse_connectapi(data)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 2)
  expect_type(result$bundle_id, "character")
  expect_identical(result$bundle_id, c(NA_character_, "123"))
})

test_that("parse_connectapi handles mixed null/non-null datetime strings", {
  data <- list(
    list(guid = "aaa", active_time = NULL),
    list(guid = "bbb", active_time = "2023-08-22T14:13:14Z")
  )

  result <- parse_connectapi(data)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 2)
  expect_type(result$active_time, "character")
  expect_identical(result$active_time, c(NA_character_, "2023-08-22T14:13:14Z"))
})

test_that("parse_connectapi handles mixed null/non-null integer timestamps", {
  data <- list(
    list(key = "abc", start_time = 1732573574, end_time = NULL),
    list(key = "def", start_time = 1732553145, end_time = 1732556770)
  )

  result <- parse_connectapi(data)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 2)
  expect_type(result$end_time, "double")
  expect_identical(result$end_time, c(NA_real_, 1732556770))
})

test_that("ensure_column coerces integer to character when ptype expects character", {
  # Old Connect returns integer IDs
  int_data <- tibble::tibble(id = c(100L, 200L, 300L))
  result <- ensure_column(int_data, NA_character_, "id")
  expect_type(result$id, "character")
  expect_identical(result$id, c("100", "200", "300"))

  # New Connect returns character IDs

  char_data <- tibble::tibble(id = c("100", "200", "300"))
  result <- ensure_column(char_data, NA_character_, "id")
  expect_type(result$id, "character")
  expect_identical(result$id, c("100", "200", "300"))
})

test_that("parse_connectapi_typed produces character id for usage with integer input", {
  # Simulates old Connect returning integer IDs
  int_hits <- list(
    list(id = 123L, user_guid = "abc", content_guid = "def",
         timestamp = "2025-04-30T12:00:00Z", data = list(path = "/test")),
    list(id = 456L, user_guid = "ghi", content_guid = "jkl",
         timestamp = "2025-04-30T13:00:00Z", data = list(path = "/other"))
  )

  result <- parse_connectapi_typed(int_hits, connectapi_ptypes$usage)
  expect_type(result$id, "character")
  expect_identical(result$id, c("123", "456"))
})

test_that("parse_connectapi_typed produces character id for usage with character input", {
  # Simulates new Connect returning character IDs
  char_hits <- list(
    list(id = "123", user_guid = "abc", content_guid = "def",
         timestamp = "2025-04-30T12:00:00Z", data = list(path = "/test")),
    list(id = "456", user_guid = "ghi", content_guid = "jkl",
         timestamp = "2025-04-30T13:00:00Z", data = list(path = "/other"))
  )

  result <- parse_connectapi_typed(char_hits, connectapi_ptypes$usage)
  expect_type(result$id, "character")
  expect_identical(result$id, c("123", "456"))
})

structure(
  list(
    url = "__api__/v1/content/01234567/thumbnail",
    status_code = 404L,
    content = charToRaw("404 page not found"),
    headers = structure(
      list(`content-type` = "text/plain; charset=utf-8"),
      class = "insensitive"
    )
  ),
  class = "response"
)

#' Get content item thumbnail
#'
#' Download the thumbnail for a content item on Connect to a file on your
#' computer.
#'
#' @param content A content item.
#' @param path Optional. A path to a file used to write the thumbnail image. If
#' no path is provided, a temporary file with the correct file extension is
#' created.
#'
#' @returns The path to the downloaded image file, if `content` has a thumbnail; otherwise `NA`.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
#' thumbnail <- get_thumbnail(item)
#' }
#'
#' @family thumbnail functions
#' @family content functions
#' @include utils.R
#' @export
get_thumbnail <- function(content, path = NULL) {
  validate_R6_class(content, "Content")
  guid <- content$content$guid
  con <- content$connect

  # Connect 2024.09.0 introduced public endpoints for content thumbnails. We
  # prefer those, falling back to the unversioned endpoints if unavailable.
  # The "content/{guid}/__thumbnail__" URL is not under "__api__".
  res <- con$GET(
    url = con$server_url("content", guid, "__thumbnail__"),
    parser = NULL
  )
  if (httr::status_code(res) == 404) {
    res <- con$GET(
      unversioned_fallback_url("applications", guid, "image"),
      parser = NULL
    )
  }
  con$raise_error(res)

  if (httr::status_code(res) == 204) {
    return(NA_character_)
  }

  # Guess file extension
  if (is.null(path)) {
    path <- tempfile(
      pattern = glue::glue("content_image_{content$content$guid}_")
    )
  }
  ct <- httr::headers(res)$`content-type`
  if (ct %in% mime::mimemap) {
    exts <- names(mime::mimemap[mime::mimemap == ct])
    # Append the correct extension to the path if it does not match any of the
    # MIME type's valid file extensions.
    if (!any(sapply(exts, function(x) endsWith(path, x)))) {
      path <- paste(path, exts[1], sep = ".")
    }
  } else {
    warning(glue::glue(
      "Could not infer file extension from content type: {ct}."
    ))
  }

  writeBin(httr::content(res, as = "raw"), path)

  return(path)
}


#' Delete content item thumbnail
#'
#' Delete the thumbnail from a content item on Connect.
#'
#' @param content A content item.
#'
#' @returns The content item (invisibly).
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
#' thumbnail <- get_thumbnail(item)
#' }
#'
#' @family thumbnail functions
#' @family content functions
#' @export
delete_thumbnail <- function(content) {
  validate_R6_class(content, "Content")
  guid <- content$content$guid
  con <- content$connect

  # Connect 2024.09.0 introduced public endpoints for content thumbnails. We
  # prefer those, falling back to the unversioned endpoints if unavailable.
  res <- con$DELETE(
    v1_url("content", guid, "thumbnail"),
    parser = NULL
  )
  # API error code 17 indicates that the request was successful but the thumbnail does not exist.
  # In this case, we don't need to make another request.
  # https://docs.posit.co/connect/api/#overview--api-error-codes
  if (httr::status_code(res) == 404 && error_code(res) != 17) {
    res <- con$DELETE(
      unversioned_fallback_url("applications", guid, "image"),
      parser = NULL
    )
  }

  # API error code 17 indicates that the request was successful but the thumbnail does not exist.
  # We do not want to throw an error in this case.
  # https://docs.posit.co/connect/api/#overview--api-error-codes
  if (httr::status_code(res) == 404 && error_code(res) != 17) {
    con$raise_error(res)
  }

  invisible(content)
}

#' Check content item thumbnail
#'
#' Check whether a content item has a thumbnail.
#'
#' @param content A content item.
#'
#' @returns `TRUE` if the content item has a thumbnail, otherwise `FALSE`.
#'   Throws an error if you do not have permission to view the thumbnail.
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
#' has_thumbnail(item)
#' }
#'
#' @family thumbnail functions
#' @family content functions
#' @export
has_thumbnail <- function(content) {
  validate_R6_class(content, "Content")
  guid <- content$content$guid
  con <- content$connect

  # Connect 2024.09.0 introduced public endpoints for content thumbnails. We
  # prefer those, falling back to the unversioned endpoints if unavailable.
  # The "content/{guid}/__thumbnail__" URL is not under "__api__".
  res <- con$GET(
    url = con$server_url("content", guid, "__thumbnail__"),
    parser = NULL
  )
  if (httr::status_code(res) == 404) {
    res <- con$GET(
      unversioned_fallback_url("applications", guid, "image"),
      parser = NULL
    )
  }
  con$raise_error(res)
  httr::status_code(res) != 204
}

#' Set content item thumbnail
#'
#' Set the thumbnail for a content item.
#'
#' @param content A content item.
#' @param path Either a path to a local file or a URL to an image available over
#' HTTP/HTTPS. If `path` is an HTTP or HTTPS URL, the image will first
#' be downloaded.
#'
#' @returns The content item (invisibly).
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
#' set_thumbnail(item, "resources/image.png")
#' }
#'
#' @family thumbnail functions
#' @family content functions
#' @export
set_thumbnail <- function(content, path) {
  validate_R6_class(content, "Content")

  valid_path <- NULL
  if (file.exists(path)) {
    valid_path <- path
  } else {
    parsed <- httr::parse_url(path)
    if (!is.null(parsed$scheme) && parsed$scheme %in% c("http", "https")) {
      valid_path <- tempfile(
        pattern = "image",
        fileext = paste0(".", tools::file_ext(parsed[["path"]]))
      )
      res <- httr::GET(path, httr::write_disk(valid_path))
      on.exit(unlink(valid_path))
      if (httr::http_error(res)) {
        stop(glue::glue(
          "Could not download image from {path}: {httr::http_status(res)$message}"
        ))
      }
    }
  }
  if (is.null(valid_path)) {
    stop(glue::glue("Could not locate image at path: {path}"))
  }

  guid <- content$content$guid
  con <- content$connect

  # Connect 2024.09.0 introduced public endpoints for content thumbnails. We
  # prefer those, falling back to the unversioned endpoints if unavailable.
  res <- con$PUT(
    path = v1_url("content", guid, "thumbnail"),
    body = httr::upload_file(valid_path),
    parser = NULL
  )
  if (httr::status_code(res) == 404) {
    res <- con$POST(
      path = unversioned_fallback_url("applications", guid, "image"),
      body = httr::upload_file(valid_path),
      parser = NULL
    )
  }
  con$raise_error(res)

  # return the input (in case it inherits more than just Content)
  invisible(content)
}

mock_dirs <- c(
  "unversioned" = "2024.08.0",
  "v1" = "2024.09.0"
)

# The mocked content items are as follows:
# - 01234567:
#   - Returns 200 and a tiny jpeg when getting the thumbnail
#   - Returns 204 (success) when setting and deleting
# - 12345678: Returns 204 when getting the thumbnail (no thumbnail)
# - 23456789: Returns 404 for all endpoints
# - non-connect/missing-image/image.R: a 404 response

for (api_ver in names(mock_dirs)) {
  mock_dir <- mock_dirs[[api_ver]]
  with_mock_dir(mock_dir, {
    with_mock_api({
      client <- connect(server = "https://connect.example", api_key = "fake")
      test_that(
        glue::glue(
          "get_thumbnail() gets the thumbnail ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")

          # User-specified path
          user_path <- tempfile("thumbnail_", fileext = ".jpg")
          received_path <- get_thumbnail(item, user_path)
          received <- readBin(received_path, "raw", n = 8)
          expected <- as.raw(c(0x4e, 0x41))
          expect_equal(user_path, received_path)
          expect_equal(received, expected)

          # User-specified path (wrong extension)
          user_path <- tempfile("thumbnail_", fileext = ".png")
          received_path <- get_thumbnail(item, user_path)
          received <- readBin(received_path, "raw", n = 8)
          expected <- as.raw(c(0x4e, 0x41))
          expect_equal(paste0(user_path, ".jpeg"), received_path)
          expect_equal(received, expected)

          # Automatic path
          received_path <- get_thumbnail(item)
          received <- readBin(received_path, "raw", n = 8)
          expected <- as.raw(c(0x4e, 0x41))
          expect_equal(
            substring(received_path, nchar(received_path) - 4),
            ".jpeg"
          )
          expect_equal(received, expected)
        }
      )

      test_that(
        glue::glue(
          "get_thumbnail() returns NA_character_ for 204 status codes ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "12345678")

          # User-specified path
          user_path <- tempfile("thumbnail_", fileext = ".jpg")
          received_path <- get_thumbnail(item, user_path)
          expect_equal(received_path, NA_character_)

          # Automatic path
          received_path <- get_thumbnail(item)
          expect_equal(received_path, NA_character_)
        }
      )

      test_that(
        glue::glue(
          "get_thumbnail() errors with 404 status codes ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "23456789")
          expect_error(
            get_thumbnail(item),
            "request failed with Client error: \\(404\\) Not Found"
          )
        }
      )

      test_that(
        glue::glue(
          "has_thumbnail() returns TRUE when the item has a thumbnail ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")
          expect_true(has_thumbnail(item))
        }
      )

      test_that(
        glue::glue(
          "has_thumbnail() returns FALSE when the status code is 204 ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "12345678")
          expect_false(has_thumbnail(item))
        }
      )

      test_that(
        glue::glue(
          "has_thumbnail() errors with 404 status codes ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "23456789")
          expect_error(
            has_thumbnail(item),
            "request failed with Client error: \\(404\\) Not Found"
          )
        }
      )

      test_that(
        glue::glue(
          "set_thumbnail() returns returns the content item when successful ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")
          received <- set_thumbnail(item, "resources/smol.jpg")
          expect_true(validate_R6_class(received, "Content"))
          expect_identical(item, received)
        }
      )

      test_that(
        glue::glue(
          "set_thumbnail() raises an error when the endpoint returns a 404 ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "23456789")
          expect_error(
            set_thumbnail(item, "resources/smol.jpg"),
            "request failed with Client error: \\(404\\) Not Found"
          )
        }
      )

      test_that(
        glue::glue(
          "set_thumbnail() works with remote images ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")
          expect_GET(
            set_thumbnail(
              item,
              "https://other.server/non-connect/working-image/not-an-image"
            ),
            "https://other.server/non-connect/working-image/not-an-image"
          )
          # We're only asserting that the remote image hits GET. Because we
          # httr::write_disk() and then using what is written doesn't work
          # (https://github.com/nealrichardson/httptest/issues/86) the rest of
          # this function is tested elsewhere, so we have confidence that so long
          # as getting and writing to disk works, we are good.
        }
      )

      test_that(
        glue::glue(
          "set_thumbnail() returns an error when the remote image cannot be found ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")
          expect_error(
            set_thumbnail(
              item,
              "https://other.server/non-connect/missing-image/image.png"
            ),
            "Could not download image from https"
          )
        }
      )

      test_that(
        glue::glue(
          "delete_thumbnail() returns the content item when delete works ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "01234567")
          expect_identical(delete_thumbnail(item), item)
        }
      )

      test_that(
        glue::glue(
          "delete_thumbnail() throws an error for other 404 errors ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "23456789")
          expect_error(
            delete_thumbnail(item),
            "request failed with Client error: \\(404\\) Not Found"
          )
        }
      )

      test_that(
        glue::glue(
          "delete_thumbnail() returns the content for 404s indicating no thumbnail ({api_ver} - {mock_dir})"
        ),
        {
          item <- content_item(client, "34567890")
          expect_identical(delete_thumbnail(item), item)
        }
      )
    })
  })
}

#' Class representing a Connect API client
#'
#' @name PositConnect
#'
#' @section Usage:
#' \preformatted{
#' client <- Connect$new(server = 'connect.example.com',
#'   api_key = 'mysecretkey')
#'
#' get_content(client)
#' client$get_tags()
#' }
#'
#' @section Details:
#'
#' This class allows a user to interact with a Connect server via the Connect
#' API. Authentication is done by providing an API key.
#'
#' @importFrom utils capture.output
#'
#' @family R6 classes
#'
#' @export
Connect <- R6::R6Class(
  "Connect",
  public = list(
    #' @field server The base URL of your Posit Connect server.
    server = NULL,
    #' @field api_key Your Posit Connect API key.
    api_key = NULL,
    #' @field tags The initial set of tags.
    tags = NULL,
    #' @field tag_map The initial tag map.
    tag_map = NULL,
    #' @field httr_additions An initial set of `httr` configuration added to each HTTP call.
    httr_additions = list(),
    #' @field using_auth Indicates that the API key is added to each HTTP call.
    using_auth = TRUE,

    #' @description Initialize a new connect.
    #' @param server The base URL of your Posit Connect server.
    #' @param api_key Your Posit Connect API key.
    initialize = function(server, api_key) {
      message_if_not_testing(glue::glue(
        "Defining Connect with server: {server}"
      ))
      if (is.null(httr::parse_url(server)$scheme)) {
        stop(glue::glue(
          "ERROR: Please provide a protocol (http / https). You gave: {server}"
        ))
      }
      self$server <- base::sub("^(.*)/$", "\\1", server)
      self$api_key <- api_key
    },

    #' @description Set additional `httr` configuration that is added to each HTTP call.
    #' @param ... Set of httr configurations.
    httr_config = function(...) {
      self$httr_additions <- rlang::list2(...)
      invisible(self)
    },

    # helpers ----------------------------------------------------------

    #' @description Print details about this instance.
    #' @param ... Ignored.
    print = function(...) {
      cat("Posit Connect API Client: \n")
      cat("  Posit Connect Server: ", self$server, "\n", sep = "")
      cat(
        "  Posit Connect API Key: ",
        paste0(
          strrep("*", 11),
          substr(self$api_key, nchar(self$api_key) - 3, nchar(self$api_key))
        ),
        "\n",
        sep = ""
      )
      # TODO: something about API key... role... ?
      # TODO: point to docs on methods... how to see methods?
      cat("\n")
      invisible(self)
    },

    #' @description Raise an error when the HTTP result is an HTTP error.
    #' @param res HTTP result.
    raise_error = function(res) {
      if (httr::http_error(res)) {
        connect_error_details <- tryCatch(
          {
            cont <- httr::content(res)
            code <- sprintf("code: %d", cont$code)
            error <- sprintf("error: %s", cont$error)
            if (length(code) == 0 & length(error) == 0) {
              ""
            } else {
              paste0("(", paste0(c(code, error), collapse = ", "), ")")
            }
          },
          error = function(e) ""
        )

        err <- sprintf(
          "%s request failed with %s %s",
          res$request$url,
          httr::http_status(res)$message,
          connect_error_details
        )

        stop(err, call. = FALSE)
      }
    },

    #' @description Returns HTTP authorization headers, or NULL when none are used.
    add_auth = function() {
      if (self$using_auth) {
        httr::add_headers(Authorization = paste0("Key ", self$api_key))
      } else {
        NULL
      }
    },

    #' @description Build a URL relative to the API root
    #' @param ... path segments
    api_url = function(...) {
      self$server_url("__api__", ...)
    },

    #' @description Build a URL relative to the server root
    #' @param ... path segments
    server_url = function(...) {
      paste(self$server, ..., sep = "/")
    },

    #' @description General wrapper around `httr` verbs
    #' @param method HTTP request method
    #' @param url URL to request
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    #' @param ... Additional arguments passed to the request function
    request = function(method, url, ..., parser = "parsed") {
      old_opt <- options(scipen = 999)
      on.exit(options(old_opt), add = TRUE)

      httr_verb <- get(method, envir = asNamespace("httr"))
      res <- rlang::exec(
        httr_verb,
        !!!c(
          list(
            url,
            self$add_auth()
          ),
          rlang::list2(...),
          self$httr_additions
        )
      )
      check_debug(res)

      if (is.null(parser)) {
        res
      } else {
        self$raise_error(res)
        httr::content(res, as = parser)
      }
    },

    #' @description Perform an HTTP GET request of the named API path.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param ... Arguments to `httr::GET()`
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    GET = function(path, ..., url = self$api_url(path), parser = "parsed") {
      self$request("GET", url, parser = parser, ...)
    },

    #' @description Perform an HTTP PUT request of the named API path.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param body The HTTP payload.
    #' @param ... Arguments to `httr::PUT()`
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' @param encode How the payload is encoded.
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    PUT = function(
      path,
      body = "{}",
      ...,
      url = self$api_url(path),
      encode = "json",
      parser = "parsed"
    ) {
      self$request(
        "PUT",
        url,
        parser = parser,
        body = body,
        encode = encode,
        ...
      )
    },

    #' @description Perform an HTTP HEAD request of the named API path.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param ... Arguments to `httr::HEAD()`
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' `httr::content(res, as = parser)`.
    HEAD = function(path, ..., url = self$api_url(path)) {
      self$request("HEAD", url, parser = NULL, ...)
    },

    #' @description Perform an HTTP DELETE request of the named API path. Returns the HTTP response object.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param ... Arguments to `httr::DELETE()`
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    DELETE = function(path, ..., url = self$api_url(path), parser = NULL) {
      self$request("DELETE", url, parser = parser, ...)
    },

    #' @description Perform an HTTP PATCH request of the named API path.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param ... Arguments to `httr::PATCH()`
    #' @param body The HTTP payload.
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' @param encode How the payload is encoded.
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    PATCH = function(
      path,
      body = "{}",
      ...,
      url = self$api_url(path),
      encode = "json",
      parser = "parsed"
    ) {
      self$request(
        "PATCH",
        url,
        parser = parser,
        body = body,
        encode = encode,
        ...
      )
    },

    #' @description Perform an HTTP POST request of the named API path.
    #' @param path API path relative to the server's `/__api__` root.
    #' @param ... Arguments to `httr::POST()`
    #' @param body The HTTP payload.
    #' @param url Target URL. Default uses `path`, but provide `url` to request
    #' a server resource that is not under `/__api__`
    #' @param encode How the payload is encoded.
    #' @param parser How the response is parsed. If `NULL`, the `httr_response`
    #' will be returned. Otherwise, the argument is forwarded to
    #' `httr::content(res, as = parser)`.
    POST = function(
      path,
      body = "{}",
      ...,
      url = self$api_url(path),
      encode = "json",
      parser = "parsed"
    ) {
      self$request(
        "POST",
        url,
        parser = parser,
        body = body,
        encode = encode,
        ...
      )
    },

    #' @description Perform an HTTP GET request of the "me" server endpoint.
    me = function() {
      self$GET(unversioned_url("me"))
    },

    #' @description Return the base URL of the Connect server.
    get_dashboard_url = function() {
      self$server
    },

    # tags ----------------------------------------------------------

    #' @description Return all tags.
    #' @param use_cache Indicates that a cached set of tags is used.
    get_tags = function(use_cache = FALSE) {
      error_if_less_than(self$version, "1.8.6")
      # TODO: check cache "age"?
      if (is.null(self$tags) || !use_cache) {
        self$tags <- self$tag()
      }
      # TODO: deprecate tag_map? by removing all of the things that use it...
      # caching is hard
      self$tag_map <- data.frame(
        id = sapply(self$tags, function(x) {
          as.character(x$id)
        }),
        name = sapply(self$tags, function(x) {
          x$name
        }),
        stringsAsFactors = FALSE
      )
      self$tag_map
    },

    #' @description Get the identifier for the named tag.
    #' @param tagname The name of the tag.
    get_tag_id = function(tagname) {
      self$get_tags()
      if (!any(self$tag_map$name == tagname)) {
        stop(sprintf("Tag %s not found on server %s", tagname, self$server))
      }
      self$tag_map[which(self$tag_map$name == tagname), "id"]
    },

    #' @description Get the tag tree.
    get_tag_tree = function() {
      raw_tags <- self$tag()
      tag_tree_parse_data(raw_tags)
    },

    #' @description Create a tag.
    #' @param name The tag name.
    #' @param parent_id The parent identifier.
    tag_create_safe = function(name, parent_id = NULL) {
      tt <- get_tags(self)

      tag_exists_id <- recursive_find_tag(tt, name, parent_id)
      if (is.na(tag_exists_id)) {
        self$tag_create(name, parent_id)
      } else {
        self$tag(tag_exists_id)
      }
    },

    #' @description Create a tag.
    #' @param name The tag name.
    #' @param parent_id The parent identifier.
    tag_create = function(name, parent_id = NULL) {
      error_if_less_than(self$version, "1.8.6")
      dat <- list(
        name = name
      )
      if (!is.null(parent_id)) {
        if (is.numeric(parent_id)) {
          warn_once("Converting `tag parent_id` to character", "tag_parent_id")
          parent_id <- as.character(parent_id)
        }
        dat <- c(
          dat,
          parent_id = parent_id
        )
      }
      self$POST(
        v1_url("tags"),
        body = dat
      )
    },

    #' @description Get a tag.
    #' @param id The tag identifier.
    tag = function(id = NULL) {
      error_if_less_than(self$version, "1.8.6")
      if (is.null(id)) {
        path <- v1_url("tags")
      } else {
        path <- v1_url("tags", id)
      }
      self$GET(path)
    },

    #' @description Delete a tag.
    #' @param id The tag identifier.
    tag_delete = function(id) {
      invisible(self$DELETE(v1_url("tags", id)))
    },

    #' @description Get a schedule.
    #' @param schedule_id The schedule identifier.
    get_schedule = function(schedule_id) {
      path <- unversioned_url("schedules", schedule_id)
      self$GET(path)
    },

    # content ----------------------------------------------------------

    #' @description Create content.
    #' @param name The content name.
    #' @param title The content title.
    #' @param ... Other content fields.
    content_create = function(name, title = name, ...) {
      path <- v1_url("content")
      other_params <- rlang::dots_list(...)

      verify_content_name(name)
      self$POST(
        path,
        body = c(
          list(name = name, title = title),
          other_params
        )
      )
    },

    #' @description Upload a content bundle.
    #' @param bundle_path The path to the bundle archive.
    #' @param guid The content GUID.
    content_upload = function(bundle_path, guid) {
      # todo : add X-Content-Checksum
      self$POST(
        v1_url("content", guid, "bundles"),
        body = httr::upload_file(bundle_path),
        encode = "raw"
      )
    },

    #' @description Deploy a content bundle.
    #' @param guid The content GUID.
    #' @param bundle_id The bundle identifier.
    content_deploy = function(guid, bundle_id) {
      path <- v1_url("content", guid, "deploy")
      self$POST(path, body = list(bundle_id = as.character(bundle_id)))
    },

    #' @description Get a content item.
    #' @param guid The content GUID.
    #' @param owner_guid The target content owner.
    #' @param name The target name.
    #' @param include Additional response fields.
    content = function(
      guid = NULL,
      owner_guid = NULL,
      name = NULL,
      include = "tags,owner"
    ) {
      if (!is.null(guid)) {
        return(self$GET(
          v1_url("content", guid),
          query = list(include = include)
        ))
      }

      query <- list(
        owner_guid = owner_guid,
        name = name,
        include = include
      )
      path <- v1_url("content")
      self$GET(path, query = query)
    },

    #' @description Get a task.
    #' @param task_id The task identifier.
    #' @param first The initial status position.
    #' @param wait Maximum time to wait for update.
    task = function(task_id, first = 0, wait = 5) {
      path <- v1_url("tasks", task_id)
      self$GET(path, query = list(first = first, wait = wait))
    },

    #' @description Set a tag for a content item.
    #' @param content_id The content identifier.
    #' @param tag_id The tag identifier.
    set_content_tag = function(content_id, tag_id) {
      self$POST(
        path = v1_url("content", content_id, "tags"),
        body = list(
          tag_id = as.character(tag_id)
        )
      )
    },

    #' @description Remove a tag from a content item.
    #' @param content_id The content identifier.
    #' @param tag_id The tag identifier.
    remove_content_tag = function(content_id, tag_id) {
      invisible(self$DELETE(
        path = v1_url("content", content_id, "tags", tag_id)
      ))
    },

    # users -----------------------------------------------

    #' @description Get user details.
    #' @param guid The user GUID.
    user = function(guid) {
      self$GET(v1_url("users", guid))
    },

    #' @description Get users.
    #' @param page_number The page number.
    #' @param prefix The search term.
    #' @param page_size The page size.
    #' @param user_role Filter by user role.
    #' @param account_status Filter by account status.
    users = function(
      page_number = 1,
      prefix = NULL,
      page_size = 500,
      user_role = NULL,
      account_status = NULL
    ) {
      path <- v1_url("users")
      if (!is.null(user_role)) {
        user_role <- paste(user_role, collapse = "|")
      }
      if (!is.null(account_status)) {
        account_status <- paste(account_status, collapse = "|")
      }
      query <- list(
        page_number = page_number,
        page_size = valid_page_size(page_size),
        prefix = prefix,
        user_role = user_role,
        account_status = account_status
      )
      self$GET(path, query = query)
    },

    #' @description Get remote users.
    #' @param prefix The search term.
    users_remote = function(prefix) {
      # No pagination here?
      path <- v1_url("users", "remote")
      query <- list(prefix = prefix)
      self$GET(path, query = query)
    },

    #' @description Create a user.
    #' @param username The username.
    #' @param email Email address.
    #' @param first_name First name.
    #' @param last_name Last name.
    #' @param password The password.
    #' @param user_must_set_password Indicates that user sets password on first login.
    #' @param user_role Role for user.
    #' @param unique_id Identifier for user.
    users_create = function(
      username,
      email,
      first_name = NULL,
      last_name = NULL,
      password = NULL,
      user_must_set_password = NULL,
      user_role = NULL,
      unique_id = NULL
    ) {
      path <- v1_url("users")
      self$POST(
        path = path,
        body = list(
          email = email,
          first_name = first_name,
          last_name = last_name,
          password = password,
          user_must_set_password = user_must_set_password,
          user_role = user_role,
          username = username,
          unique_id = unique_id
        )
      )
    },

    #' @description Create a remote user.
    #' @param temp_ticket Ticket identifying target remote user.
    users_create_remote = function(temp_ticket) {
      path <- v1_url("users")
      self$PUT(
        path = path,
        body = list(temp_ticket = temp_ticket)
      )
    },

    #' @description Lock a user.
    #' @param user_guid User GUID.
    users_lock = function(user_guid) {
      path <- v1_url("users", user_guid, "lock")
      message(path)
      self$POST(
        path = path,
        body = list(locked = TRUE)
      )
    },

    #' @description Unlock a user.
    #' @param user_guid User GUID.
    users_unlock = function(user_guid) {
      path <- v1_url("users", user_guid, "lock")
      self$POST(
        path = path,
        body = list(locked = FALSE)
      )
    },

    #' @description Update a user.
    #' @param user_guid User GUID.
    #' @param ... User fields.
    users_update = function(user_guid, ...) {
      path <- v1_url("users", user_guid)
      self$PUT(
        path = path,
        body = rlang::list2(...)
      )
    },

    # groups -----------------------------------------------------

    #' @description Get groups.
    #' @param page_number The page number.
    #' @param prefix The search term.
    #' @param page_size The page size.
    groups = function(page_number = 1, prefix = NULL, page_size = 500) {
      path <- v1_url("groups")
      query <- list(
        page_number = page_number,
        page_size = valid_page_size(page_size),
        prefix = prefix
      )
      self$GET(path, query = query)
    },

    #' @description Get group members.
    #' @param guid The group GUID.
    group_members = function(guid) {
      path <- v1_url("groups", guid, "members")
      self$GET(path)
    },

    #' @description Add a group member.
    #' @param group_guid The group GUID.
    #' @param user_guid The user GUID.
    group_member_add = function(group_guid, user_guid) {
      path <- v1_url("groups", group_guid, "members")
      self$POST(path, body = list(user_guid = user_guid))
    },

    #' @description Remove a group member.
    #' @param group_guid The group GUID.
    #' @param user_guid The user GUID.
    group_member_remove = function(group_guid, user_guid) {
      path <- v1_url("groups", group_guid, "members", user_guid)
      self$DELETE(path)
    },

    #' @description Create a group.
    #' @param name The group name.
    groups_create = function(name) {
      path <- v1_url("groups")
      self$POST(
        path = path,
        body = list(name = name)
      )
    },

    #' @description Create a remote group.
    #' @param temp_ticket Ticket identifying target remote group.
    groups_create_remote = function(temp_ticket) {
      path <- v1_url("groups")
      self$PUT(
        path = path,
        body = list(temp_ticket = temp_ticket)
      )
    },

    #' @description Get remote groups.
    #' @param prefix The search term.
    #' @param limit The maximal result set size.
    groups_remote = function(prefix = NULL, limit = 500) {
      path <- v1_url("groups", "remote")
      query <- list(
        limit = valid_page_size(limit),
        prefix = prefix
      )
      self$GET(path, query = query)
    },

    #' @description Get content to which a group has access
    #' @param guid The group GUID.
    group_content = function(guid) {
      path <- v1_url("experimental", "groups", guid, "content")
      self$GET(path)
    },

    # instrumentation --------------------------------------------

    #' @description Get (non-interactive) content visits.
    #' @param content_guid Content GUID.
    #' @param min_data_version Data version for request.
    #' @param from Start of range.
    #' @param to End of range.
    #' @param limit Result set size.
    #' @param previous Previous item.
    #' @param nxt Next item.
    #' @param asc_order Indicates ascending result order.
    inst_content_visits = function(
      content_guid = NULL,
      min_data_version = NULL,
      from = NULL,
      to = NULL,
      limit = 500,
      previous = NULL,
      nxt = NULL,
      asc_order = TRUE
    ) {
      path <- v1_url("instrumentation", "content", "visits")
      query <- list(
        content_guid = content_guid,
        min_data_version = min_data_version,
        from = make_timestamp(from),
        to = make_timestamp(to),
        limit = valid_page_size(limit),
        previous = previous,
        asc_order = tolower(as.character(asc_order))
      )
      if (length(content_guid)) {
        query$content_guid <- paste(content_guid, collapse = "|")
      }
      # This is funky because next is a reserved word in R
      query[["next"]] <- nxt
      self$GET(path, query = query)
    },

    #' @description Get interactive content visits.
    #' @description Get (non-interactive) content visits.
    #' @param content_guid Content GUID.
    #' @param min_data_version Data version for request.
    #' @param from Start of range.
    #' @param to End of range.
    #' @param limit Result set size.
    #' @param previous Previous item.
    #' @param nxt Next item.
    #' @param asc_order Indicates ascending result order.
    inst_shiny_usage = function(
      content_guid = NULL,
      min_data_version = NULL,
      from = NULL,
      to = NULL,
      limit = 500,
      previous = NULL,
      nxt = NULL,
      asc_order = TRUE
    ) {
      path <- v1_url("instrumentation", "shiny", "usage")
      query <- list(
        content_guid = content_guid,
        min_data_version = min_data_version,
        from = make_timestamp(from),
        to = make_timestamp(to),
        limit = valid_page_size(limit),
        previous = previous,
        asc_order = tolower(as.character(asc_order))
      )
      if (length(content_guid)) {
        query$content_guid <- paste(content_guid, collapse = "|")
      }
      # This is funky because next is a reserved word in R
      query[["next"]] <- nxt
      self$GET(path, query = query)
    },

    #' @description Get running processes.
    procs = function() {
      warn_experimental("procs")
      path <- unversioned_url("metrics", "procs")
      self$GET(path)
    },

    # repo ------------------------------------------------------

    #' @description Determine if Git repository is associated with authorization.
    #' @param host Repository URL.
    repo_account = function(host) {
      warn_experimental("repo_account")
      parsed_url <- httr::parse_url(host)
      if (is.null(parsed_url$scheme) || is.null(parsed_url$hostname)) {
        stop(glue::glue(
          "Scheme and hostname must be provided (i.e. 'https://github.com'). You provided '{host}'"
        ))
      }
      host <- glue::glue(parsed_url$scheme, "://", parsed_url$hostname)
      path <- unversioned_url("repo", "account")
      self$GET(path, query = list(url = host))
    },

    #' @description Get Git repository branches.
    #' @param repo Repository URL.
    repo_branches = function(repo) {
      warn_experimental("repo_branches")
      path <- unversioned_url("repo", "branches")
      self$GET(path, query = list(url = repo))
    },

    #' @description Get Git repository directories.
    #' @param repo Repository URL.
    #' @param branch Repository branch.
    repo_manifest_dirs = function(repo, branch) {
      warn_experimental("repo_manifest_dirs")
      path <- unversioned_url("repo", "manifest-dirs")
      self$GET(path, query = list(url = repo, branch = branch))
    },

    # schedule --------------------------------------------------

    #' @description Get schedules.
    #' @param start Starting time.
    #' @param end Ending time.
    #' @param detailed Indicates detailed schedule information.
    schedules = function(
      start = Sys.time(),
      end = Sys.time() + 60 * 60 * 24 * 7,
      detailed = FALSE
    ) {
      warn_experimental("schedules")
      url <- v1_url("experimental", "schedules")
      query_params <- list(
        detailed = tolower(detailed),
        start = datetime_to_rfc3339(start),
        end = datetime_to_rfc3339(end)
      )
      res <- self$GET(url, query = query_params)
      return(res[["schedules"]])
    },

    # packages --------------------------------------------------

    #' @description Get packages. This endpoint is paginated.
    #' @param name The package name to filter by.
    #' @param page_number Page number.
    #' @param page_size Page size, default 100000.
    packages = function(name = NULL, page_number = 1, page_size = 100000) {
      url <- v1_url("packages")
      query_params <- list(
        name = name,
        page_number = page_number,
        page_size = page_size
      )
      self$GET(url, query = query_params)
    },

    # misc utilities --------------------------------------------

    #' @description Get documentation.
    #' @param docs Named document.
    #' @param browse Open a browser.
    docs = function(docs = "api", browse = TRUE) {
      stopifnot(docs %in% c("admin", "user", "api"))
      url <- paste0(self$server, "/__docs__/", docs)
      if (browse) {
        utils::browseURL(url)
      }
      url
    },

    #' @description Get auditing.
    #' @param limit Result set size.
    #' @param previous Previous item.
    #' @param nxt Next item.
    #' @param asc_order Indicates ascending result order.
    audit_logs = function(
      limit = 500,
      previous = NULL,
      nxt = NULL,
      asc_order = TRUE
    ) {
      path <- v1_url("audit_logs")
      query <- list(
        limit = valid_page_size(limit),
        previous = previous,
        ascOrder = tolower(as.character(asc_order))
      )
      # This is funky because next is a reserved word in R
      query[["next"]] <- nxt
      self$GET(path = path, query = query)
    },

    #' @description Get all vanity URLs
    vanities = function() {
      path <- v1_url("vanities")
      self$GET(path = path)
    },

    #' @description Get server settings.
    server_settings = function() {
      self$GET(unversioned_url("server_settings"))
    }

    # end --------------------------------------------------------
  ),
  private = list(
    .version = NULL,
    .timezones = NULL
  ),
  active = list(
    #' @field version The server version.
    version = function() {
      if (is.null(private$.version)) {
        private$.version <- safe_server_version(self)
      }
      private$.version
    },
    #' @field timezones The server timezones.
    timezones = function() {
      if (is.null(private$.timezones)) {
        private$.timezones <- tryCatch(
          self$GET(v1_url("timezones")),
          error = function(e) {
            self$GET(unversioned_fallback_url("timezones"))
          }
        )
      }
      private$.timezones
    }
  )
)

#' Create a connection to Posit Connect
#'
#' Creates a connection to Posit Connect using the server URL and an API key.
#' Validates the connection and checks that the version of the server is
#' compatible with the current version of the package.
#'
#' When running on Connect, the client's environment will contain default
#' `CONNECT_SERVER` and `CONNECT_API_KEY` variables. The API key's permissions
#' are scoped to the publishing user's account.
#'
#' To create a client with permissions scoped to the content visitor's account,
#' call `connect()` passing a user session token from content session headers
#' to the `token` argument. To do this, you must first add a Connect API
#' integration in your published content's Access sidebar.
#'
#' @param server The URL for accessing Posit Connect. Defaults to environment
#'   variable CONNECT_SERVER
#' @param api_key The API Key to authenticate to Posit Connect with. Defaults
#'   to environment variable CONNECT_API_KEY
#' @param token Optional. A user session token. When running on a Connect server,
#'   creates a client using the content visitor's account. Running locally, the
#'   created client uses the provided API key.
#' @param token_local_testing_key Optional. Only used when not running on
#'   Connect and a `token` is provided. By default, the function returns a
#'   `Connect` object using the `api_key`. By providing a different
#'   key here you can test a visitor client with differently-scoped
#'   permissions.
#' @param audience Optional. The GUID of a Connect API integration associated with this piece of content.
#' @param ... Additional arguments. Not used at present
#' @param .check_is_fatal Whether to fail if "check" requests fail. Useful in
#'   rare cases where more http request customization is needed for requests to
#'   succeed.
#' @return A Posit Connect R6 object that can be passed along to methods
#'
#' @rdname connect
#'
#' @examples
#' \dontrun{
#' client <- connect()
#'
#' # Running in Connect, create a client using the content visitor's account.
#' # This example assumes code is being executed in a Shiny app's `server`
#' # function with a `session` object available.
#' token <- session$request$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
#' client <- connect(token = token)
#'
#' # Test locally with an API key using a different role.
#' fallback_key <- Sys.getenv("VIEWER_ROLE_API_KEY")
#' client <- connect(token = token, token_local_testing_key = fallback_key)
#' }
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' # default is to read CONNECT_SERVER and CONNECT_API_KEY environment variables
#' connect()
#'
#' @export
connect <- function(
  server = Sys.getenv("CONNECT_SERVER"),
  api_key = Sys.getenv("CONNECT_API_KEY"),
  token,
  token_local_testing_key = api_key,
  audience = NULL,
  ...,
  .check_is_fatal = TRUE
) {
  if (
    missing(token) && is.null(api_key) || is.na(api_key) || nchar(api_key) == 0
  ) {
    msg <- "Invalid (empty) API key. Please provide a valid API key"
    if (.check_is_fatal) {
      stop(glue::glue("ERROR: {msg}"))
    } else {
      message(msg)
    }
  }
  con <- Connect$new(server = server, api_key = api_key)

  if (!missing(token)) {
    error_if_less_than(con$version, "2025.01.0")
    if (on_connect() || on_connect_cloud()) {
      visitor_creds <- get_oauth_credentials(
        con,
        user_session_token = token,
        requested_token_type = "urn:posit:connect:api-key",
        audience = audience
      )
      con <- connect(server = server, api_key = visitor_creds$access_token)
    } else {
      con <- connect(server = server, api_key = token_local_testing_key)
      message(paste0(
        "Called with `token` but not running on Connect or Connect Cloud. ",
        "Continuing with fallback API key."
      ))
    }
  }

  tryCatch(
    {
      check_connect_license(con)
      warn_untested_connect(using_version = safe_server_version(con))
    },
    error = function(err) {
      if (.check_is_fatal) {
        stop(err)
      } else {
        if (inherits(err, "error")) {
          message(err$message)
        } else {
          message(err)
        }
      }
    }
  )

  con
}

check_debug <- function(res) {
  # Check for deprecation warnings from the server.
  # You might get these if you've upgraded the Connect server but not connectapi.
  # connectapi will make the right request based on the version of the server,
  # but if you have an old version of the package, it won't know the new URL
  # to request.
  if ("X-Deprecated-Endpoint" %in% names(httr::headers(res))) {
    msg <- paste(
      res[["request"]][["url"]],
      "is deprecated and will be removed in a future version of Connect.",
      "Please upgrade `connectapi` in order to use the new APIs."
    )
    warn_once(msg, id = "X-Deprecated-Endpoint", class = "deprecatedWarning")
  }
  if (getOption("connect.debug", FALSE)) {
    message(paste(res[["request"]][["method"]], res[["request"]][["url"]]))
    message(paste("Response", res[["status_code"]]))
    message(httr::content(res, as = "text"))
  }
}

test_that("users_create works", {
  ss <- client$server_settings()
  if (ss$authentication$name %in% c("LDAP")) {
    skip("not implemented for this authentication provider")
  }
  username <- create_random_name()
  password <- uuid::UUIDgenerate(use.time = TRUE)
  res <- client$users_create(
    username = username,
    first_name = "Test",
    last_name = "User",
    user_role = "publisher",
    email = "test@example.com",
    password = password
  )

  expect_equal(res$username, username)
})

test_that("users works", {
  users <- client$users()

  expect_gt(length(users$results), 0)
})

test_that("user_guid_from_username works", {
  expect_error(
    user_guid_from_username(client, "this-user-prefix-does-not-exist"),
    "user not found"
  )

  user_username <- create_random_name(20)
  user_res <- client$users_create(
    user_username,
    "example@example.com",
    password = user_username
  )

  user_username_2 <- paste0(user_username, "X")
  user_res_2 <- client$users_create(
    user_username_2,
    "example@example.com",
    password = user_username_2
  )

  expect_warning(
    user_guid_from_username(client, substr(user_username, 0, 19)),
    "multiple users found"
  )

  expect_equal(
    user_guid_from_username(client, user_username),
    user_res$guid
  )
  expect_equal(
    user_guid_from_username(client, user_username_2),
    user_res_2$guid
  )
})

cont1_name <- uuid::UUIDgenerate()
cont1_title <- "Test Content 1"
cont1_guid <- NULL
cont1_bundle <- NULL

test_that("can create content", {
  cont1 <- client$content_create(name = cont1_name, title = cont1_title)
  expect_equal(cont1$name, cont1_name)
  expect_equal(cont1$title, cont1_title)

  # Use include = NULL because we aren't passing the include in content_create()
  get_cont1 <- client$content(guid = cont1$guid, include = NULL)
  expect_identical(get_cont1, cont1)
  cont1_guid <<- cont1$guid
})

test_that("can upload and deploy content", {
  cont1_bundle <<- bundle_dir(
    rprojroot::find_package_root_file("tests/testthat/test-plot")
  )
  expect_true(fs::file_exists(cont1_bundle$path))

  res <- client$content_upload(
    bundle_path = cont1_bundle$path,
    guid = cont1_guid
  )
  expect_false(is.null(res))
  expect_silent(as.integer(res[["bundle_id"]]))

  task <- client$content_deploy(
    guid = cont1_guid,
    bundle_id = res[["bundle_id"]]
  )
  expect_type(task[["task_id"]], "character")
})

test_that("can promote content to another server", {
  # In practice, you would have this be a different server, but we should be
  # able to test using the same server.
  client2 <- connect()

  # TODO : Intermittent failures here... with a 404 response on GET
  # during the download_bundle... connect.R:154
  res <- promote(
    from = Sys.getenv("CONNECT_SERVER"),
    from_key = Sys.getenv("CONNECT_API_KEY"),
    to = Sys.getenv("CONNECT_SERVER"),
    to_key = Sys.getenv("CONNECT_API_KEY"),
    name = cont1_name
  )

  expect_type(res, "character")

  cont1_2 <- content_ensure(
    connect = client2,
    name = cont1_name
  )

  expect_identical(cont1_name, cont1_2[["name"]])
})

test_that("content_ensure works with guid", {
  c1 <- content_ensure(client, guid = cont1_guid)
  expect_identical(c1[["guid"]], cont1_guid)

  fake_guid <- paste0(cont1_guid, "-does-not-exist")
  expect_warning({
    c2 <- content_ensure(client, guid = fake_guid)
  })
  expect_false(identical(c2[["guid"]], cont1_guid))
})

test_that("content_ensure works with name", {
  expect_message(c_new <- content_ensure(client))
  expect_type(c_new[["guid"]], "character")

  expect_message(
    c_same <- content_ensure(client, name = c_new[["name"]])
  )

  expect_identical(c_new[["name"]], c_same[["name"]])
  expect_identical(c_new[["guid"]], c_same[["guid"]])

  c_newname <- paste0(c_new[["name"]], "-alternate")
  c_title <- "Some Title"
  c_desc <- "Some Description"
  expect_message(
    c_diff <- content_ensure(
      client,
      name = c_newname,
      title = c_title,
      description = c_desc
    )
  )

  expect_false(identical(c_new[["name"]], c_diff[["name"]]))
  expect_false(identical(c_new[["guid"]], c_diff[["guid"]]))
  expect_identical(c_newname, c_diff[["name"]])
  expect_identical(c_title, c_diff[["title"]])
  expect_identical(c_desc, c_diff[["description"]])
})

test_that("content_ensure fails when not permitted", {
  permit_guid <- uuid::UUIDgenerate()
  permit_name <- create_random_name()

  # duplicates to ensure it does not create
  expect_error(
    content_ensure(client, guid = permit_guid, .permitted = c("existing")),
    "not found on"
  )
  expect_error(
    content_ensure(client, guid = permit_guid, .permitted = c("existing")),
    "not found on"
  )

  # duplicates to ensure it does not create
  expect_error(
    content_ensure(client, name = permit_name, .permitted = c("existing")),
    "not found on"
  )
  expect_error(
    content_ensure(client, name = permit_name, .permitted = c("existing")),
    "not found on"
  )

  # actually create
  invisible(content_ensure(client, name = permit_name))

  # error because we expect new
  expect_error(
    content_ensure(client, name = permit_name, .permitted = c("new")),
    "already exists"
  )
})

# This whole suite assumes CONNECT_SERVER and CONNECT_API_KEY env vars are set
test_that("connect() works", {
  expect_true(validate_R6_class(connect(), "Connect"))
})

test_that("connect fails for good server, bad api key", {
  expect_error({
    connect(
      api_key = "bogus"
    )
  })
})

without_internet({
  visits_url <- function(querypath) {
    glue::glue(
      "https://connect.example/__api__/v1/instrumentation/content/visits?{querypath}"
    )
  }
  usage_url <- function(querypath) {
    glue::glue(
      "https://connect.example/__api__/v1/instrumentation/shiny/usage?{querypath}"
    )
  }

  test_that("Query params to inst_content_visits", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$inst_content_visits(),
      visits_url("limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_content_visits(
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
      ),
      visits_url(
        "content_guid=f2f37341-e21d-3d80-c698-a935ad614066&limit=500&asc_order=true"
      )
    )
    # Multiple GUIDs are | separated
    expect_GET(
      # Use short ids for readability
      con$inst_content_visits(content_guid = c("1234", "5678")),
      visits_url("content_guid=1234%7C5678&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_content_visits(asc_order = FALSE),
      visits_url("limit=500&asc_order=false")
    )

    expect_GET(
      con$inst_content_visits(limit = 10, previous = "asdf", nxt = "qwer"),
      visits_url("limit=10&previous=asdf&asc_order=true&next=qwer")
    )

    expect_GET(
      con$inst_content_visits(from = "2022-01-01", to = "2022-01-31"),
      visits_url("from=2022-01-01&to=2022-01-31&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_content_visits(min_data_version = 0),
      visits_url("min_data_version=0&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_content_visits(limit = 1000),
      visits_url("limit=500&asc_order=true")
    )
  })

  test_that("Query params to inst_shiny_usage", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$inst_shiny_usage(),
      usage_url("limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_shiny_usage(
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
      ),
      usage_url(
        "content_guid=f2f37341-e21d-3d80-c698-a935ad614066&limit=500&asc_order=true"
      )
    )
    # Multiple GUIDs are | separated
    expect_GET(
      # Use short ids for readability
      con$inst_shiny_usage(content_guid = c("1234", "5678")),
      usage_url("content_guid=1234%7C5678&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_shiny_usage(asc_order = FALSE),
      usage_url("limit=500&asc_order=false")
    )

    expect_GET(
      con$inst_shiny_usage(limit = 10, previous = "asdf", nxt = "qwer"),
      usage_url("limit=10&previous=asdf&asc_order=true&next=qwer")
    )

    expect_GET(
      con$inst_shiny_usage(from = "2022-01-01", to = "2022-01-31"),
      usage_url("from=2022-01-01&to=2022-01-31&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_shiny_usage(min_data_version = 0),
      usage_url("min_data_version=0&limit=500&asc_order=true")
    )

    expect_GET(
      con$inst_shiny_usage(limit = 1000),
      usage_url("limit=500&asc_order=true")
    )
  })
})

test_that("audit_logs work", {
  logs <- client$audit_logs()
  expect_gt(length(logs$results), 0)

  logs2 <- client$audit_logs(nxt = logs$paging$cursors$`next`)
  expect_gt(length(logs2$results), 0)
})

test_that("server_settings work", {
  ss <- client$server_settings()

  expect_gt(length(ss), 0)
})

without_internet({
  test_that("Querying groups by prefix", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      client$groups(prefix = "TEST"),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=500&prefix=TEST"
    )

    # Now with spaces
    expect_GET(
      client$groups(prefix = "A Group Name", page_size = 5),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=5&prefix=A%20Group%20Name"
    )

    # Now with too big page size: it automatically truncates (maybe that's bad)
    expect_GET(
      client$groups(page_size = 1000),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=500"
    )
  })

  test_that("Querying remote groups", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      client$groups_remote(),
      "https://connect.example/__api__/v1/groups/remote?limit=500"
    )
    expect_GET(
      client$groups_remote(prefix = "A group name"),
      "https://connect.example/__api__/v1/groups/remote?limit=500&prefix=A%20group%20name"
    )
    expect_GET(
      client$groups_remote(limit = 1000),
      "https://connect.example/__api__/v1/groups/remote?limit=500"
    )
  })
})

with_mock_dir("2024.08.0", {
  client <- connect(server = "https://connect.example", api_key = "fake")
  test_that("get_groups() paginates with no prefix", {
    # To get this result, the code has to paginate through two API requests.
    # groups-4eaf46.json
    # groups-125d47.json

    result <- get_groups(client, page_size = 5, limit = 10)
    expected_names <- c(
      "~!@#$%^&*()_+",
      "1111",
      "2_viewer_group",
      "amanda_test_group",
      "a_new_group",
      "azurepipelines",
      "cgGroup01",
      "chris_test_group",
      "connect_dev",
      "cool_kids_of_the_dmv"
    )
    expect_identical(result$name, expected_names)
  })

  test_that("get_groups() does not paginate when called with a prefix on old Connect", {
    # Only one response exists for this query; by succeeding this test verifies
    # that the pagination behavior is not engaged, and that the request is made
    # with page_size=2 so that you get everything you asked for in one page.
    # groups-deae1f.json

    result <- get_groups(client, page_size = 1, limit = 2, prefix = "c")
    expect_identical(result$name, c("connect_dev", "cool_kids_of_the_dmv"))

    # Just to be explicit, here you can see that page_number and page_size are
    # aligned and you get page_size = 2
    expect_GET(
      get_groups(client, page_size = 1, limit = 2, prefix = "asdf"),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=2&prefix=asdf"
    )
    # Max mage size is 500, so that's the best we can do
    expect_GET(
      get_groups(client, prefix = "asdf"),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=500&prefix=asdf"
    )
  })
})

with_mock_dir("2025.04.0", {
  test_that("get_groups() does paginate when called with a prefix on new Connect", {
    client <- connect(server = "https://connect.example", api_key = "fake")

    # This will have to paginate
    result <- get_groups(client, page_size = 1, limit = 2, prefix = "c")
    expect_identical(result$name, c("connect_dev", "cool_kids_of_the_dmv"))

    # Just to be explicit, here you can see that page_number and page_size are
    # not aligned and you get page_size = 1 as requested
    expect_GET(
      get_groups(client, page_size = 1, limit = 2, prefix = "asdf"),
      "https://connect.example/__api__/v1/groups?page_number=1&page_size=1&prefix=asdf"
    )
  })
})

with_mock_api({
  client <- Connect$new(
    server = "https://connect.example",
    api_key = "not-a-key"
  )

  test_that("get_group_content() works", {
    group_guids <- c(
      "a6fb5cea",
      "ae5c3b2c"
    )

    expect_snapshot(get_group_content(client, group_guids))
  })

  test_that("get_group_content() returns an empty data frame when no content exists", {
    group_guid <- "a6fb5cff"

    expect_snapshot(get_group_content(client, group_guid))
  })
})

with_mock_api({
  test_that("we can retrieve the users list", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    users <- get_users(con)
    expect_s3_class(users, "data.frame")
    expect_equal(nrow(users), 3)
  })

  test_that("we can retrieve a user by id", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    a_user <- con$user("20a79ce3-6e87-4522-9faf-be24228800a4")
    # There is no User class?
    expect_type(a_user, "list")
    expect_equal(a_user$first_name, "Carlos")
  })
})

without_internet({
  test_that("Querying users by prefix", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      client$users(prefix = "TEST"),
      "https://connect.example/__api__/v1/users?page_number=1&page_size=500&prefix=TEST"
    )

    # Now with spaces
    expect_GET(
      client$users(prefix = "A User Name", page_size = 5),
      "https://connect.example/__api__/v1/users?page_number=1&page_size=5&prefix=A%20User%20Name"
    )

    # Now with too big page size: it automatically truncates (maybe that's bad)
    expect_GET(
      client$users(page_size = 1000),
      "https://connect.example/__api__/v1/users?page_number=1&page_size=500"
    )
  })

  test_that("Querying remote users", {
    client <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      client$users_remote(prefix = "A user name"),
      "https://connect.example/__api__/v1/users/remote?prefix=A%20user%20name"
    )
  })
})

user_guid <- NULL
group_guid <- NULL

test_that("groups_create works", {
  ss <- client$server_settings()
  if (ss$authentication$name %in% c("LDAP")) {
    skip("not implemented for this authentication provider")
  }
  groupname <- create_random_name()
  res <- client$groups_create(
    name = groupname
  )

  group_guid <<- res$guid

  expect_equal(res$name, groupname)
})

test_that("groups works", {
  groups <- client$groups()

  expect_gt(length(groups$results), 0)
})

test_that("group_member_add works", {
  user_guid <<- client$users_create(
    username = paste0("group_member", create_random_name()),
    email = "test@example.com",
    user_must_set_password = TRUE
  )$guid

  res <- client$group_member_add(group_guid, user_guid)

  members <- client$group_members(group_guid)

  expect_true(user_guid %in% purrr::map_chr(members$results, ~ .x$guid))
})

test_that("group_member_remove works", {
  res <- client$group_member_remove(group_guid, user_guid)

  members <- client$group_members(group_guid)

  expect_false(user_guid %in% purrr::map_chr(members$results, ~ .x$guid))
})

test_that("Connect R6 class preserves provided values", {
  server <- "http://myhost.example.com"
  api_key <- "fake"

  con <- Connect$new(server = server, api_key = api_key)

  expect_identical(con$server, server)
  expect_identical(con$api_key, api_key)
})

test_that("trailing slash removed from server", {
  con <- Connect$new(server = "http://myhost.example.com/", api_key = "fake")
  con2 <- Connect$new(server = "http://myhost.example.com", api_key = "fake")

  expect_identical(con$server, con2$server)
})

test_that("error if protocol not defined", {
  con <- Connect$new(server = "https://myhost.example.com", api_key = "fake")

  expect_error(
    Connect$new("test.example.com", "fake"),
    "protocol"
  )

  expect_error(
    Connect$new("://test.example.com", "fake"),
    "protocol"
  )
})

test_that("connect() errors if API key is empty", {
  expect_error(
    connect(api_key = ""),
    "provide a valid API key"
  )

  expect_error(
    connect(api_key = NA_character_),
    "provide a valid API key"
  )

  expect_error(
    connect(api_key = NULL),
    "provide a valid API key"
  )
})

test_that("connect() just warns on empty API key if .check_is_fatal=FALSE", {
  # Use without_internet() because for some reason it doesn't return early
  # if there is no API key.
  without_internet({
    expect_message(
      client <- connect(
        server = "http://posit.co",
        api_key = "",
        .check_is_fatal = FALSE
      ),
      "provide a valid API key"
    )
    expect_true(
      validate_R6_class(client, "Connect")
    )
  })
})

test_that("connect() fails for bad server", {
  without_internet({
    expect_error(
      connect(server = "does-not-exist.rstudio.com", api_key = "bogus"),
      "Please provide a protocol"
    )
    # This is how without_internet() errors making the ping request
    expect_error(
      connect(server = "http://does-not-exist.rstudio.com", api_key = "bogus"),
      "GET http://does-not-exist.rstudio.com/__ping__"
    )
  })
})

test_that(".check_is_fatal toggle handles server validation", {
  without_internet({
    # See above. But the error is caught and returned as a message in this case.
    expect_message(
      client <- connect(
        server = "http://fake-value.example.com",
        api_key = "fake-value",
        .check_is_fatal = FALSE
      ),
      "GET http://fake-value.example.com/__ping__"
    )
    expect_true(
      validate_R6_class(client, "Connect")
    )
  })
})

test_that("version is validated", {
  skip("not implemented yet")
})

test_that("Handling error responses", {
  con <- Connect$new(server = "https://connect.example", api_key = "fake")
  resp <- fake_response(
    "https://connect.example/__api__/",
    headers = list(`Content-Type` = "application/json; charset=utf-8"),
    status_code = 400L,
    content = list(code = 3, error = "Invalid GUID: abc", payload = NULL)
  )
  expect_error(
    con$raise_error(resp),
    "request failed with Client error: \\(400\\) Bad Request \\(code: 3, error: Invalid GUID: abc\\)"
  )
})

test_that("Handling error responses without all expected fields", {
  con <- Connect$new(server = "https://connect.example", api_key = "fake")
  resp1 <- fake_response(
    "https://connect.example/__api__/",
    headers = list(`Content-Type` = "application/json; charset=utf-8"),
    status_code = 400L,
    content = list(code = 3L)
  )
  resp2 <- fake_response(
    "https://connect.example/__api__/",
    headers = list(`Content-Type` = "application/json; charset=utf-8"),
    status_code = 400L,
    content = list(error = "Invalid GUID: abc")
  )
  resp3 <- fake_response(
    "https://connect.example/__api__/",
    headers = list(`Content-Type` = "application/json; charset=utf-8"),
    status_code = 400L,
    content = NULL
  )
  expect_error(
    con$raise_error(resp1),
    "request failed with Client error: \\(400\\) Bad Request \\(code: 3\\)"
  )
  expect_error(
    con$raise_error(resp2),
    "request failed with Client error: \\(400\\) Bad Request \\(error: Invalid GUID: abc\\)"
  )
  expect_error(
    con$raise_error(resp3),
    "request failed with Client error: \\(400\\) Bad Request $"
  )
})

test_that("Handling deprecation warnings", {
  rlang::reset_warning_verbosity("X-Deprecated-Endpoint")
  on.exit(rlang::reset_warning_verbosity("X-Deprecated-Endpoint"))

  # No warning here
  resp <- fake_response(
    "https://connect.example/__api__/",
    headers = list(
      `Content-Type` = "application/json; charset=utf-8"
    )
  )
  expect_warning(check_debug(resp), NA)

  withr::with_options(
    list(rlib_warning_verbosity = "default"),
    {
      # Yes warning here
      resp <- fake_response(
        "https://connect.example/__api__/",
        headers = list(
          `X-Deprecated-Endpoint` = "/v1",
          `Content-Type` = "application/json; charset=utf-8"
        )
      )
      expect_warning(
        check_debug(resp),
        paste(
          "https://connect.example/__api__/ is deprecated and will be removed in a",
          "future version of Connect. Please upgrade `connectapi` in order to use",
          "the new APIs."
        ),
        class = "deprecatedWarning"
      )
      # No warning if you do it again because we only warn the first time
      expect_warning(check_debug(resp), NA)
    }
  )
})

with_mock_api({
  test_that("browse URLs", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    # Inject into this function something other than utils::browseURL
    # so we can assert that it is being called without actually trying to open a browser
    suppressMessages(trace(
      "browse_url",
      where = connectapi::browse_solo,
      tracer = quote({
        browseURL <- # nolint: object_name_linter
          function(x) {
            warning(paste("Opening", x))
          }
      }),
      at = 1,
      print = FALSE
    ))
    expect_warning(
      browse_connect(con),
      "Opening https://connect.example"
    )
    expect_warning(
      browse_api_docs(con),
      "Opening https://connect.example/__docs__/api"
    )
    suppressMessages(untrace("browse_url", where = connectapi::browse_solo))
  })

  test_that("httr_config", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    con$httr_config(httr::add_headers(MY_MAGIC_HEADER = "value"))
    expect_header(
      expect_GET(
        con$GET("v1/content"),
        "https://connect.example/__api__/v1/content"
      ),
      "MY_MAGIC_HEADER: value"
    )
  })

  test_that("client$version is NA when server settings lacks version info", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    v <- con$version
    expect_true(is.na(v))
  })
})

test_that("client$version is returns version when server settings exposes it", {
  with_mock_dir("2024.09.0", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_equal(con$version, "2024.09.0")
  })
})

test_that("Visitor client can successfully be created running on Connect", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake",
      RSTUDIO_PRODUCT = "CONNECT"
    )

    expect_warning(
      client <- connect(token = "my-token"),
      "This feature requires Posit Connect version"
    )

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "visitor-api-key"
    )
  })
})

test_that("Visitor client can successfully be created running on Connect (POSIT_PRODUCT)", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake",
      POSIT_PRODUCT = "CONNECT"
    )

    expect_warning(
      client <- connect(token = "my-token"),
      "This feature requires Posit Connect version"
    )

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "visitor-api-key"
    )
  })
})

test_that("Visitor client can successfully be created running on Connect Cloud", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake",
      POSIT_PRODUCT = "CONNECT_CLOUD"
    )

    expect_warning(
      client <- connect(token = "my-token"),
      "This feature requires Posit Connect version"
    )

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "visitor-api-key"
    )
  })
})

test_that("Visitor client can successfully be created running on Connect with audience", {
  with_mock_dir("2025.07.0", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake",
      RSTUDIO_PRODUCT = "CONNECT"
    )

    client <- connect(token = "my-token", audience = "audience")

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "visitor-api-key"
    )
  })
})

test_that("Visitor client uses fallback api key when running locally", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake"
    )

    # With default fallback
    expect_warning(
      expect_message(
        client <- connect(token = NULL),
        "Called with `token` but not running on Connect or Connect Cloud. Continuing with fallback API key."
      ),
      "the server version is not exposed by this Posit Connect instance"
    )

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "fake"
    )

    # With explicitly-defined fallback
    expect_warning(
      expect_message(
        client <- connect(
          token = NULL,
          token_local_testing_key = "fallback_fake"
        ),
        "Called with `token` but not running on Connect or Connect Cloud. Continuing with fallback API key."
      ),
      "the server version is not exposed by this Posit Connect instance"
    )

    expect_equal(
      client$server,
      "https://connect.example"
    )
    expect_equal(
      client$api_key,
      "fallback_fake"
    )
  })
})

test_that("Visitor client code path errs with older Connect version", {
  with_mock_dir("2024.09.0", {
    withr::local_envvar(
      CONNECT_SERVER = "https://connect.example",
      CONNECT_API_KEY = "fake",
      RSTUDIO_PRODUCT = "CONNECT"
    )

    expect_error(
      client <- connect(token = "my-token"),
      "This feature requires Posit Connect version 2025.01.0 but you are using 2024.09.0"
    )
  })
})

test_that("Scientific notation is not used for reasonable parameters", {
  with_mock_api({
    test_scipen <- 5
    withr::local_options(
      rlib_warning_verbosity = "quiet",
      scipen = test_scipen
    )

    client <- Connect$new(server = "https://connect.example", api_key = "fake")

    expect_GET(
      get_packages(client, page_size = 999999999),
      "https://connect.example/__api__/v1/packages?page_number=1&page_size=999999999"
    )
    expect_equal(getOption("scipen"), test_scipen)
  })
})

without_internet({
  test_that("Query params to get task", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$task("asdf"),
      "https://connect.example/__api__/v1/tasks/asdf?first=0&wait=5"
    )
  })
})

#' Browse
#'
#' Browse to different locations on Connect via
#' utils::browseURL
#'
#' @param content A R6 Content object
#' @param connect A R6 Connect object
#'
#' @return The url that is opened in the browser
#'
#' @rdname browse
#' @export
browse_solo <- function(content) {
  validate_R6_class(content, "Content")
  url <- content$get_url()
  browse_url(url)
}

#' @rdname browse
#' @export
browse_dashboard <- function(content) {
  validate_R6_class(content, "Content")
  url <- content$get_dashboard_url()
  browse_url(url)
}

#' @rdname browse
#' @export
browse_api_docs <- function(connect) {
  validate_R6_class(connect, "Connect")
  url <- connect$docs("api", browse = FALSE)
  browse_url(url)
}

#' @rdname browse
#' @export
browse_connect <- function(connect) {
  validate_R6_class(connect, "Connect")
  url <- connect$server
  browse_url(url)
}

#' @importFrom utils browseURL
browse_url <- function(url) {
  browseURL(url)
  invisible(url)
}

#' Promote content from one Connect server to another
#'
#' @param from The url for the server containing the content (the originating
#'   server)
#'
#' @param to   The url for the server where the content will be deployed (the
#'   destination server)
#'
#' @param to_key An API key on the destination "to" server. If the destination
#'   content is going to be updated, the API key must belong to a user with
#'   collaborator access on the content that will be updated. If the destination
#'   content is to be created new, the API key must belong to a user with
#'   publisher privileges.
#'
#' @param from_key An API key on the originating "from" server. The API key must
#'   belong to a user with collaborator access to the content to be promoted.
#'
#' @param name The name of the content on the originating "from" server.
#'   If content with the same name is found on the destination server,
#'   the content will be updated. If no content on the destination server
#'   has a matching name, a new endpoint will be created.
#'
#' @return The URL for the content on the destination "to" server
#' @export
promote <- function(from, to, to_key, from_key, name) {
  # TODO Validate Inputs

  # set up clients
  from_client <- connect(server = from, api_key = from_key)
  to_client <- connect(server = to, api_key = to_key)

  # find content on "from" server
  from_app <- from_client$content(name = name, include = NULL)
  if (length(from_app) != 1) {
    stop(sprintf(
      "Found %d apps matching app name %s on %s. Content must have a unique name.",
      length(from_app),
      name,
      from
    ))
  }
  from_app <- Content$new(connect = from_client, content = from_app[[1]])

  # download bundle
  bundle <- download_bundle(from_app)

  # find or create app to update
  to_app <- content_ensure(connect = to_client, name = name)
  to_app <- content_item(connect = to_client, guid = to_app$guid)

  task <- deploy(to_client, bundle = bundle, guid = to_app$content$guid)

  poll_task(task)

  to_app_url <- to_app$get_url()

  return(to_app_url)
}

#' Paging
#'
#' Helper functions that make paging easier in
#' the Posit Connect Server API.
#'
#' @rdname paging
#'
#' @param client A Connect client object
#' @param req For page_cursor, the output from an initial response to an API
#'   endpoint that uses cursor-based pagination. For page_offset, a request that
#'   needs to be paged.
#' @param limit A row limit
#'
#' @return The aggregated results from all requests
#'
#' @export
page_cursor <- function(client, req, limit = Inf) {
  prg <- optional_progress_bar(
    format = "downloading page :current (:tick_rate/sec) :elapsedfull",
    total = NA,
    clear = FALSE
  )

  prg$tick()
  response <- req

  # collect whole pages, then flatten once at the end
  pages <- list(response$results)
  n_items <- length(response$results)
  while (!is.null(response$paging$`next`) && n_items < limit) {
    prg$tick()

    next_url <- response$paging$`next`
    response <- client$GET(url = next_url)

    pages[[length(pages) + 1L]] <- response$results
    n_items <- n_items + length(response$results)
  }

  head(do.call(c, pages), n = limit)
}
# TODO: Decide if this `limit = Inf` is helpful or a hack...
#       it is essentially a "row limit" on paging

#' @rdname paging
#' @export
page_offset <- function(client, req, limit = Inf) {
  qreq <- rlang::enquo(req)
  qexpr <- rlang::quo_get_expr(qreq)

  prg <- optional_progress_bar(
    format = "downloading page :current (:tick_rate/sec) :elapsedfull",
    total = NA,
    clear = FALSE
  )

  # handle paging
  prg$tick()

  req_response <- rlang::eval_tidy(qreq)
  res <- req_response$results
  total_items <- req_response$total

  agg_response <- res
  agg_length <- length(agg_response)
  while (length(res) > 0 && agg_length < limit && agg_length < total_items) {
    prg$tick()

    # bump the page number
    current_page <- req_response$current_page
    new_expr <- qexpr
    new_expr$page_number <- current_page + 1
    new_req <- rlang::quo_set_expr(qreq, new_expr)

    # next request
    req_response <- rlang::eval_tidy(new_req)
    res <- req_response$results

    agg_response <- c(agg_response, res)
    agg_length <- length(agg_response)

    # clear out variables
    current_page <- NULL
    new_expr <- NULL
    new_req <- NULL
  }
  # Make sure we never return more than `limit` records
  head(agg_response, limit)
}


optional_progress_bar <- function(...) {
  if (requireNamespace("progress", quietly = TRUE)) {
    progress::progress_bar$new(...)
  } else {
    # Return a mock object that behaves enough like a progress bar object
    list(
      tick = function() {}
    )
  }
}

# TODO
# - next stop: vanity URLs?
# - figure out filtering... and such...?
# - draw diagram for understanding dbplyr execution
# - how does the op-list work... can you make "collect" happen before filter, mutate, and such?
# - need to make pagination actually work...
# - filters based on content_guid, started, ended, etc.
# - nrow should be super fast if we know how many total records there are...
# - show example usage...
#' Connect Tibble
#'
#' `r lifecycle::badge('experimental')`
#' A lazy tibble that automatically pages through API requests when `collect`ed.
#'
#' @param src The source object
#' @param from The type of tibble
#' @param ... Additional arguments that are not yet implemented
#'
#' @return A `tbl_connect` object
#'
#' @export
tbl_connect <- function(
  src,
  from = c(
    "users",
    "groups",
    "content",
    "usage_shiny",
    "usage_static",
    "audit_logs"
  ),
  ...
) {
  rlang::check_installed("dbplyr")

  validate_R6_class(src, "Connect")

  from <- arg_match(from)

  # TODO: go get the vars we should expect...
  vars <- connectapi_ptypes[[from]]
  if (is.null(vars)) vars <- character()

  # TODO: figure out number of rows...
  ops <- op_base_connect(from, vars)

  dplyr::make_tbl(c("connect", "lazy"), src = src, ops = ops)
}

# This will be registered in .onLoad if dplyr is available
collect.tbl_connect <- # nolint
  function(x, ..., n = Inf) {
    api_build(op = x[["ops"]], con = x[["src"]], n = n)
  }

api_build <- function(op, con = NULL, ..., n = NULL) {
  UseMethod("api_build")
}

#' @export
api_build.op_head <- function(op, con, ..., n) {
  n <- op$args$n
  api_build(op$x, con, ..., n = n)
}

#' @export
api_build.op_base_connect <- function(op, con, ..., n) {
  if (op$x == "users") {
    res <- page_offset(con, con$users(), limit = n)
  } else if (op$x == "groups") {
    res <- page_offset(con, con$groups(), limit = n)
  } else if (op$x == "content") {
    # TODO: no limit notion here... we just pull all of them...
    res <- con$content()
  } else if (op$x == "usage_shiny") {
    res <- con$inst_shiny_usage(limit = n) %>% page_cursor(con, ., limit = n)
  } else if (op$x == "usage_static") {
    res <- con$inst_content_visits(limit = n) %>% page_cursor(con, ., limit = n)
  } else if (op$x == "audit_logs") {
    res <- con$audit_logs(limit = n) %>% page_cursor(con, ., limit = n)
  } else {
    stop(glue::glue("'{op$x}' is not recognized"))
  }
  parse_connectapi_typed(res, op$ptype)
}

cat_line <- function(...) {
  cat(paste0(..., "\n"), sep = "")
}

#' @importFrom utils head
#' @export
head.tbl_connect <- function(x, n = 6L, ...) {
  if (inherits(x[["ops"]], "op_head")) {
    x$ops$args$n <- min(x$ops$args$n, n)
  } else {
    x$ops <- op_single("head", x = x[["ops"]], args = list(n = n))
  }
  x
}

#' @export
print.tbl_connect <- function(x, ..., n = NULL) {
  cat_line(format(x, ..., n = n))
  invisible(x)
}

#' @export
as.data.frame.tbl_connect <- function(
  x,
  row.names = NULL, # nolint
  optional = NULL,
  ...,
  n = Inf
) {
  # We don't need to check if dplyr is available here
  # because you won't have a tbl_connect without first
  # checking for dplyr.
  as.data.frame(dplyr::collect(x, n = n))
}

op_base_connect <- function(x, vars) {
  op_base(x, vars, class = "connect")
}

op_base <- function(x, vars, class = character()) {
  stopifnot(is.character(vars) || is.character(names(vars)))

  structure(
    list(
      x = x,
      vars = names(vars),
      ptype = vars
    ),
    class = c(paste0("op_base_", class), "op_base", "op")
  )
}

op_single <- function(name, x, dots = list(), args = list()) {
  structure(
    list(
      name = name,
      x = x,
      dots = dots,
      args = args
    ),
    class = c(paste0("op_", name), "op_single", "op")
  )
}

connect_vars <- function(op) UseMethod("connect_vars")
#' @export
connect_vars.op_base <- function(op) op$vars
#' @export
connect_vars.op_single <- function(op) connect_vars(op$x)
#' @export
connect_vars.tbl_connect <- function(op) connect_vars(op[["ops"]])

# important for `nrow`/`ncol` to work
#' @export
dim.tbl_connect <- function(x) {
  c(NA, length(connect_vars(x[["ops"]])))
}

# important for `colnames` to work
#' @export
dimnames.tbl_connect <- function(x) {
  list(NULL, connect_vars(x[["ops"]]))
}

library(httptest)
library(R6)

set_requester(
  function(r) {
    gsub_request(
      gsub_request(
        r,
        # Prune the domain
        "^https?://[^/]+/",
        ""
      ),
      # Truncate GUIDs to 8 characters
      "([0-9a-f]{8})\\-[0-9a-f\\-]{27}",
      "\\1"
    )
  }
)

set_redactor(
  function(r) {
    gsub_response(
      gsub_response(
        r,
        # Prune the domain
        "^https?://[^/]+/",
        ""
      ),
      # Truncate GUIDs to 8 characters
      "([0-9a-f]{8})\\-[0-9a-f\\-]{27}",
      "\\1"
    )
  }
)

# Mocks are in directories by Connect version. 2024.08.0 contains all mocks
# created before 2024.09.0, and is the default mock path.
.mockPaths("2024.08.0")

MockConnect <- R6Class(
  "MockConnect",
  inherit = Connect,
  public = list(
    initialize = function(version = NA) {
      self$server <- "https://connect.example"
      self$api_key <- "fake"
      private$.version <- version
    },
    # The request function matches the route against the routes in the names of
    # the response list. When a response is selected, it is removed from the
    # list.
    request = function(method, url, ..., parser = "parsed") {
      route <- paste(method, url)

      # Record call
      self$log_call(route)

      # Look for response
      if (!(route %in% names(self$responses))) {
        stop(glue::glue("Unexpected route: {route}"))
      }
      idx <- match(route, names(self$responses))
      res <- self$responses[[idx]]
      self$responses <- self$responses[-idx]

      if (is.null(parser)) {
        res
      } else {
        self$raise_error(res)
        httr::content(res, as = parser)
      }
    },
    responses = list(),
    # Add a response to a list of responses. The response is keyed according to
    # its route, represented as `{VERB} {URL}`. The URL is constructed as an API
    # URL for the server (this will probably have to change in the future). Each
    # response can only be used once. You can supply multiple responses for the
    # same URL.
    mock_response = function(
      method,
      path,
      content,
      status_code = 200L,
      headers = c("Content-Type" = "application/json; charset=utf-8")
    ) {
      url <- self$api_url(path)
      route <- paste(method, url)

      res <- new_mock_response(url, content, status_code, headers)

      route <- paste(method, url)
      new_response <- list(res)
      new_response <- setNames(new_response, route)

      self$responses <- append(self$responses, new_response)
    },
    call_log = character(),
    log_call = function(route) {
      self$call_log <- c(self$call_log, route)
    }
  )
)

new_mock_response <- function(
  url,
  content,
  status_code,
  headers = character()
) {
  # Headers in responses are case-insensitive lists.
  names(headers) <- tolower(names(headers))
  headers <- as.list(headers)
  headers <- structure(
    as.list(headers),
    class = c("insensitive", class(headers))
  )

  # Treat content similarly to httr::POST, with a subset of behaviors
  if (is.character(content) && length(content) == 1) {
    content <- charToRaw(content)
  } else if (is.list(content)) {
    content <- charToRaw(jsonlite::toJSON(
      content,
      auto_unbox = TRUE,
      null = "null"
    ))
  }

  structure(
    list(
      url = url,
      status_code = status_code,
      request = structure(list(url = url), class = "request"),
      headers = headers,
      content = content
    ),
    class = "response"
  )
}

#' @importFrom utils compareVersion untar
#' @importFrom lifecycle deprecate_warn
#' @importFrom rlang arg_match is_string
"_PACKAGE"

# Even if we don't currently have any references to functions in lifecycle,
# we have to include it in Imports because we use it in RdMacros

utils::globalVariables(
  c(
    ".",
    "access_type",
    "connectapi_ptypes",
    "guid",
    "last_deployed_time",
    "owner_guid",
    "parse_connectapi_typed",
    "permission",
    "r_version",
    "scoped_experimental_silence",
    "validate_R6_class",
    "warn_experimental"
  )
)

current_connect_version <- "2024.03.0"

.onLoad <- function(...) {
  vctrs::s3_register("dplyr::collect", "tbl_connect")
  vctrs::s3_register("vctrs::vec_cast", "character.integer")
  invisible()
}

#' Content
#'
#' An R6 class that represents content.
#'
#' @family R6 classes
#' @export
Content <- R6::R6Class(
  "Content",
  public = list(
    #' @field connect An R6 Connect object.
    connect = NULL,
    #' @field content The content details from Posit Connect. Properties are described in [get_content()].
    content = NULL,

    #' @description Initialize this content.
    #' @param connect The `Connect` instance.
    #' @param content The content data.
    initialize = function(connect, content) {
      validate_R6_class(connect, "Connect")
      self$connect <- connect
      # TODO: need to check that content has
      # at least guid, url, title to be functional
      self$content <- content
    },

    #' @description Obtain the content data from the Connect server.
    get_content_remote = function() {
      new_content_details <- self$connect$content(self$content$guid)
      self$content <- new_content_details
      self$content
    },
    #' @description Return the set of content bundles.
    get_bundles = function() {
      url <- v1_url("content", self$content$guid, "bundles")
      self$connect$GET(url)
    },
    #' @description Download the source archive for a content bundle.
    #' @param bundle_id The bundle identifer.
    #' @param filename Where to write the result.
    #' @param overwrite Overwrite an existing filename.
    bundle_download = function(
      bundle_id,
      filename = tempfile(pattern = "bundle", fileext = ".tar.gz"),
      overwrite = FALSE
    ) {
      url <- v1_url(
        "content",
        self$content$guid,
        "bundles",
        bundle_id,
        "download"
      )
      self$connect$GET(
        url,
        httr::write_disk(filename, overwrite = overwrite),
        parser = "raw"
      )
      return(filename)
    },
    #' @description Delete a content bundle.
    #' @param bundle_id The bundle identifer.
    bundle_delete = function(bundle_id) {
      url <- v1_url("content", self$content$guid, "bundles", bundle_id)
      self$connect$DELETE(url)
    },
    #' @description Update this content item.
    #' @param ... Content fields.
    update = function(...) {
      con <- self$connect
      error_if_less_than(con$version, "1.8.6")
      url <- v1_url("content", self$content$guid)
      body <- rlang::list2(...)
      if (length(body)) {
        # Only need to make a request if there are changes
        con$PATCH(url, body = body)
      }
      self
    },
    #' @description Delete this content item.
    danger_delete = function() {
      con <- self$connect
      url <- v1_url("content", self$content$guid)
      con$DELETE(url)
    },
    #' @description Return the URL for this content.
    get_url = function() {
      self$content$content_url
    },
    #' @description Return the URL for this content in the Posit Connect dashboard.
    #' @param pane The pane in the dashboard to link to.
    get_dashboard_url = function(pane = "") {
      url <- self$content$dashboard_url
      if (nzchar(pane)) {
        url <- paste0(url, "/", pane)
      }
      url
    },
    #' @description Return the jobs for this content
    jobs = function() {
      res <- self$connect$GET(
        v1_url("content", self$content$guid, "jobs"),
        parser = NULL
      )
      use_unversioned <- endpoint_does_not_exist(res)
      if (use_unversioned) {
        res <- self$connect$GET(
          unversioned_fallback_url("applications", self$content$guid, "jobs"),
          parser = NULL
        )
      }
      self$connect$raise_error(res)
      parsed <- httr::content(res, as = "parsed")
      if (use_unversioned) {
        # The unversioned endpoint does not contain a `status` field. Its field
        # `finalized` is `FALSE` corresponds to active jobs. The `finalized`
        # field is dropped during parsing.
        parsed <- purrr::modify_if(
          parsed,
          ~ isFALSE(.x$finalized),
          function(x) {
            x$status <- 0
            x
          }
        )
      }
      # Ensure content identifiers are included.
      # The `app_guid` field was never returned by Connect, but is added here
      # for backward compatibility with older `connectapi` versions, which added
      # a field named `app_guid` to job data.
      if (compare_connect_version(self$connect$version, "2025.01.0") < 0) {
        # Versions of Connect below 2025.01.0 only included `app_id`. We must
        # add all other fields.
        purrr::map(
          parsed,
          ~ purrr::list_modify(
            .x,
            content_id = .x$app_id,
            app_guid = self$content$guid,
            content_guid = self$content_guid
          )
        )
      } else {
        # Connect 2025.01.0 includes `content_id` and `content_guid`, and
        # retains `app_id` for backward compat. We only need to add `app_guid`
        # for `connectapi` back-compat.
        purrr::map(
          parsed,
          ~ purrr::list_modify(
            .x,
            app_guid = .x$content_guid
          )
        )
      }
    },
    #' @description Terminate a single job for this content item.
    #' @param key The job key.
    register_job_kill_order = function(key) {
      con <- self$connect
      url <- v1_url("content", self$content$guid, "jobs", key)
      res <- self$connect$DELETE(url)
      if (endpoint_does_not_exist(res)) {
        con$raise_error(res)
      }
      res
    },
    #' @description Return the variants for this content.
    variants = function() {
      warn_experimental("variants")
      guid <- self$content$guid
      url <- unversioned_url("applications", guid, "variants")
      self$connect$GET(url)
    },
    #' @description Set a tag for this content.
    #' @param tag_id The tag identifier.
    tag_set = function(tag_id) {
      self$connect$set_content_tag(
        self$content$guid,
        tag_id = tag_id
      )
    },
    #' @description Remove a tag for this content.
    #' @param tag_id The tag identifier.
    tag_delete = function(tag_id) {
      # note that deleting the parent tag deletes all children
      self$connect$remove_content_tag(
        self$content$guid,
        tag_id = tag_id
      )
    },
    #' @description The tags for this content.
    tags = function() {
      url <- v1_url("content", self$content$guid, "tags")
      self$connect$GET(url)
    },
    #' @description Add a principal to the ACL for this content.
    #' @param principal_guid GUID for the target user or group.
    #' @param principal_type Acting on user or group.
    #' @param role The kind of content access.
    permissions_add = function(principal_guid, principal_type, role) {
      url <- v1_url("content", self$content$guid, "permissions")
      self$connect$POST(
        url,
        body = list(
          principal_guid = principal_guid,
          principal_type = principal_type,
          role = role
        )
      )
    },
    #' @description Alter a principal in the ACL for this content.
    #' @param id The target identifier.
    #' @param principal_guid GUID for the target user or group.
    #' @param principal_type Acting on user or group.
    #' @param role The kind of content access.
    permissions_update = function(id, principal_guid, principal_type, role) {
      url <- v1_url("content", self$content$guid, "permissions", id)
      self$connect$PUT(
        url,
        body = list(
          principal_guid = principal_guid,
          principal_type = principal_type,
          role = role
        )
      )
    },
    #' @description Remove an entry from the ACL for this content.
    #' @param id The target identifier.
    permissions_delete = function(id) {
      url <- v1_url("content", self$content$guid, "permissions", id)
      self$connect$DELETE(url)
    },
    #' @description Obtain some or all of the ACL for this content.
    #' @param id The target identifier.
    #' @param add_owner Include the content owner in the result set.
    permissions = function(id = NULL, add_owner = FALSE) {
      guid <- self$content$guid
      if (is.null(id)) {
        url <- v1_url("content", self$content$guid, "permissions")
      } else {
        url <- v1_url("content", self$content$guid, "permissions", id)
      }
      res <- self$connect$GET(url)
      # NOTE: the default for the low-level functions is to map to the API
      # as close as possible. This differs from the "cleaner UX" functions
      if (add_owner) {
        owner_entry <- list(
          id = NA_character_,
          content_guid = guid,
          # TODO: what if groups can own content?
          principal_guid = self$content$owner_guid,
          principal_type = "user",
          role = "owner"
        )
        res <- c(res, list(owner_entry))
      }
      res
    },
    #' @description Return the environment variables set for this content.
    environment = function() {
      url <- v1_url("content", self$content$guid, "environment")
      self$connect$GET(url)
    },
    #' @description Adjust the environment variables set for this content.
    #' @param ... Environment variable names and values. Use `NA` as the value
    #' to unset variables.
    environment_set = function(...) {
      url <- v1_url("content", self$content$guid, "environment")
      # post with
      # key = NA to remove
      vals <- rlang::list2(...)
      body <- purrr::imap(vals, function(.x, .y) {
        # TODO: evaluate whether we should be coercing to character or erroring
        return(list(name = .y, value = as.character(.x)))
      })
      names(body) <- NULL

      self$connect$PATCH(path = url, body = body)
    },
    #' @description Overwrite the environment variables set for this content.
    #' @param ... Environment variable names and values.
    environment_all = function(...) {
      url <- v1_url("content", self$content$guid, "environment")

      vals <- rlang::list2(...)
      if (length(vals) == 0) {
        # Make sure we send an empty array and not an empty list
        body <- "[]"
      } else {
        body <- purrr::imap(vals, function(.x, .y) {
          # TODO: evaluate whether we should be coercing to character or erroring
          return(list(name = .y, value = as.character(.x)))
        })
        names(body) <- NULL
      }

      self$connect$PUT(path = url, body = body)
    },
    #' @description Deploy this content
    #' @param bundle_id Target bundle identifier.
    deploy = function(bundle_id = NULL) {
      body <- list(bundle_id = bundle_id)
      self$connect$POST(
        v1_url("content", self$content$guid, "deploy"),
        body = body
      )
    },
    #' @description Get Git repository details
    #' @return NULL if no repo is set, otherwise a list with fields:
    #' - repository
    #' - branch
    #' - directory
    #' - polling
    #' - last_error
    #' - last_known_commit
    repository = function() {
      con <- self$connect
      error_if_less_than(con$version, "2022.12.0")
      guid <- self$content$guid
      resp <- con$GET(
        v1_url("content", guid, "repository"),
        parser = NULL
      )
      if (httr::status_code(resp) == 404) {
        # 404 means there is no repository set
        return(NULL)
      }
      con$raise_error(resp)
      httr::content(resp, as = "parsed")
    },
    #' @description Adjust Git polling.
    #' @param polling Polling enabled.
    repo_enable = function(polling = TRUE) {
      con <- self$connect
      error_if_less_than(con$version, "2022.12.0")
      guid <- self$content$guid
      con$PATCH(
        v1_url("content", guid, "repository"),
        body = list(polling = polling)
      )
    },
    #' @description Adjust Git repository
    #' @param repository Git repository URL
    #' @param branch Git repository branch
    #' @param directory Git repository directory
    #' @param polling Whether to check for updates
    repo_set = function(
      repository,
      branch = "main",
      directory = ".",
      polling = FALSE
    ) {
      guid <- self$content$guid
      error_if_less_than(self$connect$version, "2022.12.0")
      self$connect$PUT(
        v1_url("content", guid, "repository"),
        body = list(
          repository = repository,
          branch = branch,
          directory = directory,
          polling = polling
        )
      )
    },
    #' @description Get package dependencies
    packages = function() {
      self$connect$GET(v1_url("content", self$content$guid, "packages"))
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      cat("Posit Connect Content: \n")
      cat("  Content GUID: ", self$content$guid, "\n", sep = "")
      cat(
        "  Content URL: ",
        self$content$dashboard_url,
        "\n",
        sep = ""
      )
      cat("  Content Title: ", self$content$title, "\n", sep = "")
      cat("\n")
      cat(
        'content_item(client, guid = "',
        self$content$guid,
        '")',
        "\n",
        sep = ""
      )
      cat("\n")
      invisible(self)
    }
  ),
  active = list(
    #' @field default_variant The default variant for this object.
    default_variant = function() {
      get_variant(self, "default")
    },

    #' @field is_rendered TRUE if this is a rendered content type, otherwise FALSE.
    is_rendered = function() {
      self$content$app_mode %in%
        c("rmd-static", "jupyter-static", "quarto-static")
    },

    #' @field is_interactive TRUE if this is a rendered content type, otherwise FALSE.
    is_interactive = function() {
      interactive_app_modes <- c(
        "shiny",
        "rmd-shiny",
        "jupyter-voila",
        "python-api",
        "python-dash",
        "python-streamlit",
        "python-bokeh",
        "python-fastapi",
        "python-shiny",
        "quarto-shiny",
        "tensorflow-saved-model",
        "api"
      )
      self$content$app_mode %in% interactive_app_modes
    }
  )
)

#' Environment
#'
#' An R6 class that represents a Content's Environment Variables
#'
#' @rdname EnvironmentR6
#'
#' @family R6 classes
#' @export
Environment <- R6::R6Class(
  "Environment",
  inherit = Content,
  public = list(
    #' @field env_raw The (raw) set of environment variables.
    env_raw = NULL,
    #' @field env_vars The set of environment variables.
    env_vars = NULL,

    #' @description Initialize this set of environment variables.
    #' @param connect The `Connect` instance.
    #' @param content The `Content` instance.
    initialize = function(connect, content) {
      super$initialize(connect = connect, content = content)
      self$env_refresh()
    },
    #' @description Fetch the set of environment variables.
    environment = function() {
      res <- super$environment()
      env_raw <- res
      env_vars <- res
      return(res)
    },
    #' @description Update the set of environment variables.
    #' @param ... Environment variable names and values.
    environment_set = function(...) {
      res <- super$environment_set(...)
      env_raw <- res
      env_vars <- res
      return(res)
    },
    #' @description Overwrite the set of environment variables.
    #' @param ... Environment variable names and values.
    environment_all = function(...) {
      res <- super$environment_all(...)
      env_raw <- res
      env_vars <- res
      return(res)
    },
    #' @description Fetch the set o environment variables.
    env_refresh = function() {
      # mutates the existing instance, so future
      # references have the right version
      self$env_raw <- self$environment()
      self$env_vars <- self$env_raw
      return(self)
    },
    #' @description Print this object.
    #' @param ... Unused.
    print = function(...) {
      super$print(...)
      cat("Environment Variables:\n")
      cat("  vctrs::vec_c(\n")
      purrr::map(self$env_vars, ~ cat(paste0('    "', .x, '",\n')))
      cat("  )\n")
      cat("\n")
      invisible(self)
    }
  )
)

# does it make more sense to automatically "get the latest"
# or to force the user to do that?
#' Manage Environment Variables
#'
#' Manage Environment Variables for a piece of content.
#'
#' `get_environment()` returns an Environment object for use with "setter" methods
#'
#' `set_environment_new()` updates environment values (either creating new
#' values or updating existing). Set `NA` as the value to remove a variable.
#'
#' `set_environment_remove()` is a wrapper on `set_environment_new()` that
#' allows removing named / listed variables quickly
#'
#' `set_environment_all()` sets _all_ environment variable values (will remove
#' variables not specified)
#'
#' @param content An R6 Content object as returned by `content_item()`
#' @param env An R6 Environment object as returned by `get_environment()`
#' @param ... name = value pairs of environment variable names and values
#'
#' @family content functions
#' @export
#'
#' @rdname environment
get_environment <- function(content) {
  validate_R6_class(content, "Content")
  content_data <- content$get_content_remote()
  connect_client <- content$connect
  return(Environment$new(connect_client, content_data))
}

#' @rdname environment
#' @export
set_environment_new <- function(env, ...) {
  validate_R6_class(env, "Content")

  if (!inherits(env, "Environment")) {
    env <- get_environment(env)
  }

  # update existing env vars with new ones
  new_env_vars <- rlang::dots_list(...)

  env$environment_set(!!!new_env_vars)

  env$env_refresh()
}

#' @rdname environment
#' @export
set_environment_remove <- function(env, ...) {
  to_remove <- rlang::enexprs(...)
  to_remove_names <- c(names(to_remove), as.character(unlist(to_remove)))
  to_remove_names <- to_remove_names[nchar(to_remove_names) > 0]
  to_remove_final <- rlang::set_names(
    rep(NA, length(to_remove_names)),
    to_remove_names
  )

  set_environment_new(env, !!!to_remove_final)
}

#' @rdname environment
#' @export
set_environment_all <- function(env, ...) {
  validate_R6_class(env, "Content")

  if (!inherits(env, "Environment")) {
    env <- get_environment(env)
  }

  # set all environment variables
  new_env_vars <- rlang::dots_list(...)

  env$environment_all(!!!new_env_vars)

  env$env_refresh()
}

#' Get Content Item
#'
#' Returns a single content item based on guid
#'
#' @param connect A Connect object
#' @param guid The GUID for the content item to be retrieved
#'
#' @return A Content object for use with other content endpoints
#'
#' @family content functions
#' @export
#' @examples
#' \dontrun{
#' connect() %>%
#'   content_item("some-guid") %>%
#'   content_update_access_type("all")
#' }
content_item <- function(connect, guid) {
  # TODO : think about how to handle if GUID does not exist
  validate_R6_class(connect, "Connect")

  res <- connect$content(guid)

  Content$new(connect = connect, content = res)
}

#' Get Content Title
#'
#' Return content title for a piece of content. If the content
#' is missing (deleted) or not visible, then returns the `default`
#'
#' @param connect A Connect object
#' @param guid The GUID for the content item to be retrieved
#' @param default The default value returned for missing or not visible content
#'
#' @return character. The title of the requested content
#'
#' @family content functions
#' @export
content_title <- function(connect, guid, default = "Unknown Content") {
  validate_R6_class(connect, "Connect")

  content_title <- tryCatch(
    {
      res <- suppressMessages(connect$content(guid))
      # TODO: What about length 0?
      if (is.null(res$title)) {
        return(default)
      }
      res$title
    },
    error = function(e) {
      return(default)
    }
  )

  return(content_title)
}

#' @importFrom uuid UUIDgenerate
content_ensure <- function(
  connect,
  name = uuid::UUIDgenerate(),
  title = name,
  guid = NULL,
  ...,
  .permitted = c("new", "existing")
) {
  if (!is.null(guid)) {
    # guid-based deployment
    # just in case we get a 404 back...
    content <- tryCatch(
      suppressMessages(connect$content(guid = guid)),
      error = function(e) {
        return(NULL)
      }
    )
    if (is.null(content)) {
      if (!"new" %in% .permitted) {
        stop(glue::glue("guid {guid} was not found on {connect$server}"))
      }
      warning(glue::glue(
        "guid {guid} was not found on {connect$server}.",
        "Creating new content with name {name}"
      ))
      content <- connect$content_create(
        name = name,
        title = title,
        ...
      )
    }
  } else {
    # name-based deployment
    content <- connect$content(name = name, include = NULL)
    if (length(content) > 1) {
      stop(glue::glue(
        "Found {length(content)} content items ",
        "matching {name} on {connect$server}",
        ", content must have a unique name."
      ))
    } else if (length(content) == 0) {
      if (!"new" %in% .permitted) {
        stop(glue::glue(
          "Content with name {name} was not found on {connect$server}"
        ))
      }
      message(glue::glue(
        "Creating NEW content {content$guid} ",
        "with name {name} on {connect$server}"
      ))
      # create app
      content <- connect$content_create(
        name = name,
        title = title,
        ...
      )
    } else {
      content <- content[[1]]
      if (!"existing" %in% .permitted) {
        stop(glue::glue(
          "Content with name {name} already exists at {content$dashboard_url}"
        ))
      }
      message(glue::glue(
        "Found EXISTING content {content$guid} with ",
        "name {name} on {connect$server}"
      ))
      # TODO: update values...? need a PUT endpoint
    }
  }
  return(content)
}

#' Get Jobs
#'
#' Retrieve details about server processes associated with a `content_item`,
#' such as a FastAPI app or a Quarto render.
#'
#' Note that Connect versions below 2022.10.0 use a legacy endpoint, and will
#' not return the complete set of information provided by newer versions.
#'
#' `get_jobs()` returns job data as a data frame, whereas `get_jobs_list()`
#' returns job data in a list.
#'
#' You might get job data as a data frame if you want to perform some
#' calculations about job data (e.g. counting server processes over time), or if
#' you want to filter jobs to find a specific key.
#'
#' The objects in list returned by `get_jobs_list()` are useful if you want to
#' take an action on a job, such as getting its process log with
#' `get_log()`.
#'
#' @param content A Content object, as returned by `content_item()`
#'
#' @return
#'
#' - `get_jobs()`: A data frame with a row representing each job.
#' - `get_job_list()`: A list with each element representing a job.
#'
#' Jobs contain the following fields:
#'
#' - `id`: The job identifier.
#' - `ppid`: The job's parent process identifier (see Note 1).
#' - `pid`: The job's process identifier.
#' - `key`: The job's unique key identifier.
#' - `remote_id`: The job's identifier for off-host execution configurations
#' (see Note 1).
#' - `app_id`: The job's parent content identifier; deprecated in favor of `content_id`.
#' - `app_guid`: The job's parent content GUID; deprecated in favor of `content_guid`.
#' - `content_id`: The job's parent content identifier.
#' - `content_guid`: The job's parent content GUID.
#' - `variant_id`: The identifier of the variant owning this job.
#' - `bundle_id`: The identifier of a content bundle linked to this job.
#' - `start_time`: The timestamp (RFC3339) indicating when this job started.
#' - `end_time`: The timestamp (RFC3339) indicating when this job finished.
#' - `last_heartbeat_time`: The timestamp (RFC3339) indicating the last time
#' this job was observed to be running (see Note 1).
#' - `queued_time`: The timestamp (RFC3339) indicating when this job was added
#' to the queue to be processed. Only scheduled reports will present a value
#' for this field (see Note 1).
#' - `queue_name`: The name of the queue which processes the job. Only
#' scheduled reports will present a value for this field (see Note 1).
#' - `tag`: A tag to identify the nature of the job.
#' - `exit_code`: The job's exit code. Present only when job is finished.
#' - `status`: The current status of the job. On Connect 2022.10.0 and newer,
#' one of Active: 0, Finished: 1, Finalized: 2; on earlier versions, Active:
#' 0, otherwise `NA`.
#' - `hostname`: The name of the node which processes the job.
#' - `cluster`: The location where this content runs. Content running on the
#' same server as Connect will have either a null value or the string Local.
#' Gives the name of the cluster when run external to the Connect host
#' (see Note 1).
#' - `image`: The location where this content runs. Content running on
#' the same server as Connect will have either a null value or the string
#' Local. References the name of the target image when content runs in
#' a clustered environment such as Kubernetes (see Note 1).
#' - `run_as`: The UNIX user that executed this job.
#'
#' @note
#' 1. On Connect instances earlier than 2022.10.0, these fields will contain `NA` values.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
#' jobs <- get_jobs(item)
#' job_list <- get_job_list(item)
#' }
#'
#' @family job functions
#' @family content functions
#' @rdname get_jobs
#' @export
get_jobs <- function(content) {
  validate_R6_class(content, "Content")

  jobs <- content$jobs()
  parse_connectapi_typed(jobs, connectapi_ptypes$jobs, strict = TRUE)
}

#' Terminate Jobs
#'
#' Register a job kill order for one or more jobs associated with a content
#' item. Requires Connect 2022.10.0 or newer.
#'
#' @param content A Content object, as returned by `content_item()`
#' @param keys Optional. One or more job keys, which can be obtained using
#' `get_jobs(content)`. If no keys are provided, will terminate all active
#' jobs for the provided content item.
#'
#' @return A data frame with the status of each termination request.
#'
#' - `app_id`: The content item's identifier.
#' - `app_guid`: The content item's GUID.
#' - `job_key`: The job key.
#' - `job_id`: The job's identifier.
#' - `result`: The result string returned by Connect.
#' - `code`: An error code, `NA` if the request was successful.
#' - `error`: An error message, `NA` if the result was successful.
#'
#' Note that `app_id`, `app_guid`, `job_id`, and `result` are `NA` if the
#' request returns an error.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
#' result <- terminate_jobs(item)
#' }
#'
#' @family job functions
#' @family content functions
#' @export
terminate_jobs <- function(content, keys = NULL) {
  validate_R6_class(content, "Content")

  if (is.null(keys)) {
    all_jobs <- get_jobs(content)
    keys <- all_jobs[all_jobs$status == 0, ]$key
    if (length(keys) == 0) {
      message("No active jobs found.")
      return(vctrs::vec_ptype(connectapi_ptypes$job_termination))
    }
  }

  res <- purrr::map(keys, content$register_job_kill_order)
  res_content <- purrr::map(res, httr::content)
  res_df <- tibble::tibble(
    parse_connectapi_typed(
      res_content,
      connectapi_ptypes$job_termination,
      strict = TRUE
    )
  )
  # Errors will not have the job_key.
  res_df$job_key <- keys
  res_df
}

#' @rdname get_jobs
#' @export
get_job_list <- function(content) {
  validate_R6_class(content, "Content")

  purrr::map(content$jobs(), ~ purrr::list_modify(.x, client = content$connect))
}

#' Get Job Log
#'
#' Get the log output for a job. Requires Connect 2022.10.0 or newer.
#'
#' Note: The output of `get_jobs()` cannot be used with `get_log()`.
#' Please use an object from the list returned by `get_job_list()`.
#'
#' @param job A job, represented by an element from the list returned by `get_job_list()`.
#' @param max_log_lines Optional. An integer indicating the maximum number of
#' log lines to return. If `NULL` (default), Connect returns a maximum of 5000
#' lines.
#'
#' @return A data frame with the requested log. Each row represents an entry.
#'
#' - `source`: `stdout` or `stderr`
#' - `timestamp`: The time of the entry.
#' - `data`: The logged text.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
#' jobs <- get_job_list(item)
#' log <- get_log(jobs[[1]])
#' }
#'
#'
#' @family job functions
#' @family content functions
#' @export
get_log <- function(job, max_log_lines = NULL) {
  error_if_less_than(job$client$version, "2022.10.0")

  query <- list(maxLogLines = max_log_lines)
  res <- job$client$GET(
    v1_url("content", job$app_guid, "jobs", job$key, "log"),
    query = query
  )
  parse_connectapi_typed(res$entries, connectapi_ptypes$job_log)
}

#' Set RunAs User
#'
#' Set the `RunAs` user for a piece of content.
#' The `run_as_current_user` flag only does anything if:
#'
#' - PAM is the authentication method
#' - `Applications.RunAsCurrentUser` is enabled on the server
#'
#' Also worth noting that the `run_as` user must exist on the Posit Connect
#' server (as a linux user) and have appropriate group memberships, or you will
#' get a `400: Bad Request`. Set to `NULL` to use the default RunAs user / unset
#' any current configuration.
#'
#' To "read" the current RunAs user, use the `Content` object or `get_content()` function.
#'
#' @param content an R6 Content item
#' @param run_as The RunAs user to use for this content
#' @param run_as_current_user Whether to run this content as the viewer of the application
#'
#' @return a Content object, updated with new details
#'
#' @seealso connectapi::content_update
#'
#' @family content functions
#' @export
set_run_as <- function(content, run_as, run_as_current_user = FALSE) {
  validate_R6_class(content, "Content")

  content$update(run_as = run_as, run_as_current_user = run_as_current_user)

  invisible(content$get_content_remote())

  return(content)
}


#' Delete Content
#'
#' Delete a content item. WARNING: This action deletes all history, configuration,
#' logs, and resources about a content item. It _cannot_ be undone.
#'
#' @param content an R6 content item
#' @param force Optional. A boolean that determines whether we should prompt in interactive sessions
#'
#' @return The R6 Content item. The item is deleted, but information about it is cached locally
#'
#' @family content functions
#' @export
content_delete <- function(content, force = FALSE) {
  validate_R6_class(content, "Content")

  cn <- content$get_content_remote()
  if (!force) {
    if (interactive()) {
      cat(glue::glue(
        "WARNING: Are you sure you want to delete '{cn$title}' ({cn$guid})?"
      ))
      if (utils::menu(c("Yes", "No")) == 2) {
        stop("'No' selected. Aborting content delete")
      }
    }
  }

  message(glue::glue("Deleting content '{cn$title}' ({cn$guid})"))
  res <- content$danger_delete()
  content$connect$raise_error(res)

  return(content)
}

#' Update Content
#'
#' Update settings for a content item. For a list of all settings, see the
#' [latest
#' documentation](https://docs.posit.co/connect/api/#patch-/v1/content/{guid})
#' or the documentation for your server via `connectapi::browse_api_docs()`.
#'
#' Popular selections are `content_update(access_type="all")`,
#' `content_update(access_type="logged_in")` or
#' `content_update(access_type="acl")`, process settings, title, description,
#' etc.
#'
#' - `content_update_access_type()` is a helper to make it easier to change access_type
#' - `content_update_owner()` is a helper to make it easier to change owner
#'
#' @param content An R6 content item
#' @param ... Settings up update that are passed along to Posit Connect
#' @param access_type One of "all", "logged_in", or "acl"
#' @param owner_guid The GUID of a user who is a publisher, so that they can
#'   become the new owner of the content
#'
#' @return An R6 content item
#'
#' @family content functions
#' @export
content_update <- function(content, ...) {
  validate_R6_class(content, "Content")

  content$update(...)

  content$get_content_remote()

  return(content)
}

#' @rdname content_update
#' @export
content_update_access_type <- function(
  content,
  access_type = c("all", "logged_in", "acl")
) {
  if (
    length(access_type) > 1 || !access_type %in% c("all", "logged_in", "acl")
  ) {
    stop("Please select one of 'all', 'logged_in', or 'acl'.")
  }
  content_update(content = content, access_type = access_type)
}

#' @rdname content_update
#' @export
content_update_owner <- function(content, owner_guid) {
  content_update(content = content, owner_guid = owner_guid)
}

#' Lock or Unlock Content
#'
#' Lock or unlock a content item. When content is locked, all processes are
#' terminated, rendering is disabled, and new bundles cannot be deployed.
#'
#' `lock_content()` locks a content item with an optional message displayed to
#' visitors (supports Markdown).
#'
#' `unlock_content()` unlocks a content item, reverting the effects of locking.
#'
#' @param content An R6 content item
#' @param locked_message Optional. A custom message that is displayed by the
#' content item when locked. It is possible to format this message using Markdown.
#'
#' @return An R6 content item
#'
#' @family content functions
#' @rdname lock_content
#' @export
#' @examples
#' \dontrun{
#' # Lock content with a message
#' client <- connect()
#' content <- content_item(client, "content-guid")
#' content <- lock_content(content, locked_message = "Ah ah ah! You didn't say the magic word!")
#'
#' # Lock content without a message
#' content <- lock_content(content)
#'
#' # Unlock content
#' content <- unlock_content(content)
#' }
lock_content <- function(content, locked_message = "") {
  validate_R6_class(content, "Content")
  error_if_less_than(content$connect$version, "2024.08.0")

  update_params <- list(locked = TRUE)
  if (!is.null(locked_message)) {
    update_params$locked_message <- locked_message
  }

  content$update(!!!update_params)
  content$get_content_remote()

  return(content)
}

#' @rdname lock_content
#' @export
unlock_content <- function(content) {
  validate_R6_class(content, "Content")
  error_if_less_than(content$connect$version, "2024.08.0")

  content$update(locked = FALSE, locked_message = "")
  content$get_content_remote()

  return(content)
}


#' Verify Content Name
#'
#' Ensures that a content name fits the specifications / requirements of Posit
#' Connect. Throws an error if content name is invalid. Content names (as of the
#' time of writing) must be between 3 and 64 alphanumeric characters, dashes,
#' and underscores
#'
#' @param name The proposed content name
#'
#' @return The name (or an error if invalid)
#'
#' @seealso connectapi::create_random_name
#' @family content functions
#' @export
verify_content_name <- function(name) {
  if (
    grepl("[^\\-\\_a-zA-Z0-9]", name, perl = TRUE) ||
      nchar(name) < 3 ||
      nchar(name) > 64
  ) {
    stop(glue::glue(
      "ERROR: content name '{name}' must be between 3 and 64 alphanumeric characters, ",
      "dashes, and underscores"
    ))
  }
  return(name)
}

#' Create Random Name
#'
#' Creates a random name from the LETTERS dataset
#'
#' @param length Optional. The length of the random name. Defaults to 25
#'
#' @return The randomly generated name
#'
#' @seealso connectapi::verify_content_name
#' @family content functions
#' @export
create_random_name <- function(length = 25) {
  tolower(paste(sample(LETTERS, length, replace = TRUE), collapse = ""))
}

#' Get Bundles
#'
#' Lists bundles for a content item
#'
#' @param content A R6 Content item, as returned by `content_item()`
#'
#' @rdname get_bundles
#' @param bundle_id A specific bundle ID for a content item
#' @family content functions
#' @export
get_bundles <- function(content) {
  validate_R6_class(content, "Content")
  bundles <- content$get_bundles()

  parse_connectapi_typed(bundles, connectapi_ptypes$bundles)
}

#' @rdname get_bundles
#' @family content functions
#' @export
delete_bundle <- function(content, bundle_id) {
  validate_R6_class(content, "Content")
  cn <- content$get_content_remote()
  message(glue::glue(
    "Deleting bundle {bundle_id} for content '{cn$title}' ({cn$guid})"
  ))
  res <- content$bundle_delete(bundle_id)
  content$connect$raise_error(res)
  return(content)
}


#' Content permissions
#'
#' Get or set content permissions for a content item
#'
#' Permission modification:
#' - `content_add_*` adds a permission to the content
#' - `content_delete_*` removes a permission from the content
#'
#' Permission retrieval:
#' - `get_content_permissions()` lists permissions
#' - `get_my_permission()` gets the permission associated with the caller.
#' - `get_user_permission()` gets the permissions associated with a given user.
#'   It does not evaluate group memberships
#' - `get_group_permission()` gets the permissions associated with a given
#'   group.
#'
#' NOTE: by default, the owner is injected with an "NA_character_" permission id.
#' This makes it easier to find / isolate this record.
#'
#' @param content An R6 content object
#' @param guid The guid associated with either a user (for `content_add_user`) or group (for `content_add_group`)
#' @param role The role to assign to a user. Either "viewer" or "owner." Defaults to "viewer"
#' @param add_owner Optional. Whether to include the owner in returned
#'   permission sets. Default is TRUE. The owner will have an NA_character_
#'   permission "id"
#'
#' @name permissions
#' @rdname permissions
#' @family content functions
#' @export
content_add_user <- function(content, guid, role = c("viewer", "owner")) {
  validate_R6_class(content, "Content")
  role <- .define_role(role)

  purrr::map(guid, ~ .content_add_permission_impl(content, "user", .x, role))

  return(content)
}

#' @rdname permissions
#' @export
content_add_group <- function(content, guid, role = c("viewer", "owner")) {
  validate_R6_class(content, "Content")
  role <- .define_role(role)

  purrr::map(
    guid,
    ~ .content_add_permission_impl(
      content = content,
      type = "group",
      guid = .x,
      role = role
    )
  )

  return(content)
}

.content_delete_permission_impl <- function(content, type, guid) {
  res <- .get_permission(content, type, guid)
  if (length(res) > 0) {
    message(glue::glue("Removing {type} permission for '{guid}'"))
    remove_permission <- content$permissions_delete(res[[1]]$id)
    return(remove_permission)
  } else {
    message(glue::glue(
      "{type} '{guid}' already does not have access. No permission being removed"
    ))
    return(NULL)
  }
}

.content_add_permission_impl <- function(content, type, guid, role) {
  existing <- .get_permission(content, type, guid)
  if (length(existing) > 0) {
    message(glue::glue(
      "Updating permission for {type} '{guid}' with role '{role}'"
    ))
    res <- content$permissions_update(
      id = existing[[1]]$id,
      principal_guid = guid,
      principal_type = type,
      role = role
    )
  } else {
    message(glue::glue(
      "Adding permission for {type} '{guid}' with role '{role}'"
    ))
    res <- content$permissions_add(
      principal_guid = guid,
      principal_type = type,
      role = role
    )
  }
  return(res)
}

#' @rdname permissions
#' @export
content_delete_user <- function(content, guid) {
  validate_R6_class(content, "Content")
  purrr::map(
    guid,
    ~ .content_delete_permission_impl(
      content = content,
      type = "user",
      guid = .x
    )
  )
  return(content)
}

#' @rdname permissions
#' @export
content_delete_group <- function(content, guid) {
  validate_R6_class(content, "Content")
  purrr::map(
    guid,
    ~ .content_delete_permission_impl(
      content = content,
      type = "group",
      guid = .x
    )
  )
  return(content)
}

.define_role <- function(role) {
  if (length(role) > 1) {
    # use default
    return("viewer")
  } else {
    if (role %in% c("viewer", "owner")) {
      return(role)
    } else {
      stop(glue::glue(
        "ERROR: invalid role. Expected 'viewer' or 'owner', instead got {{ role }}"
      ))
    }
  }
}

.get_permission <- function(content, type, guid, add_owner = TRUE) {
  res <- content$permissions(add_owner = add_owner)
  purrr::keep(
    res,
    ~ identical(.x$principal_type, type) && identical(.x$principal_guid, guid)
  )
}

#' @rdname permissions
#' @export
get_user_permission <- function(content, guid, add_owner = TRUE) {
  validate_R6_class(content, "Content")
  res <- .get_permission(content, "user", guid, add_owner = add_owner)
  if (length(res) > 0) {
    return(res[[1]])
  } else {
    return(NULL)
  }
}

#' @rdname permissions
#' @export
get_my_permission <- function(content, add_owner = TRUE) {
  my_guid <- content$connect$GET("me")$guid
  get_user_permission(content, my_guid, add_owner = add_owner)
}

#' @rdname permissions
#' @export
get_group_permission <- function(content, guid) {
  validate_R6_class(content, "Content")
  # do not add_owner, because groups cannot own content
  res <- .get_permission(content, "group", guid, add_owner = FALSE)
  if (length(res) > 0) {
    return(res[[1]])
  } else {
    return(NULL)
  }
}


#' @rdname permissions
#' @export
get_content_permissions <- function(content, add_owner = TRUE) {
  validate_R6_class(content, "Content")
  res <- content$permissions(add_owner = add_owner)
  parse_connectapi_typed(res, connectapi_ptypes$permissions)
}

#' Render a content item.
#'
#' @description Submit a request to render a content item. Once submitted, the
#' server runs an asynchronous process to render the content. This might be
#' useful if content needs to be updated after its source data has changed,
#' especially if this doesn't happen on a regular schedule.
#'
#' Only valid for rendered content (e.g., most Quarto documents, Jupyter
#' notebooks, R Markdown reports).
#'
#' @param content The content item you wish to render.
#' @param variant_key If a variant key is provided, render that variant. Otherwise, render the default variant.
#' @return A [VariantTask] object that can be used to track completion of the render.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
#' task <- content_render(item)
#' poll_task(task)
#' }
#'
#' @export
content_render <- function(content, variant_key = NULL) {
  scoped_experimental_silence()
  validate_R6_class(content, "Content")
  if (!content$is_rendered) {
    stop(glue::glue(
      "Render not supported for application mode: {content$content$app_mode}. ",
      "Did you mean content_restart()?"
    ))
  }
  if (is.null(variant_key)) {
    target_variant <- get_variant(content, "default")
  } else {
    target_variant <- get_variant(content, variant_key)
  }
  render_task <- target_variant$render()

  VariantTask$new(
    connect = content$connect,
    content = content$content,
    key = target_variant$key,
    task = render_task
  )
}

#' Restart a content item.
#'
#' @description Submit a request to restart a content item. Once submitted, the
#' server performs an asynchronous request to kill all processes associated with
#' the content item, starting new processes as needed. This might be useful if
#' the application relies on data that is loaded at startup, or if its memory
#' usage has grown over time.
#'
#' Note that users interacting with certain types of applications may have their
#' workflows interrupted.
#'
#' Only valid for interactive content (e.g., applications, APIs).
#'
#' @param content The content item you wish to restart.
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
#' content_restart(item)
#' }
#'
#' @importFrom rlang :=
#' @export
content_restart <- function(content) {
  validate_R6_class(content, "Content")
  if (!content$is_interactive) {
    stop(glue::glue(
      "Restart not supported for application mode: {content$content$app_mode}. ",
      "Did you mean content_render()?"
    ))
  }
  unix_epoch_in_seconds <- as.integer(Sys.time())
  # nolint start: object_usage_linter, object_name_linter
  # https://rlang.r-lib.org/reference/glue-operators.html#using-glue-syntax-in-packages
  env_var_name <- glue::glue("_CONNECT_RESTART_{unix_epoch_in_seconds}")
  content$environment_set("{env_var_name}" := unix_epoch_in_seconds)
  tryCatch(
    {
      content$environment_set("{env_var_name}" := NA)
    },
    error = function(e) {
      # we sometimes see the rapid set/unset lead to failures, which leave the
      # env var in place. If we get a 500 error, wait a second and try one more
      # time
      if (grepl("500", conditionMessage(e))) {
        Sys.sleep(1)
        tryCatch(
          content$environment_set("{env_var_name}" := NA),
          error = function(e) {
            warning(
              glue::glue(
                "Restarted content by setting environment variable {env_var_name}, but was unable to clean it up. See ?set_environment_remove for how to manually remove this variable."
              ),
              call. = FALSE
            )
          }
        )
      } else {
        warning(
          glue::glue(
            "Restarted content by setting environment variable {env_var_name}, but was unable to clean it up. See ?set_environment_remove for how to manually remove this variable."
          ),
          call. = FALSE
        )
      }
    }
  )

  # nolint end
  invisible(NULL)
}

#' Package dependencies for a content item
#'
#' @description Get a data frame of package dependencies used by a content item.
#'
#' @param content A content item
#'
#' @return A data frame with the following columns:
#'
#' - `language` : Language ecosystem the package belongs to (`r` or `python`)
#' - `name`: The package name
#' - `version`: The package version
#' - `hash`: For R packages, the package `DESCRIPTION` hash
#'
#' @examples
#' \dontrun{
#' client <- connect()
#' item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
#' packages <- get_content_packages(item)
#' }
#'
#' @family packages functions
#' @export
get_content_packages <- function(content) {
  error_if_less_than(content$connect$version, "2025.01.0")
  res <- content$packages()
  parse_connectapi_typed(res, connectapi_ptypes$content_packages)
}

#' Search for content on the Connect server
#'
#' @param client A Connect object
#' @param q The search query, using the syntax described in the Connect
#'   documentation on [content search
#'   terms](https://docs.posit.co/connect/user/viewing-content/#searching-content)
#' @param include Comma-separated character string of values indicating additional
#'   details to include in the response. Values can be `owner` and `vanity_url`;
#'   both are included by default.
#' @param page_size The number of items to fetch per page. Maximum is 500.
#' @param limit Maximum number of items to return overall. Defaults to `Inf` (all items).
#' @param ... Additional query parameters passed to the API for future expansion.
#'   Note: If you pass `page_number` here, it will affect the *starting* page
#'   for pagination, but all subsequent pages will still be fetched. This is
#'   usually not what you want.
#'
#' @return
#' A list of [Content] objects, of class "connect_content_list"
#'
#' @details
#' Please see https://docs.posit.co/connect/api/#get-/v1/search/content for more
#' information.
#'
#' @examples
#' \dontrun{
#' library(connectapi)
#' client <- connect()
#'
#' my_content <- search_content(client, q = "owner:@me")
#'
#' shiny_content <- purrr::keep(my_content, function(x) {
#'   x$content$app_mode == "rmd-shiny"
#' })
#'
#' purrr::map(shiny_content, lock_content)
#' }
#'
#' @family content functions
#' @export
search_content <- function(
  client,
  q = NULL,
  include = "owner,vanity_url",
  page_size = 500,
  limit = Inf,
  ...
) {
  error_if_less_than(client$version, "2024.04.0")

  res <- page_offset(
    client,
    req = .search_content(
      client,
      q = q,
      include = include,
      page_size = page_size,
      ...
    ),
    limit = limit
  )

  content_list <- purrr::map(res, function(x) {
    Content$new(client, x)
  })

  class(content_list) <- c("connect_content_list", class(content_list))
  content_list
}

.search_content <- function(
  client,
  q,
  include,
  page_number = 1,
  page_size = 500,
  ...
) {
  path <- v1_url("search", "content")

  query <- list(
    q = q,
    page_number = page_number,
    page_size = page_size,
    include = include,
    ...
  )

  client$GET(path, query = query)
}

#' Convert content list to a data frame
#'
#' @description
#' Converts a list returned by [search_content()] into a data frame.
#'
#' @param x A `connect_content_list` object (from [search_content()]).
#' @param row.names Passed to [base::as.data.frame()].
#' @param optional Passed to [base::as.data.frame()].
#' @param ... Passed to [base::as.data.frame()].
#'
#' @return A `data.frame` with one row per content item.
#' @export
as.data.frame.connect_content_list <- function(
  x,
  row.names = NULL, # nolint
  optional = FALSE,
  ...
) {
  content_tbl <- as_tibble(x)
  as.data.frame(
    content_tbl,
    row.names = row.names,
    optional = optional,
    ...
  )
}

#' Convert integration list to a tibble
#'
#' @description
#' Converts a list returned by [search_content()] to a tibble.
#'
#' @param x A `connect_content_list` object.
#' @param ... Unused.
#'
#' @return A tibble with one row per content item.
#' @export
as_tibble.connect_content_list <- function(x, ...) {
  content_data <- purrr::map(x, "content")
  parse_connectapi_typed(content_data, connectapi_ptypes$content)
}

# Setup ----------------------------------------------------

collab_guid <- NULL
viewer_guid <- NULL

# deploy content
cont1_title <- "Test Content 1"
cont1_content <- deploy_example(client, "static", title = cont1_title)
cont1_guid <- cont1_content$content$guid


# Metadata Tests ----------------------------------------------------

test_that("content_item works", {
  cont1_tmp <- client %>% content_item(guid = cont1_guid)

  expect_true(validate_R6_class(cont1_tmp, "Content"))
  expect_equal(cont1_tmp$content$guid, cont1_guid)
})

test_that("content_title works in a simple example", {
  test_title <- content_title(client, cont1_guid)
  expect_identical(test_title, cont1_title)
})

test_that("content_title handles missing content gracefully", {
  null_title <- content_title(client, "not_a_real_guid")
  expect_identical(null_title, "Unknown Content")

  null_title_custom <- content_title(
    client,
    "not_a_real_guid",
    "other-default"
  )
  expect_identical(null_title_custom, "other-default")
})

test_that("content_title handles NULL titles gracefully", {
  c2_name <- uuid::UUIDgenerate()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  c2 <- deploy(connect = client, bundle = bnd, name = c2_name, title = NA)
  expect_null(c2$content$title)
  null_title <- content_title(client, c2$content$guid, "Test Title")
  expect_identical(null_title, "Test Title")
})

test_that("content_update_owner works", {
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )
  myc <- deploy(client, bnd)

  new_user <- client$users_create(
    username = glue::glue("test_admin_{create_random_name()}"),
    email = "example@example.com",
    user_role = "administrator",
    user_must_set_password = TRUE
  )

  expect_equal(myc$get_content_remote()$owner_guid, client$me()$guid)

  res <- content_update_owner(myc, new_user$guid)

  expect_equal(
    myc$get_content_remote()$owner_guid,
    new_user$guid
  )

  # permissions do not remain
  expect_null(get_user_permission(myc, client$me()$guid))

  # switch back (as an admin)
  res2 <- content_update_owner(myc, client$me()$guid)

  expect_equal(myc$get_content_remote()$owner_guid, client$me()$guid)

  # permissions do not remain
  expect_null(get_user_permission(myc, new_user$guid))

  # viewer cannot be made an owner
  viewer_user <- client$users_create(
    username = glue::glue("test_viewer_{create_random_name()}"),
    email = "viewer@example.com",
    user_role = "viewer",
    user_must_set_password = TRUE
  )

  expect_error(
    expect_message(
      content_update_owner(myc, viewer_user$guid),
      "permission to publish"
    ),
    "403"
  )
})

test_that("content_update_access_type works", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)

  # returns as expected
  tsk <- content_update_access_type(tsk, "all")
  expect_equal(tsk$content$access_type, "all")

  # modifies the R6 object in place
  content_update_access_type(tsk, "logged_in")
  expect_equal(tsk$content$access_type, "logged_in")

  # works twice
  content_update_access_type(tsk, "acl")
  content_update_access_type(tsk, "acl")
  expect_equal(tsk$content$access_type, "acl")

  expect_error(content_update_access_type(tsk), "one of")
})

test_that("content_update works", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)

  content_update(tsk, title = "test content_update")
  expect_equal(tsk$content$title, "test content_update")

  # should not change or error with empty input
  expect_equal(content_update(tsk)$content$title, "test content_update")

  expect_equal(
    content_update(tsk, title = "test content_update2")$content$title,
    "test content_update2"
  )
})

test_that("content_delete works", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)

  tsk <- deploy(connect = client, bundle = bund)

  expect_message(res <- content_delete(tsk, force = TRUE), "Deleting content")
  expect_true(validate_R6_class(res, "Content"))

  expect_error(res$get_content_remote(), "404")
})

test_that("content_delete prompts and errors", {
  skip("not sure how to test this")
})

# Bundles ---------------------------------------------------

test_that("get_bundles and delete_bundle work", {
  bnd_name <- create_random_name()
  bnd <- bundle_static(
    path = rprojroot::find_package_root_file(
      "tests/testthat/examples/static/test.png"
    )
  )

  bc1 <- deploy(client, bnd, bnd_name)
  bc1 <- deploy(client, bnd, bnd_name)
  bc1 <- deploy(client, bnd, bnd_name)

  bnd_dat <- get_bundles(bc1)
  expect_equal(nrow(bnd_dat), 3)
  expect_s3_class(bnd_dat, "tbl_df")

  not_active_bundles <- bnd_dat[!bnd_dat$active, ]

  bnd_del <- delete_bundle(bc1, not_active_bundles[["id"]][[1]])
  expect_true(validate_R6_class(bnd_del, "Content"))

  bnd_dat2 <- get_bundles(bc1)
  expect_equal(nrow(bnd_dat2), 2)
})

# Permissions ---------------------------------------

test_that("returns owner permission", {
  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)
  tsk <- deploy(connect = client, bundle = bund)
  my_guid <- client$GET("me")$guid

  prm <- get_content_permissions(tsk)
  expect_length(prm[["id"]], 1)
  expect_equal(prm[["principal_guid"]], my_guid)

  my_prm <- get_user_permission(tsk, my_guid)
  expect_equal(my_prm$role, "owner")

  # NOTE: this NA shows that owner was injected
  expect_equal(my_prm$id, NA_character_)

  my_prm <- get_my_permission(tsk)
  expect_equal(my_prm$role, "owner")

  # add_owner = FALSE gives the previous behavior
  expect_length(get_my_permission(tsk, add_owner = FALSE), 0)
})

test_that("add a collaborator works", {
  # create a user
  collab <- client$users_create(
    username = glue::glue("test_collab{create_random_name()}"),
    email = "collab@example.com",
    user_must_set_password = TRUE,
    user_role = "publisher"
  )
  collab_guid <<- collab$guid

  # no permission at first
  expect_null(get_user_permission(cont1_content, collab_guid))

  # add a collaborator
  invisible(content_add_user(cont1_content, collab_guid, "owner"))

  expect_equal(get_user_permission(cont1_content, collab_guid)$role, "owner")
})

test_that("add collaborator twice works", {
  # add a collaborator
  invisible(content_add_user(cont1_content, collab_guid, "owner"))
  invisible(content_add_user(cont1_content, collab_guid, "owner"))

  # get acl
  acls <- get_content_permissions(cont1_content)

  which_match <- purrr::map2_lgl(
    acls$principal_guid,
    acls$role,
    function(.x, .y) {
      .x == collab_guid && .y == "owner"
    }
  )
  expect_true(any(which_match))
  expect_equal(sum(which_match), 1)
})

test_that("add a viewer works", {
  # create a user
  view_user <- client$users_create(
    username = glue::glue("test_viewer{create_random_name()}"),
    email = "viewer@example.com",
    user_must_set_password = TRUE,
    user_role = "viewer"
  )
  viewer_guid <<- view_user$guid

  # no permission at first
  expect_null(get_user_permission(cont1_content, viewer_guid))

  # add a viewer
  invisible(content_add_user(cont1_content, viewer_guid, "viewer"))

  # get acl
  acls <- get_content_permissions(cont1_content)

  which_match <- purrr::map2_lgl(
    acls$principal_guid,
    acls$role,
    function(.x, .y) {
      .x == viewer_guid && .y == "viewer"
    }
  )
  expect_true(any(which_match))
  expect_equal(sum(which_match), 1)
})

test_that("add a viewer twice works", {
  # add a viewer
  invisible(content_add_user(cont1_content, viewer_guid, "viewer"))
  invisible(content_add_user(cont1_content, viewer_guid, "viewer"))

  # get acl
  acls <- get_content_permissions(cont1_content)

  which_match <- purrr::map2_lgl(
    acls$principal_guid,
    acls$role,
    function(.x, .y) {
      .x == viewer_guid && .y == "viewer"
    }
  )
  expect_true(any(which_match))
  expect_equal(sum(which_match), 1)
})

test_that("remove a collaborator works", {
  # remove a collaborator
  invisible(content_delete_user(cont1_content, collab_guid))

  # get acl
  acls <- get_content_permissions(cont1_content)

  which_match <- purrr::map2_lgl(
    acls$principal_guid,
    acls$role,
    function(.x, .y) {
      .x == collab_guid && .y == "owner"
    }
  )
  expect_false(any(which_match))
})

test_that("remove a collaborator twice works", {
  # remove a collaborator
  invisible(content_delete_user(cont1_content, collab_guid))
  invisible(content_delete_user(cont1_content, collab_guid))

  # get acl
  acls <- get_content_permissions(cont1_content)

  which_match <- purrr::map2_lgl(
    acls$principal_guid,
    acls$role,
    function(.x, .y) {
      .x == collab_guid && .y == "owner"
    }
  )
  expect_false(any(which_match))
})

# Lock / Unlock -----------------------------------------

test_that("lock and unlock content works", {
  skip_if_connect_older_than(client, "2024.08.0")

  tar_path <- rprojroot::find_package_root_file(
    "tests/testthat/examples/static.tar.gz"
  )
  bund <- bundle_path(path = tar_path)
  tsk <- deploy(connect = client, bundle = bund)

  # Lock with message
  lock_content(tsk, locked_message = "Maintenance in progress")
  expect_true(tsk$content$locked)
  expect_equal(tsk$content$locked_message, "Maintenance in progress")

  # Lock again with different message
  lock_content(tsk, locked_message = "Still under maintenance")
  expect_true(tsk$content$locked)
  expect_equal(tsk$content$locked_message, "Still under maintenance")

  # Lock again without message clears it
  lock_content(tsk)
  expect_true(tsk$content$locked)
  expect_equal(tsk$content$locked_message, "")

  # Unlock
  unlock_content(tsk)
  expect_false(tsk$content$locked)
  expect_equal(tsk$content$locked_message, "")
})

without_internet({
  scoped_experimental_silence()

  repo_path <- function(subpath) {
    glue::glue("https://connect.example/__api__/repo/{subpath}")
  }

  test_that("Query params to repo_account", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$repo_account("https://github.com/posit-dev/connectapi"),
      repo_path("account?url=https%3A%2F%2Fgithub.com")
    )
    expect_error(
      con$repo_account("asdf"),
      "Scheme and hostname must be provided"
    )
  })

  test_that("Query params to repo_branches", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$repo_branches("https://github.com/posit-dev/connectapi"),
      repo_path(
        "branches?url=https%3A%2F%2Fgithub.com%2Fposit-dev%2Fconnectapi"
      )
    )
  })

  test_that("Query params to repo_manifest_dirs", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$repo_manifest_dirs("https://github.com/posit-dev/connectapi", "main"),
      repo_path(
        "manifest-dirs?url=https%3A%2F%2Fgithub.com%2Fposit-dev%2Fconnectapi&branch=main"
      )
    )
  })
})

with_mock_api({
  test_that("we can retrieve a repository information if it exists", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_true(item$repository()$polling)
  })

  test_that("repository is null if it is not set", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "c3426b0b-e21d-3d80-c698-a935ad614066")
    expect_null(item$repository())
  })

  test_that("we can set a repository", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "c3426b0b-e21d-3d80-c698-a935ad614066")
    expect_PUT(
      item$repo_set(repository = "https://github.com/posit-dev/connectapi"),
      "https://connect.example/__api__/v1/content/c3426b0b/repository",
      '{"repository":"https://github.com/posit-dev/connectapi",',
      '"branch":"main","directory":".","polling":false}'
    )
  })

  test_that("we can enable polling", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_PATCH(
      item$repo_enable(TRUE),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/repository",
      '{"polling":true}'
    )
  })
})

# Temporarily skip these on newer Connect versions because the manifest for this
# content specifies an old version of R not found in newer images.
skip_if(safe_server_version(client) > "2024.03.0")

# ensure that RSPM is being used so these do not take eternity
rmd_content <- deploy_example(client, "rmarkdown")

test_that("get_environment works with no environment variables", {
  env <- get_environment(rmd_content)
  curr_vers <- env$env_version
  env <- get_environment(rmd_content)

  expect_identical(env$env_vars, list())
})

test_that("set_environment works", {
  scoped_experimental_silence()
  env <- get_environment(rmd_content)

  new_env <- set_environment_new(
    env,
    test = "value",
    test1 = TRUE,
    test2 = 4567
  )

  expect_equal(
    new_env$env_vars,
    list(
      "test",
      "test1",
      "test2"
    )
  )

  new_env1 <- set_environment_new(env, test1 = "another")
  expect_equal(
    new_env$env_vars,
    list(
      "test",
      "test1",
      "test2"
    )
  )

  # remove a value multiple times
  rm1 <- set_environment_remove(env, test)
  expect_equal(rm1$env_vars, list("test1", "test2"))
  rm2 <- set_environment_remove(env, test)
  expect_equal(rm1$env_vars, list("test1", "test2"))

  rm3 <- set_environment_remove(env, "test1")
  expect_equal(rm3$env_vars, list("test2"))

  rm4 <- set_environment_remove(env, "test2")
  expect_equal(rm4$env_vars, list())
})

test_that("add environment variable works", {
  res <- set_environment_new(rmd_content, MYVAR = "hi")
  expect_true("MYVAR" %in% res$env_vars)
  expect_true(validate_R6_class(res, "Environment"))
})

test_that("edit environment variable works", {
  res <- set_environment_new(rmd_content, MYVAR = "hi", OTHER = "other")
  res2 <- set_environment_new(rmd_content, MYVAR = "hi2")
  expect_true("MYVAR" %in% res2$env_vars)
  expect_true("OTHER" %in% res2$env_vars)
  expect_true(validate_R6_class(res2, "Environment"))
})

test_that("get environment works", {
  res <- set_environment_new(rmd_content, MYVAR = "hi")
  res <- get_environment(rmd_content)
  expect_true("MYVAR" %in% res$env_vars)
  expect_true(validate_R6_class(res, "Environment"))
})

test_that("remove environment variable works", {
  res <- set_environment_new(rmd_content, MYVAR = "hi", ANOTHER = "how")
  rem <- set_environment_new(rmd_content, MYVAR = NA)

  expect_false("MYVAR" %in% rem$env_vars)
  expect_true("ANOTHER" %in% rem$env_vars)
  expect_true(validate_R6_class(res, "Environment"))

  res <- set_environment_remove(rmd_content, ANOTHER)
  expect_false("ANOTHER" %in% res$env_vars)

  res <- set_environment_new(rmd_content, MYVAR = "hi", ANOTHER = "how")
  myvar <- c("MYVAR", "ANOTHER")
  res <- set_environment_remove(rmd_content, !!myvar)
  expect_false("ANOTHER" %in% res$env_vars)
  expect_false("MYVAR" %in% res$env_vars)

  res <- set_environment_new(rmd_content, MYVAR = "hi", ANOTHER = "how")
  myvar <- c("MYVAR", "ANOTHER")
  res <- set_environment_remove(rmd_content, !!!myvar)
  expect_false("ANOTHER" %in% res$env_vars)
  expect_false("MYVAR" %in% res$env_vars)

  res <- set_environment_new(rmd_content, MYVAR = "hi", ANOTHER = "how")
  myvar <- c("MYVAR" = "1", "ANOTHER" = "2")
  res <- set_environment_remove(rmd_content, !!!myvar)
  expect_false("ANOTHER" %in% res$env_vars)
  expect_false("MYVAR" %in% res$env_vars)

  res <- set_environment_new(rmd_content, MYVAR = "hi", ANOTHER = "how")
  myvar <- c("MYVAR" = "1", "ANOTHER" = "2")
  res <- set_environment_remove(rmd_content, !!myvar)
  expect_true("ANOTHER" %in% res$env_vars)
  expect_true("MYVAR" %in% res$env_vars)
})

test_that("set all environment variables works", {
  res <- set_environment_all(rmd_content)
  expect_true(length(res$env_vars) == 0)
  expect_true(validate_R6_class(res, "Environment"))

  res <- set_environment_all(rmd_content, "MYVAR" = "ONE", "ANOTHER" = "HELLO")
  expect_equal(res$env_vars, list("ANOTHER", "MYVAR"))
  expect_true(validate_R6_class(res, "Environment"))

  res <- set_environment_all(rmd_content, "MYVAR" = "TWO", "HELLO" = "AGAIN")
  expect_equal(res$env_vars, list("HELLO", "MYVAR"))
  expect_true(validate_R6_class(res, "Environment"))

  res <- set_environment_all(rmd_content)
  expect_true(length(res$env_vars) == 0)
  expect_true(validate_R6_class(res, "Environment"))
})

with_mock_api({
  test_that("content environment vars", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")

    expect_GET(
      item$environment(),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/environment"
    )
    expect_PATCH(
      item$environment_set(VAR_NAME = "new_value"),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/environment",
      '[{"name":"VAR_NAME","value":"new_value"}]'
    )
    expect_PATCH(
      item$environment_set(VAR_NAME = NA),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/environment",
      '[{"name":"VAR_NAME","value":null}]'
    )
    expect_PUT(
      item$environment_all(VAR_NAME = "new_value"),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/environment",
      '[{"name":"VAR_NAME","value":"new_value"}]'
    )
    expect_PUT(
      item$environment_all(),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/environment",
      "[]"
    )
  })

  test_that("env var function wrappers", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "34567890-e21d-3d80-c698-a935ad614066")

    expect_PATCH(
      set_environment_remove(item, var_to_remove),
      "https://connect.example/__api__/v1/content/34567890/environment",
      '[{"name":"var_to_remove","value":null}]'
    )
    expect_PATCH(
      set_environment_remove(item, one, another),
      "https://connect.example/__api__/v1/content/34567890/environment",
      '[{"name":"one","value":null},{"name":"another","value":null}]'
    )
    expect_PATCH(
      set_environment_remove(item, "var_to_remove"),
      "https://connect.example/__api__/v1/content/34567890/environment",
      '[{"name":"var_to_remove","value":null}]'
    )
  })
})

with_mock_dir("2025.09.0", {
  client <- Connect$new(
    server = "https://connect.example",
    api_key = "not-a-key"
  )

  test_that("page_offset respects a specified 'limit' ", {
    # These mocks return four results across two pages.
    res <- search_content(client, q = "blobfish", limit = 3)
    expect_equal(length(res), 3)
    expect_equal(
      purrr::map_chr(res, list("content", "title")),
      c("blobfish dashboard", "blobfish api", "blobfish report")
    )
  })
})

test_that("verify_content_name works with valid names", {
  short_name <- "123"
  expect_equal(verify_content_name(short_name), short_name)

  long_name <- "abcdefghijklmnopqrstuvwxyabcdefghijklmnopqrstuvwxyabcdefghijklmn"
  expect_equal(verify_content_name(long_name), long_name)

  uuid <- uuid::UUIDgenerate()
  expect_equal(verify_content_name(uuid), uuid)
})

test_that("verify_content_name fails for invalid names", {
  expect_error(verify_content_name("a"))
  expect_error(verify_content_name(NA))
  # 65 characters
  expect_error(verify_content_name(
    "abcdefghijklmnopqrstuvwxyabcdefghijklmnopqrstuvwxyabcdefghijklmno"
  ))
  expect_error(verify_content_name(NULL))
  expect_error(verify_content_name("abc!@#$"))
  expect_error(verify_content_name("123 abc"))
})

test_that("create_random_name works with no length", {
  expect_type(create_random_name(), "character")
})

test_that("create_random_name works with length", {
  expect_equal(nchar(create_random_name(200)), 200)
  expect_equal(nchar(create_random_name(1)), 1)
})

with_mock_api({
  test_that("we can retrieve the content list", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_true(validate_R6_class(con, "Connect"))
    content_list <- get_content(con)
    expect_s3_class(content_list, "data.frame")
    expect_equal(nrow(content_list), 3)
  })

  test_that("we can retrieve a content item", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_true(validate_R6_class(item, "Content"))
    expect_equal(
      item$get_url(),
      "https://connect.example/content/f2f37341-e21d-3d80-c698-a935ad614066/"
    )

    expect_snapshot(print(item))
  })

  test_that("browse URLs", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    # Inject into this function something other than utils::browseURL
    # so we can assert that it is being called without actually trying to open a browser
    suppressMessages(trace(
      "browse_url",
      where = connectapi::browse_solo,
      tracer = quote({
        browseURL <- # nolint: object_name_linter
          function(x) {
            warning(paste("Opening", x))
          }
      }),
      at = 1,
      print = FALSE
    ))
    expect_warning(
      browse_solo(item),
      "Opening https://connect.example/content/f2f37341-e21d-3d80-c698-a935ad614066/"
    )
    expect_warning(
      browse_dashboard(item),
      "Opening https://connect.example/connect/#/apps/f2f37341-e21d-3d80-c698-a935ad614066"
    )
    suppressMessages(untrace("browse_url", where = connectapi::browse_solo))
  })

  test_that("we can modify a content item", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))

    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_warning(
      expect_PATCH(
        item$update(description = "new description"),
        "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
        '{"description":"new description"}'
      ),
      "the server version is not exposed by this Posit Connect instance"
    )
  })

  test_that("we can delete a content item", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    expect_DELETE(item$danger_delete())
  })

  test_that("we can get the content item's permissions", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")
    perms <- item$permissions()
    expect_type(perms, "list")
    expect_equal(perms[[1]]$id, 94)

    # Now get perms with id specified
    one_perm <- item$permissions(id = 94)
    expect_equal(one_perm, perms[[1]])
  })

  test_that("we can modify the content item's permissions", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    item <- content_item(con, "f2f37341-e21d-3d80-c698-a935ad614066")

    # Add one
    expect_POST(
      item$permissions_add(
        "a01792e3-2e67-402e-99af-be04a48da074",
        "user",
        "viewer"
      ),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/permissions",
      '{"principal_guid":"a01792e3-2e67-402e-99af-be04a48da074","principal_type":"user","role":"viewer"}'
    )

    # Update one
    expect_PUT(
      item$permissions_update(
        94,
        "a01792e3-2e67-402e-99af-be04a48da074",
        "user",
        "editor"
      ),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/permissions/94",
      '{"principal_guid":"a01792e3-2e67-402e-99af-be04a48da074","principal_type":"user","role":"editor"}'
    )

    expect_DELETE(
      item$permissions_delete("a01792e3-2e67-402e-99af-be04a48da074"),
      "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/permissions"
    )
  })
})

without_internet({
  test_that("Query params to connect$content()", {
    con <- Connect$new(server = "https://connect.example", api_key = "fake")
    expect_GET(
      con$content(),
      "https://connect.example/__api__/v1/content?include=tags%2Cowner"
    )

    expect_GET(
      con$content("a01792e3-2e67-402e-99af-be04a48da074"),
      "https://connect.example/__api__/v1/content/a01792e3-2e67-402e-99af-be04a48da074"
    )

    expect_GET(
      con$content(include = NULL),
      "https://connect.example/__api__/v1/content"
    )

    expect_GET(
      con$content(owner_guid = "a01792e3-2e67-402e-99af-be04a48da074"),
      "https://connect.example/__api__/v1/content?owner_guid=a01792e3-2e67-402e-99af-be04a48da074&include=tags%2Cowner"
    )

    expect_GET(
      con$content(name = "A name for content"),
      "https://connect.example/__api__/v1/content?name=A%20name%20for%20content&include=tags%2Cowner"
    )

    expect_GET(
      con$content(guid = 1),
      "https://connect.example/__api__/v1/content/1?include=tags%2Cowner"
    )

    expect_GET(
      con$content(guid = 1, include = "owner,tags"),
      "https://connect.example/__api__/v1/content/1?include=owner%2Ctags"
    )
  })
})

with_mock_api({
  test_that("content_render() calls the correct endpoint, returns task on success", {
    client <- Connect$new(
      server = "https://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
    render_task <- content_render(x)
    expect_equal(render_task$task[["id"]], "v9XYo7OKkAQJPraI")
    expect_equal(render_task$connect, client)
    # TODO think about how to get variant key into response
  })

  test_that("content_render() can render a non-default variant", {
    client <- Connect$new(
      server = "https://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
    render_task <- content_render(x, variant_key = "SECOND")
    expect_equal(render_task$task[["id"]], "variant2_task_id")
    expect_equal(render_task$connect, client)
    # TODO think about how to get variant key into response
  })

  test_that("content_render() raises an error when called on interactive content", {
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
    expect_error(
      content_render(x),
      "Render not supported for application mode: shiny. Did you mean content_restart()?",
      fixed = TRUE
    )
  })

  test_that("content_restart() calls the correct endpoint", {
    client <- Connect$new(
      server = "https://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
    expect_PATCH(
      content_restart(x),
      url = "https://connect.example/__api__/v1/content/8f37d6e0/environment"
    )
  })

  test_that("content_restart() raises an error when called on interactive content", {
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
    expect_error(
      content_restart(x),
      "Restart not supported for application mode: quarto-static. Did you mean content_render()?",
      fixed = TRUE
    )
  })

  test_that("content$default_variant gets the default variant", {
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    x <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
    expect_warning(
      v <- x$default_variant,
      "experimental"
    )
    expect_identical(v$key, "WrEKKa77")
  })
})

# jobs -----

test_that("get_jobs() using the old and new endpoints returns sensible results", {
  with_mock_api({
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )

    item <- content_item(client, "8f37d6e0")
    jobs_v1 <- get_jobs(item)
  })

  with_mock_dir("2024.07.0", {
    jobs_v0 <- get_jobs(item)
  })

  # Connect 2025.01.0 also provides `content_id` and `content_guid` columns.
  # For older versions, `connectapi` adds them, so objects should all be identical.
  with_mock_dir("2025.01.0", {
    jobs_v1_2025_01_0 <- get_jobs(item)
  })

  # Columns we expect to be identical
  common_cols <- c(
    "id",
    "pid",
    "key",
    "app_id",
    "app_guid",
    "content_id",
    "content_guid",
    "variant_id",
    "bundle_id",
    "start_time",
    "end_time",
    "tag",
    "exit_code",
    "hostname"
  )
  expect_identical(
    jobs_v1[common_cols],
    jobs_v0[common_cols]
  )
  expect_identical(
    jobs_v1[common_cols],
    jobs_v1_2025_01_0[common_cols]
  )

  # Status columns line up as expected
  expect_equal(jobs_v1$status, c(0L, 0L, 2L, 2L, 2L, 2L))
  expect_equal(jobs_v0$status, c(0L, 0L, NA, NA, NA, NA))
  expect_equal(jobs_v1_2025_01_0$status, c(0L, 0L, 2L, 2L, 2L, 2L))
})

with_mock_api({
  client <- Connect$new(
    server = "http://connect.example",
    api_key = "not-a-key"
  )
  test_that("get_job_list() returns expected data", {
    item <- content_item(client, "8f37d6e0")
    job_list <- get_job_list(item)

    expect_equal(
      purrr::map_chr(job_list, "id"),
      c("40793542", "40669829", "40097386", "40096649", "40080413", "39368207")
    )

    expect_equal(
      purrr::map_chr(job_list, "app_guid"),
      rep("8f37d6e0", 6)
    )

    expect_equal(
      purrr::map(job_list, "client"),
      list(client, client, client, client, client, client)
    )
  })

  test_that("terminate_jobs() returns expected data when active jobs exist", {
    item <- content_item(client, "8f37d6e0")
    expect_equal(
      terminate_jobs(item),
      tibble::tibble(
        app_id = c(NA, 52389L),
        app_guid = c(NA, "8f37d6e0"),
        job_key = c("waaTO7v75I84S1hQ", "k3sHkEoWJNwQim7g"),
        job_id = c(NA, "40669829"),
        result = c(NA, "Order to kill job registered"),
        code = c(163L, NA),
        error = c(
          "The specified job cannot be terminated because it is not active",
          NA
        )
      )
    )
  })

  test_that("terminate_jobs() functions as expected with no active jobs", {
    item <- content_item(client, "01234567")
    expect_message(
      expect_equal(
        terminate_jobs(item),
        tibble::tibble(
          app_id = integer(),
          app_guid = character(),
          job_key = character(),
          job_id = character(),
          result = character(),
          code = integer(),
          error = character()
        )
      ),
      "No active jobs found."
    )
  })
})

test_that("an error is raised when terminate_jobs() calls a bad URL", {
  with_mock_api({
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    item <- content_item(client, "8f37d6e0")
  })

  with_mock_dir("2024.07.0", {
    expect_error(
      terminate_jobs(item, "waaTO7v75I84S1hQ")
    )
  })
})

test_that("get_log() gets job logs", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))

    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    item <- content_item(client, "8f37d6e0")
    job_list <- get_job_list(item)
    # This job's log is present at {mock_dir}/v1/content/8f37d6e0/jobs/mxPGVOMVk6f8dso2/log.json.
    job <- purrr::keep(job_list, ~ .x$key == "mxPGVOMVk6f8dso2")[[1]]
    expect_warning(
      log <- get_log(job),
      "the server version is not exposed by this Posit Connect instance"
    )
    expect_identical(
      log,
      tibble::tibble(
        source = c("stderr", "stderr", "stderr"),
        timestamp = structure(
          c(1733512169.9480169, 1733512169.9480703, 1733512169.9480758),
          tzone = Sys.timezone(),
          class = c("POSIXct", "POSIXt")
        ),
        data = c(
          "[rsc-session] Content GUID: 8f37d6e0",
          "[rsc-session] Content ID: 52389",
          "[rsc-session] Bundle ID: 127015"
        )
      )
    )

    expect_GET(
      expect_warning(get_log(job, max_log_lines = 10)),
      "http://connect.example/__api__/v1/content/8f37d6e0/jobs/mxPGVOMVk6f8dso2/log?maxLogLines=10"
    )
  })
})

test_that("get_content_packages() gets packages", {
  with_mock_api({
    withr::local_options(list(rlib_warning_verbosity = "verbose"))
    client <- Connect$new(
      server = "http://connect.example",
      api_key = "not-a-key"
    )
    item <- content_item(client, "8f37d6e0")
    expect_warning(
      expect_identical(
        get_content_packages(item),

        tibble::tibble(
          language = c("r", "r", "r"),
          name = c("askpass", "backports", "base64enc"),
          version = c("1.2.1", "1.5.0", "0.1-3"),
          hash = c(
            "449d6f179d9b3b5df1dc44fad2930b83",
            "6c5e6997f9a6836358ea459b93f2bc39",
            "c590d29e555926af053055e23ee79efb"
          )
        )
      ),
      "the server version is not exposed by this Posit Connect instance"
    )
  })
})

# content search ----

with_mock_dir("2025.09.0", {
  client <- Connect$new(
    server = "https://connect.example",
    api_key = "not-a-key"
  )

  test_that("content search returns the expected list of content", {
    res <- search_content(client, q = "sea bream")
    expect_equal(
      class(res[[1]]),
      c("Content", "R6")
    )
    expect_equal(
      purrr::map_chr(res, list("content", "guid")),
      c("c9f68287", "53032a0e")
    )
    expect_equal(
      purrr::map_chr(res, list("content", "title")),
      c("sea bream report", "sea bream dashboard")
    )
  })

  test_that("content search fetches multiple pages correctly", {
    res <- search_content(client, q = "blobfish")
    expect_equal(length(res), 4)
    expect_equal(
      purrr::map_chr(res, list("content", "title")),
      c(
        "blobfish dashboard",
        "blobfish api",
        "blobfish report",
        "blobfish report (v2)"
      )
    )
  })

  test_that("search_content() can be converted to a data frame correctly", {
    content_df <- search_content(client, q = "sea bream") |>
      as_tibble()
    expect_named(
      content_df,
      c(
        "guid",
        "name",
        "title",
        "description",
        "access_type",
        "connection_timeout",
        "read_timeout",
        "init_timeout",
        "idle_timeout",
        "max_processes",
        "min_processes",
        "max_conns_per_process",
        "load_factor",
        "created_time",
        "last_deployed_time",
        "bundle_id",
        "app_mode",
        "content_category",
        "parameterized",
        "cluster_name",
        "image_name",
        "r_version",
        "py_version",
        "quarto_version",
        "run_as",
        "run_as_current_user",
        "owner_guid",
        "content_url",
        "dashboard_url",
        "app_role",
        "vanity_url",
        "id",
        "owner",
        "tags"
      )
    )
    expect_equal(
      content_df$title,
      c("sea bream report", "sea bream dashboard")
    )
    expect_equal(
      content_df$guid,
      c("c9f68287", "53032a0e")
    )
  })
})

test_that("content search uses default page_size of 500 and page_number of 1", {
  without_internet(
    expect_GET(
      search_content(client, q = "bream"),
      "https://connect.example/__api__/v1/search/content?q=bream&page_number=1&page_size=500&include=owner%2Cvanity_url" #nolint
    )
  )
})

test_that("content search passes arbitrary parameters through ... to query string", {
  without_internet(
    expect_GET(
      search_content(
        client,
        q = "bream",
        future_param = "value"
      ),
      "https://connect.example/__api__/v1/search/content?q=bream&page_number=1&page_size=500&include=owner%2Cvanity_url&future_param=value" #nolint
    )
  )
})

test_that("the inner .search_content() func calls the endpoint correctly", {
  without_internet(
    expect_GET(
      .search_content(
        client,
        q = "bream",
        page_number = 2,
        page_size = 20,
        include = "owner"
      ),
      "https://connect.example/__api__/v1/search/content?q=bream&page_number=2&page_size=20&include=owner"
    )
  )
})

test_that("content search errors on Connect < 2024.04.0", {
  client <- MockConnect$new("2024.04.0")
  client$mock_response(
    "GET",
    "v1/search/content",
    content = list()
  )
  expect_no_error(search_content(client))

  client <- MockConnect$new("2024.03.0")
  expect_error(
    search_content(client),
    "ERROR: This feature requires Posit Connect version 2024.04.0 but you are using 2024.03.0."
  )
})

# lock content ----

with_mock_dir("2025.09.0", {
  test_that("lock_content() and unlock_content() make the expected requests", {
    client <- Connect$new(
      server = "https://connect.example",
      api_key = "not-a-key"
    )
    client$version # Hydrate version
    item <- content_item(client, "6632a162")
    without_internet({
      expect_PATCH(
        lock_content(item),
        "https://connect.example/__api__/v1/content/6632a162",
        '{"locked":true,"locked_message":""}'
      )
      expect_PATCH(
        lock_content(item, "ACCESS DENIED"),
        "https://connect.example/__api__/v1/content/6632a162",
        '{"locked":true,"locked_message":"ACCESS DENIED"}'
      )
      expect_PATCH(
        unlock_content(item),
        "https://connect.example/__api__/v1/content/6632a162",
        '{"locked":false,"locked_message":""}'
      )
    })
  })
})

test_that("lock_content() and unlock_content() error on Connect < 2024.08.0", {
  client <- MockConnect$new("2024.07.0")
  client$mock_response(
    "GET",
    "v1/content/test-guid",
    content = list(
      guid = "test-guid",
      title = "Test Content",
      app_mode = "static"
    )
  )
  item <- content_item(client, "test-guid")
  expect_error(
    lock_content(item),
    "ERROR: This feature requires Posit Connect version 2024.08.0 but you are using 2024.07.0."
  )
  expect_error(
    unlock_content(item),
    "ERROR: This feature requires Posit Connect version 2024.08.0 but you are using 2024.07.0."
  )
})

test_that("content_restart retries if delete fails", {
  mock_connect <- MockConnect$new("2025.09.0")
  mock_connect$mock_response(
    "GET",
    "v1/content/test-guid",
    content = list(
      guid = "test-guid",
      name = "test-content",
      title = "Test Content",
      app_mode = "shiny",
      content_url = "https://connect.example/content/test-guid/",
      dashboard_url = "https://connect.example/connect/#/apps/test-guid"
    )
  )

  # First PATCH to set the env var (succeeds)
  mock_connect$mock_response(
    "PATCH",
    "v1/content/test-guid/environment",
    content = list(),
    status_code = 200L
  )

  # Second PATCH to delete the env var (fails with 500)
  mock_connect$mock_response(
    "PATCH",
    "v1/content/test-guid/environment",
    content = list(error = "Internal server error"),
    status_code = 500L
  )

  # Third PATCH to delete the env var (retry succeeds)
  mock_connect$mock_response(
    "PATCH",
    "v1/content/test-guid/environment",
    content = list(),
    status_code = 200L
  )

  # Get the content item
  item <- content_item(mock_connect, "test-guid")

  expect_no_error(content_restart(item))
  expect_equal(
    mock_connect$call_log,
    c(
      "GET https://connect.example/__api__/v1/content/test-guid",
      "PATCH https://connect.example/__api__/v1/content/test-guid/environment",
      "PATCH https://connect.example/__api__/v1/content/test-guid/environment",
      "PATCH https://connect.example/__api__/v1/content/test-guid/environment"
    )
  )
})

test_that("content_restart doesn't retry if error isn't a 500", {
  mock_connect <- MockConnect$new("2025.09.0")
  mock_connect$mock_response(
    "GET",
    "v1/content/test-guid",
    content = list(
      guid = "test-guid",
      name = "test-content",
      title = "Test Content",
      app_mode = "shiny",
      content_url = "https://connect.example/content/test-guid/",
      dashboard_url = "https://connect.example/connect/#/apps/test-guid"
    )
  )

  mock_connect$mock_response(
    "PATCH",
    "v1/content/test-guid/environment",
    content = list(),
    status_code = 200L
  )

  mock_connect$mock_response(
    "PATCH",
    "v1/content/test-guid/environment",
    content = list(error = "Bad request"),
    status_code = 400L
  )

  item <- content_item(mock_connect, "test-guid")
  expect_warning(content_restart(item))
  # only 2 PATCHes
  expect_equal(
    mock_connect$call_log,
    c(
      "GET https://connect.example/__api__/v1/content/test-guid",
      "PATCH https://connect.example/__api__/v1/content/test-guid/environment",
      "PATCH https://connect.example/__api__/v1/content/test-guid/environment"
    )
  )
})

# TODO: This testing requires an LDAP server
# TODO: use the same underlying implementation somehow...

test_that("users_create_remote errors if prefix returns zero matches", {
  skip("not implemented")
})
test_that("users_create_remote errors if prefix returns more than one matches", {
  skip("not implemented")
})
test_that("users_create_remote errors if expect is greater than one", {
  skip("not implemented")
})
test_that("users_create_remote works if match already exists", {
  # should message
  skip("not implemented")
})
test_that("users_create_remote creates match if one is found", {
  skip("not implemented")
})
test_that("users_create_remote creates multiples if multiple are found", {
  skip("not implemented")
})
test_that("users_create_remote avoids existing matches if multiple are found", {
  # fails today
  skip("not implemented")
})

test_that("groups_create_remote errors if prefix returns zero matches", {
  skip("not implemented")
})
test_that("groups_create_remote errors if prefix returns more than one matches", {
  skip("not implemented")
})
test_that("groups_create_remote errors if expect is greater than one", {
  skip("not implemented")
})
test_that("groups_create_remote works if match already exists", {
  # should message
  skip("not implemented")
})
test_that("groups_create_remote creates match if one is found", {
  skip("not implemented")
})
test_that("groups_create_remote creates multiples if multiple are found", {
  skip("not implemented")
})
test_that("groups_create_remote avoids existing matches if multiple are found", {
  # fails today
  skip("not implemented")
})

# this (nonstandard) test checks output against a file
# thus allowing an easy preview of expected output
# AND some level of standard / consistency / expectation

# nolint start: commented_code_linter
#
# regenerate test printing file with the following
# writeLines(
#   connectapi:::generate_R6_print_output(),
#   rprojroot::find_testthat_root_file("test-print-output.txt")
# )
#
# nolint end

test_that("output matches previous expectation", {
  withr::with_envvar(
    list(TESTTHAT = ""),
    expect_snapshot(generate_R6_print_output())
  )
})

---
title: "Customize HTTP Requests"
author: "Cole Arendt"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Customize HTTP Requests}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Sometimes when using `connectapi`, customizing HTTP requests is desirable. For
instance, some common use cases are:

- using a custom certificate bundle
- turn off certificate trust verification (**ONLY FOR TESTING**)
- setting custom headers
- using Kerberos (SPNEGO)

This is possible with `connectapi` thanks to the underlying library in use,
[`httr`](https://httr.r-lib.org/).

## Getting Started

When you initialize a `connectapi` API client, you implicitly create a `httr`
HTTP client. The `httr` package allows you to configure your HTTP requests
globally using `httr::set_config()` or in a scoped variant `httr::with_config`.
We will walk through a few examples below.

### Custom CA Bundle

```r
library(httr)
library(connectapi)

client <- connect()

# notice that TLS verification fails
get_users(client)

# use a custom Certificate Authority to verify SSL/TLS requests
httr::set_config(httr::config(cainfo = "/path/to/my.pem"))

# now it should succeed!
get_users(client)
```

### Turn Off Certificate Trust Verification - ONLY FOR TESTING

Sometimes when first setting up a server, it is common to use self-signed
certificates. This is generally bad for reliable communication and security (as
there is no reason for any computer to trust this server as a "self-declared"
trustworthy actor).

However, it can be useful while the organization's Certificate Authority (CA) is
in the process of issuing a valid certificate, or while a certificate is
procured from a public CA.

```r
# disabling certificate trust (can allow man-in-the-middle attacks, etc.)
httr::set_config(httr::config(ssl_verifypeer = 0, ssl_verifyhost = 0))

# should work
client <- connect()
get_users(client)
```

You can also do this in a more scoped fashion:

```r
httr::with_config(
  httr::config(ssl_verifypeer = 0, ssl_verifyhost = 0),
  {
    client <- connect()
    get_users(client)
  }
)
```

### Custom Headers, Cookies, Proxy, etc.

`httr` has some helpers for common tasks like `httr::add_headers()`,
`httr::set_cookies()`, `httr::use_proxy()`, etc. Using them is a bit tricky,
but can be done by way of the `client$httr_config()` function.

Pass any usual `httr` arguments to `client$httr_config()`, and those arguments
will then be saved and passed to any subsequent `GET`, `PUT`, `POST`, `PATCH`,
`HEAD`, `DELETE` requests you send with that client.

```r
# for instance, to set custom headers (i.e. to get through a proxy)
client$httr_config(httr::add_headers(MY_MAGIC_HEADER = "value"))

# or to clear sticky cookies if you want to switch nodes in an HA cluster
client <- connect()
client$server_settings()$hostname
client$httr_config(handle = httr::handle(""))

# now you have a chance to get a new host
client$server_settings()$hostname

# use an outbound proxy
client$httr_config(httr::use_proxy("http://myproxy.example.com"))
```

> NOTE: these values are completely overwritten each time you call
>`client$httr_config()`, so ensure that you pass all desired values at the same
> time

### Using Kerberos

Suffice it to say that effectively using Kerberos for HTTP is a bit of an
advanced topic. However, it _is_ possible with `httr`.

It is worth noting that today, this interferes with API key authentication, which we
are hoping to improve in a future release of Posit Connect.

```r
# disables authentication header that is included by default
client$using_auth <- FALSE

# use Kerberos authentication mechanism (requires local credential cache)
client$httr_config(httr::authenticate(":", "", type = "gssnegotiate"))
```

---
title: "Getting Started with connectapi"
author: "Cole Arendt"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with connectapi}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Connecting

The first thing that you need to accomplish when using the `connectapi` is
getting connected to your Posit Connect server. This requires the URL to your
server and an API key. Directions on how to get an API key are
[here](https://docs.posit.co/connect/user/api-keys/#api-keys).

Once you have your API key, you can create a Connect object in your code, like so:

```r
library(connectapi)
client <- connect(
  server = "http://example.com:3939",
  api_key = "aihgaahegiahgg"
)
```

The alternative is to define a `.Renviron` file, which specifies environment variables:

```
CONNECT_SERVER=http://example.com:3939
CONNECT_API_KEY=aihgaahegiahgg
```

Then, those can be used in your code:

```r
library(connectapi)

# this happens by default if you restart your R session
readRenviron(".Renviron")

client <- connect()
```

## Content Management

The `connectapi` package has tools to help programmatically deploy and manage
content. To get started, you need to use the `rsconnect` package, specifically
the `rsconnect::writeManifest()` function, to build a `manifest.json` file for a
directory of content that you want to deploy.

Once the `manifest.json` file is present, you can reference the directory and deploy
it directly:

```r
bnd <- bundle_dir("./my/directory")

# name must be unique on the server
# if you do not specify it, a random value will be provided
content_1 <- client %>%
  deploy(bnd, name = "my-content", title = "Amazing Report!!")
```

The content object (`content_1`) includes information about the deployment that you
just requested. This can be explored with:

```r
content_1 %>% poll_task()
```

Alternatively, you can immediately begin altering the settings of the content object
while you wait for deployment to complete:

```r
content_1 %>%
  set_image_url("https://gph.is/29vyb0s") %>%
  set_vanity_url("/my_clever_content")

# ensure the vanity URL is set as expected
content_1 %>%
  get_vanity_url()
```

If you have content that already exists and you want to update it, you can do so using the content GUID
(which you can find within Posit Connect in the Info pane).

```r
content_2 <- client %>%
  deploy(bnd, guid = "d78ba9f8-bb57-422e-b164-9ecd8e4c4fd6") %>%
  poll_task()
```

You can also use the Posit Connect content GUID to retrieve information about
existing content (if you want to examine or change settings)

```r
content_3 <- client %>%
  content_item(guid = "96532fbc-725e-441b-9cc6-a5622535241b")
```

---
title: "Content Permissions"
author: "Cole Arendt"
date: "8/9/2021"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(connectapi)
client <- connect()
options("connectapi_disable_experimental_warnings" = TRUE)
```

# Getting Started

To filter content by permissions, you first need a "baseline set of content." This
could be all content on the server, all content in a particular tag, etc.

> NOTE: performance will depend heavily on the size of this baseline set of
content, because the permissions API today requires enumeration. To improve
performance for large sets of content, you can use `pins` or caching on disk to
reduce how often the requests must be re-executed.

We will start by deploying a few pieces of test content, two test users, set
access controls, and tags:

```{r deploy_baseline}
bnd <- bundle_static(system.file("logo.png", package = "connectapi"))

content_1 <- deploy(client, bnd, title = "App 1")
content_2 <- deploy(client, bnd, title = "App 2")

user_restricted <- client$users_create("example_restricted", "restricted@example.com", password = create_random_name())
user_all <- client$users_create("example_all", "all@example.com", password = create_random_name())

invisible(create_tag_tree(client, "Example", "Permissions"))
tags <- get_tags(client)
tag_1 <- tags$Example$Permissions

set_content_tags(content_1, tag_1)
set_content_tags(content_2, tag_1)

content_add_user(content_1, user_restricted$guid, role = "viewer")
content_add_user(content_1, user_all$guid, "viewer")
content_add_user(content_2, user_all$guid, "viewer")
```


# Retrieve the Content List

The `content_list_with_permissions()` function is the core of what we want. However, it defaults to return _all content on the server._ For some servers, this is very expensive (and can take 30 minutes or more).

Instead, we recommend using the `.p` argument to define a "predicate" function
(in the style of `purrr::keep()`) that determines which records to keep. Since all this predicate has access to is the "content list" itself, we will retrieve a list of Content GUIDs first.

```{r list}
my_tag_content <- content_list_by_tag(client, tag_1)
content_guids <- my_tag_content$guid

c_with_p <- content_list_with_permissions(client, .p = ~ .x$guid %in% content_guids)

# another approach, with a function
c_with_p <- content_list_with_permissions(client, .p = function(.x) {
  .x$guid %in% content_guids
})

# notice the "permission" column:
c_with_p$permission
```

# Filter the Content List

We added a helper function to the package that should filter the content list for you: `content_list_guid_has_access()`

In a Shiny application or other personalized context (i.e. using `session$user`), you then
filter the content list to only what a user should see (using the `permissions` column returned above)
```{r filter}
# restricted has access
content_list_guid_has_access(c_with_p, user_restricted$guid) %>% .$title

# "all" has access
content_list_guid_has_access(c_with_p, user_all$guid) %>% .$title
```

# Display the Content List

We plan to build a full example in Shiny (and to show example code below).
However, suffice it to say that for RStudio Connect version 1.9.0 or newer,
[`connectwidgets`](https://pkgs.rstudio.com/connectwidgets/) is a great way to
plan to display your data, and provides several helpers for doing so!

---
title: "Getting Started with connectapi"
author: "Cole Arendt"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with connectapi}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Connecting

The first thing that you need to accomplish when using the `connectapi` is
getting connected to your Posit Connect server. This requires the URL to your
server and an API key. Directions on how to get an API key are
[here](https://docs.posit.co/connect/user/api-keys/#api-keys).

Once you have your API key, you can create a Connect object in your code, like so:

```r
library(connectapi)
client <- connect(
  server = "http://example.com:3939",
  api_key = "aihgaahegiahgg"
)
```

The alternative is to define a `.Renviron` file, which specifies environment variables:

```
CONNECT_SERVER=http://example.com:3939
CONNECT_API_KEY=aihgaahegiahgg
```

Then, those can be used in your code:

```r
library(connectapi)

# this happens by default if you restart your R session
readRenviron(".Renviron")

client <- connect()
```

## Content Management

The `connectapi` package has tools to help programmatically deploy and manage
content. To get started, you need to use the `rsconnect` package, specifically
the `rsconnect::writeManifest()` function, to build a `manifest.json` file for a
directory of content that you want to deploy.

Once the `manifest.json` file is present, you can reference the directory and deploy
it directly:

```r
bnd <- bundle_dir("./my/directory")

# name must be unique on the server
# if you do not specify it, a random value will be provided
content_1 <- client %>%
  deploy(bnd, name = "my-content", title = "Amazing Report!!")
```

The content object (`content_1`) includes information about the deployment that you
just requested. This can be explored with:

```r
content_1 %>% poll_task()
```

Alternatively, you can immediately begin altering the settings of the content object
while you wait for deployment to complete:

```r
content_1 %>%
  set_thumbnail("https://gph.is/29vyb0s") %>%
  set_vanity_url("/my_clever_content")

# ensure the vanity URL is set as expected
content_1 %>%
  get_vanity_url()
```

If you have content that already exists and you want to update it, you can do so using the content GUID
(which you can find within Posit Connect in the Info pane).

```r
content_2 <- client %>%
  deploy(bnd, guid = "d78ba9f8-bb57-422e-b164-9ecd8e4c4fd6") %>%
  poll_task()
```

You can also use the Posit Connect content GUID to retrieve information about
existing content (if you want to examine or change settings)

```r
content_3 <- client %>%
  content_item(guid = "96532fbc-725e-441b-9cc6-a5622535241b")
```

---
title: "connectapi Tags"
author: "Cole Arendt"
date: "7/12/2020"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(connectapi)
client <- connect()
options("connectapi_disable_experimental_warnings" = TRUE)
```

# Getting Started

To understand how tags work with `connectapi`, you must first understand tags in
RStudio Connect.

- Tags in RStudio Connect consist of multiple "tag trees," each with a "Category"
as its parent.
- The tag hierarchy / structure is created by administrators
- Publishers and administrators can associate any non-Category tag with content
- A tag being a member of a "child" tag doe _not_ automatically make it a member
of the "parent" tag

As always, get started by defining the `CONNECT_SERVER` and `CONNECT_API_KEY`
variables, then initialize an API client.

```{r eval=FALSE}
library(connectapi)
client <- connect()
```

**NOTE:** This example report will create a few tag hierarchies and then use
them. As a result, to use this example verbatim requires admin privileges and
_will_ create tags on your server.

# Create the Tag Tree(s)

To get started, we will create a tag tree to show how things work. The
`create_tag()` helper allows you to create a singular tag (by specifying its
parent, etc.). For our purposes, `create_tag_tree()` is easier, since it
creates the entire tree specified at once.

```{r get_tags}
start_tags <- get_tags(client)
start_tags
```

```{r check_tag_exist, include=FALSE}
# Protect against tags already existing
if ("DemoProject" %in% names(start_tags) || "DemoAudience" %in% names(start_tags)) {
  stop("ERROR: One of the demo tags already exist for you! Beware lest they be deleted by this demo")
}
```

```{r create_demo_tags}
tree_project_1 <- create_tag_tree(client, "DemoProject", "project_1")
tree_project_1

tmp_tags <- get_tags(client)

tree_project_2 <- create_tag(client, name = "project_2", parent = tmp_tags$DemoProject)
tree_project_2

tree_audience_1 <- create_tag_tree(client, "DemoAudience", "Sales")
tree_audience_2 <- create_tag_tree(client, "DemoAudience", "Finance")

get_tags(client)
```

# Content tags

In order to show how tags work, we need some content to work with.

```{r content_deploy}
bnd <- bundle_static(system.file("logo.png", package = "connectapi"))

content_1 <- deploy(client, bnd)
content_2 <- deploy(client, bnd)
```

## Set Tags

Content 1 is for `project_1` and `Sales`, so let's set the tags!
There are a few ways to do so.

```{r content_1_tags}
all_tags <- get_tags(client)

set_content_tag_tree(content_1, "DemoProject", "project_1")

set_content_tags(content_1, all_tags$DemoAudience$Sales)
```
Content 2 is for `project_2` and both Audiences (`Sales` and `Finance`).

```{r content_2_tags}
set_content_tags(
  content_2,
  all_tags$DemoProject$project_2,
  all_tags$DemoAudience$Sales,
  all_tags$DemoAudience$Finance
)
```

## See the tags associated with content

In order to see the tags associated with content, use `get_content_tags()`.
The data structure is the same as `get_tags()`, but the "whole list" is filtered
to only the tags that are associated with a piece of content.

```{r}
c1_tags <- get_content_tags(content_1)
c1_tags

c2_tags <- get_content_tags(content_2)
c2_tags
```

# List all content associated with a tag

Once tags have been defined, you can also search for all of the content
associated with a tag.

```{r}
content_list_by_tag(client, all_tags$DemoAudience$Sales)

content_list_by_tag(client, all_tags$DemoProject$project_1)

content_list_by_tag(client, all_tags$DemoProject)
```

# Cleanup

Now we will clean up the demo tags that we created.

```{r cleanup}
# Protect against tags already existing
if ("DemoProject" %in% names(start_tags) || "DemoAudience" %in% names(start_tags)) {
  stop("ERROR: One of the demo tags already exist for you! Beware lest they be deleted by this demo")
}

latest_tags <- get_tags(client)
delete_tag(client, latest_tags$DemoProject)
delete_tag(client, latest_tags$DemoAudience)

# TODO: delete content
```

---
title: "Lifecycle Statuses"
author: "Cole Arendt"
date: "5/21/2021"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(dplyr)
```

```{r}
lifecycle::pkg_lifecycle_statuses("connectapi") %>%
  distinct(package, fun, lifecycle) %>%
  rmarkdown::paged_table()
```

# we can retrieve a content item

    Code
      print(item)
    Output
      Posit Connect Content: 
        Content GUID: f2f37341-e21d-3d80-c698-a935ad614066
        Content URL: https://connect.example/connect/#/apps/f2f37341-e21d-3d80-c698-a935ad614066
        Content Title: Performance Data
      
      content_item(client, guid = "f2f37341-e21d-3d80-c698-a935ad614066")
      

# get_group_content() works

    Code
      get_group_content(client, group_guids)
    Output
      # A tibble: 4 x 7
        group_guid group_name  content_guid content_name     content_title access_type
        <chr>      <chr>       <chr>        <chr>            <chr>         <chr>      
      1 a6fb5cea   connect_dev 8b57f54e     app-1197-9825-t~ app-1197-982~ acl        
      2 a6fb5cea   connect_dev 8bf70c85     quarto-email-de~ quarto-email~ acl        
      3 a6fb5cea   connect_dev fcad1958     top-queries      top-queries   logged_in  
      4 ae5c3b2c   group12     46fb83eb     forecast-email-~ forecast-ema~ logged_in  
      # i 1 more variable: role <chr>

# get_group_content() returns an empty data frame when no content exists

    Code
      get_group_content(client, group_guid)
    Output
      # A tibble: 1 x 7
        group_guid group_name content_guid content_name content_title access_type
        <chr>      <chr>      <chr>        <chr>        <chr>         <chr>      
      1 <NA>       <NA>       <NA>         <NA>         <NA>          <NA>       
      # i 1 more variable: role <chr>

# warn_untested_connect warning snapshot

    Code
      capture_warning(warn_untested_connect("2022.02", "2022.01"))
    Output
      NULL

---

    Code
      capture_warning(warn_untested_connect("2022.01", "2022.02"))
    Output
      <warning/rlang_warning>
      Warning:
      You are using an older version of Posit Connect (2022.01) than is tested (2022.02). Some APIs may not function as expected.
      This warning is displayed once per session.

---
title: "test-doc"
author: "Cole Arendt"
date: "5/31/2019"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## R Markdown

This is an R Markdown document. Markdown is a simple formatting syntax for authoring HTML, PDF, and MS Word documents. For more details on using R Markdown see <http://rmarkdown.rstudio.com>.

When you click the **Knit** button a document will be generated that includes both content as well as the output of any embedded R code chunks within the document. You can embed an R code chunk like this:

```{r cars}
summary(cars)
```

## Including Plots

You can also embed plots, for example:

```{r pressure, echo=FALSE}
plot(pressure)
```

Note that the `echo = FALSE` parameter was added to the code chunk to prevent printing of the R code that generated the plot.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{bundle_dir}
\alias{bundle_dir}
\title{Define a bundle from a Directory}
\usage{
bundle_dir(
  path = ".",
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz")
)
}
\arguments{
\item{path}{The path to the directory to be bundled}

\item{filename}{The output bundle path}
}
\value{
Bundle A bundle object
}
\description{
Creates a bundle from a target directory.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

bundle_dir(system.file("tests/testthat/examples/shiny/", package = "connectapi"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other deployment functions: 
\code{\link{bundle_path}()},
\code{\link{bundle_static}()},
\code{\link{deploy}()},
\code{\link{download_bundle}()},
\code{\link{poll_task}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{Environment}
\alias{Environment}
\title{Environment}
\description{
Environment

Environment
}
\details{
An R6 class that represents a Content's Environment Variables
}
\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{PositConnect}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Super class}{
\code{\link[connectapi:Content]{connectapi::Content}} -> \code{Environment}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{env_raw}}{The (raw) set of environment variables.}

\item{\code{env_vars}}{The set of environment variables.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Environment-new}{\code{Environment$new()}}
\item \href{#method-Environment-environment}{\code{Environment$environment()}}
\item \href{#method-Environment-environment_set}{\code{Environment$environment_set()}}
\item \href{#method-Environment-environment_all}{\code{Environment$environment_all()}}
\item \href{#method-Environment-env_refresh}{\code{Environment$env_refresh()}}
\item \href{#method-Environment-print}{\code{Environment$print()}}
\item \href{#method-Environment-clone}{\code{Environment$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_delete"><a href='../../connectapi/html/Content.html#method-Content-bundle_delete'><code>connectapi::Content$bundle_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_download"><a href='../../connectapi/html/Content.html#method-Content-bundle_download'><code>connectapi::Content$bundle_download()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="danger_delete"><a href='../../connectapi/html/Content.html#method-Content-danger_delete'><code>connectapi::Content$danger_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="deploy"><a href='../../connectapi/html/Content.html#method-Content-deploy'><code>connectapi::Content$deploy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_bundles"><a href='../../connectapi/html/Content.html#method-Content-get_bundles'><code>connectapi::Content$get_bundles()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_content_remote"><a href='../../connectapi/html/Content.html#method-Content-get_content_remote'><code>connectapi::Content$get_content_remote()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_dashboard_url"><a href='../../connectapi/html/Content.html#method-Content-get_dashboard_url'><code>connectapi::Content$get_dashboard_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_url"><a href='../../connectapi/html/Content.html#method-Content-get_url'><code>connectapi::Content$get_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="jobs"><a href='../../connectapi/html/Content.html#method-Content-jobs'><code>connectapi::Content$jobs()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="packages"><a href='../../connectapi/html/Content.html#method-Content-packages'><code>connectapi::Content$packages()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions"><a href='../../connectapi/html/Content.html#method-Content-permissions'><code>connectapi::Content$permissions()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_add"><a href='../../connectapi/html/Content.html#method-Content-permissions_add'><code>connectapi::Content$permissions_add()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_delete"><a href='../../connectapi/html/Content.html#method-Content-permissions_delete'><code>connectapi::Content$permissions_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_update"><a href='../../connectapi/html/Content.html#method-Content-permissions_update'><code>connectapi::Content$permissions_update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="register_job_kill_order"><a href='../../connectapi/html/Content.html#method-Content-register_job_kill_order'><code>connectapi::Content$register_job_kill_order()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_enable"><a href='../../connectapi/html/Content.html#method-Content-repo_enable'><code>connectapi::Content$repo_enable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_set"><a href='../../connectapi/html/Content.html#method-Content-repo_set'><code>connectapi::Content$repo_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repository"><a href='../../connectapi/html/Content.html#method-Content-repository'><code>connectapi::Content$repository()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_delete"><a href='../../connectapi/html/Content.html#method-Content-tag_delete'><code>connectapi::Content$tag_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_set"><a href='../../connectapi/html/Content.html#method-Content-tag_set'><code>connectapi::Content$tag_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tags"><a href='../../connectapi/html/Content.html#method-Content-tags'><code>connectapi::Content$tags()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="update"><a href='../../connectapi/html/Content.html#method-Content-update'><code>connectapi::Content$update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="variants"><a href='../../connectapi/html/Content.html#method-Content-variants'><code>connectapi::Content$variants()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-new"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-new}{}}}
\subsection{Method \code{new()}}{
Initialize this set of environment variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$new(connect, content)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{connect}}{The \code{Connect} instance.}

\item{\code{content}}{The \code{Content} instance.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-environment"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-environment}{}}}
\subsection{Method \code{environment()}}{
Fetch the set of environment variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$environment()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-environment_set"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-environment_set}{}}}
\subsection{Method \code{environment_set()}}{
Update the set of environment variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$environment_set(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Environment variable names and values.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-environment_all"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-environment_all}{}}}
\subsection{Method \code{environment_all()}}{
Overwrite the set of environment variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$environment_all(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Environment variable names and values.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-env_refresh"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-env_refresh}{}}}
\subsection{Method \code{env_refresh()}}{
Fetch the set o environment variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$env_refresh()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-print"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Environment-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Environment-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Environment$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{deploy}
\alias{deploy}
\alias{deploy_current}
\title{Deploy a bundle}
\usage{
deploy(
  connect,
  bundle,
  name = create_random_name(),
  title = name,
  guid = NULL,
  ...,
  .pre_deploy = {
 }
)

deploy_current(content)
}
\arguments{
\item{connect}{A Connect object}

\item{bundle}{A Bundle object}

\item{name}{The unique name for the content on the server}

\item{title}{optional The title to be used for the content on the server}

\item{guid}{optional The GUID if the content already exists on the server}

\item{...}{Additional arguments passed along to the content creation}

\item{.pre_deploy}{An expression to execute before deploying the new bundle.
The variables \code{content} and \code{bundle_id} are supplied}

\item{content}{A Content object}
}
\value{
Task A task object
}
\description{
Deploys a bundle (tarball) to an Posit Connect server. If not provided,
\code{name} (a unique identifier) will be an auto-generated alphabetic string. If
deploying to an existing endpoint, you can set \code{name} or \code{guid} to the
desired content.
}
\details{
This function accepts the same arguments as \code{connectapi::content_update()}.

\code{deploy_current()} is a helper to easily redeploy the currently active bundle
for an existing content item.
}
\examples{
\dontrun{
client <- connect()

# beware bundling big directories, like `renv/`, `data/`, etc.
bnd <- bundle_dir(".")

deploy(client, bnd)
}
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

client <- connect()
bnd <- bundle_path(system.file("tests/testthat/examples/static.tar.gz", package = "connectapi"))
deploy(client, bnd)
\dontshow{\}) # examplesIf}
}
\seealso{
connectapi::content_update

Other deployment functions: 
\code{\link{bundle_dir}()},
\code{\link{bundle_path}()},
\code{\link{bundle_static}()},
\code{\link{download_bundle}()},
\code{\link{poll_task}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audits.R
\name{audit_r_versions}
\alias{audit_r_versions}
\title{Audit R Versions}
\usage{
audit_r_versions(content)
}
\arguments{
\item{content}{\code{data.frame} of content information, as from \code{\link[=get_content]{get_content()}}}
}
\value{
A plot that shows the R version used by content over time and in
aggregate.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\seealso{
Other audit functions: 
\code{\link{audit_access_open}()},
\code{\link{audit_runas}()},
\code{\link{vanity_is_available}()}
}
\concept{audit functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user.R
\name{user_guid_from_username}
\alias{user_guid_from_username}
\title{User}
\usage{
user_guid_from_username(client, username)
}
\arguments{
\item{client}{A Connect R6 object}

\item{username}{The user's username}
}
\description{
Get user details
}
\details{
\code{user_guid_from_username()} is a helper to retrieve a user GUID, given the
user's username. It is useful in Shiny applications for using \code{session$user}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git.R
\name{deploy_repo}
\alias{deploy_repo}
\alias{deploy_repo_enable}
\alias{deploy_repo_update}
\title{Deploy a Git Repository}
\usage{
deploy_repo(
  client,
  repository,
  branch,
  subdirectory,
  name = create_random_name(),
  title = name,
  ...
)

deploy_repo_enable(content, enabled = TRUE)

deploy_repo_update(content)
}
\arguments{
\item{client}{A Connect R6 object}

\item{repository}{The git repository to deploy}

\item{branch}{The git branch to deploy}

\item{subdirectory}{The subdirectory to deploy (must contain a \code{manifest.json})}

\item{name}{The "name" / unique identifier for the content. Defaults to a random character string}

\item{title}{The "title" of the content}

\item{...}{Additional options for defining / specifying content attributes}

\item{content}{An R6 Content object (i.e. the result of \code{content_item()})}

\item{enabled}{Whether Connect will enable automatic polling for repository updates}
}
\value{
A ContentTask object, for use with \code{poll_task()} (if you want to follow the logs)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Deploy a git repository directly to Posit Connect,
using Posit Connect's "pull-based" "git-polling" method of deployment.
}
\details{
\itemize{
\item \code{deploy_repo_enable()} enables (or disables) Posit Connect's git polling for a piece of content
\item \code{deploy_repo_update()} triggers an update of the content from its git
repository, if any are present
}
}
\seealso{
connectapi::poll_task, connectapi::repo_check_branches, connectapi::repo_check_manifest_dirs

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/browse.R
\name{browse_solo}
\alias{browse_solo}
\alias{browse_dashboard}
\alias{browse_api_docs}
\alias{browse_connect}
\title{Browse}
\usage{
browse_solo(content)

browse_dashboard(content)

browse_api_docs(connect)

browse_connect(connect)
}
\arguments{
\item{content}{A R6 Content object}

\item{connect}{A R6 Connect object}
}
\value{
The url that is opened in the browser
}
\description{
Browse to different locations on Connect via
utils::browseURL
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{lock_content}
\alias{lock_content}
\alias{unlock_content}
\title{Lock or Unlock Content}
\usage{
lock_content(content, locked_message = "")

unlock_content(content)
}
\arguments{
\item{content}{An R6 content item}

\item{locked_message}{Optional. A custom message that is displayed by the
content item when locked. It is possible to format this message using Markdown.}
}
\value{
An R6 content item
}
\description{
Lock or unlock a content item. When content is locked, all processes are
terminated, rendering is disabled, and new bundles cannot be deployed.
}
\details{
\code{lock_content()} locks a content item with an optional message displayed to
visitors (supports Markdown).

\code{unlock_content()} unlocks a content item, reverting the effects of locking.
}
\examples{
\dontrun{
# Lock content with a message
client <- connect()
content <- content_item(client, "content-guid")
content <- lock_content(content, locked_message = "Ah ah ah! You didn't say the magic word!")

# Lock content without a message
content <- lock_content(content)

# Unlock content
content <- unlock_content(content)
}
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_content}
\alias{get_content}
\title{Get information about content on the Posit Connect server}
\usage{
get_content(src, guid = NULL, owner_guid = NULL, name = NULL, ..., .p = NULL)
}
\arguments{
\item{src}{A Connect object}

\item{guid}{The guid for a particular content item}

\item{owner_guid}{The unique identifier of the user who owns the content}

\item{name}{The content name specified when the content was created}

\item{...}{Extra arguments. Currently not used.}

\item{.p}{Optional. A predicate function, passed as-is to \code{purrr::keep()}
before turning the response into a tibble. Can be useful for performance.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{guid}: The unique identifier of this content item.
\item \code{name}: A simple, URL-friendly identifier. Allows alpha-numeric
characters, hyphens ("-"), and underscores ("_").
\item \code{title}: The title of this content.
\item \code{description}: A rich description of this content
\item \code{access_type}: Access type describes how this content manages its
viewers. It may have a value of \code{all}, \code{logged_in} or \code{acl}.
The value \code{all} is the most permissive; any visitor to Posit
Connect will be able to view this content. The value \code{logged_in}
indicates that all Posit Connect accounts may view the content. The
\code{acl} value lets specifically enumerated users and groups view the
content. Users configured as collaborators may always view content.
\item \code{connection_timeout}: Maximum number of seconds allowed without data
sent or received across a client connection. A value of 0 means
connections will never time-out (not recommended). When null, the
default \code{Scheduler.ConnectionTimeout} is used. Applies only to content
types that are executed on demand.
\item \code{read_timeout}: Maximum number of seconds allowed without data received
from a client connection. A value of 0 means a lack of client (browser)
interaction never causes the connection to close. When null, the default
\code{Scheduler.ReadTimeout} is used. Applies only to content types that are
executed on demand.
\item \code{init_timeout}: The maximum number of seconds allowed for an interactive
application to start. Posit Connect must be able to connect
to a newly launched Shiny application, for example, before this threshold
has elapsed. When null, the default \code{Scheduler.InitTimeout} is
used. Applies only to content types that are executed on demand.
\item \code{idle_timeout}: The maximum number of seconds a worker process
for an interactive application to remain alive after it goes idle (no
active connections). When null, the default \code{Scheduler.IdleTimeout}
is used. Applies only to content types that are executed on demand.
\item \code{max_processes}: Specifies the total number of concurrent processes
allowed for a single interactive application. When null, the
default \code{Scheduler.MaxProcesses} setting is used. Applies only to
content types that are executed on demand.
\item \code{min_processes}: Specifies the minimum number of concurrent
processes allowed for a single interactive application. When null, the
default \code{Scheduler.MinProcesses} is used. Applies only to content types
that are executed on demand.
\item \code{max_conns_per_process}: Specifies the maximum number of
client connections allowed to an individual process. Incoming connections
which will exceed this limit are routed to a new process or rejected.
When null, the default \code{Scheduler.MaxConnsPerProcess} is used. Applies
only to content types that are executed on demand.
\item \code{load_factor}: Controls how aggressively new processes are spawned.
When null, the default \code{Scheduler.LoadFactor} is used. Applies only to
content types that are executed on demand.
\item \code{created_time}: The timestamp (RFC3339) indicating when this
content was created.
\item \code{last_deployed_time}: The timestamp (RFC3339) indicating when
this content last had a successful bundle deployment performed.
\item \code{bundle_id}: The identifier for the active deployment bundle.
Automatically assigned upon the successful deployment of that bundle.
\item \code{app_mode}: The runtime model for this content. Has a value
of \code{unknown} before data is deployed to this item. Automatically assigned
upon the first successful bundle deployment. Allowed: \code{api},
\code{jupyter-static}, \code{python-api}, \code{python-bokeh}, \code{python-dash},
\code{python-streamlit}, \code{rmd-shiny}, \code{rmd-static}, \code{shiny}, \code{static},
\code{tensorflow-saved-model}, \code{unknown}.
\item \code{content_category}: Describes the specialization of the content
runtime model. Automatically assigned upon the first successful bundle
deployment.
\item \code{parameterized}: True when R Markdown rendered content
allows parameter configuration. Automatically assigned upon the first
successful bundle deployment. Applies only to content with an app_mode
of rmd-static.
\item \code{r_version}: The version of the R interpreter associated
with this content. The value null represents that an R interpreter is
not used by this content or that the R package environment has not been
successfully restored. Automatically assigned upon the successful
deployment of a bundle.
\item \code{py_version}: The version of the Python interpreter
associated with this content. The value null represents that a Python
interpreter is not used by this content or that the Python package
environment has not been successfully restored. Automatically assigned
upon the successful deployment of a bundle.
\item \code{run_as}: The UNIX user that executes this content.
When null, the default Applications.RunAs is used. Applies
only to executable content types - not static.
\item \code{run_as_current_user}: Indicates if this content is allowed
to execute as the logged-in user when using PAM authentication.
Applies only to executable content types - not static.
\item \code{owner_guid}: The unique identifier for the owner
\item \code{content_url}: The URL associated with this content. Computed
from the GUID for this content.
\item \code{dashboard_url}: The URL within the Connect dashboard where
this content can be configured. Computed from the GUID for this content.
\item \code{role}: The relationship of the accessing user to this
content. A value of owner is returned for the content owner. editor
indicates a collaborator. The viewer value is given to users who are
permitted to view the content. A none role is returned for
administrators who cannot view the content but are permitted to view
its configuration. Computed at the time of the request.
\item \code{vanity_url}: The vanity URL associated with this content item.
\item \code{id}: The internal numeric identifier of this content item.
\item \code{tags}: Tags associated with this content item. Each entry is a list
with the following fields:
\itemize{
\item \code{id}: The identifier for the tag.
\item \code{name}: The name of the tag.
\item \code{parent_id}: The identifier for the parent tag. Null if the tag is a
top-level tag.
\item \code{created_time}: The timestamp (RFC3339) indicating when the tag was
created.
\item \code{updated_time}: The timestamp (RFC3339) indicating when the tag was
last updated.
}
\item \code{owner}: Basic details about the owner of this content item. Each entry
is a list with the following fields:
\itemize{
\item \code{guid}: The user's GUID, or unique identifier, in UUID RFC4122 format.
\item \code{username}: The user's username.
\item \code{first_name}: The user's first name.
\item \code{last_name}: The user's last name.
}
}
}
\description{
Get information about content on the Posit Connect server
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/content for more
information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

get_content(client)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thumbnail.R
\name{set_thumbnail}
\alias{set_thumbnail}
\title{Set content item thumbnail}
\usage{
set_thumbnail(content, path)
}
\arguments{
\item{content}{A content item.}

\item{path}{Either a path to a local file or a URL to an image available over
HTTP/HTTPS. If \code{path} is an HTTP or HTTPS URL, the image will first
be downloaded.}
}
\value{
The content item (invisibly).
}
\description{
Set the thumbnail for a content item.
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
set_thumbnail(item, "resources/image.png")
}

}
\seealso{
Other thumbnail functions: 
\code{\link{delete_thumbnail}()},
\code{\link{get_thumbnail}()},
\code{\link{has_thumbnail}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{thumbnail functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{as_tibble.connect_integration_list}
\alias{as_tibble.connect_integration_list}
\title{Convert integration list to a tibble}
\usage{
\method{as_tibble}{connect_integration_list}(x, ...)
}
\arguments{
\item{x}{A \code{connect_integration_list} object.}

\item{...}{Unused.}
}
\value{
A tibble with one row per integration.
}
\description{
Converts a list returned by \code{\link[=get_integrations]{get_integrations()}} to a tibble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{as_tibble.connect_content_list}
\alias{as_tibble.connect_content_list}
\title{Convert integration list to a tibble}
\usage{
\method{as_tibble}{connect_content_list}(x, ...)
}
\arguments{
\item{x}{A \code{connect_content_list} object.}

\item{...}{Unused.}
}
\value{
A tibble with one row per content item.
}
\description{
Converts a list returned by \code{\link[=search_content]{search_content()}} to a tibble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_procs}
\alias{get_procs}
\title{Get Real-Time Process Data}
\usage{
get_procs(src)
}
\arguments{
\item{src}{The source object}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{pid}: The PID of the current process
\item \code{appId}: The application ID
\item \code{appGuid}: The application GUID
\item \code{appName}: The application name
\item \code{appUrl}: The application URL
\item \code{appRunAs}: The application RunAs user
\item \code{type}: The type of process
\item \code{cpuCurrent}: The current CPU usage
\item \code{cpuTotal}: The total CPU usage
\item \code{ram}: The current RAM usage
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
This returns real-time process data from the Posit Connect API. It requires
administrator privileges to use. NOTE that this only returns data for the
server that responds to the request (i.e. in a Highly Available cluster)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect.R
\name{PositConnect}
\alias{PositConnect}
\alias{Connect}
\title{Class representing a Connect API client}
\description{
Class representing a Connect API client

Class representing a Connect API client
}
\section{Usage}{

\preformatted{
client <- Connect$new(server = 'connect.example.com',
  api_key = 'mysecretkey')

get_content(client)
client$get_tags()
}
}

\section{Details}{


This class allows a user to interact with a Connect server via the Connect
API. Authentication is done by providing an API key.
}

\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{server}}{The base URL of your Posit Connect server.}

\item{\code{api_key}}{Your Posit Connect API key.}

\item{\code{tags}}{The initial set of tags.}

\item{\code{tag_map}}{The initial tag map.}

\item{\code{httr_additions}}{An initial set of \code{httr} configuration added to each HTTP call.}

\item{\code{using_auth}}{Indicates that the API key is added to each HTTP call.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{version}}{The server version.}

\item{\code{timezones}}{The server timezones.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Connect-new}{\code{Connect$new()}}
\item \href{#method-Connect-httr_config}{\code{Connect$httr_config()}}
\item \href{#method-Connect-print}{\code{Connect$print()}}
\item \href{#method-Connect-raise_error}{\code{Connect$raise_error()}}
\item \href{#method-Connect-add_auth}{\code{Connect$add_auth()}}
\item \href{#method-Connect-api_url}{\code{Connect$api_url()}}
\item \href{#method-Connect-server_url}{\code{Connect$server_url()}}
\item \href{#method-Connect-request}{\code{Connect$request()}}
\item \href{#method-Connect-GET}{\code{Connect$GET()}}
\item \href{#method-Connect-PUT}{\code{Connect$PUT()}}
\item \href{#method-Connect-HEAD}{\code{Connect$HEAD()}}
\item \href{#method-Connect-DELETE}{\code{Connect$DELETE()}}
\item \href{#method-Connect-PATCH}{\code{Connect$PATCH()}}
\item \href{#method-Connect-POST}{\code{Connect$POST()}}
\item \href{#method-Connect-me}{\code{Connect$me()}}
\item \href{#method-Connect-get_dashboard_url}{\code{Connect$get_dashboard_url()}}
\item \href{#method-Connect-get_tags}{\code{Connect$get_tags()}}
\item \href{#method-Connect-get_tag_id}{\code{Connect$get_tag_id()}}
\item \href{#method-Connect-get_tag_tree}{\code{Connect$get_tag_tree()}}
\item \href{#method-Connect-tag_create_safe}{\code{Connect$tag_create_safe()}}
\item \href{#method-Connect-tag_create}{\code{Connect$tag_create()}}
\item \href{#method-Connect-tag}{\code{Connect$tag()}}
\item \href{#method-Connect-tag_delete}{\code{Connect$tag_delete()}}
\item \href{#method-Connect-get_schedule}{\code{Connect$get_schedule()}}
\item \href{#method-Connect-content_create}{\code{Connect$content_create()}}
\item \href{#method-Connect-content_upload}{\code{Connect$content_upload()}}
\item \href{#method-Connect-content_deploy}{\code{Connect$content_deploy()}}
\item \href{#method-Connect-content}{\code{Connect$content()}}
\item \href{#method-Connect-task}{\code{Connect$task()}}
\item \href{#method-Connect-set_content_tag}{\code{Connect$set_content_tag()}}
\item \href{#method-Connect-remove_content_tag}{\code{Connect$remove_content_tag()}}
\item \href{#method-Connect-user}{\code{Connect$user()}}
\item \href{#method-Connect-users}{\code{Connect$users()}}
\item \href{#method-Connect-users_remote}{\code{Connect$users_remote()}}
\item \href{#method-Connect-users_create}{\code{Connect$users_create()}}
\item \href{#method-Connect-users_create_remote}{\code{Connect$users_create_remote()}}
\item \href{#method-Connect-users_lock}{\code{Connect$users_lock()}}
\item \href{#method-Connect-users_unlock}{\code{Connect$users_unlock()}}
\item \href{#method-Connect-users_update}{\code{Connect$users_update()}}
\item \href{#method-Connect-groups}{\code{Connect$groups()}}
\item \href{#method-Connect-group_members}{\code{Connect$group_members()}}
\item \href{#method-Connect-group_member_add}{\code{Connect$group_member_add()}}
\item \href{#method-Connect-group_member_remove}{\code{Connect$group_member_remove()}}
\item \href{#method-Connect-groups_create}{\code{Connect$groups_create()}}
\item \href{#method-Connect-groups_create_remote}{\code{Connect$groups_create_remote()}}
\item \href{#method-Connect-groups_remote}{\code{Connect$groups_remote()}}
\item \href{#method-Connect-group_content}{\code{Connect$group_content()}}
\item \href{#method-Connect-inst_content_visits}{\code{Connect$inst_content_visits()}}
\item \href{#method-Connect-inst_shiny_usage}{\code{Connect$inst_shiny_usage()}}
\item \href{#method-Connect-procs}{\code{Connect$procs()}}
\item \href{#method-Connect-repo_account}{\code{Connect$repo_account()}}
\item \href{#method-Connect-repo_branches}{\code{Connect$repo_branches()}}
\item \href{#method-Connect-repo_manifest_dirs}{\code{Connect$repo_manifest_dirs()}}
\item \href{#method-Connect-schedules}{\code{Connect$schedules()}}
\item \href{#method-Connect-packages}{\code{Connect$packages()}}
\item \href{#method-Connect-docs}{\code{Connect$docs()}}
\item \href{#method-Connect-audit_logs}{\code{Connect$audit_logs()}}
\item \href{#method-Connect-vanities}{\code{Connect$vanities()}}
\item \href{#method-Connect-server_settings}{\code{Connect$server_settings()}}
\item \href{#method-Connect-clone}{\code{Connect$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-new"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-new}{}}}
\subsection{Method \code{new()}}{
Initialize a new connect.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$new(server, api_key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{server}}{The base URL of your Posit Connect server.}

\item{\code{api_key}}{Your Posit Connect API key.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-httr_config"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-httr_config}{}}}
\subsection{Method \code{httr_config()}}{
Set additional \code{httr} configuration that is added to each HTTP call.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$httr_config(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Set of httr configurations.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-print"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-print}{}}}
\subsection{Method \code{print()}}{
Print details about this instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-raise_error"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-raise_error}{}}}
\subsection{Method \code{raise_error()}}{
Raise an error when the HTTP result is an HTTP error.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$raise_error(res)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{res}}{HTTP result.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-add_auth"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-add_auth}{}}}
\subsection{Method \code{add_auth()}}{
Returns HTTP authorization headers, or NULL when none are used.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$add_auth()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-api_url"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-api_url}{}}}
\subsection{Method \code{api_url()}}{
Build a URL relative to the API root
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$api_url(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{path segments}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-server_url"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-server_url}{}}}
\subsection{Method \code{server_url()}}{
Build a URL relative to the server root
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$server_url(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{path segments}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-request"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-request}{}}}
\subsection{Method \code{request()}}{
General wrapper around \code{httr} verbs
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$request(method, url, ..., parser = "parsed")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{method}}{HTTP request method}

\item{\code{url}}{URL to request}

\item{\code{...}}{Additional arguments passed to the request function}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-GET"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-GET}{}}}
\subsection{Method \code{GET()}}{
Perform an HTTP GET request of the named API path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$GET(path, ..., url = self$api_url(path), parser = "parsed")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{...}}{Arguments to \code{httr::GET()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-PUT"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-PUT}{}}}
\subsection{Method \code{PUT()}}{
Perform an HTTP PUT request of the named API path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$PUT(
  path,
  body = "{}",
  ...,
  url = self$api_url(path),
  encode = "json",
  parser = "parsed"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{body}}{The HTTP payload.}

\item{\code{...}}{Arguments to \code{httr::PUT()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}}

\item{\code{encode}}{How the payload is encoded.}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-HEAD"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-HEAD}{}}}
\subsection{Method \code{HEAD()}}{
Perform an HTTP HEAD request of the named API path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$HEAD(path, ..., url = self$api_url(path))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{...}}{Arguments to \code{httr::HEAD()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-DELETE"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-DELETE}{}}}
\subsection{Method \code{DELETE()}}{
Perform an HTTP DELETE request of the named API path. Returns the HTTP response object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$DELETE(path, ..., url = self$api_url(path), parser = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{...}}{Arguments to \code{httr::DELETE()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-PATCH"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-PATCH}{}}}
\subsection{Method \code{PATCH()}}{
Perform an HTTP PATCH request of the named API path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$PATCH(
  path,
  body = "{}",
  ...,
  url = self$api_url(path),
  encode = "json",
  parser = "parsed"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{body}}{The HTTP payload.}

\item{\code{...}}{Arguments to \code{httr::PATCH()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}}

\item{\code{encode}}{How the payload is encoded.}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-POST"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-POST}{}}}
\subsection{Method \code{POST()}}{
Perform an HTTP POST request of the named API path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$POST(
  path,
  body = "{}",
  ...,
  url = self$api_url(path),
  encode = "json",
  parser = "parsed"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{API path relative to the server's \verb{/__api__} root.}

\item{\code{body}}{The HTTP payload.}

\item{\code{...}}{Arguments to \code{httr::POST()}}

\item{\code{url}}{Target URL. Default uses \code{path}, but provide \code{url} to request
a server resource that is not under \verb{/__api__}}

\item{\code{encode}}{How the payload is encoded.}

\item{\code{parser}}{How the response is parsed. If \code{NULL}, the \code{httr_response}
will be returned. Otherwise, the argument is forwarded to
\code{httr::content(res, as = parser)}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-me"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-me}{}}}
\subsection{Method \code{me()}}{
Perform an HTTP GET request of the "me" server endpoint.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$me()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-get_dashboard_url"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-get_dashboard_url}{}}}
\subsection{Method \code{get_dashboard_url()}}{
Return the base URL of the Connect server.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$get_dashboard_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-get_tags"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-get_tags}{}}}
\subsection{Method \code{get_tags()}}{
Return all tags.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$get_tags(use_cache = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{use_cache}}{Indicates that a cached set of tags is used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-get_tag_id"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-get_tag_id}{}}}
\subsection{Method \code{get_tag_id()}}{
Get the identifier for the named tag.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$get_tag_id(tagname)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tagname}}{The name of the tag.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-get_tag_tree"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-get_tag_tree}{}}}
\subsection{Method \code{get_tag_tree()}}{
Get the tag tree.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$get_tag_tree()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-tag_create_safe"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-tag_create_safe}{}}}
\subsection{Method \code{tag_create_safe()}}{
Create a tag.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$tag_create_safe(name, parent_id = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{The tag name.}

\item{\code{parent_id}}{The parent identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-tag_create"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-tag_create}{}}}
\subsection{Method \code{tag_create()}}{
Create a tag.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$tag_create(name, parent_id = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{The tag name.}

\item{\code{parent_id}}{The parent identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-tag"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-tag}{}}}
\subsection{Method \code{tag()}}{
Get a tag.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$tag(id = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-tag_delete"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-tag_delete}{}}}
\subsection{Method \code{tag_delete()}}{
Delete a tag.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$tag_delete(id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-get_schedule"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-get_schedule}{}}}
\subsection{Method \code{get_schedule()}}{
Get a schedule.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$get_schedule(schedule_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{schedule_id}}{The schedule identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-content_create"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-content_create}{}}}
\subsection{Method \code{content_create()}}{
Create content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$content_create(name, title = name, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{The content name.}

\item{\code{title}}{The content title.}

\item{\code{...}}{Other content fields.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-content_upload"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-content_upload}{}}}
\subsection{Method \code{content_upload()}}{
Upload a content bundle.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$content_upload(bundle_path, guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bundle_path}}{The path to the bundle archive.}

\item{\code{guid}}{The content GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-content_deploy"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-content_deploy}{}}}
\subsection{Method \code{content_deploy()}}{
Deploy a content bundle.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$content_deploy(guid, bundle_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{The content GUID.}

\item{\code{bundle_id}}{The bundle identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-content"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-content}{}}}
\subsection{Method \code{content()}}{
Get a content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$content(
  guid = NULL,
  owner_guid = NULL,
  name = NULL,
  include = "tags,owner"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{The content GUID.}

\item{\code{owner_guid}}{The target content owner.}

\item{\code{name}}{The target name.}

\item{\code{include}}{Additional response fields.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-task"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-task}{}}}
\subsection{Method \code{task()}}{
Get a task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$task(task_id, first = 0, wait = 5)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{task_id}}{The task identifier.}

\item{\code{first}}{The initial status position.}

\item{\code{wait}}{Maximum time to wait for update.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-set_content_tag"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-set_content_tag}{}}}
\subsection{Method \code{set_content_tag()}}{
Set a tag for a content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$set_content_tag(content_id, tag_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{content_id}}{The content identifier.}

\item{\code{tag_id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-remove_content_tag"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-remove_content_tag}{}}}
\subsection{Method \code{remove_content_tag()}}{
Remove a tag from a content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$remove_content_tag(content_id, tag_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{content_id}}{The content identifier.}

\item{\code{tag_id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-user"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-user}{}}}
\subsection{Method \code{user()}}{
Get user details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$user(guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{The user GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users}{}}}
\subsection{Method \code{users()}}{
Get users.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users(
  page_number = 1,
  prefix = NULL,
  page_size = 500,
  user_role = NULL,
  account_status = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{page_number}}{The page number.}

\item{\code{prefix}}{The search term.}

\item{\code{page_size}}{The page size.}

\item{\code{user_role}}{Filter by user role.}

\item{\code{account_status}}{Filter by account status.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_remote}{}}}
\subsection{Method \code{users_remote()}}{
Get remote users.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_remote(prefix)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prefix}}{The search term.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_create"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_create}{}}}
\subsection{Method \code{users_create()}}{
Create a user.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_create(
  username,
  email,
  first_name = NULL,
  last_name = NULL,
  password = NULL,
  user_must_set_password = NULL,
  user_role = NULL,
  unique_id = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{username}}{The username.}

\item{\code{email}}{Email address.}

\item{\code{first_name}}{First name.}

\item{\code{last_name}}{Last name.}

\item{\code{password}}{The password.}

\item{\code{user_must_set_password}}{Indicates that user sets password on first login.}

\item{\code{user_role}}{Role for user.}

\item{\code{unique_id}}{Identifier for user.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_create_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_create_remote}{}}}
\subsection{Method \code{users_create_remote()}}{
Create a remote user.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_create_remote(temp_ticket)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{temp_ticket}}{Ticket identifying target remote user.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_lock"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_lock}{}}}
\subsection{Method \code{users_lock()}}{
Lock a user.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_lock(user_guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{user_guid}}{User GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_unlock"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_unlock}{}}}
\subsection{Method \code{users_unlock()}}{
Unlock a user.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_unlock(user_guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{user_guid}}{User GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-users_update"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-users_update}{}}}
\subsection{Method \code{users_update()}}{
Update a user.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$users_update(user_guid, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{user_guid}}{User GUID.}

\item{\code{...}}{User fields.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-groups"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-groups}{}}}
\subsection{Method \code{groups()}}{
Get groups.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$groups(page_number = 1, prefix = NULL, page_size = 500)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{page_number}}{The page number.}

\item{\code{prefix}}{The search term.}

\item{\code{page_size}}{The page size.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-group_members"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-group_members}{}}}
\subsection{Method \code{group_members()}}{
Get group members.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$group_members(guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{The group GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-group_member_add"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-group_member_add}{}}}
\subsection{Method \code{group_member_add()}}{
Add a group member.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$group_member_add(group_guid, user_guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{group_guid}}{The group GUID.}

\item{\code{user_guid}}{The user GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-group_member_remove"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-group_member_remove}{}}}
\subsection{Method \code{group_member_remove()}}{
Remove a group member.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$group_member_remove(group_guid, user_guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{group_guid}}{The group GUID.}

\item{\code{user_guid}}{The user GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-groups_create"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-groups_create}{}}}
\subsection{Method \code{groups_create()}}{
Create a group.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$groups_create(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{The group name.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-groups_create_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-groups_create_remote}{}}}
\subsection{Method \code{groups_create_remote()}}{
Create a remote group.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$groups_create_remote(temp_ticket)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{temp_ticket}}{Ticket identifying target remote group.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-groups_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-groups_remote}{}}}
\subsection{Method \code{groups_remote()}}{
Get remote groups.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$groups_remote(prefix = NULL, limit = 500)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prefix}}{The search term.}

\item{\code{limit}}{The maximal result set size.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-group_content"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-group_content}{}}}
\subsection{Method \code{group_content()}}{
Get content to which a group has access
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$group_content(guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{The group GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-inst_content_visits"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-inst_content_visits}{}}}
\subsection{Method \code{inst_content_visits()}}{
Get (non-interactive) content visits.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$inst_content_visits(
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{content_guid}}{Content GUID.}

\item{\code{min_data_version}}{Data version for request.}

\item{\code{from}}{Start of range.}

\item{\code{to}}{End of range.}

\item{\code{limit}}{Result set size.}

\item{\code{previous}}{Previous item.}

\item{\code{nxt}}{Next item.}

\item{\code{asc_order}}{Indicates ascending result order.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-inst_shiny_usage"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-inst_shiny_usage}{}}}
\subsection{Method \code{inst_shiny_usage()}}{
Get interactive content visits.


Get (non-interactive) content visits.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$inst_shiny_usage(
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{content_guid}}{Content GUID.}

\item{\code{min_data_version}}{Data version for request.}

\item{\code{from}}{Start of range.}

\item{\code{to}}{End of range.}

\item{\code{limit}}{Result set size.}

\item{\code{previous}}{Previous item.}

\item{\code{nxt}}{Next item.}

\item{\code{asc_order}}{Indicates ascending result order.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-procs"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-procs}{}}}
\subsection{Method \code{procs()}}{
Get running processes.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$procs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-repo_account"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-repo_account}{}}}
\subsection{Method \code{repo_account()}}{
Determine if Git repository is associated with authorization.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$repo_account(host)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{host}}{Repository URL.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-repo_branches"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-repo_branches}{}}}
\subsection{Method \code{repo_branches()}}{
Get Git repository branches.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$repo_branches(repo)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{repo}}{Repository URL.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-repo_manifest_dirs"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-repo_manifest_dirs}{}}}
\subsection{Method \code{repo_manifest_dirs()}}{
Get Git repository directories.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$repo_manifest_dirs(repo, branch)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{repo}}{Repository URL.}

\item{\code{branch}}{Repository branch.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-schedules"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-schedules}{}}}
\subsection{Method \code{schedules()}}{
Get schedules.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$schedules(
  start = Sys.time(),
  end = Sys.time() + 60 * 60 * 24 * 7,
  detailed = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{start}}{Starting time.}

\item{\code{end}}{Ending time.}

\item{\code{detailed}}{Indicates detailed schedule information.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-packages"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-packages}{}}}
\subsection{Method \code{packages()}}{
Get packages. This endpoint is paginated.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$packages(name = NULL, page_number = 1, page_size = 1e+05)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{The package name to filter by.}

\item{\code{page_number}}{Page number.}

\item{\code{page_size}}{Page size, default 100000.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-docs"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-docs}{}}}
\subsection{Method \code{docs()}}{
Get documentation.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$docs(docs = "api", browse = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{docs}}{Named document.}

\item{\code{browse}}{Open a browser.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-audit_logs"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-audit_logs}{}}}
\subsection{Method \code{audit_logs()}}{
Get auditing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$audit_logs(limit = 500, previous = NULL, nxt = NULL, asc_order = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{limit}}{Result set size.}

\item{\code{previous}}{Previous item.}

\item{\code{nxt}}{Next item.}

\item{\code{asc_order}}{Indicates ascending result order.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-vanities"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-vanities}{}}}
\subsection{Method \code{vanities()}}{
Get all vanity URLs
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$vanities()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-server_settings"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-server_settings}{}}}
\subsection{Method \code{server_settings()}}{
Get server settings.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$server_settings()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Connect-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Connect-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Connect$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content_delete}
\alias{content_delete}
\title{Delete Content}
\usage{
content_delete(content, force = FALSE)
}
\arguments{
\item{content}{an R6 content item}

\item{force}{Optional. A boolean that determines whether we should prompt in interactive sessions}
}
\value{
The R6 Content item. The item is deleted, but information about it is cached locally
}
\description{
Delete a content item. WARNING: This action deletes all history, configuration,
logs, and resources about a content item. It \emph{cannot} be undone.
}
\seealso{
Other content functions: 
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_usage}
\alias{get_usage}
\title{Get usage information for deployed content}
\usage{
get_usage(client, content_guid = NULL, from = NULL, to = NULL)
}
\arguments{
\item{client}{A \code{Connect} R6 client object.}

\item{content_guid}{Optional. A single content GUID or a character vector of
GUIDs to filter results. When multiple GUIDs are provided they are
collapsed with \code{"|"}.}

\item{from}{Optional date-time (\code{POSIXct} or \code{POSIXlt}). Only
records after this time are returned. If not provided, records
are returned back to the first record available.}

\item{to}{Optional date-time (\code{POSIXct} or \code{POSIXlt}). Only records
before this time are returned. If not provided, all records up to
the most recent are returned.}
}
\value{
A list of usage records. Each record is a list with all elements
as character strings unless otherwise specified.
\itemize{
\item \code{id}: A string identifier for the hit.
\item \code{user_guid}: The user GUID if the visitor is logged-in, \code{NULL} for
anonymous hits.
\item \code{content_guid}: The GUID of the visited content.
\item \code{timestamp}: The time of the hit in RFC3339 format.
\item \code{data}: A nested list with optional fields:
\itemize{
\item \code{path}: The request path (if recorded).
\item \code{user_agent}: The user agent string (if available).
}
}

Use \code{\link[=as.data.frame]{as.data.frame()}} or \code{\link[tibble:as_tibble]{tibble::as_tibble()}} to convert to a flat
table with parsed types. In the resulting data frame:
\itemize{
\item \code{timestamp} is parsed to \code{POSIXct}.
\item \code{path} and \code{user_agent} are extracted from the nested \code{data} field.
}

By default, \code{\link[=as.data.frame]{as.data.frame()}} attempts to extract the nested fields using
the \pkg{tidyr} package. If \pkg{tidyr} is not available, or if you want to
skip unnesting, call \code{as.data.frame(x, unnest = FALSE)} to leave \code{data} as
a list-column.
}
\description{
Retrieve content hits for all available content on the server. Available
content depends on the user whose API key is in use. Administrator accounts
will receive data for all content on the server. Publishers will receive data
for all content they own or collaborate on.

If no date-times are provided, all usage data will be returned.
}
\details{
The data returned by \code{get_usage()} includes all content types. For Shiny
content, the \code{timestamp} indicates the \emph{start} of the Shiny session.
Additional fields for Shiny and non-Shiny are available respectively from
\code{\link[=get_usage_shiny]{get_usage_shiny()}} and \code{\link[=get_usage_static]{get_usage_static()}}. \code{get_usage_shiny()} includes a
field for the session end time; \code{get_usage_static()} includes variant,
rendering, and bundle identifiers for the visited content.

When possible, however, we recommend using \code{get_usage()} over
\code{get_usage_static()} or \code{get_usage_shiny()}, as it is faster and more efficient.
}
\examples{
\dontrun{
client <- connect()

# Fetch the last 2 days of hits
usage <- get_usage(client, from = Sys.Date() - 2, to = Sys.Date())

# Fetch usage after a specified date and convert to a data frame.
usage <- get_usage(
  client,
  from = as.POSIXct("2025-05-02 12:40:00", tz = "UTC")
)

# Fetch usage for a specific content item
usage <- get_usage(client, content_guid = "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Fetch all usage
usage <- get_usage(client)

# Convert to tibble or data frame
usage_df <- tibble::as_tibble(usage)

# Skip unnesting if tidyr is not installed
usage_df <- as.data.frame(usage, unnest = FALSE)
}

}
\seealso{
\code{\link[=as.data.frame.connect_list_hits]{as.data.frame.connect_list_hits()}}, \code{\link[=as_tibble.connect_list_hits]{as_tibble.connect_list_hits()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{verify_content_name}
\alias{verify_content_name}
\title{Verify Content Name}
\usage{
verify_content_name(name)
}
\arguments{
\item{name}{The proposed content name}
}
\value{
The name (or an error if invalid)
}
\description{
Ensures that a content name fits the specifications / requirements of Posit
Connect. Throws an error if content name is invalid. Content names (as of the
time of writing) must be between 3 and 64 alphanumeric characters, dashes,
and underscores
}
\seealso{
connectapi::create_random_name

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git.R
\name{git}
\alias{git}
\alias{repo_check_account}
\alias{repo_check_branches}
\alias{repo_check_branches_ref}
\alias{repo_check_manifest_dirs}
\title{Git Repository Helpers}
\usage{
repo_check_account(client, host)

repo_check_branches(client, repository)

repo_check_branches_ref(client, repository)

repo_check_manifest_dirs(client, repository, branch)
}
\arguments{
\item{client}{A Connect R6 object}

\item{host}{The git repository host (with schema). For example, "https://github.com"}

\item{repository}{The git repository to explore or consider deploying}

\item{branch}{The git branch to explore for subdirectories}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} These functions help use Posit Connect's
configured authorization to query available branches and subdirectories for
deployment using \code{deploy_repo()}
}
\details{
\itemize{
\item \code{repo_check_account()} messages whether an account is in use, and then returns that account
\item \code{repo_check_branches()} retrieves which branches are available, returning in a named list
\item \code{repo_check_manifest_dirs()} retrieves which directories contain a
\code{manifest.json}, returning in a named list
}
}
\seealso{
connectapi::deploy_repo

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{poll_task}
\alias{poll_task}
\title{Poll Task}
\usage{
poll_task(task, wait = 1, callback = message)
}
\arguments{
\item{task}{A Task object}

\item{wait}{The interval to wait between polling}

\item{callback}{A function to be called for each message received. Set to NULL for no callback}
}
\value{
Task The Task object that was input
}
\description{
Polls a task, waiting for information about a deployment. If the task has
results, the output will be a modified "Task" object with \code{task$get_data()}
available to retrieve the results.
}
\details{
For a simple way to silence messages, set \code{callback = NULL}
}
\seealso{
Other deployment functions: 
\code{\link{bundle_dir}()},
\code{\link{bundle_path}()},
\code{\link{bundle_static}()},
\code{\link{deploy}()},
\code{\link{download_bundle}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groups.R
\name{get_groups}
\alias{get_groups}
\title{Get group information from the Posit Connect server}
\usage{
get_groups(src, page_size = 500, prefix = NULL, limit = Inf)
}
\arguments{
\item{src}{The source object.}

\item{page_size}{The number of records to return per page (max 500).}

\item{prefix}{Filters groups by prefix (group name).
The filter is case insensitive.}

\item{limit}{The number of groups to retrieve before paging stops. Default
is to return all results; however, for Connect server versions older than
2025.04.0, \code{limit} is capped at 500 when \code{prefix} is provided.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{guid}: The unique identifier of the group
\item \code{name}: The group name
\item \code{owner_guid}: The group owner's unique identifier. When using LDAP or
Proxied authentication with group provisioning enabled this property
will always be null.
}
}
\description{
Get group information from the Posit Connect server
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/groups for more information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

# get all groups
get_groups(client, limit = Inf)
}

}
\seealso{
Other groups functions: 
\code{\link{get_group_content}()},
\code{\link{get_group_members}()}
}
\concept{groups functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{permissions}
\alias{permissions}
\alias{content_add_user}
\alias{content_add_group}
\alias{content_delete_user}
\alias{content_delete_group}
\alias{get_user_permission}
\alias{get_my_permission}
\alias{get_group_permission}
\alias{get_content_permissions}
\title{Content permissions}
\usage{
content_add_user(content, guid, role = c("viewer", "owner"))

content_add_group(content, guid, role = c("viewer", "owner"))

content_delete_user(content, guid)

content_delete_group(content, guid)

get_user_permission(content, guid, add_owner = TRUE)

get_my_permission(content, add_owner = TRUE)

get_group_permission(content, guid)

get_content_permissions(content, add_owner = TRUE)
}
\arguments{
\item{content}{An R6 content object}

\item{guid}{The guid associated with either a user (for \code{content_add_user}) or group (for \code{content_add_group})}

\item{role}{The role to assign to a user. Either "viewer" or "owner." Defaults to "viewer"}

\item{add_owner}{Optional. Whether to include the owner in returned
permission sets. Default is TRUE. The owner will have an NA_character_
permission "id"}
}
\description{
Get or set content permissions for a content item
}
\details{
Permission modification:
\itemize{
\item \verb{content_add_*} adds a permission to the content
\item \verb{content_delete_*} removes a permission from the content
}

Permission retrieval:
\itemize{
\item \code{get_content_permissions()} lists permissions
\item \code{get_my_permission()} gets the permission associated with the caller.
\item \code{get_user_permission()} gets the permissions associated with a given user.
It does not evaluate group memberships
\item \code{get_group_permission()} gets the permissions associated with a given
group.
}

NOTE: by default, the owner is injected with an "NA_character_" permission id.
This makes it easier to find / isolate this record.
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variant.R
\name{Variant}
\alias{Variant}
\title{Variant}
\description{
Variant

Variant
}
\details{
An R6 class that represents a Variant
}
\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{PositConnect}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Super class}{
\code{\link[connectapi:Content]{connectapi::Content}} -> \code{Variant}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{key}}{The variant key.}

\item{\code{variant}}{The variant.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Variant-get_variant_remote}{\code{Variant$get_variant_remote()}}
\item \href{#method-Variant-new}{\code{Variant$new()}}
\item \href{#method-Variant-send_mail}{\code{Variant$send_mail()}}
\item \href{#method-Variant-get_schedule}{\code{Variant$get_schedule()}}
\item \href{#method-Variant-get_schedule_remote}{\code{Variant$get_schedule_remote()}}
\item \href{#method-Variant-get_subscribers}{\code{Variant$get_subscribers()}}
\item \href{#method-Variant-remove_subscriber}{\code{Variant$remove_subscriber()}}
\item \href{#method-Variant-add_subscribers}{\code{Variant$add_subscribers()}}
\item \href{#method-Variant-render}{\code{Variant$render()}}
\item \href{#method-Variant-renderings}{\code{Variant$renderings()}}
\item \href{#method-Variant-update_variant}{\code{Variant$update_variant()}}
\item \href{#method-Variant-jobs}{\code{Variant$jobs()}}
\item \href{#method-Variant-get_url}{\code{Variant$get_url()}}
\item \href{#method-Variant-get_url_rev}{\code{Variant$get_url_rev()}}
\item \href{#method-Variant-get_dashboard_url}{\code{Variant$get_dashboard_url()}}
\item \href{#method-Variant-print}{\code{Variant$print()}}
\item \href{#method-Variant-clone}{\code{Variant$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_delete"><a href='../../connectapi/html/Content.html#method-Content-bundle_delete'><code>connectapi::Content$bundle_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_download"><a href='../../connectapi/html/Content.html#method-Content-bundle_download'><code>connectapi::Content$bundle_download()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="danger_delete"><a href='../../connectapi/html/Content.html#method-Content-danger_delete'><code>connectapi::Content$danger_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="deploy"><a href='../../connectapi/html/Content.html#method-Content-deploy'><code>connectapi::Content$deploy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment"><a href='../../connectapi/html/Content.html#method-Content-environment'><code>connectapi::Content$environment()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment_all"><a href='../../connectapi/html/Content.html#method-Content-environment_all'><code>connectapi::Content$environment_all()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment_set"><a href='../../connectapi/html/Content.html#method-Content-environment_set'><code>connectapi::Content$environment_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_bundles"><a href='../../connectapi/html/Content.html#method-Content-get_bundles'><code>connectapi::Content$get_bundles()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_content_remote"><a href='../../connectapi/html/Content.html#method-Content-get_content_remote'><code>connectapi::Content$get_content_remote()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="packages"><a href='../../connectapi/html/Content.html#method-Content-packages'><code>connectapi::Content$packages()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions"><a href='../../connectapi/html/Content.html#method-Content-permissions'><code>connectapi::Content$permissions()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_add"><a href='../../connectapi/html/Content.html#method-Content-permissions_add'><code>connectapi::Content$permissions_add()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_delete"><a href='../../connectapi/html/Content.html#method-Content-permissions_delete'><code>connectapi::Content$permissions_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_update"><a href='../../connectapi/html/Content.html#method-Content-permissions_update'><code>connectapi::Content$permissions_update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="register_job_kill_order"><a href='../../connectapi/html/Content.html#method-Content-register_job_kill_order'><code>connectapi::Content$register_job_kill_order()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_enable"><a href='../../connectapi/html/Content.html#method-Content-repo_enable'><code>connectapi::Content$repo_enable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_set"><a href='../../connectapi/html/Content.html#method-Content-repo_set'><code>connectapi::Content$repo_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repository"><a href='../../connectapi/html/Content.html#method-Content-repository'><code>connectapi::Content$repository()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_delete"><a href='../../connectapi/html/Content.html#method-Content-tag_delete'><code>connectapi::Content$tag_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_set"><a href='../../connectapi/html/Content.html#method-Content-tag_set'><code>connectapi::Content$tag_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tags"><a href='../../connectapi/html/Content.html#method-Content-tags'><code>connectapi::Content$tags()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="update"><a href='../../connectapi/html/Content.html#method-Content-update'><code>connectapi::Content$update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="variants"><a href='../../connectapi/html/Content.html#method-Content-variants'><code>connectapi::Content$variants()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_variant_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_variant_remote}{}}}
\subsection{Method \code{get_variant_remote()}}{
Get the underlying variant data.


Get and store the (remote) variant data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_variant_remote()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-new"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-new}{}}}
\subsection{Method \code{new()}}{
Initialize this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$new(connect, content, key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{connect}}{The \code{Connect} instance.}

\item{\code{content}}{The \code{Content} instance.}

\item{\code{key}}{The variant key.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-send_mail"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-send_mail}{}}}
\subsection{Method \code{send_mail()}}{
Mail previously rendered content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$send_mail(to = c("me", "collaborators", "collaborators_viewers"))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{to}}{Targeting.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_schedule"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_schedule}{}}}
\subsection{Method \code{get_schedule()}}{
Get the (remote) schedule data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_schedule()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_schedule_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_schedule_remote}{}}}
\subsection{Method \code{get_schedule_remote()}}{
Get the (remote) schedule data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_schedule_remote()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_subscribers"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_subscribers}{}}}
\subsection{Method \code{get_subscribers()}}{
Get the subscribers.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_subscribers()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-remove_subscriber"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-remove_subscriber}{}}}
\subsection{Method \code{remove_subscriber()}}{
Remove a named subscriber.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$remove_subscriber(guid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guid}}{User GUID.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-add_subscribers"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-add_subscribers}{}}}
\subsection{Method \code{add_subscribers()}}{
Add named subscribers.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$add_subscribers(guids)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{guids}}{User GUIDs.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-render"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-render}{}}}
\subsection{Method \code{render()}}{
Render this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$render()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-renderings"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-renderings}{}}}
\subsection{Method \code{renderings()}}{
List the renderings of this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$renderings()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-update_variant"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-update_variant}{}}}
\subsection{Method \code{update_variant()}}{
Update this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$update_variant(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Target fields and values.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-jobs"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-jobs}{}}}
\subsection{Method \code{jobs()}}{
Jobs for this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$jobs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_url"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_url}{}}}
\subsection{Method \code{get_url()}}{
Return the URL for this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_url_rev"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_url_rev}{}}}
\subsection{Method \code{get_url_rev()}}{
Return the URL associated with one rendering for this variant.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_url_rev(rev)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rev}}{Rendering identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-get_dashboard_url"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-get_dashboard_url}{}}}
\subsection{Method \code{get_dashboard_url()}}{
Return the URL for this variant in the Posit Connect dashboard.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$get_dashboard_url(pane = "access")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{pane}}{The pane in the dashboard to link to.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-print"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Variant-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Variant-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Variant$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{content_list_by_tag}
\alias{content_list_by_tag}
\title{Content List}
\usage{
content_list_by_tag(src, tag)
}
\arguments{
\item{src}{An R6 Connect object}

\item{tag}{A \code{connect_tag_tree} object or tag ID}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Get a content list
}
\details{
\code{content_list_by_tag()} retrieves a content list by tag
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{users_create_remote}
\alias{users_create_remote}
\title{Create a Remote User}
\usage{
users_create_remote(connect, prefix, expect = 1, check = TRUE, exact = FALSE)
}
\arguments{
\item{connect}{An R6 Connect object.}

\item{prefix}{character. The prefix of the user name to search for.}

\item{expect}{number. Optional. The number of responses to expect for this search.}

\item{check}{boolean. Optional. Whether to check for local existence first.}

\item{exact}{boolean. Optional. Whether to only create users whose username
exactly matches the provided \code{prefix}.}
}
\value{
The results of creating the users.
}
\description{
The remote user creation workflow involves authentication providers like LDAP
that involve a queryable identity store. This helper wraps the API calls
necessary to retrieve information about and then create such a user. It
functions with a "fuzzy match" \code{prefix} by default, but if you want to
instantiate users directly, you should set \code{exact = TRUE}.
}
\details{
NOTE: there can be problems with usernames that are not unique. Please open
an issue if you run into any problems.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{Content}
\alias{Content}
\title{Content}
\description{
Content

Content
}
\details{
An R6 class that represents content.
}
\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{PositConnect}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{connect}}{An R6 Connect object.}

\item{\code{content}}{The content details from Posit Connect. Properties are described in \code{\link[=get_content]{get_content()}}.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{default_variant}}{The default variant for this object.}

\item{\code{is_rendered}}{TRUE if this is a rendered content type, otherwise FALSE.}

\item{\code{is_interactive}}{TRUE if this is a rendered content type, otherwise FALSE.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Content-new}{\code{Content$new()}}
\item \href{#method-Content-get_content_remote}{\code{Content$get_content_remote()}}
\item \href{#method-Content-get_bundles}{\code{Content$get_bundles()}}
\item \href{#method-Content-bundle_download}{\code{Content$bundle_download()}}
\item \href{#method-Content-bundle_delete}{\code{Content$bundle_delete()}}
\item \href{#method-Content-update}{\code{Content$update()}}
\item \href{#method-Content-danger_delete}{\code{Content$danger_delete()}}
\item \href{#method-Content-get_url}{\code{Content$get_url()}}
\item \href{#method-Content-get_dashboard_url}{\code{Content$get_dashboard_url()}}
\item \href{#method-Content-jobs}{\code{Content$jobs()}}
\item \href{#method-Content-register_job_kill_order}{\code{Content$register_job_kill_order()}}
\item \href{#method-Content-variants}{\code{Content$variants()}}
\item \href{#method-Content-tag_set}{\code{Content$tag_set()}}
\item \href{#method-Content-tag_delete}{\code{Content$tag_delete()}}
\item \href{#method-Content-tags}{\code{Content$tags()}}
\item \href{#method-Content-permissions_add}{\code{Content$permissions_add()}}
\item \href{#method-Content-permissions_update}{\code{Content$permissions_update()}}
\item \href{#method-Content-permissions_delete}{\code{Content$permissions_delete()}}
\item \href{#method-Content-permissions}{\code{Content$permissions()}}
\item \href{#method-Content-environment}{\code{Content$environment()}}
\item \href{#method-Content-environment_set}{\code{Content$environment_set()}}
\item \href{#method-Content-environment_all}{\code{Content$environment_all()}}
\item \href{#method-Content-deploy}{\code{Content$deploy()}}
\item \href{#method-Content-repository}{\code{Content$repository()}}
\item \href{#method-Content-repo_enable}{\code{Content$repo_enable()}}
\item \href{#method-Content-repo_set}{\code{Content$repo_set()}}
\item \href{#method-Content-packages}{\code{Content$packages()}}
\item \href{#method-Content-print}{\code{Content$print()}}
\item \href{#method-Content-clone}{\code{Content$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-new"></a>}}
\if{latex}{\out{\hypertarget{method-Content-new}{}}}
\subsection{Method \code{new()}}{
Initialize this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$new(connect, content)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{connect}}{The \code{Connect} instance.}

\item{\code{content}}{The content data.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-get_content_remote"></a>}}
\if{latex}{\out{\hypertarget{method-Content-get_content_remote}{}}}
\subsection{Method \code{get_content_remote()}}{
Obtain the content data from the Connect server.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$get_content_remote()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-get_bundles"></a>}}
\if{latex}{\out{\hypertarget{method-Content-get_bundles}{}}}
\subsection{Method \code{get_bundles()}}{
Return the set of content bundles.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$get_bundles()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-bundle_download"></a>}}
\if{latex}{\out{\hypertarget{method-Content-bundle_download}{}}}
\subsection{Method \code{bundle_download()}}{
Download the source archive for a content bundle.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$bundle_download(
  bundle_id,
  filename = tempfile(pattern = "bundle", fileext = ".tar.gz"),
  overwrite = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bundle_id}}{The bundle identifer.}

\item{\code{filename}}{Where to write the result.}

\item{\code{overwrite}}{Overwrite an existing filename.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-bundle_delete"></a>}}
\if{latex}{\out{\hypertarget{method-Content-bundle_delete}{}}}
\subsection{Method \code{bundle_delete()}}{
Delete a content bundle.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$bundle_delete(bundle_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bundle_id}}{The bundle identifer.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-update"></a>}}
\if{latex}{\out{\hypertarget{method-Content-update}{}}}
\subsection{Method \code{update()}}{
Update this content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$update(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Content fields.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-danger_delete"></a>}}
\if{latex}{\out{\hypertarget{method-Content-danger_delete}{}}}
\subsection{Method \code{danger_delete()}}{
Delete this content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$danger_delete()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-get_url"></a>}}
\if{latex}{\out{\hypertarget{method-Content-get_url}{}}}
\subsection{Method \code{get_url()}}{
Return the URL for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$get_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-get_dashboard_url"></a>}}
\if{latex}{\out{\hypertarget{method-Content-get_dashboard_url}{}}}
\subsection{Method \code{get_dashboard_url()}}{
Return the URL for this content in the Posit Connect dashboard.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$get_dashboard_url(pane = "")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{pane}}{The pane in the dashboard to link to.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-jobs"></a>}}
\if{latex}{\out{\hypertarget{method-Content-jobs}{}}}
\subsection{Method \code{jobs()}}{
Return the jobs for this content
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$jobs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-register_job_kill_order"></a>}}
\if{latex}{\out{\hypertarget{method-Content-register_job_kill_order}{}}}
\subsection{Method \code{register_job_kill_order()}}{
Terminate a single job for this content item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$register_job_kill_order(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The job key.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-variants"></a>}}
\if{latex}{\out{\hypertarget{method-Content-variants}{}}}
\subsection{Method \code{variants()}}{
Return the variants for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$variants()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-tag_set"></a>}}
\if{latex}{\out{\hypertarget{method-Content-tag_set}{}}}
\subsection{Method \code{tag_set()}}{
Set a tag for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$tag_set(tag_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tag_id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-tag_delete"></a>}}
\if{latex}{\out{\hypertarget{method-Content-tag_delete}{}}}
\subsection{Method \code{tag_delete()}}{
Remove a tag for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$tag_delete(tag_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tag_id}}{The tag identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-tags"></a>}}
\if{latex}{\out{\hypertarget{method-Content-tags}{}}}
\subsection{Method \code{tags()}}{
The tags for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$tags()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-permissions_add"></a>}}
\if{latex}{\out{\hypertarget{method-Content-permissions_add}{}}}
\subsection{Method \code{permissions_add()}}{
Add a principal to the ACL for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$permissions_add(principal_guid, principal_type, role)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{principal_guid}}{GUID for the target user or group.}

\item{\code{principal_type}}{Acting on user or group.}

\item{\code{role}}{The kind of content access.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-permissions_update"></a>}}
\if{latex}{\out{\hypertarget{method-Content-permissions_update}{}}}
\subsection{Method \code{permissions_update()}}{
Alter a principal in the ACL for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$permissions_update(id, principal_guid, principal_type, role)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{id}}{The target identifier.}

\item{\code{principal_guid}}{GUID for the target user or group.}

\item{\code{principal_type}}{Acting on user or group.}

\item{\code{role}}{The kind of content access.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-permissions_delete"></a>}}
\if{latex}{\out{\hypertarget{method-Content-permissions_delete}{}}}
\subsection{Method \code{permissions_delete()}}{
Remove an entry from the ACL for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$permissions_delete(id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{id}}{The target identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-permissions"></a>}}
\if{latex}{\out{\hypertarget{method-Content-permissions}{}}}
\subsection{Method \code{permissions()}}{
Obtain some or all of the ACL for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$permissions(id = NULL, add_owner = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{id}}{The target identifier.}

\item{\code{add_owner}}{Include the content owner in the result set.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-environment"></a>}}
\if{latex}{\out{\hypertarget{method-Content-environment}{}}}
\subsection{Method \code{environment()}}{
Return the environment variables set for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$environment()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-environment_set"></a>}}
\if{latex}{\out{\hypertarget{method-Content-environment_set}{}}}
\subsection{Method \code{environment_set()}}{
Adjust the environment variables set for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$environment_set(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Environment variable names and values. Use \code{NA} as the value
to unset variables.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-environment_all"></a>}}
\if{latex}{\out{\hypertarget{method-Content-environment_all}{}}}
\subsection{Method \code{environment_all()}}{
Overwrite the environment variables set for this content.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$environment_all(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Environment variable names and values.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-deploy"></a>}}
\if{latex}{\out{\hypertarget{method-Content-deploy}{}}}
\subsection{Method \code{deploy()}}{
Deploy this content
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$deploy(bundle_id = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bundle_id}}{Target bundle identifier.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-repository"></a>}}
\if{latex}{\out{\hypertarget{method-Content-repository}{}}}
\subsection{Method \code{repository()}}{
Get Git repository details
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$repository()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
NULL if no repo is set, otherwise a list with fields:
\itemize{
\item repository
\item branch
\item directory
\item polling
\item last_error
\item last_known_commit
}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-repo_enable"></a>}}
\if{latex}{\out{\hypertarget{method-Content-repo_enable}{}}}
\subsection{Method \code{repo_enable()}}{
Adjust Git polling.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$repo_enable(polling = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{polling}}{Polling enabled.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-repo_set"></a>}}
\if{latex}{\out{\hypertarget{method-Content-repo_set}{}}}
\subsection{Method \code{repo_set()}}{
Adjust Git repository
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$repo_set(repository, branch = "main", directory = ".", polling = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{repository}}{Git repository URL}

\item{\code{branch}}{Git repository branch}

\item{\code{directory}}{Git repository directory}

\item{\code{polling}}{Whether to check for updates}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-packages"></a>}}
\if{latex}{\out{\hypertarget{method-Content-packages}{}}}
\subsection{Method \code{packages()}}{
Get package dependencies
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$packages()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-print"></a>}}
\if{latex}{\out{\hypertarget{method-Content-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Content-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Content-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Content$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content_update}
\alias{content_update}
\alias{content_update_access_type}
\alias{content_update_owner}
\title{Update Content}
\usage{
content_update(content, ...)

content_update_access_type(content, access_type = c("all", "logged_in", "acl"))

content_update_owner(content, owner_guid)
}
\arguments{
\item{content}{An R6 content item}

\item{...}{Settings up update that are passed along to Posit Connect}

\item{access_type}{One of "all", "logged_in", or "acl"}

\item{owner_guid}{The GUID of a user who is a publisher, so that they can
become the new owner of the content}
}
\value{
An R6 content item
}
\description{
Update settings for a content item. For a list of all settings, see the
\href{https://docs.posit.co/connect/api/#patch-/v1/content/{guid}}{latest documentation}
or the documentation for your server via \code{connectapi::browse_api_docs()}.
}
\details{
Popular selections are \code{content_update(access_type="all")},
\code{content_update(access_type="logged_in")} or
\code{content_update(access_type="acl")}, process settings, title, description,
etc.
\itemize{
\item \code{content_update_access_type()} is a helper to make it easier to change access_type
\item \code{content_update_owner()} is a helper to make it easier to change owner
}
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audits.R
\name{vanity_is_available}
\alias{vanity_is_available}
\title{Check to see if a vanity URL is currently in use}
\usage{
vanity_is_available(connect, vanity)
}
\arguments{
\item{connect}{A Connect R6 object}

\item{vanity}{string of the vanity URL to check}
}
\value{
logical indicating if the vanity URL is available.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\seealso{
Other audit functions: 
\code{\link{audit_access_open}()},
\code{\link{audit_r_versions}()},
\code{\link{audit_runas}()}
}
\concept{audit functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content_restart}
\alias{content_restart}
\title{Restart a content item.}
\usage{
content_restart(content)
}
\arguments{
\item{content}{The content item you wish to restart.}
}
\description{
Submit a request to restart a content item. Once submitted, the
server performs an asynchronous request to kill all processes associated with
the content item, starting new processes as needed. This might be useful if
the application relies on data that is loaded at startup, or if its memory
usage has grown over time.

Note that users interacting with certain types of applications may have their
workflows interrupted.

Only valid for interactive content (e.g., applications, APIs).
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "8f37d6e0-3395-4a2c-aa6a-d7f2fe1babd0")
content_restart(item)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runtime-caches.R
\name{get_runtime_caches}
\alias{get_runtime_caches}
\title{Get runtime caches}
\usage{
get_runtime_caches(client)
}
\arguments{
\item{client}{A \code{Connect} object.}
}
\value{
A tibble of runtime caches on the server, showing \code{language},
\code{version} and \code{image_name}. For Connect servers not using off-host execution,
\code{image_name} is \code{"Local"}.
}
\description{
View the runtime caches on a Connect server.
Requires Administrator privileges.
}
\examples{
\dontrun{
client <- connect()
get_runtime_caches(client)
}

}
\seealso{
\code{\link[=delete_runtime_cache]{delete_runtime_cache()}}

Other server management functions: 
\code{\link{delete_runtime_cache}()}
}
\concept{server management functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groups.R
\name{get_group_content}
\alias{get_group_content}
\title{Get content access permissions for a group or groups}
\usage{
get_group_content(src, guids)
}
\arguments{
\item{src}{A Connect client object}

\item{guids}{A character vector of group guids}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{group_guid}: The group's GUID
\item \code{group_name}: The group's name
\item \code{content_guid}: The content item's GUID
\item \code{content_name}: The content item's name
\item \code{content_title}: The content item's title
\item \code{access_type}: The access type of the content item ("all", "logged_in", or "acl")
\item \code{role}: The access type that members of the group have to the
content item, "publisher" or "viewer".
}
}
\description{
Get content access permissions for a group or groups
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

# Get a data frame of groups
groups <- get_groups(client)

# Get permissions for a single group by passing in the corresponding row.
get_group_content(client, groups[1, "guid"])
dplyr::filter(groups, name = "research_scientists") \%>\%
  dplyr::pull(guid) \%>\%
  get_group_content(client, .)

# Get permissions for all groups by passing in all group guids.
get_group_content(client, groups$guid)
}

}
\seealso{
Other groups functions: 
\code{\link{get_group_members}()},
\code{\link{get_groups}()}
}
\concept{groups functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_usage_shiny}
\alias{get_usage_shiny}
\title{Get usage information for deployed shiny applications}
\usage{
get_usage_shiny(
  src,
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
)
}
\arguments{
\item{src}{the source object}

\item{content_guid}{Filter results by content GUID}

\item{min_data_version}{Filter by data version. Records with a data version
lower than the given value will be excluded from the set of results.}

\item{from}{The timestamp that starts the time window of interest. Any usage
information that ends prior to this timestamp will not be returned.
Individual records may contain a starting time that is before this if they
end after it or have not finished. Must be of class Date or POSIX}

\item{to}{The timestamp that ends the time window of interest. Any usage
information that starts after this timestamp will not be returned.
Individual records may contain an ending time that is after this
(or no ending time) if they start before it. Must be of class Date or
POSIX}

\item{limit}{The number of records to return.}

\item{previous}{Retrieve the previous page of Shiny application usage
logs relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{nxt}{Retrieve the next page of Shiny application usage logs
relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{asc_order}{Defaults to TRUE; Determines if the response records
should be listed in ascending or descending order within the response.
Ordering is by the started timestamp field.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{content_guid}: The GUID, in RFC4122 format, of the
Shiny application this information pertains to.
\item \code{user_guid}: The GUID, in RFC4122 format, of the user
that visited the application.
\item \code{started}: The timestamp, in RFC3339 format, when the
user opened the application.
\item \code{ended}: The timestamp, in RFC3339 format, when the
user left the application.
\item \code{data_version}: The data version the record was recorded
with. The Shiny Application Events section of the Posit Connect Admin
Guide explains how to interpret data_version values.
}
}
\description{
Get usage information for deployed shiny applications
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/instrumentation/shiny/usage
for more information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

from <- Sys.Date() - lubridate::days(5)
get_usage_shiny(client, limit = 20, from = from)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{set_vanity_url}
\alias{set_vanity_url}
\title{Set the Vanity URL}
\usage{
set_vanity_url(content, url, force = FALSE)
}
\arguments{
\item{content}{A Content object}

\item{url}{The path component of the URL}

\item{force}{optional. Default FALSE. Whether to force-reassign a vanity URL that already exists}
}
\value{
An updated Content object
}
\description{
Set the vanity URL for a piece of content.
}
\examples{
\dontrun{
bnd <- bundle_dir("~/my/directory")
connect() \%>\%
  deploy(bnd) \%>\%
  set_vanity_url("a/vanity/url")
}

}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schedule.R
\name{set_schedule}
\alias{set_schedule}
\alias{set_schedule_minute}
\alias{set_schedule_hour}
\alias{set_schedule_day}
\alias{set_schedule_weekday}
\alias{set_schedule_week}
\alias{set_schedule_dayofweek}
\alias{set_schedule_semimonth}
\alias{set_schedule_dayofmonth}
\alias{set_schedule_dayweekofmonth}
\alias{set_schedule_year}
\alias{set_schedule_remove}
\alias{schedule_describe}
\title{Set a Schedule}
\usage{
set_schedule(.schedule, ...)

set_schedule_minute(
  .schedule,
  n = 30,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_hour(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_day(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_weekday(
  .schedule,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_week(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_dayofweek(
  .schedule,
  days,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_semimonth(
  .schedule,
  first = TRUE,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_dayofmonth(
  .schedule,
  n = 1,
  day = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_dayweekofmonth(
  .schedule,
  n = 1,
  day = 1,
  week = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_year(
  .schedule,
  n = 1,
  start_time = Sys.time(),
  activate = TRUE,
  email = FALSE,
  timezone = Sys.timezone()
)

set_schedule_remove(.schedule)

schedule_describe(.schedule)
}
\arguments{
\item{.schedule}{A schedule object. As returned by \code{get_variant_schedule()}}

\item{...}{Scheduling parameters}

\item{n}{The "number of" iterations}

\item{start_time}{The start time of the schedule}

\item{activate}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This parameter no longer
has any effect due to changes in the Connect API and will be removed in a
future release.}

\item{email}{Whether to send emails on this schedule}

\item{timezone}{The timezone to use for setting the schedule. Defaults to \code{Sys.timezone()}}

\item{days}{The days of the week (0-6)}

\item{first}{\link{logical} Whether to execute on the 1st and 15th (TRUE) or 14th and last (FALSE)}

\item{day}{The day of the week (0-6) or day of the month (0-31)}

\item{week}{The week of the month (0-5)}

\item{schedule}{A JSON blob (as a string) describing the schedule. See "More Details"}
}
\value{
An updated Schedule object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Sets the schedule for a given Variant. Requires a
\code{Schedule} object (as returned by \code{get_variant_schedule()})
}
\details{
\itemize{
\item \code{set_schedule()} is a raw interface to Posit Connect's \code{schedule} API
\item \verb{set_schedule_*()} functions provide handy wrappers around \code{set_schedule()}
\item \code{set_schedule_remove()} removes a schedule / un-schedules a variant
}

Beware, using \code{set_schedule()} currently uses the Posit Connect \code{schedule} API
directly, and so can be a little clunky. Using the \verb{set_schedule_*()} is generally
recommended.
}
\seealso{
Other schedule functions: 
\code{\link{get_timezones}()},
\code{\link{get_variant_schedule}()}
}
\concept{schedule functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_aws_content_credentials}
\alias{get_aws_content_credentials}
\title{Obtain AWS credentials for your content.}
\usage{
get_aws_content_credentials(
  connect,
  content_session_token = NULL,
  audience = NULL
)
}
\arguments{
\item{connect}{A Connect R6 object.}

\item{content_session_token}{Optional. The content session token. This token
can only be obtained when the content is running on a Connect server. The
token identifies the service account integration previously configured by
the publisher on the Connect server. Defaults to the value from the
environment variable: \code{CONNECT_CONTENT_SESSION_TOKEN}}

\item{audience}{Optional. The GUID of an OAuth integration associated with
this piece of content.}
}
\value{
The AWS credentials as a list with fields named \code{access_key_id},
\code{secret_access_key}, \code{session_token}, and \code{expiration}.
}
\description{
Obtain AWS credentials for your content.
}
\details{
Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-service-account-aws-credentials
for more information. See the example below of using this function with
\code{paws} to access S3. Any library that allows you to pass AWS credentials
will be able to utilize the credentials returned from this function call.
}
\examples{
\dontrun{
library(connectapi)
library(paws)

client <- connect()
# Pulls the content session token from the environment
# when deployed into Connect.
aws_credentials <- get_aws_content_credentials(client)

# Create S3 client with AWS credentials from Connect
svc <- paws::s3(
  credentials = list(
    creds = list(
      access_key_id = aws_credentials$access_key_id,
      secret_access_key = aws_credentials$secret_access_key,
      session_token = aws_credentials$session_token
    )
  )
)

# Get object from S3
obj <- svc$get_object(
  Bucket = "my-bucket",
  Key = "my-data.csv"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{get_log}
\alias{get_log}
\title{Get Job Log}
\usage{
get_log(job, max_log_lines = NULL)
}
\arguments{
\item{job}{A job, represented by an element from the list returned by \code{get_job_list()}.}

\item{max_log_lines}{Optional. An integer indicating the maximum number of
log lines to return. If \code{NULL} (default), Connect returns a maximum of 5000
lines.}
}
\value{
A data frame with the requested log. Each row represents an entry.
\itemize{
\item \code{source}: \code{stdout} or \code{stderr}
\item \code{timestamp}: The time of the entry.
\item \code{data}: The logged text.
}
}
\description{
Get the log output for a job. Requires Connect 2022.10.0 or newer.
}
\details{
Note: The output of \code{get_jobs()} cannot be used with \code{get_log()}.
Please use an object from the list returned by \code{get_job_list()}.
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
jobs <- get_job_list(item)
log <- get_log(jobs[[1]])
}


}
\seealso{
Other job functions: 
\code{\link{get_jobs}()},
\code{\link{terminate_jobs}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{job functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variant.R
\name{get_variant_renderings}
\alias{get_variant_renderings}
\alias{variant_render}
\title{Render a Variant}
\usage{
get_variant_renderings(variant)

variant_render(variant)
}
\arguments{
\item{variant}{An R6 Variant object. As returned by \code{get_variant()} or \code{get_variant_default()}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Get details about renderings (i.e. render history)
or execute a variant on demand
}
\details{
\itemize{
\item \code{get_variant_renderings()} returns all renderings / content for a particular variant. Returns a \code{tibble}
\item \code{variant_render()} executes a variant on demand. Returns a \code{VariantTask} object
}
}
\seealso{
Other variant functions: 
\code{\link{get_variants}()}
}
\concept{variant functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{swap_vanity_urls}
\alias{swap_vanity_urls}
\title{Swap Vanity URLs}
\usage{
swap_vanity_urls(content_a, content_b)
}
\arguments{
\item{content_a}{A Content object}

\item{content_b}{A Content object}
}
\value{
A list of the new vanity URLs for \code{content_a} and \code{content_b}
}
\description{
Swap the vanity URLs of two pieces of content.
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_vanity_urls}
\alias{get_vanity_urls}
\title{Get all vanity URLs}
\usage{
get_vanity_urls(client)
}
\arguments{
\item{client}{A \code{Connect} object.}
}
\value{
A tibble with columns for \code{content_guid}, \code{path}, and
\code{created_time}.
}
\description{
Get a table of all vanity URLs on the server. Requires administrator
privileges.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()
get_vanity_urls(client)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{as.data.frame.connect_integration_list}
\alias{as.data.frame.connect_integration_list}
\title{Convert integrations list to a data frame}
\usage{
\method{as.data.frame}{connect_integration_list}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{A \code{connect_integration_list} object (from \code{\link[=get_integrations]{get_integrations()}}).}

\item{row.names}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}

\item{optional}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}

\item{...}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}
}
\value{
A \code{data.frame} with one row per integration.
}
\description{
Converts a list returned by \code{\link[=get_integrations]{get_integrations()}} into a data frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_oauth_credentials}
\alias{get_oauth_credentials}
\title{Perform an OAuth credential exchange to obtain a visitor's OAuth access token.}
\usage{
get_oauth_credentials(
  connect,
  user_session_token,
  requested_token_type = NULL,
  audience = NULL
)
}
\arguments{
\item{connect}{A Connect R6 object.}

\item{user_session_token}{The content visitor's session token. This token
can only be obtained when the content is running on a Connect server. The token
identifies the user who is viewing the content interactively on the Connect server.
Read this value from the HTTP header: \code{Posit-Connect-User-Session-Token}}

\item{requested_token_type}{Optional. The requested token type. If unset, will
default to \code{urn:ietf:params:oauth:token-type:access_token}. Otherwise, this can
be set to \code{urn:ietf:params:aws:token-type:credentials} for AWS integrations or
\code{urn:posit:connect:api-key} for Connect API Key integrations.}

\item{audience}{Optional. The GUID of an OAuth integration associated with
this piece of content.}
}
\value{
The OAuth credential exchange response.
}
\description{
Perform an OAuth credential exchange to obtain a visitor's OAuth access token.
}
\details{
Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-a-viewer-oauth-access-token
for more information.
}
\examples{
\dontrun{
library(connectapi)
library(plumber)
client <- connect()

#* @get /do
function(req) {
  user_session_token <- req$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
  credentials <- get_oauth_credentials(client, user_session_token)

  # ... do something with `credentials$access_token` ...

  "done"
}
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_oauth_content_credentials]{get_oauth_content_credentials()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{get_tags}
\alias{get_tags}
\alias{get_tag_data}
\alias{create_tag}
\alias{create_tag_tree}
\alias{delete_tag}
\alias{get_content_tags}
\alias{set_content_tag_tree}
\alias{set_content_tags}
\alias{filter_tag_tree_id}
\alias{filter_tag_tree_chr}
\title{Get all Tags on the server}
\usage{
get_tags(src)

get_tag_data(src)

create_tag(src, name, parent = NULL)

create_tag_tree(src, ...)

delete_tag(src, tag)

get_content_tags(content)

set_content_tag_tree(content, ...)

set_content_tags(content, ...)

filter_tag_tree_id(tags, ids)

filter_tag_tree_chr(tags, pattern)
}
\arguments{
\item{src}{The source object}

\item{name}{The name of the tag to create}

\item{parent}{optional. A \code{connect_tag_tree} object (as returned by \code{get_tags()}) pointed at the parent tag}

\item{...}{Additional arguments

Manage tags (requires Administrator role):
\itemize{
\item \code{get_tags()} - returns a "tag tree" object that can be traversed with \code{tag_tree$tag1$childtag}
\item \code{get_tag_data()} - returns a tibble of tag data
\item \code{create_tag()} - create a tag by specifying the Parent directly
\item \code{create_tag_tree()} - create tag(s) by specifying the "desired" tag tree
hierarchy
\item \code{delete_tag()} - delete a tag (and its children). WARNING: will
disassociate any content automatically
}

Manage content tags:
\itemize{
\item \code{get_content_tags()} - return a \code{connect_tag_tree} object corresponding to
the tags for a piece of content.
\item \code{set_content_tag_tree()} - attach a tag to content by specifying the
desired tag tree
\item \code{set_content_tags()} - Set multiple tags at once by providing
\code{connect_tag_tree} objects
}

Search a tag tree:
\itemize{
\item \code{filter_tag_tree_chr()} - filters a tag tree based on a regex
\item \code{filter_tag_tree_id()} - filters a tag tree based on an id
}}

\item{tag}{A \code{connect_tag_tree} object (as returned by \code{get_tags()})}

\item{content}{An R6 Content object, as returned by \code{content_item()}}

\item{tags}{A \code{connect_tag_tree} object (as returned by \code{get_tags()})}

\item{ids}{A list of \code{id}s to filter the tag tree by}

\item{pattern}{A regex to filter the tag tree by (it is passed to \code{grepl})}
}
\description{
Tag manipulation and assignment functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{groups_create_remote}
\alias{groups_create_remote}
\title{Create a Remote Group}
\usage{
groups_create_remote(connect, prefix, expect = 1, check = TRUE, exact = FALSE)
}
\arguments{
\item{connect}{An R6 Connect object.}

\item{prefix}{character. The prefix of the user name to search for.}

\item{expect}{number. Optional. The number of responses to expect for this search.}

\item{check}{boolean. Optional. Whether to check for local existence first.}

\item{exact}{boolean. Optional. Whether to only create groups whose name
exactly matches the provided \code{prefix}.}
}
\value{
The results of creating the groups.
}
\description{
Create a Remote Group
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{get_jobs}
\alias{get_jobs}
\alias{get_job_list}
\title{Get Jobs}
\usage{
get_jobs(content)

get_job_list(content)
}
\arguments{
\item{content}{A Content object, as returned by \code{content_item()}}
}
\value{
\itemize{
\item \code{get_jobs()}: A data frame with a row representing each job.
\item \code{get_job_list()}: A list with each element representing a job.
}

Jobs contain the following fields:
\itemize{
\item \code{id}: The job identifier.
\item \code{ppid}: The job's parent process identifier (see Note 1).
\item \code{pid}: The job's process identifier.
\item \code{key}: The job's unique key identifier.
\item \code{remote_id}: The job's identifier for off-host execution configurations
(see Note 1).
\item \code{app_id}: The job's parent content identifier; deprecated in favor of \code{content_id}.
\item \code{app_guid}: The job's parent content GUID; deprecated in favor of \code{content_guid}.
\item \code{content_id}: The job's parent content identifier.
\item \code{content_guid}: The job's parent content GUID.
\item \code{variant_id}: The identifier of the variant owning this job.
\item \code{bundle_id}: The identifier of a content bundle linked to this job.
\item \code{start_time}: The timestamp (RFC3339) indicating when this job started.
\item \code{end_time}: The timestamp (RFC3339) indicating when this job finished.
\item \code{last_heartbeat_time}: The timestamp (RFC3339) indicating the last time
this job was observed to be running (see Note 1).
\item \code{queued_time}: The timestamp (RFC3339) indicating when this job was added
to the queue to be processed. Only scheduled reports will present a value
for this field (see Note 1).
\item \code{queue_name}: The name of the queue which processes the job. Only
scheduled reports will present a value for this field (see Note 1).
\item \code{tag}: A tag to identify the nature of the job.
\item \code{exit_code}: The job's exit code. Present only when job is finished.
\item \code{status}: The current status of the job. On Connect 2022.10.0 and newer,
one of Active: 0, Finished: 1, Finalized: 2; on earlier versions, Active:
0, otherwise \code{NA}.
\item \code{hostname}: The name of the node which processes the job.
\item \code{cluster}: The location where this content runs. Content running on the
same server as Connect will have either a null value or the string Local.
Gives the name of the cluster when run external to the Connect host
(see Note 1).
\item \code{image}: The location where this content runs. Content running on
the same server as Connect will have either a null value or the string
Local. References the name of the target image when content runs in
a clustered environment such as Kubernetes (see Note 1).
\item \code{run_as}: The UNIX user that executed this job.
}
}
\description{
Retrieve details about server processes associated with a \code{content_item},
such as a FastAPI app or a Quarto render.
}
\details{
Note that Connect versions below 2022.10.0 use a legacy endpoint, and will
not return the complete set of information provided by newer versions.

\code{get_jobs()} returns job data as a data frame, whereas \code{get_jobs_list()}
returns job data in a list.

You might get job data as a data frame if you want to perform some
calculations about job data (e.g. counting server processes over time), or if
you want to filter jobs to find a specific key.

The objects in list returned by \code{get_jobs_list()} are useful if you want to
take an action on a job, such as getting its process log with
\code{get_log()}.
}
\note{
\enumerate{
\item On Connect instances earlier than 2022.10.0, these fields will contain \code{NA} values.
}
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
jobs <- get_jobs(item)
job_list <- get_job_list(item)
}

}
\seealso{
Other job functions: 
\code{\link{get_log}()},
\code{\link{terminate_jobs}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{job functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{dashboard_url}
\alias{dashboard_url}
\title{Build a Dashboard URL from a Content Item}
\usage{
dashboard_url(content, pane = "")
}
\arguments{
\item{content}{\link{Content} A Content object}

\item{pane}{character The pane in the dashboard to link to}
}
\value{
character The dashboard URL for the content provided
}
\description{
Returns the URL for the content dashboard (opened to the selected pane).
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variant.R
\name{VariantTask}
\alias{VariantTask}
\title{VariantTask}
\description{
VariantTask

VariantTask
}
\details{
An R6 class that represents a Variant Task
}
\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{PositConnect}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}}
}
\concept{R6 classes}
\section{Super classes}{
\code{\link[connectapi:Content]{connectapi::Content}} -> \code{\link[connectapi:Variant]{connectapi::Variant}} -> \code{VariantTask}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{task}}{The task.}

\item{\code{data}}{The variant data.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-VariantTask-new}{\code{VariantTask$new()}}
\item \href{#method-VariantTask-get_task}{\code{VariantTask$get_task()}}
\item \href{#method-VariantTask-add_data}{\code{VariantTask$add_data()}}
\item \href{#method-VariantTask-get_data}{\code{VariantTask$get_data()}}
\item \href{#method-VariantTask-print}{\code{VariantTask$print()}}
\item \href{#method-VariantTask-clone}{\code{VariantTask$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_delete"><a href='../../connectapi/html/Content.html#method-Content-bundle_delete'><code>connectapi::Content$bundle_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="bundle_download"><a href='../../connectapi/html/Content.html#method-Content-bundle_download'><code>connectapi::Content$bundle_download()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="danger_delete"><a href='../../connectapi/html/Content.html#method-Content-danger_delete'><code>connectapi::Content$danger_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="deploy"><a href='../../connectapi/html/Content.html#method-Content-deploy'><code>connectapi::Content$deploy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment"><a href='../../connectapi/html/Content.html#method-Content-environment'><code>connectapi::Content$environment()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment_all"><a href='../../connectapi/html/Content.html#method-Content-environment_all'><code>connectapi::Content$environment_all()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="environment_set"><a href='../../connectapi/html/Content.html#method-Content-environment_set'><code>connectapi::Content$environment_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_bundles"><a href='../../connectapi/html/Content.html#method-Content-get_bundles'><code>connectapi::Content$get_bundles()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="get_content_remote"><a href='../../connectapi/html/Content.html#method-Content-get_content_remote'><code>connectapi::Content$get_content_remote()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="packages"><a href='../../connectapi/html/Content.html#method-Content-packages'><code>connectapi::Content$packages()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions"><a href='../../connectapi/html/Content.html#method-Content-permissions'><code>connectapi::Content$permissions()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_add"><a href='../../connectapi/html/Content.html#method-Content-permissions_add'><code>connectapi::Content$permissions_add()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_delete"><a href='../../connectapi/html/Content.html#method-Content-permissions_delete'><code>connectapi::Content$permissions_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="permissions_update"><a href='../../connectapi/html/Content.html#method-Content-permissions_update'><code>connectapi::Content$permissions_update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="register_job_kill_order"><a href='../../connectapi/html/Content.html#method-Content-register_job_kill_order'><code>connectapi::Content$register_job_kill_order()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_enable"><a href='../../connectapi/html/Content.html#method-Content-repo_enable'><code>connectapi::Content$repo_enable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repo_set"><a href='../../connectapi/html/Content.html#method-Content-repo_set'><code>connectapi::Content$repo_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="repository"><a href='../../connectapi/html/Content.html#method-Content-repository'><code>connectapi::Content$repository()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_delete"><a href='../../connectapi/html/Content.html#method-Content-tag_delete'><code>connectapi::Content$tag_delete()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tag_set"><a href='../../connectapi/html/Content.html#method-Content-tag_set'><code>connectapi::Content$tag_set()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="tags"><a href='../../connectapi/html/Content.html#method-Content-tags'><code>connectapi::Content$tags()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="update"><a href='../../connectapi/html/Content.html#method-Content-update'><code>connectapi::Content$update()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Content" data-id="variants"><a href='../../connectapi/html/Content.html#method-Content-variants'><code>connectapi::Content$variants()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="add_subscribers"><a href='../../connectapi/html/Variant.html#method-Variant-add_subscribers'><code>connectapi::Variant$add_subscribers()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_dashboard_url"><a href='../../connectapi/html/Variant.html#method-Variant-get_dashboard_url'><code>connectapi::Variant$get_dashboard_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_schedule"><a href='../../connectapi/html/Variant.html#method-Variant-get_schedule'><code>connectapi::Variant$get_schedule()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_schedule_remote"><a href='../../connectapi/html/Variant.html#method-Variant-get_schedule_remote'><code>connectapi::Variant$get_schedule_remote()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_subscribers"><a href='../../connectapi/html/Variant.html#method-Variant-get_subscribers'><code>connectapi::Variant$get_subscribers()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_url"><a href='../../connectapi/html/Variant.html#method-Variant-get_url'><code>connectapi::Variant$get_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_url_rev"><a href='../../connectapi/html/Variant.html#method-Variant-get_url_rev'><code>connectapi::Variant$get_url_rev()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="get_variant_remote"><a href='../../connectapi/html/Variant.html#method-Variant-get_variant_remote'><code>connectapi::Variant$get_variant_remote()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="jobs"><a href='../../connectapi/html/Variant.html#method-Variant-jobs'><code>connectapi::Variant$jobs()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="remove_subscriber"><a href='../../connectapi/html/Variant.html#method-Variant-remove_subscriber'><code>connectapi::Variant$remove_subscriber()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="render"><a href='../../connectapi/html/Variant.html#method-Variant-render'><code>connectapi::Variant$render()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="renderings"><a href='../../connectapi/html/Variant.html#method-Variant-renderings'><code>connectapi::Variant$renderings()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="send_mail"><a href='../../connectapi/html/Variant.html#method-Variant-send_mail'><code>connectapi::Variant$send_mail()</code></a></span></li>
<li><span class="pkg-link" data-pkg="connectapi" data-topic="Variant" data-id="update_variant"><a href='../../connectapi/html/Variant.html#method-Variant-update_variant'><code>connectapi::Variant$update_variant()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-new"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-new}{}}}
\subsection{Method \code{new()}}{
Initialize this variant task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$new(connect, content, key, task)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{connect}}{The \code{Connect} instance.}

\item{\code{content}}{The \code{Content} instance.}

\item{\code{key}}{The variant key.}

\item{\code{task}}{The task data.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-get_task"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-get_task}{}}}
\subsection{Method \code{get_task()}}{
Return the underlying task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$get_task()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-add_data"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-add_data}{}}}
\subsection{Method \code{add_data()}}{
Set the data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$add_data(data)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{The data.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-get_data"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-get_data}{}}}
\subsection{Method \code{get_data()}}{
Get the data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$get_data()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-print"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VariantTask-clone"></a>}}
\if{latex}{\out{\hypertarget{method-VariantTask-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VariantTask$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{create_integration}
\alias{create_integration}
\title{Create an OAuth integration}
\usage{
create_integration(client, name, description = NULL, template, config)
}
\arguments{
\item{client}{A \code{Connect} R6 client object.}

\item{name}{A descriptive name to identify the integration.}

\item{description}{Optional, default \code{NULL.} A brief description of the integration.}

\item{template}{The template to use to configure this integration (e.g.,
"custom", "github", "google", "connect").}

\item{config}{A list containing the configuration for the integration. The
required fields vary depending on the template selected.}
}
\value{
A \code{connect_integration} object representing the newly created
integration. See \code{\link[=get_integration]{get_integration()}} for details on the returned object.
}
\description{
Creates a new OAuth integration on the Posit Connect server. OAuth integrations
allow content to access external resources using OAuth credentials.

You must have administrator privileges to perform this action.

See the Posit Connect documentation on
\href{https://docs.posit.co/connect/admin/integrations/oauth-integrations/}{OAuth integrations} for
more information.
}
\examples{
\dontrun{
client <- connect()

# Create a GitHub OAuth integration
github_integration <- create_integration(
  client,
  name = "GitHub Integration",
  description = "Integration with GitHub for OAuth access",
  template = "github",
  config = list(
    client_id = "your-client-id",
    client_secret = "your-client-secret"
  )
)

# Create a custom OAuth integration
custom_integration <- create_integration(
  client,
  name = "Custom API Integration",
  description = "Integration with our custom API",
  template = "custom",
  config = list(
    auth_mode = "Confidential",
    auth_type = "Viewer",
    authorization_uri = "https://api.example.com/oauth/authorize",
    client_id = "your-client-id",
    client_secret = "your-client-secret",
    token_uri = "https://api.example.com/oauth/token"
  )
)
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_integration]{get_integration()}}, \code{\link[=update_integration]{update_integration()}},
\code{\link[=delete_integration]{delete_integration()}}

Other oauth integration functions: 
\code{\link{delete_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integration}()},
\code{\link{get_integrations}()},
\code{\link{set_integrations}()},
\code{\link{update_integration}()}
}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content_title}
\alias{content_title}
\title{Get Content Title}
\usage{
content_title(connect, guid, default = "Unknown Content")
}
\arguments{
\item{connect}{A Connect object}

\item{guid}{The GUID for the content item to be retrieved}

\item{default}{The default value returned for missing or not visible content}
}
\value{
character. The title of the requested content
}
\description{
Return content title for a piece of content. If the content
is missing (deleted) or not visible, then returns the \code{default}
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{Bundle}
\alias{Bundle}
\title{Bundle}
\description{
Bundle

Bundle
}
\details{
An R6 class that represents a bundle
}
\seealso{
Other R6 classes: 
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{PositConnect}},
\code{\link{Task}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{path}}{The bundle path on disk.}

\item{\code{size}}{The size of the bundle.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Bundle-new}{\code{Bundle$new()}}
\item \href{#method-Bundle-print}{\code{Bundle$print()}}
\item \href{#method-Bundle-clone}{\code{Bundle$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Bundle-new"></a>}}
\if{latex}{\out{\hypertarget{method-Bundle-new}{}}}
\subsection{Method \code{new()}}{
Initialize this content bundle.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Bundle$new(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{The bundle path on disk.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Bundle-print"></a>}}
\if{latex}{\out{\hypertarget{method-Bundle-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Bundle$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Bundle-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Bundle-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Bundle$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{get_environment}
\alias{get_environment}
\alias{set_environment_new}
\alias{set_environment_remove}
\alias{set_environment_all}
\title{Manage Environment Variables}
\usage{
get_environment(content)

set_environment_new(env, ...)

set_environment_remove(env, ...)

set_environment_all(env, ...)
}
\arguments{
\item{content}{An R6 Content object as returned by \code{content_item()}}

\item{env}{An R6 Environment object as returned by \code{get_environment()}}

\item{...}{name = value pairs of environment variable names and values}
}
\description{
Manage Environment Variables for a piece of content.
}
\details{
\code{get_environment()} returns an Environment object for use with "setter" methods

\code{set_environment_new()} updates environment values (either creating new
values or updating existing). Set \code{NA} as the value to remove a variable.

\code{set_environment_remove()} is a wrapper on \code{set_environment_new()} that
allows removing named / listed variables quickly

\code{set_environment_all()} sets \emph{all} environment variable values (will remove
variables not specified)
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variant.R
\name{get_variants}
\alias{get_variants}
\alias{get_variant}
\alias{get_variant_default}
\title{Get Variant}
\usage{
get_variants(content)

get_variant(content, key)

get_variant_default(content)
}
\arguments{
\item{content}{An R6 Content object. Returned from \code{content_item()}}

\item{key}{The Variant key for a specific variant}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Work with variants
}
\details{
\itemize{
\item \code{get_variants()} returns a \code{tibble} with variant data for a \code{content_item}
\item \code{get_variant_default()} returns the default variant for a \code{content_item}
\item \code{get_variant()} returns a specific variant for a \code{content_item} (specified by \code{key})
}
}
\seealso{
Other variant functions: 
\code{\link{get_variant_renderings}()}

Other variant functions: 
\code{\link{get_variant_renderings}()}

Other variant functions: 
\code{\link{get_variant_renderings}()}
}
\concept{variant functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{download_bundle}
\alias{download_bundle}
\title{Download a Bundle from Deployed Connect Content}
\usage{
download_bundle(
  content,
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz"),
  bundle_id = NULL,
  overwrite = FALSE
)
}
\arguments{
\item{content}{A Content object}

\item{filename}{Optional. The output bundle path}

\item{bundle_id}{Optional. A string representing the bundle_id to download.
If NULL, will use the currently active bundle.}

\item{overwrite}{Optional. Default FALSE. Whether to overwrite the target location if it already exists}
}
\value{
Bundle A bundle object
}
\description{
Downloads a Content item's active bundle, or (optionally) one of its other bundles.
}
\seealso{
Other deployment functions: 
\code{\link{bundle_dir}()},
\code{\link{bundle_path}()},
\code{\link{bundle_static}()},
\code{\link{deploy}()},
\code{\link{poll_task}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_usage_static}
\alias{get_usage_static}
\title{Get usage information from deployed static content}
\usage{
get_usage_static(
  src,
  content_guid = NULL,
  min_data_version = NULL,
  from = NULL,
  to = NULL,
  limit = 500,
  previous = NULL,
  nxt = NULL,
  asc_order = TRUE
)
}
\arguments{
\item{src}{the source object}

\item{content_guid}{Filter results by content GUID}

\item{min_data_version}{Filter by data version. Records with a data version
lower than the given value will be excluded from the set of results.}

\item{from}{The timestamp that starts the time window of interest. Any usage
information that ends prior to this timestamp will not be returned.
Individual records may contain a starting time that is before this if they
end after it or have not finished. Must be of class Date or POSIX}

\item{to}{The timestamp that ends the time window of interest. Any usage
information that starts after this timestamp will not be returned.
Individual records may contain an ending time that is after this
(or no ending time) if they start before it. Must be of class Date or
POSIX}

\item{limit}{The number of records to return.}

\item{previous}{Retrieve the previous page of Shiny application usage
logs relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{nxt}{Retrieve the next page of Shiny application usage logs
relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{asc_order}{Defaults to TRUE; Determines if the response records
should be listed in ascending or descending order within the response.
Ordering is by the started timestamp field.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{content_guid}: The GUID, in RFC4122 format, of the Shiny
application this information pertains to.
\item \code{user_guid}: The GUID, in RFC4122 format, of the user that
visited the application.
\item \code{variant_key}: The key of the variant the user visited.
This will be null for static content.
\item \code{time}: The timestamp, in RFC3339 format, when the user
visited the content.
\item \code{rendering_id}: The ID of the rendering the user visited.
This will be null for static content.
\item \code{bundle_id}: The ID of the particular bundle used.
\item \code{data_version}: The data version the record was recorded
with. The Rendered and Static Content Visit Events section of the
Posit Connect Admin Guide explains how to interpret data_version
values.
}
}
\description{
This function retrieves usage information from static content
on the Posit Connect server (e.g. Rmarkdown, Jupyter Notebooks)
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/instrumentation/content/visits
for more information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

from <- Sys.Date() - lubridate::days(5)
get_usage_static(client, limit = 20, from = from)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_cursor}
\alias{page_cursor}
\alias{page_offset}
\title{Paging}
\usage{
page_cursor(client, req, limit = Inf)

page_offset(client, req, limit = Inf)
}
\arguments{
\item{client}{A Connect client object}

\item{req}{For page_cursor, the output from an initial response to an API
endpoint that uses cursor-based pagination. For page_offset, a request that
needs to be paged.}

\item{limit}{A row limit}
}
\value{
The aggregated results from all requests
}
\description{
Helper functions that make paging easier in
the Posit Connect Server API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schedule.R
\name{get_timezones}
\alias{get_timezones}
\title{Get TimeZones}
\usage{
get_timezones(connect)
}
\arguments{
\item{connect}{An R6 Connect object}
}
\value{
A TimeZone vector to be used for setting time zones
}
\description{
Get the available timezones from the server.
}
\seealso{
Other schedule functions: 
\code{\link{get_variant_schedule}()},
\code{\link{set_schedule}()}
}
\concept{schedule functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_runtimes}
\alias{get_runtimes}
\title{Get available runtimes on server}
\usage{
get_runtimes(client, runtimes = NULL)
}
\arguments{
\item{client}{A \code{Connect} object.}

\item{runtimes}{Optional. A character vector of runtimes to include. Must be
some combination of \code{"r"}, \code{"python"}, \code{"quarto"}, and \code{"tensorflow"}. Quarto
is only supported on Connect >= 2021.08.0, and Tensorflow is only supported
on Connect >= 2024.03.0.}
}
\value{
A tibble with columns for \code{runtime}, \code{version}, and \code{cluster_name}
and \code{image_name}. Cluster name and image name are only meaningful on Connect
instances running off-host execution.
}
\description{
Get a table showing available versions of R, Python, Quarto, and Tensorflow
on the Connect server.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()
get_runtimes(client, runtimes = c("r", "python", "tensorflow"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connectapi-package.R, R/connectapi.R
\docType{package}
\name{connectapi-package}
\alias{connectapi-package}
\title{connectapi: Utilities for Interacting with the 'Posit Connect' Server API}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a helpful 'R6' class and methods for interacting with the 'Posit Connect' Server API along with some meaningful utility functions for regular tasks. API documentation varies by 'Posit Connect' installation and version, but the latest documentation is also hosted publicly at \url{https://docs.posit.co/connect/api/}.

\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a helpful 'R6' class and methods for interacting with the 'Posit Connect' Server API along with some meaningful utility functions for regular tasks. API documentation varies by 'Posit Connect' installation and version, but the latest documentation is also hosted publicly at \url{https://docs.posit.co/connect/api/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://posit-dev.github.io/connectapi/}
  \item \url{https://github.com/posit-dev/connectapi}
  \item Report bugs at \url{https://github.com/posit-dev/connectapi/issues}
}


Useful links:
\itemize{
  \item \url{https://posit-dev.github.io/connectapi/}
  \item \url{https://github.com/posit-dev/connectapi}
  \item Report bugs at \url{https://github.com/posit-dev/connectapi/issues}
}

}
\author{
\strong{Maintainer}: Kara Woo \email{kara.woo@posit.co}

Authors:
\itemize{
  \item Toph Allen \email{toph@posit.co}
  \item Neal Richardson
  \item Sean Lopp
  \item Cole Arendt
}

Other contributors:
\itemize{
  \item Posit, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{get_bundles}
\alias{get_bundles}
\alias{delete_bundle}
\title{Get Bundles}
\usage{
get_bundles(content)

delete_bundle(content, bundle_id)
}
\arguments{
\item{content}{A R6 Content item, as returned by \code{content_item()}}

\item{bundle_id}{A specific bundle ID for a content item}
}
\description{
Lists bundles for a content item
}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{as_tibble.connect_list_hits}
\alias{as_tibble.connect_list_hits}
\title{Convert usage data to a tibble}
\usage{
\method{as_tibble}{connect_list_hits}(x, ...)
}
\arguments{
\item{x}{A \code{connect_list_hits} object.}

\item{...}{Passed to \code{\link[=as.data.frame]{as.data.frame()}}.}
}
\value{
A tibble with one row per usage record.
}
\description{
Converts an object returned by \code{\link[=get_usage]{get_usage()}} to a tibble via
\code{\link[=as.data.frame.connect_list_hits]{as.data.frame.connect_list_hits()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{as.data.frame.connect_list_hits}
\alias{as.data.frame.connect_list_hits}
\title{Convert usage data to a data frame}
\usage{
\method{as.data.frame}{connect_list_hits}(x, row.names = NULL, optional = FALSE, ..., unnest = TRUE)
}
\arguments{
\item{x}{A \code{connect_list_hits} object (from \code{\link[=get_usage]{get_usage()}}).}

\item{row.names}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}

\item{optional}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}

\item{...}{Passed to \code{\link[base:as.data.frame]{base::as.data.frame()}}.}

\item{unnest}{Logical; if \code{TRUE} (default), extracts nested fields using
\pkg{tidyr}. Set to \code{FALSE} to skip unnesting.}
}
\value{
A \code{data.frame} with one row per usage record.
}
\description{
Converts an object returned by \code{\link[=get_usage]{get_usage()}} into a data frame with parsed
column types. By default, extracts \code{path} and \code{user_agent} from the \code{data}
field, if available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groups.R
\name{get_group_members}
\alias{get_group_members}
\title{Get users within a specific group}
\usage{
get_group_members(src, guid)
}
\arguments{
\item{src}{A Connect client object}

\item{guid}{A group GUID identifier}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{email}: The user's email
\item \code{username}: The user's username
\item \code{first_name}: The user's first name
\item \code{last_name}: The user's last name
\item \code{user_role}: The user's role. It may have a value of administrator,
publisher or viewer.
\item \code{created_time}: The timestamp (in RFC3339 format) when the user
was created in the Posit Connect server
\item \code{updated_time}: The timestamp (in RFC3339 format) when the user
was last updated in the Posit Connect server
\item \code{active_time}: The timestamp (in RFC3339 format) when the user
was last active on the Posit Connect server
\item \code{confirmed}: When false, the created user must confirm their
account through an email. This feature is unique to password
authentication.
\item \code{locked}: Whether or not the user is locked
\item \code{guid}: The user's GUID, or unique identifier, in UUID RFC4122 format
}
}
\description{
Get users within a specific group
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/groups/-group_guid-/members
for more information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

# get the first 20 groups
groups <- get_groups(client)

group_guid <- groups$guid[1]

get_group_members(client, guid = group_guid)
}

}
\seealso{
Other groups functions: 
\code{\link{get_group_content}()},
\code{\link{get_groups}()}
}
\concept{groups functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promote.R
\name{promote}
\alias{promote}
\title{Promote content from one Connect server to another}
\usage{
promote(from, to, to_key, from_key, name)
}
\arguments{
\item{from}{The url for the server containing the content (the originating
server)}

\item{to}{The url for the server where the content will be deployed (the
destination server)}

\item{to_key}{An API key on the destination "to" server. If the destination
content is going to be updated, the API key must belong to a user with
collaborator access on the content that will be updated. If the destination
content is to be created new, the API key must belong to a user with
publisher privileges.}

\item{from_key}{An API key on the originating "from" server. The API key must
belong to a user with collaborator access to the content to be promoted.}

\item{name}{The name of the content on the originating "from" server.
If content with the same name is found on the destination server,
the content will be updated. If no content on the destination server
has a matching name, a new endpoint will be created.}
}
\value{
The URL for the content on the destination "to" server
}
\description{
Promote content from one Connect server to another
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{get_associations}
\alias{get_associations}
\title{Get OAuth associations for a piece of content}
\usage{
get_associations(x)
}
\arguments{
\item{x}{A \code{Content} object}
}
\value{
A list of OAuth integration associations. Each association includes details such as:
\itemize{
\item \code{app_guid}: The content item's GUID (deprecated, use \code{content_guid} instead).
\item \code{content_guid}: The content item's GUID.
\item \code{oauth_integration_guid}: The GUID of the OAuth integration.
\item \code{oauth_integration_name}: The name of the OAuth integration.
\item \code{oauth_integration_description}: A description of the OAuth integration.
\item \code{oauth_integration_template}: The template used for this OAuth integration.
\item \code{oauth_integration_auth_type}: The authentication type (e.g., "Viewer" or "Service Account").
\item \code{created_time}: The timestamp when the association was created.
}
}
\description{
Given a \code{Content} object, retrieves a list of its
OAuth associations. An association contains a content GUID and an association
GUID, and indicates that the integration can be used by the content when it
runs.
}
\examples{
\dontrun{
client <- connect()

# Get OAuth associations for an app.
my_app <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")
my_app_associations <- get_associations(my_app)

# Given those associations, retrieve the integrations themselves.
my_app_integrations <- purrr::map(
  my_app_associations,
  ~ get_integration(client, .x$oauth_integration_guid)
)
}

}
\seealso{
\code{\link[=set_integrations]{set_integrations()}}, \code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_integration]{get_integration()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{delete_integration}()},
\code{\link{get_integration}()},
\code{\link{get_integrations}()},
\code{\link{set_integrations}()},
\code{\link{update_integration}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{set_run_as}
\alias{set_run_as}
\title{Set RunAs User}
\usage{
set_run_as(content, run_as, run_as_current_user = FALSE)
}
\arguments{
\item{content}{an R6 Content item}

\item{run_as}{The RunAs user to use for this content}

\item{run_as_current_user}{Whether to run this content as the viewer of the application}
}
\value{
a Content object, updated with new details
}
\description{
Set the \code{RunAs} user for a piece of content.
The \code{run_as_current_user} flag only does anything if:
}
\details{
\itemize{
\item PAM is the authentication method
\item \code{Applications.RunAsCurrentUser} is enabled on the server
}

Also worth noting that the \code{run_as} user must exist on the Posit Connect
server (as a linux user) and have appropriate group memberships, or you will
get a \verb{400: Bad Request}. Set to \code{NULL} to use the default RunAs user / unset
any current configuration.

To "read" the current RunAs user, use the \code{Content} object or \code{get_content()} function.
}
\seealso{
connectapi::content_update

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{get_integrations}
\alias{get_integrations}
\title{Get OAuth integrations}
\usage{
get_integrations(x)
}
\arguments{
\item{x}{A \code{Connect} or \code{Content} R6 object.}
}
\value{
A list of class \code{connect_integration_list}, where each element is a \code{connect_integration} object
with the following fields. (Raw API fields are character strings unless noted otherwise):
\itemize{
\item \code{id}: The internal identifier of this OAuth integration.
\item \code{guid}: The GUID of this OAuth integration.
\item \code{created_time}: Timestamp (RFC3339) when the integration was created.
\item \code{updated_time}: Timestamp (RFC3339) when the integration was last updated.
\item \code{name}: A descriptive name.
\item \code{description}: A brief description.
\item \code{template}: The template used to configure the integration.
\item \code{auth_type}: The OAuth flow used.
\item \code{config}: A list with integration-specific config fields.
}

Use \code{\link[=as.data.frame]{as.data.frame()}} or \code{\link[tibble:as_tibble]{tibble::as_tibble()}} to convert the result to a data frame with parsed types.
}
\description{
Retrieve OAuth integrations either from the Connect server or associated with a specific content item.

If \code{x} is a \code{Connect} object, this function lists all OAuth integrations on the server.
If \code{x} is a \code{Content} object, it returns the integrations associated with that content item.

You must have administrator or publisher privileges to use this function.
}
\examples{
\dontrun{
# From a Connect client
client <- connect()
integrations <- get_integrations(client)

# Filter or update specific ones
github_integration <- purrr::keep(integrations, \(x) x$template == "github")[[1]]

json_payload <- jsonlite::toJSON(list(
  description = "Updated Description",
  config = list(client_secret = "new-secret")
), auto_unbox = TRUE)

client$PATCH(
  paste0("v1/oauth/integrations/", github_integration$guid),
  body = json_payload
)

# From a Content item
content <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")
content_integrations <- get_integrations(content)

# Filter content integrations
snowflake_integrations <- purrr::keep(content_integrations, ~ .x$template == "snowflake")
}

}
\seealso{
\code{\link[=get_integration]{get_integration()}}, \code{\link[=set_integrations]{set_integrations()}}, \code{\link[=get_associations]{get_associations()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{delete_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integration}()},
\code{\link{set_integrations}()},
\code{\link{update_integration}()}
}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{create_random_name}
\alias{create_random_name}
\title{Create Random Name}
\usage{
create_random_name(length = 25)
}
\arguments{
\item{length}{Optional. The length of the random name. Defaults to 25}
}
\value{
The randomly generated name
}
\description{
Creates a random name from the LETTERS dataset
}
\seealso{
connectapi::verify_content_name

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{update_integration}
\alias{update_integration}
\title{Update an OAuth integration}
\usage{
update_integration(
  integration,
  name = NULL,
  description = NULL,
  template = NULL,
  config = NULL
)
}
\arguments{
\item{integration}{A \code{connect_integration} object (as returned by \code{\link[=get_integrations]{get_integrations()}},
\code{\link[=get_integration]{get_integration()}}, or \code{\link[=create_integration]{create_integration()}}).}

\item{name}{A new name for the integration.}

\item{description}{A new description for the integration.}

\item{template}{The template to use (generally not changed after creation).}

\item{config}{A list with updated OAuth integration configuration. If \code{NULL}
(default), the configuration remains unchanged. You can update individual
configuration fields without affecting others.}
}
\value{
A \code{connect_integration} object representing the updated OAuth
integration. See \code{\link[=get_integration]{get_integration()}} for details on the returned object.
}
\description{
Updates an existing OAuth integration. All fields except \code{integration} are optional,
and are unchanged if not provided.

You must have administrator privileges to perform this action.

See the Posit Connect documentation on
\href{https://docs.posit.co/connect/admin/integrations/oauth-integrations/}{OAuth integrations} for
more information.
}
\examples{
\dontrun{
client <- connect()

# Get an existing integration
integration <- get_integration(client, "your-integration-guid")

# Update the integration's name and description
updated_integration <- update_integration(
  integration,
  name = "Updated GitHub Integration",
  description = "A more descriptive description."
)

# Update only the client secret in the configuration
updated_integration <- update_integration(
  integration,
  config = list(
    client_secret = "your-new-client-secret"
  )
)
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_integration]{get_integration()}}, \code{\link[=create_integration]{create_integration()}},
\code{\link[=delete_integration]{delete_integration()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{delete_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integration}()},
\code{\link{get_integrations}()},
\code{\link{set_integrations}()}
}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{bundle_path}
\alias{bundle_path}
\title{Define a bundle from a path (a path directly to a tar.gz file)}
\usage{
bundle_path(path)
}
\arguments{
\item{path}{The path to a .tar.gz file}
}
\value{
Bundle A bundle object
}
\description{
Define a bundle from a path (a path directly to a tar.gz file)
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

bundle_path(system.file("tests/testthat/examples/static.tar.gz", package = "connectapi"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other deployment functions: 
\code{\link{bundle_dir}()},
\code{\link{bundle_static}()},
\code{\link{deploy}()},
\code{\link{download_bundle}()},
\code{\link{poll_task}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{as_integration}
\alias{as_integration}
\title{Convert objects to integration class}
\usage{
as_integration(x, client)
}
\arguments{
\item{x}{An object to convert to an integration.}

\item{client}{The Connect client object where the integration comes from.}
}
\value{
An integration object. The object has all the fields from the
integrations endpoint (see \code{\link[=get_integrations]{get_integrations()}}) and a Connect client as a
\code{client} attribute (\code{attr(x, "client")})
}
\description{
Convert objects to integration class
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{get_integration}
\alias{get_integration}
\title{Get the details of an OAuth integration}
\usage{
get_integration(client, guid)
}
\arguments{
\item{client}{A \code{Connect} R6 client object.}

\item{guid}{The GUID of an integration available on the Connect server.}
}
\value{
A \code{connect_integration} object representing an OAuth integration,
which has the following fields:
\itemize{
\item \code{id}: The internal identifier of this OAuth integration.
\item \code{guid}: The GUID of this OAuth integration.
\item \code{created_time}: The timestamp (RFC3339) indicating when this integration
was created.
\item \code{updated_time}: The timestamp (RFC3339) indicating when this integration
was last updated
\item \code{name}: A descriptive name to identify the OAuth integration.
\item \code{description}: A brief text to describe the OAuth integration.
\item \code{template}: The template used to configure this OAuth integration.
\item \code{auth_type}: The authentication type indicates which OAuth flow is used by
this integration.
\item \code{config}: A list with the OAuth integration configuration. Fields
differ between integrations.
}
}
\description{
Given the GUID of an OAuth integration available on a Connect server, retrieve
its details. You must have administrator or publisher privileges to perform
this action.
}
\examples{
\dontrun{
client <- connect()
x <- get_integration(client, guid)
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_associations]{get_associations()}}, \code{\link[=set_integrations]{set_integrations()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{delete_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integrations}()},
\code{\link{set_integrations}()},
\code{\link{update_integration}()}
}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lazy.R
\name{tbl_connect}
\alias{tbl_connect}
\title{Connect Tibble}
\usage{
tbl_connect(
  src,
  from = c("users", "groups", "content", "usage_shiny", "usage_static", "audit_logs"),
  ...
)
}
\arguments{
\item{src}{The source object}

\item{from}{The type of tibble}

\item{...}{Additional arguments that are not yet implemented}
}
\value{
A \code{tbl_connect} object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
A lazy tibble that automatically pages through API requests when \code{collect}ed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{search_content}
\alias{search_content}
\title{Search for content on the Connect server}
\usage{
search_content(
  client,
  q = NULL,
  include = "owner,vanity_url",
  page_size = 500,
  limit = Inf,
  ...
)
}
\arguments{
\item{client}{A Connect object}

\item{q}{The search query, using the syntax described in the Connect
documentation on \href{https://docs.posit.co/connect/user/viewing-content/#searching-content}{content search terms}}

\item{include}{Comma-separated character string of values indicating additional
details to include in the response. Values can be \code{owner} and \code{vanity_url};
both are included by default.}

\item{page_size}{The number of items to fetch per page. Maximum is 500.}

\item{limit}{Maximum number of items to return overall. Defaults to \code{Inf} (all items).}

\item{...}{Additional query parameters passed to the API for future expansion.
Note: If you pass \code{page_number} here, it will affect the \emph{starting} page
for pagination, but all subsequent pages will still be fetched. This is
usually not what you want.}
}
\value{
A list of \link{Content} objects, of class "connect_content_list"
}
\description{
Search for content on the Connect server
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/search/content for more
information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

my_content <- search_content(client, q = "owner:@me")

shiny_content <- purrr::keep(my_content, function(x) {
  x$content$app_mode == "rmd-shiny"
})

purrr::map(shiny_content, lock_content)
}

}
\seealso{
Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{terminate_jobs}
\alias{terminate_jobs}
\title{Terminate Jobs}
\usage{
terminate_jobs(content, keys = NULL)
}
\arguments{
\item{content}{A Content object, as returned by \code{content_item()}}

\item{keys}{Optional. One or more job keys, which can be obtained using
\code{get_jobs(content)}. If no keys are provided, will terminate all active
jobs for the provided content item.}
}
\value{
A data frame with the status of each termination request.
\itemize{
\item \code{app_id}: The content item's identifier.
\item \code{app_guid}: The content item's GUID.
\item \code{job_key}: The job key.
\item \code{job_id}: The job's identifier.
\item \code{result}: The result string returned by Connect.
\item \code{code}: An error code, \code{NA} if the request was successful.
\item \code{error}: An error message, \code{NA} if the result was successful.
}

Note that \code{app_id}, \code{app_guid}, \code{job_id}, and \code{result} are \code{NA} if the
request returns an error.
}
\description{
Register a job kill order for one or more jobs associated with a content
item. Requires Connect 2022.10.0 or newer.
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
result <- terminate_jobs(item)
}

}
\seealso{
Other job functions: 
\code{\link{get_jobs}()},
\code{\link{get_log}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_integrations}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{job functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content_render}
\alias{content_render}
\title{Render a content item.}
\usage{
content_render(content, variant_key = NULL)
}
\arguments{
\item{content}{The content item you wish to render.}

\item{variant_key}{If a variant key is provided, render that variant. Otherwise, render the default variant.}
}
\value{
A \link{VariantTask} object that can be used to track completion of the render.
}
\description{
Submit a request to render a content item. Once submitted, the
server runs an asynchronous process to render the content. This might be
useful if content needs to be updated after its source data has changed,
especially if this doesn't happen on a regular schedule.

Only valid for rendered content (e.g., most Quarto documents, Jupyter
notebooks, R Markdown reports).
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
task <- content_render(item)
poll_task(task)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_audit_logs}
\alias{get_audit_logs}
\title{Get Audit Logs from Posit Connect Server}
\usage{
get_audit_logs(src, limit = 500, previous = NULL, nxt = NULL, asc_order = TRUE)
}
\arguments{
\item{src}{The source object}

\item{limit}{The number of records to return.}

\item{previous}{Retrieve the previous page of Shiny application usage
logs relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{nxt}{Retrieve the next page of Shiny application usage logs
relative to the provided value. This value corresponds to an internal
reference within the server and should be sourced from the appropriate
attribute within the paging object of a previous response.}

\item{asc_order}{Defaults to TRUE; Determines if the response records
should be listed in ascending or descending order within the response.
Ordering is by the started timestamp field.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{id}: ID of the audit action
\item \code{time}: Timestamp in RFC3339 format when action was taken
\item \code{user_id}: User ID of the actor who made the audit action
\item \code{user_description}: Description of the actor
\item \code{action}: Audit action taken
\item \code{event_description}: Description of action
}
}
\description{
Get Audit Logs from Posit Connect Server
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/audit_logs for more
information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

# get the last 20 audit logs
get_audit_logs(client, limit = 20, asc_order = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runtime-caches.R
\name{delete_runtime_cache}
\alias{delete_runtime_cache}
\title{Delete a runtime cache}
\usage{
delete_runtime_cache(
  client,
  language,
  version,
  image_name = "Local",
  dry_run = FALSE
)
}
\arguments{
\item{client}{A \code{Connect} object.}

\item{language}{The language of the cache, either "R" or "Python".}

\item{version}{The version of the cache, e.g. "4.3.3".}

\item{image_name}{Optional. The name of the off-host execution image for
the cache, or "Local" (the default) for native execution caches.}

\item{dry_run}{Optional, default \code{FALSE}. If true, perform a dry run of
the deletion.}
}
\value{
A \code{Task} object representing the deletion task. If \code{dry_run} is
\code{TRUE}, returns \code{NULL} or throws an error if the deletion would fail.
}
\description{
Delete a runtime cache from a Connect server.
Requires Administrator privileges.
}
\examples{
\dontrun{
client <- connect()
task <- delete_runtime_cache(client, "R", "4.3.3")
poll_task(task)
}

}
\seealso{
\code{\link[=get_runtime_caches]{get_runtime_caches()}}

Other server management functions: 
\code{\link{get_runtime_caches}()}
}
\concept{server management functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{bundle_static}
\alias{bundle_static}
\title{Define a bundle from a static file (or files)}
\usage{
bundle_static(
  path,
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz")
)
}
\arguments{
\item{path}{The path to a file (or files) that will be used for the static bundle}

\item{filename}{The output bundle path}
}
\value{
Bundle A bundle object
}
\description{
Defines a bundle from static files. It copies all files to a temporary
directory, generates a basic manifest file (using the first file as the
"primary"), and bundles the directory.
}
\details{
NOTE: the \code{rsconnect} package is required for this function to work properly.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

bundle_static(system.file("logo.png", package = "connectapi"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other deployment functions: 
\code{\link{bundle_dir}()},
\code{\link{bundle_path}()},
\code{\link{deploy}()},
\code{\link{download_bundle}()},
\code{\link{poll_task}()}
}
\concept{deployment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_packages}
\alias{get_packages}
\title{All package dependencies on the server}
\usage{
get_packages(src, name = NULL, page_size = 100000, limit = Inf)
}
\arguments{
\item{src}{A \code{Connect} client object.}

\item{name}{Optional package name to filter by. Python package are normalized
during matching; R package names must match exactly.}

\item{page_size}{Optional. Integer specifying page size for API
paging.}

\item{limit}{Optional. Specify the maximum number of records after which
to cease paging.}
}
\value{
A data frame with the following columns:
\itemize{
\item \code{language}: Language ecosystem the package belongs to (\code{r} or \code{python})
\item \code{language_version}: Version of R or Python used by the content
\item \code{name}: Package name
\item \code{version}: Package version
\item \code{hash}: Package description hash for R packages
\item \code{bundle_id}: Identifier for the bundle that depends on this package
\item \code{content_id}: Numeric identifier for the content that depends on this
package
\item \code{content_guid}: The unique identifier of the content item that depends on
this package
}
}
\description{
Get a data frame of package dependencies used by all content
items on the server.
}
\examples{
\dontrun{
client <- connect()
packages <- get_packages(client)
}

}
\seealso{
Other packages functions: 
\code{\link{get_content_packages}()}
}
\concept{packages functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{content_list_with_permissions}
\alias{content_list_with_permissions}
\alias{content_list_guid_has_access}
\title{Get Content List with Permissions}
\usage{
content_list_with_permissions(src, ..., .p = NULL)

content_list_guid_has_access(content_list, guid)
}
\arguments{
\item{src}{A Connect R6 object}

\item{...}{Extra arguments. Currently not used}

\item{.p}{Optional. A predicate function, passed as-is to \code{purrr::keep()}. See
\code{get_content()} for more details. Can greatly help performance by reducing
how many items to get permissions for}

\item{content_list}{A "content list with permissions" as returned by \code{content_list_with_permissions()}}

\item{guid}{A user or group GUID to filter the content list by whether they have access}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} These functions are experimental placeholders until the API supports
this behavior.
}
\details{
\code{content_list_with_permissions} loops through content and retrieves
permissions for each item (with a progress bar). This can take a long time
for lots of content! Make sure to use the optional \code{.p} argument as a predicate
function that filters the content list before it is transformed.

\code{content_list_guid_has_access} works with a \code{content_list_with_permissions}
dataset by checking whether a given GUID (either user or group) has access to
the content by:
\itemize{
\item checking if the content has access_type == "all"
\item checking if the content has access_type == "logged_in"
\item checking if the provided guid is the content owner
\item checking if the provided guid is in the list of content permissions (in the "permissions" column)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_aws_credentials}
\alias{get_aws_credentials}
\title{Obtain a visitor's AWS credentials}
\usage{
get_aws_credentials(connect, user_session_token, audience = NULL)
}
\arguments{
\item{connect}{A Connect R6 object.}

\item{user_session_token}{The content visitor's session token. This token
can only be obtained when the content is running on a Connect server. The token
identifies the user who is viewing the content interactively on the Connect server.
Read this value from the HTTP header: \code{Posit-Connect-User-Session-Token}}

\item{audience}{Optional. The GUID of an OAuth integration associated with
this piece of content.}
}
\value{
The AWS credentials as a list with fields named \code{access_key_id},
\code{secret_access_key}, \code{session_token}, and \code{expiration}.
}
\description{
Obtain a visitor's AWS credentials
}
\details{
Please see https://docs.posit.co/connect/user/oauth-integrations/#obtaining-service-account-aws-credentials
for more information. See the example below of using this function in a
Plumber API with paws to access S3. Any library that allows you to pass
AWS credentials will be able to utilize the credentials returned from
this function call.
}
\examples{
\dontrun{
library(connectapi)
library(plumber)
library(paws)
client <- connect()

#* @get /do
function(req) {
  user_session_token <- req$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
  aws_credentials <- get_aws_credentials(client, user_session_token)

  # Create S3 client with AWS credentials from Connect
  svc <- paws::s3(
    credentials = list(
      creds = list(
        access_key_id = aws_credentials$access_key_id,
        secret_access_key = aws_credentials$secret_access_key,
        session_token = aws_credentials$session_token
      )
    )
  )

  # Get object from S3
  obj <- svc$get_object(
    Bucket = "my-bucket",
    Key = "my-data.csv"
  )

  "done"
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audits.R
\name{audit_runas}
\alias{audit_runas}
\title{Audit Run As Settings}
\usage{
audit_runas(content)
}
\arguments{
\item{content}{\code{data.frame} of content information, as from \code{\link[=get_content]{get_content()}}}
}
\value{
A data frame with the app name and the Run As user if the Run As user
is not the default
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\seealso{
Other audit functions: 
\code{\link{audit_access_open}()},
\code{\link{audit_r_versions}()},
\code{\link{vanity_is_available}()}
}
\concept{audit functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{delete_integration}
\alias{delete_integration}
\title{Delete an OAuth integration}
\usage{
delete_integration(integration)
}
\arguments{
\item{integration}{A \code{connect_integration} object (as returned by \code{\link[=get_integrations]{get_integrations()}},
\code{\link[=get_integration]{get_integration()}}, or \code{\link[=create_integration]{create_integration()}}).}
}
\value{
Returns \code{NULL} invisibly if successful.
}
\description{
Deletes an OAuth integration from the Posit Connect server. This permanently
removes the integration and any associated content associations.

You must have administrator privileges to perform this action.

See the Posit Connect documentation on
\href{https://docs.posit.co/connect/admin/integrations/oauth-integrations/}{OAuth integrations} for
more information.
}
\examples{
\dontrun{
client <- connect()

# Get an integration to delete
integration <- get_integration(client, "your-integration-guid")

# Delete the integration
delete_integration(integration)
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_integration]{get_integration()}}, \code{\link[=create_integration]{create_integration()}},
\code{\link[=update_integration]{update_integration()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integration}()},
\code{\link{get_integrations}()},
\code{\link{set_integrations}()},
\code{\link{update_integration}()}
}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect.R
\name{connect}
\alias{connect}
\title{Create a connection to Posit Connect}
\usage{
connect(
  server = Sys.getenv("CONNECT_SERVER"),
  api_key = Sys.getenv("CONNECT_API_KEY"),
  token,
  token_local_testing_key = api_key,
  audience = NULL,
  ...,
  .check_is_fatal = TRUE
)
}
\arguments{
\item{server}{The URL for accessing Posit Connect. Defaults to environment
variable CONNECT_SERVER}

\item{api_key}{The API Key to authenticate to Posit Connect with. Defaults
to environment variable CONNECT_API_KEY}

\item{token}{Optional. A user session token. When running on a Connect server,
creates a client using the content visitor's account. Running locally, the
created client uses the provided API key.}

\item{token_local_testing_key}{Optional. Only used when not running on
Connect and a \code{token} is provided. By default, the function returns a
\code{Connect} object using the \code{api_key}. By providing a different
key here you can test a visitor client with differently-scoped
permissions.}

\item{audience}{Optional. The GUID of a Connect API integration associated with this piece of content.}

\item{...}{Additional arguments. Not used at present}

\item{.check_is_fatal}{Whether to fail if "check" requests fail. Useful in
rare cases where more http request customization is needed for requests to
succeed.}
}
\value{
A Posit Connect R6 object that can be passed along to methods
}
\description{
Creates a connection to Posit Connect using the server URL and an API key.
Validates the connection and checks that the version of the server is
compatible with the current version of the package.
}
\details{
When running on Connect, the client's environment will contain default
\code{CONNECT_SERVER} and \code{CONNECT_API_KEY} variables. The API key's permissions
are scoped to the publishing user's account.

To create a client with permissions scoped to the content visitor's account,
call \code{connect()} passing a user session token from content session headers
to the \code{token} argument. To do this, you must first add a Connect API
integration in your published content's Access sidebar.
}
\examples{
\dontrun{
client <- connect()

# Running in Connect, create a client using the content visitor's account.
# This example assumes code is being executed in a Shiny app's `server`
# function with a `session` object available.
token <- session$request$HTTP_POSIT_CONNECT_USER_SESSION_TOKEN
client <- connect(token = token)

# Test locally with an API key using a different role.
fallback_key <- Sys.getenv("VIEWER_ROLE_API_KEY")
client <- connect(token = token, token_local_testing_key = fallback_key)
}

\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

# default is to read CONNECT_SERVER and CONNECT_API_KEY environment variables
connect()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrations.R
\name{set_integrations}
\alias{set_integrations}
\title{Set all OAuth integrations for a content item}
\usage{
set_integrations(content, integrations)
}
\arguments{
\item{content}{A \code{Content} R6 object representing the content item to modify.}

\item{integrations}{The complete set of integrations to be associated with the
content. May be a single \code{connect_integration} object, a list of
\code{connect_integration} objects, or \code{NULL}. Passing in an empty list or
explicitly passing \code{NULL} will remove all associated integrations from the
content.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Removes all existing OAuth integrations associated with a content item, and
creates associations with the integrations provided. You must have
administrator or publisher privileges to perform this action.
}
\examples{
\dontrun{
client <- connect()

content <- content_item(client, "12345678-90ab-cdef-1234-567890abcdef")

integrations <- get_integrations(client)

# Associate a single integration
github_integration <- purrr::keep(integrations, \(x) x$template == "github")[[1]]
set_integrations(content, github_integration)

# Associate multiple integrations at once
selected_integrations <- integrations[1:2]
set_integrations(content, selected_integrations)

# Unset integrations
set_integrations(content, NULL)
}

}
\seealso{
\code{\link[=get_integrations]{get_integrations()}}, \code{\link[=get_integration]{get_integration()}}, \code{\link[=get_associations]{get_associations()}}, \code{\link[=content_item]{content_item()}}

Other oauth integration functions: 
\code{\link{create_integration}()},
\code{\link{delete_integration}()},
\code{\link{get_associations}()},
\code{\link{get_integration}()},
\code{\link{get_integrations}()},
\code{\link{update_integration}()}

Other content functions: 
\code{\link{content_delete}()},
\code{\link{content_item}()},
\code{\link{content_title}()},
\code{\link{content_update}()},
\code{\link{create_random_name}()},
\code{\link{dashboard_url}()},
\code{\link{delete_thumbnail}()},
\code{\link{delete_vanity_url}()},
\code{\link{deploy_repo}()},
\code{\link{get_associations}()},
\code{\link{get_bundles}()},
\code{\link{get_environment}()},
\code{\link{get_jobs}()},
\code{\link{get_log}()},
\code{\link{get_thumbnail}()},
\code{\link{get_vanity_url}()},
\code{\link{git}},
\code{\link{has_thumbnail}()},
\code{\link{lock_content}()},
\code{\link{permissions}},
\code{\link{search_content}()},
\code{\link{set_run_as}()},
\code{\link{set_thumbnail}()},
\code{\link{set_vanity_url}()},
\code{\link{swap_vanity_urls}()},
\code{\link{terminate_jobs}()},
\code{\link{verify_content_name}()}
}
\concept{content functions}
\concept{oauth integration functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{Task}
\alias{Task}
\title{Task}
\description{
Task

Task
}
\details{
An R6 class that represents a Task
}
\seealso{
Other R6 classes: 
\code{\link{Bundle}},
\code{\link{Content}},
\code{\link{ContentTask}},
\code{\link{Environment}},
\code{\link{PositConnect}},
\code{\link{Vanity}},
\code{\link{Variant}},
\code{\link{VariantSchedule}},
\code{\link{VariantTask}}
}
\concept{R6 classes}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{connect}}{The Connect instance.}

\item{\code{task}}{The task.}

\item{\code{data}}{The task data.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Task-new}{\code{Task$new()}}
\item \href{#method-Task-get_task}{\code{Task$get_task()}}
\item \href{#method-Task-add_data}{\code{Task$add_data()}}
\item \href{#method-Task-get_data}{\code{Task$get_data()}}
\item \href{#method-Task-print}{\code{Task$print()}}
\item \href{#method-Task-clone}{\code{Task$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-new"></a>}}
\if{latex}{\out{\hypertarget{method-Task-new}{}}}
\subsection{Method \code{new()}}{
Initialize this task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$new(connect, task)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{connect}}{The \code{Connect} instance.}

\item{\code{task}}{The task data.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-get_task"></a>}}
\if{latex}{\out{\hypertarget{method-Task-get_task}{}}}
\subsection{Method \code{get_task()}}{
Return the underlying task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$get_task()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-add_data"></a>}}
\if{latex}{\out{\hypertarget{method-Task-add_data}{}}}
\subsection{Method \code{add_data()}}{
Set the data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$add_data(data)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{The data.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-get_data"></a>}}
\if{latex}{\out{\hypertarget{method-Task-get_data}{}}}
\subsection{Method \code{get_data()}}{
Get the data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$get_data()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-print"></a>}}
\if{latex}{\out{\hypertarget{method-Task-print}{}}}
\subsection{Method \code{print()}}{
Print this object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Unused.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Task-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_users}
\alias{get_users}
\title{Get user information from the Posit Connect server}
\usage{
get_users(
  src,
  page_size = 500,
  prefix = NULL,
  limit = Inf,
  user_role = NULL,
  account_status = NULL
)
}
\arguments{
\item{src}{The source object}

\item{page_size}{the number of records to return per page (max 500)}

\item{prefix}{Filters users by prefix (username, first name, or last name).
The filter is case insensitive.}

\item{limit}{The max number of records to return}

\item{user_role}{Optionally filter by user role ("administrator",
"publisher", "viewer"). Pass in a vector of multiple roles to match any value
(boolean OR). When \code{NULL} (the default), results are not filtered.}

\item{account_status}{Optionally filter by account status ("locked",
"licensed", "inactive"). Pass a vector of multiple statuses to match any
value (boolean OR). When \code{NULL} (the default), results are not filtered.}
}
\value{
A tibble with the following columns:
\itemize{
\item \code{email}: The user's email
\item \code{username}: The user's username
\item \code{first_name}: The user's first name
\item \code{last_name}: The user's last name
\item \code{user_role}: The user's role. It may have a value of administrator,
publisher or viewer.
\item \code{created_time}: The timestamp (in RFC3339 format) when the user was
created in the Posit Connect server
\item \code{updated_time}: The timestamp (in RFC3339 format) when the user was last
updated in the Posit Connect server
\item \code{active_time}: The timestamp (in RFC3339 format) when the user was last
active on the Posit Connect server
\item \code{confirmed}: When false, the created user must confirm their account
through an email. This feature is unique to password authentication.
\item \code{locked}: Whether or not the user is locked
\item \code{guid}: The user's GUID, or unique identifier, in UUID RFC4122 format
}
}
\description{
Get user information from the Posit Connect server
}
\details{
Please see https://docs.posit.co/connect/api/#get-/v1/users for more information.
}
\examples{
\dontrun{
library(connectapi)
client <- connect()

# Get all users
get_users(client)

# Get all licensed users
get_users(client, account_status = "licensed")

# Get all users who are administrators or publishers
get_users(client, user_role = c("administrator", "publisher"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schedule.R
\name{get_variant_schedule}
\alias{get_variant_schedule}
\title{Get a Variant Schedule}
\usage{
get_variant_schedule(variant)
}
\arguments{
\item{variant}{A Variant object, as returned by \code{get_variant()} or \code{get_variant_default()}}
}
\value{
A VariantSchedule object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Gets the schedule associated with a Variant.
}
\seealso{
Other schedule functions: 
\code{\link{get_timezones}()},
\code{\link{set_schedule}()}
}
\concept{schedule functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{get_content_packages}
\alias{get_content_packages}
\title{Package dependencies for a content item}
\usage{
get_content_packages(content)
}
\arguments{
\item{content}{A content item}
}
\value{
A data frame with the following columns:
\itemize{
\item \code{language} : Language ecosystem the package belongs to (\code{r} or \code{python})
\item \code{name}: The package name
\item \code{version}: The package version
\item \code{hash}: For R packages, the package \code{DESCRIPTION} hash
}
}
\description{
Get a data frame of package dependencies used by a content item.
}
\examples{
\dontrun{
client <- connect()
item <- content_item(client, "951bf3ad-82d0-4bca-bba8-9b27e35c49fa")
packages <- get_content_packages(item)
}

}
\seealso{
Other packages functions: 
\code{\link{get_packages}()}
}
\concept{packages functions}
